/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef COMMON_HPP
#define COMMON_HPP

#define _USE_MATH_DEFINES
#include <cstdlib>
#include <cstdio>
#include <math.h>
#include <iostream>
#include <limits>
#include <tvmet/Vector.h>
#include <tvmet/Matrix.h>
#include <ctime>
#include <vector>
#ifdef _WIN32
#include <algorithm>
#include <conio.h>
#include <windows.h>
#undef min
#undef max
inline float fminf(float x, float y) {return std::min<float>(x,y);}
inline float fmaxf(float x, float y) {return std::max<float>(x,y);}
#endif

using namespace tvmet;

#define uint unsigned int

typedef Vector<float,2> Vec2f;
typedef Matrix<float,2,2> Mat2f;

typedef Vector<float,3> Vec3f;
typedef Matrix<float,3,3> Mat3f;

inline float sqf (float x) {return x*x;}
inline float randf () {return (float)rand()/RAND_MAX;}
#ifdef WIN32
inline long int lroundf (float x) {return (long int)(x+0.5f);}
inline float cbrtf (float x) {return pow(x,1/3.f);}
#endif
inline float clampf (float x, float a, float b) {return x<a ? a : x>b ? b : x;}
inline float reflectf (float x, float a, float b) {return x<a ? 2*a-x : x>b ? 2*b-x : x;}

inline float lerp (float t, float a, float b) {return a + t*(b-a);}
inline float lerp (float f0, float f1, float a, float b) {return f0*a + f1*b;}

inline uint nearest_power (float base, float target) {return powf(base, lroundf(logf(target)/logf(base)));}

static const float not_a_number = std::numeric_limits<float>::quiet_NaN();
static const float infinity = std::numeric_limits<float>::infinity();
static const float epsilon = std::numeric_limits<float>::epsilon();

// diagonal matrix from vector of diagonal entries
template<class T, std::size_t Sz> inline Matrix<T,Sz,Sz> diagonal (const Vector<T,Sz>& d) {
	Matrix<T,Sz,Sz> m(0);
	for (uint i = 0; i < Sz; i++)
		m(i,i) = d(i);
	return m;
}
// outer (tensor) product of two vectors
template<class T, std::size_t Sz> inline Matrix<T,Sz,Sz> outer (const Vector<T,Sz>& u, const Vector<T,Sz>& v) {
	Matrix<T,Sz,Sz> m;
	for (uint j = 0; j < Sz; j++) for (uint i = 0; i < Sz; i++)
		m(i,j) = u(i)*v(j);
	return m;
}
// determinant of matrix
template<class T> inline T det (const Matrix<T,2,2> &A) {
	return A(0,0)*A(1,1) - A(0,1)*A(1,0);
}
template<class T> inline T det (const Matrix<T,3,3> &A) {
	return dot(col(A,0), cross(col(A,1),col(A,2)));
}
// inverse of matrix
template<class T> inline Matrix<T,2,2> inverse (const Matrix<T,2,2>& A) {
	Matrix<T,2,2> A_inv;
	A_inv(0,0) = A(1,1);
	A_inv(0,1) = -A(0,1);
	A_inv(1,0) = -A(1,0);
	A_inv(1,1) = A(0,0);
	A_inv /= det(A);
	return A_inv;
}
template<class T> inline Matrix<T,3,3> inverse (const Matrix<T,3,3>& A) {
	Matrix<T,3,3> A_inv;
	A_inv(0,0) = A(1,1)*A(2,2) - A(1,2)*A(2,1);
	A_inv(0,1) = A(0,2)*A(2,1) - A(0,1)*A(2,2);
	A_inv(0,2) = A(0,1)*A(1,2) - A(0,2)*A(1,1);
	A_inv(1,0) = A(1,2)*A(2,0) - A(1,0)*A(2,2);
	A_inv(1,1) = A(0,0)*A(2,2) - A(0,2)*A(2,0);
	A_inv(1,2) = A(0,2)*A(1,0) - A(0,0)*A(1,2);
	A_inv(2,0) = A(1,0)*A(2,1) - A(1,1)*A(2,0);
	A_inv(2,1) = A(0,1)*A(2,0) - A(0,0)*A(2,1);
	A_inv(2,2) = A(0,0)*A(1,1) - A(0,1)*A(1,0);
	A_inv /= det(A);
	return A_inv;
}

template<class T>
inline T clamp (T x, T a, T b) {return x<a ? a : x>b ? b : x;}
template <class T> bool isnanl(T n) {
	return n != n;
}

/// Find an element in an array
template <class T> int find(T val, std::vector<T> &array) {
	for (int i = 0; i < array.size(); i++) {
		if (val == array[i]) return i;
	}
	return -1;
}

#ifdef WIN32
#include <float.h>
//#define isnan(x) _isnan(x)
class Timer {
public:
	__int64 then;
	float tick () {
		__int64 now, freq;
		QueryPerformanceCounter((LARGE_INTEGER*)&now);
		QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
		float retval=(float)(now - then)/freq;
		then = now;
		return retval;
	}
};
#else
#include <ctime>
class Timer {
public:
	struct timespec then;
	float tick () {
		struct timespec now;
		clock_gettime(CLOCK_REALTIME, &now);
		float retval = (now.tv_sec - then.tv_sec) + (now.tv_nsec - then.tv_nsec)*1e-9f;
		then = now;
		return retval;
	}
};
#endif

#endif
