/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <fstream>
#include <stdint.h>
#define TIXML_USE_STL
#include <tinyxml.h>
#include <cstring>
#include <cstdlib>
#include <string>

#include "io.hpp"
#include "iohelpers.hpp"
#include "../sim/particle.hpp"
#include "../sim/scripted.hpp"
#include "../model/primitive.hpp"

#include <btBulletWorldImporter.h>

using namespace std;

// ---------------- //
// Helper functions //
// ---------------- //

template <size_t d> struct Size;
template <> struct Size<2> {int32_t m, n;};
template <> struct Size<3> {int32_t m, n, o;};

/*template <size_t d> void resize (Sand<d> &sand, Size<d> s, float dx);
template <> inline void resize (Sand<2> &sand, Size<2> s, float dx) {sand.resize(s.m,s.n, dx);}
template <> inline void resize (Sand<3> &sand, Size<3> s, float dx) {sand.resize(s.m,s.n,s.o, dx);}*/

template <size_t d, typename T> void resize (T &thing, Size<d> s, float dx);
template <> inline void resize (Sand<2> &sand, Size<2> s, float dx) {sand.resize(s.m,s.n, dx);}
template <> inline void resize (Sand<3> &sand, Size<3> s, float dx) {sand.resize(s.m,s.n,s.o, dx);}
template <> inline void resize (ScalarField<2> &field, Size<2> s, float dx) {field.resize(s.m,s.n, dx);}
template <> inline void resize (ScalarField<3> &field, Size<3> s, float dx) {field.resize(s.m,s.n,s.o, dx);}

template <size_t d, typename T> Size<d> size (const T &thing);
template <> inline Size<2> size (const Sand<2> &sand) {Size<2> s; s.m = sand.m; s.n = sand.n; return s;}
template <> inline Size<3> size (const Sand<3> &sand) {Size<3> s; s.m = sand.m; s.n = sand.n; s.o = sand.o; return s;}
template <> inline Size<2> size (const ScalarField<2> &field) {Size<2> s; s.m = field.m; s.n = field.n; return s;}
template <> inline Size<3> size (const ScalarField<3> &field) {Size<3> s; s.m = field.m; s.n = field.n; s.o = field.o; return s;}

template <size_t d> Size<d> lroundf (const Vector<float,d> &v);
template <> Size<2> lroundf (const Vec2f &v) {Size<2> s; s.m = lroundf(v[0]); s.n = lroundf(v[1]); return s;}
template <> Size<3> lroundf (const Vec3f &v) {Size<3> s; s.m = lroundf(v[0]); s.n = lroundf(v[1]); s.o = lroundf(v[2]); return s;}

// ---------------- //
//   Sand sources   //
// ---------------- //

template <size_t d> void add_particles_box (vector<Particle<d> > &particles, float r, Vector<float,d> x0, Vector<float,d> x1, float rho, Vector<float,d> v);
template <> void add_particles_box<2> (vector<Particle<2> > &particles, float r, Vec2f x0, Vec2f x1, float rho, Vec2f v) {
	if (x0[0] > x1[0]) std::swap(x0[0], x1[0]);
	if (x0[1] > x1[1]) std::swap(x0[1], x1[1]);
	float wx = x1[0]-x0[0], wy = x1[1]-x0[1];
	int m, n;
	if (wx > wy) {m = lroundf(wx/wy); n = 1;}
	else {m = 1; n = lroundf(wy/wx);}
	float wxb = wx/m, wyb = wy/n;
	float volb = wxb*wyb;
	// particles per side
	int pps = (int)ceil(sqrtf(volb)/(2*r));
	float dx = wxb/pps, dy = wyb/pps;
	Mat2f r2(diagonal(Vec2f(dx*dx,dy*dy))*0.25);
	for (int block = 0; block < m*n; block++) {
		float x0b = x0[0] + (m>1 ? block*wxb : 0);
		float y0b = x0[1] + (n>1 ? block*wyb : 0);
		for (int j = 0; j < pps; j++) for (int i = 0; i < pps; i++) {
			Particle<2> p;
			p.x[0] = x0b + dx*(i + randf()/**0.5f + 0.25f*/);
			p.x[1] = y0b + dy*(j + randf()/**0.5f + 0.25f*/);
			p.v = v;
			p.m = dx*dy*rho;
			p.r2 = r2;
			particles.push_back(p);
		}
	}
}
template <> void add_particles_box<3> (vector<Particle<3> > &particles, float r, Vec3f x0, Vec3f x1, float rho, Vec3f v) {
	if (x0[0] > x1[0]) std::swap(x0[0], x1[0]);
	if (x0[1] > x1[1]) std::swap(x0[1], x1[1]);
	if (x0[2] > x1[2]) std::swap(x0[2], x1[2]);
	float wx = x1[0]-x0[0], wy = x1[1]-x0[1], wz = x1[2]-x0[2];
	int m, n, o;
	if (wz < wx && wz < wy) {m = lroundf(wx/wz); n = lroundf(wy/wz); o = 1;}
	else if (wy < wx) {m = lroundf(wx/wy); n = 1; o = lroundf(wz/wy);}
	else {m = 1; n = lroundf(wy/wx); o = lroundf(wz/wx);}
	float wxb = wx/m, wyb = wy/n, wzb = wz/o;
	float volb = wxb*wyb*wzb;
	// particles per side
	int pps = (int)ceil(pow(volb,1/3.f)/(2*r));
	float dx = wxb/pps, dy = wyb/pps, dz = wzb/pps;
	Mat3f r2(diagonal(Vec3f(dx*dx,dy*dy,dz*dz))*0.25);
	for (int block = 0; block < m*n*o; block++) {
		int bi = block%m, bj = (block/m)%n, bk = (block/m)/n;
		float x0b = x0[0] + bi*wxb;
		float y0b = x0[1] + bj*wyb;
		float z0b = x0[2] + bk*wzb;
		for (int k = 0; k < pps; k++) for (int j = 0; j < pps; j++) for (int i = 0; i < pps; i++) {
			Particle<3> p;
			p.x[0] = x0b + dx*(i + randf()/**0.5f + 0.25f*/);
			p.x[1] = y0b + dy*(j + randf()/**0.5f + 0.25f*/);
			p.x[2] = z0b + dz*(k + randf()/**0.5f + 0.25f*/);
			p.v = v;
			p.m = dx*dy*dz*rho;
			p.r2 = r2;
			particles.push_back(p);
		}
	}
}

template <size_t d> void add_particles_sphere (vector<Particle<d> > &particles, float r, Vector<float,d> c, float radius, float rho, Vector<float,d> v);
template <> void add_particles_sphere<2> (vector<Particle<2> > &particles, float r, Vec2f c, float radius, float rho, Vec2f v) {
	// divisions per radius
	int nr = (int)lroundf(radius/(2*r));
	float dx = radius/nr;
	Mat2f r2(identity<float,2,2>()*dx*dx*0.25f);
	for (int ir = 0; ir < nr; ir++) {
		int ntheta = lroundf(2*M_PI*(ir+1));
		for (int itheta = 0; itheta < ntheta; itheta++) {
			Particle<2> p;
			float dist = dx*sqrtf(sqf(ir) + (sqf(ir+1) - sqf(ir))*(randf()/**0.5f + 0.25f*/));
			float theta = 2*M_PI*(itheta + randf()/**0.5f + 0.25f*/)/ntheta;
			p.x[0] = c[0] + dist*cos(theta);
			p.x[1] = c[1] + dist*sin(theta);
			p.v = v;
			p.m = M_PI*sqf(radius) * (sqf(ir+1) - sqf(ir))/sqf(nr) * 1/ntheta * rho;
			p.r2 = r2;
			particles.push_back(p);
		}
	}
}
inline float cbf (float x) {return x*x*x;}
template <> void add_particles_sphere<3> (vector<Particle<3> > &particles, float r, Vec3f c, float radius, float rho, Vec3f v) {
	// divisions per radius
	int nr = (int)lroundf(radius/(2*r));
	float dx = radius/nr;
	Mat3f r2(identity<float,3,3>()*dx*dx*0.25f);
	for (int ir = 0; ir < nr; ir++) {
		int ntheta = lroundf(M_PI*(ir+1));
		for (int itheta = 0; itheta < ntheta; itheta++) {
			float r_ = (ir+1)*sin(M_PI*clampf(ntheta*0.5f, itheta,itheta+1)/ntheta);
			int nphi = lroundf(2*M_PI*r_);
			for (int iphi = 0; iphi < nphi; iphi++) {
				Particle<3> p;
				float dist = dx*cbrtf(cbf(ir) + (cbf(ir+1) - cbf(ir))*(randf()/**0.5f + 0.25f*/));
				float theta = acos(cos(M_PI*itheta/ntheta) + (cos(M_PI*(itheta+1)/ntheta) - cos(M_PI*itheta/ntheta))*(randf()/**0.5f + 0.25f*/));
				float phi = 2*M_PI*(iphi + randf()/**0.5f + 0.25f*/)/nphi;
				p.x[0] = c[0] + dist*cos(theta);
				p.x[1] = c[1] + dist*sin(theta)*cos(phi);
				p.x[2] = c[2] + dist*sin(theta)*sin(phi);
				p.v = v;
				p.m = 4/3.f*M_PI*cbf(radius) * (cbf(ir+1) - cbf(ir))/cbf(nr) * (cos(M_PI*(itheta+1)/ntheta) - cos(M_PI*itheta/ntheta))/(-2) * 1/nphi * rho;
				p.r2 = r2;
				particles.push_back(p);
			}
		}
	}
}

static Vec2f clamp_vector (const Vec2f &x, const Vec2f &a, const Vec2f &b) {
	return Vec2f(clampf(x[0], a[0],b[0]), clampf(x[1], a[1],b[1]));
}
static Vec3f clamp_vector (const Vec3f &x, const Vec3f &a, const Vec3f &b) {
	return Vec3f(clampf(x[0], a[0],b[0]), clampf(x[1], a[1],b[1]), clampf(x[2], a[2],b[2]));
}

template <size_t d> void remove_particles_box (vector<Particle<d> > &particles, float r, Vector<float,d> x0, Vector<float,d> x1) {
	int np = particles.size();
	for (int p = 0; p < np; p++) {
		Particle<d> &particle = particles[p];
		float dist_box = norm2(particle.x - clamp_vector(particle.x, x0,x1));
		if (dist_box < r)
			particle.m = 0;
	}
	int q = 0;
	for (int p = 0; p < np; p++)
		if (particles[p].m > epsilon)
			particles[q++] = particles[p];
	particles.resize(q);
}

template <size_t d> void remove_particles_sphere (vector<Particle<d> > &particles, float r, Vector<float,d> cx, float radius) {
	int np = particles.size();
	for (int p = 0; p < np; p++) {
		Particle<d> &particle = particles[p];
		float dist = norm2(particle.x - cx) - radius;
		if (dist < r)
			particle.m = 0;
	}
	int q = 0;
	for (int p = 0; p < np; p++)
		if (particles[p].m > epsilon)
			particles[q++] = particles[p];
	particles.resize(q);
}

// ---------------- //
//    XML  input    //
// ---------------- //

// Helpers

template <size_t d> inline const char* dim_string () {return d==2 ? "2d" : d==3 ? "3d" : "";}
template <size_t d> void load_vector_xml (TiXmlElement *elem, const char *format, Vector<float,d> &v, const string &name);
template <> void load_vector_xml (TiXmlElement *elem, const char *format, Vec2f &v, const string &name) {
	if (!elem) cout << "No " << name << endl << yield;
	char dims[] = "xy";
	for (int i = 0; i < 2; i++) {
		string key = stringf(format, dims[i]);
		if (elem->QueryFloatAttribute(key.c_str(), &v[i]) != TIXML_SUCCESS)
			cout << "Error reading " << name << "/" << key << endl << yield;
	}
}
template <> void load_vector_xml (TiXmlElement *elem, const char *format, Vec3f &v, const string &name) {
	if (!elem) cout << "No " << name << endl << yield;
	char dims[] = "xyz";
	for (int i = 0; i < 3; i++) {
		string key = stringf(format, dims[i]);
		if (elem->QueryFloatAttribute(key.c_str(), &v[i]) != TIXML_SUCCESS)
			cout << "Error reading " << name << "/" << key << endl << yield;
	}
}
void load_quaternion_xml (TiXmlElement *elem, Quatf &q, const string &name) {
	if (!elem) cout << "No " << name << endl << yield;
	if (elem->QueryFloatAttribute("s", &q.s) != TIXML_SUCCESS) cout << "Error reading " << name << "/s" << endl << yield;
	load_vector_xml(elem, "%c", q.v, name);
	q = normalize(q);
}

template <size_t d> btCollisionShape* load_bullet_xml (TiXmlElement *elem);
template <> btCollisionShape* load_bullet_xml<2> (TiXmlElement *elem) {
	string shape = elem->Attribute("shape");
	if (shape.compare("sphere") == 0) {
		float r;
		if (elem->QueryFloatAttribute("r", &r) != TIXML_SUCCESS) cout << "Error reading bullet/r" << endl << yield;
		return new btCylinderShapeZ(btVector3(r,r,1));
	} else if (shape.compare("box") == 0) {
		Vec2f l;
		load_vector_xml(elem, "l%c", l, "bullet/l");
		return new btBoxShape(btVector3(l[0]/2,l[1]/2,1));
	} else {
		cout << "Unknown bullet shape " << shape << endl << yield;
		return NULL;
	}
}
template <> btCollisionShape* load_bullet_xml<3> (TiXmlElement *elem) {
	string shape = elem->Attribute("shape");
	if (shape.compare("sphere") == 0) {
		float r;
		if (elem->QueryFloatAttribute("r", &r) != TIXML_SUCCESS) cout << "Error reading bullet/r" << endl << yield;
		return new btSphereShape(r);
	} else if (shape.compare("box") == 0) {
		Vec3f l;
		load_vector_xml(elem, "l%c", l, "bullet/l");
		return new btBoxShape(btVector3(l[0]/2,l[1]/2,l[2]/2));
	} else {
		cout << "Unknown bullet shape " << shape << endl << yield;
		return NULL;
	}
}

template <size_t d> void load_body_xml (TiXmlElement *elem, RigidBody<d> &body);
template <> void load_body_xml (TiXmlElement *elem, RigidBody<2> &body) {
	float m, I;
	Vec2f c, x, v;
	float q, w;
	if (elem->QueryFloatAttribute("m", &m) != TIXML_SUCCESS) cout << "Error reading body/m" << endl << yield;
	load_vector_xml(elem, "c%c", c, "body/c");
	if (elem->QueryFloatAttribute("I", &I) != TIXML_SUCCESS) cout << "Error reading body/I" << endl << yield;
	load_vector_xml(elem->FirstChildElement("x"), "%c", x, "body/x");
	load_vector_xml(elem->FirstChildElement("v"), "%c", v, "body/v");
	if (elem->QueryFloatAttribute("q", &q) != TIXML_SUCCESS) cout << "Error reading body/q" << endl << yield;
	if (elem->QueryFloatAttribute("w", &w) != TIXML_SUCCESS) cout << "Error reading body/w" << endl << yield;
	TiXmlElement *bullet = elem->FirstChildElement("bullet");
	if (!bullet) cout << "No bullet shape" << endl << yield;
	btCollisionShape *shape = load_bullet_xml<2>(bullet);
	btTransform xform(btQuaternion(btVector3(0,0,1),q), to_bt(x));
	btTransform com(btQuaternion(0,0,0,1), to_bt(c));
	btRigidBody::btRigidBodyConstructionInfo ci(m, new btDefaultMotionState(xform,com), shape, btVector3(0,0,I));
	body.bt = new btRigidBody(ci);
	body.bt->setLinearVelocity(to_bt(v));
	body.bt->setAngularVelocity(btVector3(0,0,w));
	body.bt->setCollisionFlags(body.bt->getCollisionFlags() & (~btCollisionObject::CF_STATIC_OBJECT));
	body.aux.dp = 0.f;
	body.aux.ds = 0.f;
}
template <> void load_body_xml (TiXmlElement *elem, RigidBody<3> &body) {
	float m; Vec3f I, c;
	Vec3f x, v;
	Quatf q;
	Vec3f w;
	if (elem->QueryFloatAttribute("m", &m) != TIXML_SUCCESS) cout << "Error reading body/m" << endl << yield;
	load_vector_xml(elem, "c%c", c, "body/c");
	load_vector_xml(elem, "I%c", I, "body/I");
	load_vector_xml(elem->FirstChildElement("x"), "%c", x, "body/x");
	load_vector_xml(elem->FirstChildElement("v"), "%c", v, "body/v");
	load_quaternion_xml(elem->FirstChildElement("q"), q, "body/q");
	load_vector_xml(elem->FirstChildElement("w"), "%c", w, "body/w");
	TiXmlElement *bullet = elem->FirstChildElement("bullet");
	if (!bullet) cout << "No bullet shape" << endl << yield;
	btCollisionShape *shape = load_bullet_xml<3>(bullet);
	btTransform xform(to_bt(q), to_bt(x));
	btTransform com(btQuaternion(0,0,0,1), to_bt(c));
	btRigidBody::btRigidBodyConstructionInfo ci(m, new btDefaultMotionState(xform,com), shape, to_bt(I));
	body.bt = new btRigidBody(ci);
	body.bt->setLinearVelocity(to_bt(v));
	body.bt->setAngularVelocity(to_bt(w));
	body.bt->setCollisionFlags(body.bt->getCollisionFlags() & (~btCollisionObject::CF_STATIC_OBJECT));
	body.aux.dp = 0.f;
	body.aux.ds = 0.f;
}

template <size_t d> void load_pose_xml (TiXmlElement *elem, RigidPose<d> &pose, const string &name);
template <> void load_pose_xml (TiXmlElement *elem, RigidPose<2> &pose, const string &name) {
	load_vector_xml(elem->FirstChildElement("x"), "%c", pose.x, stringf("%s/x",name.c_str()));
	if (elem->QueryFloatAttribute("q", &pose.q) != TIXML_SUCCESS) cout << "Error reading " << name << "/q" << endl << yield;
}
template <> void load_pose_xml (TiXmlElement *elem, RigidPose<3> &pose, const string &name) {
	load_vector_xml(elem->FirstChildElement("x"), "%c", pose.x, stringf("%s/x",name.c_str()));
	load_quaternion_xml(elem->FirstChildElement("q"), pose.q, stringf("%s/q",name.c_str()));
}

template <size_t d> void load_twist_xml (TiXmlElement *elem, Screw<d> &twist, const string &name);
template <> void load_twist_xml (TiXmlElement *elem, Screw<2> &twist, const string &name) {
	Vec2f v;
	float w;
	load_vector_xml(elem->FirstChildElement("v"), "%c", v, stringf("%s/v",name.c_str()));
	if (elem->QueryFloatAttribute("w", &w) != TIXML_SUCCESS) cout << "Error reading " << name << "/w" << endl << yield;
	twist = screw(v,w);
}
template <> void load_twist_xml (TiXmlElement *elem, Screw<3> &twist, const string &name) {
	Vec3f v, w;
	load_vector_xml(elem->FirstChildElement("v"), "%c", v, stringf("%s/v",name.c_str()));
	load_vector_xml(elem->FirstChildElement("w"), "%c", w, stringf("%s/w",name.c_str()));
	twist = screw(v,w);
}

/*
bool load_renderman_setup (const string &filename, RendermanRenderer &ren) {
	try {
		TiXmlDocument doc(filename);
		if (!doc.LoadFile()) cout << "Couldn't load setup file '" << filename << "'" << endl << yield;
		TiXmlElement *pRoot = doc.RootElement();
		if (!pRoot) {cout << "No root element" << endl; return false;}
		
		TiXmlElement *pRibfiles = pRoot->FirstChildElement("ribfiles");
		if (!pRibfiles) cout << "No RIB files" << endl;
		else {
			if ( pRibfiles->Attribute("init") ) ren.ribinitfile = string(pRibfiles->Attribute("init"));
			if ( pRibfiles->Attribute("world") ) ren.ribworldfile = string(pRibfiles->Attribute("world"));
			if ( pRibfiles->Attribute("dsm") ) ren.ribdsmworldfile = string(pRibfiles->Attribute("dsm"));
			if ( pRibfiles->Attribute("dsminit") ) ren.ribdsminitfile = string(pRibfiles->Attribute("dsminit"));
		}

		TiXmlElement *pCamera = pRoot->FirstChildElement("camera");
		if (!pCamera) cout << "No Camera set" <<endl;
		else {
			//if (pCamera->QueryFloatAttribute("lat", &(ren.viewpoint.lat)) != TIXML_SUCCESS) cout << "Error reading camera/lat" << endl << yield;
			//if (pCamera->QueryFloatAttribute("lon", &(ren.viewpoint.lon)) != TIXML_SUCCESS) cout << "Error reading camera/lon" << endl << yield;
			if (pCamera->QueryFloatAttribute("posx", &(ren.camposition[0])) != TIXML_SUCCESS) cout << "Error reading camera/posx" << endl << yield;
			if (pCamera->QueryFloatAttribute("posy", &(ren.camposition[1])) != TIXML_SUCCESS) cout << "Error reading camera/posy" << endl << yield;
			if (pCamera->QueryFloatAttribute("posz", &(ren.camposition[2])) != TIXML_SUCCESS) cout << "Error reading camera/posz" << endl << yield;
			if (pCamera->QueryFloatAttribute("lookatx", &(ren.camto[0])) != TIXML_SUCCESS) cout << "Error reading camera/lookatx" << endl << yield;
			if (pCamera->QueryFloatAttribute("lookaty", &(ren.camto[1])) != TIXML_SUCCESS) cout << "Error reading camera/lookaty" << endl << yield;
			if (pCamera->QueryFloatAttribute("lookatz", &(ren.camto[2])) != TIXML_SUCCESS) cout << "Error reading camera/lookatz" << endl << yield;
			if (pCamera->QueryFloatAttribute("fov", &(ren.fov)) != TIXML_SUCCESS) cout << "Error reading camera/fov" << endl << yield;
		}
		
		TiXmlElement *pImage = pRoot->FirstChildElement("image");
		if (!pImage) cout << "Image resolution not set" << endl << yield;
		else {
			if(pImage->QueryIntAttribute("xres", &(ren.imageres[0])) !=TIXML_SUCCESS) cout << "Error reading image/xres" << endl << yield;
			if(pImage->QueryIntAttribute("yres", &(ren.imageres[1])) !=TIXML_SUCCESS) cout << "Error reading image/yres" << endl << yield;
		}

		TiXmlElement *pEnvMap = pRoot->FirstChildElement("envmap");
		if (pEnvMap) {
			if(pEnvMap->Attribute("name")) {
				ren.envmapname = string(pEnvMap->Attribute("name"));
			}
			else cout << "Environment map name not specified" << endl << yield;

			if (pEnvMap->QueryFloatAttribute("posx", &(ren.envmappos[0])) != TIXML_SUCCESS) cout << "Error reading envmap/posx" << endl << yield;
			if (pEnvMap->QueryFloatAttribute("posy", &(ren.envmappos[1])) != TIXML_SUCCESS) cout << "Error reading envmap/posy" << endl << yield;
			if (pEnvMap->QueryFloatAttribute("posz", &(ren.envmappos[2])) != TIXML_SUCCESS) cout << "Error reading envmap/posz" << endl << yield;
			if (pEnvMap->QueryIntAttribute("xres", &(ren.envmapres[0])) != TIXML_SUCCESS) cout << "Error reading envmap/xres" << endl << yield;
			if (pEnvMap->QueryIntAttribute("yres", &(ren.envmapres[1])) != TIXML_SUCCESS) cout << "Error reading envmap/yres" << endl << yield;
		}

		TiXmlElement *pSand = pRoot->FirstChildElement("sand");
		if (!pSand) cout << "Defaulting to default sand color" << endl;
		else {
			if (pSand->QueryFloatAttribute("r", &(ren.sandcolor[0])) == TIXML_SUCCESS) cout << "Setting sand red to " << ren.sandcolor[0] << endl;
			else cout << "Sand red not specified" << endl;
			if (pSand->QueryFloatAttribute("g", &(ren.sandcolor[1])) == TIXML_SUCCESS) cout << "Setting sand green to " << ren.sandcolor[1] << endl;
			else cout << "Sand green not specified" << endl;
			if (pSand->QueryFloatAttribute("b", &(ren.sandcolor[2])) == TIXML_SUCCESS) cout << "Setting sand blue to " << ren.sandcolor[2] << endl;
			else cout << "Sand blue not specified" << endl;
			if (pSand->QueryFloatAttribute("alpha", &(ren.sandopacity[0])) == TIXML_SUCCESS) {
				cout << "Setting sand alpha to " << ren.sandopacity[0] << endl;
				ren.sandopacity[1] = ren.sandopacity[2] = ren.sandopacity[0];
			}
			else cout << "Sand alpha not specified" << endl;
		}

		TiXmlElement *pShadowMap = pRoot->FirstChildElement("shadowmap");
		if (!pShadowMap) cout << "Shadow Map resolution not set" << endl;
		else {
			if(pShadowMap->QueryIntAttribute("res", &(ren.shadowmapres)) !=TIXML_SUCCESS) cout << "Error reading shadowmap/res" << endl << yield;
			if(pShadowMap->QueryFloatAttribute("fov", &(ren.shmapfov)) !=TIXML_SUCCESS) cout << "Error reading shadowmap/fov" << endl << yield;
			if(pShadowMap->Attribute("storage")) {
				ren.shmap_temp = string(pShadowMap->Attribute("storage"));
				cout << "Temporary shadowmap output set to : " << ren.shmap_temp << endl;
			}
			else {
				ren.shmap_temp = "temp.depth";
			}
		}

		TiXmlElement *pParams = pRoot->FirstChildElement("params");
		if (!pParams) cout << "No params set" << endl;
		else {
			if (pParams->QueryFloatAttribute("tol", &(ren.tolerance)) != TIXML_SUCCESS) cout << "Error reading params/tol" << endl << yield;
			if (pParams->Attribute("samplingtol") && strcmp(pParams->Attribute("samplingtol"), "yes") == 0) {
				ren.samplingtol = true;
			}
			else {
				ren.samplingtol = false;
			}
			if (pParams->QueryFloatAttribute("cutoff", &(ren.cutoff_distance)) != TIXML_SUCCESS) cout << "Error reading params/cutoff" << endl;
			if (pParams->QueryFloatAttribute("cutoffden", &(ren.cutoff_density)) != TIXML_SUCCESS) { 
				cout << "Error reading params/cutoff" << endl;
				ren.cutoff_density = ren.cutoff_distance;
			}
			if (pParams->QueryIntAttribute("sampling", &(ren.distfield_supersampling)) != TIXML_SUCCESS) cout << "Error reading params/cutoff" << endl;
			if (pParams->QueryFloatAttribute("nJitter", &(ren.nJitter)) != TIXML_SUCCESS) cout << "Error reading params/nJitter" << endl << yield;
			else ren.nJitter *= (float)M_PI / 180;
			if (pParams->QueryIntAttribute("gradientsamples", &(ren.nGradientSamples)) != TIXML_SUCCESS) {
				cout << "Number of gradient samples not specified, defaulting to 4" << endl;
				ren.nGradientSamples = 4;
			}
		}

		TiXmlElement *pParticles = pRoot->FirstChildElement("particles");
		if (!pParticles) cout << "No particle params set" << endl << yield;
		else {
			if (pParticles->QueryFloatAttribute("rmin", &(ren.rmin)) != TIXML_SUCCESS) cout << "Error reading particles/rmin" << endl << yield;
			if (pParticles->QueryFloatAttribute("rmax", &(ren.rmax)) != TIXML_SUCCESS) cout << "Error reading particles/rmax" << endl << yield;
			if (pParticles->QueryFloatAttribute("render_rad", &(ren.render_width)) != TIXML_SUCCESS) cout << "Error reading particles/render_rad" << endl << yield;
			if (pParticles->QueryIntAttribute("minRP", &(ren.minRP)) != TIXML_SUCCESS) cout << "No minimum render particles per sim particle" << endl;
			//ren.rmin *= 0.000001f;
			//ren.rmax *= 0.000001f;
		}
	} catch (Yield y) {return false;}

	return true;
}

*/
template <size_t d> bool load_distance_field (const RigidBodyInfo &info, const string &modeldir, ScalarField<d> &dist) {
	string distfile = modeldir+info.src+".dist";
	if (!load_scalar_field(distfile, dist)) {
		cout << "Error: Unable to load distance field: " << distfile << endl;
		return false;
	}
	dist.dx *= info.scale;
	for (int i = 0; i < dist.s.size(); i++) dist.s[i] -= info.offset/dist.dx;
	return true;
}

float inverse (float x) {return 1.f/x;}

btRigidBody* new_plane (const Vec3f &x, const Vec3f &n) {
	btCollisionShape* shape = new btStaticPlaneShape(to_bt(n), 0);
	btTransform xform(btQuaternion(0,0,0,1),to_bt(x));
	btRigidBody::btRigidBodyConstructionInfo ci(0, new btDefaultMotionState(xform), shape, btVector3(0,0,0));
	return new btRigidBody(ci);
}

template <size_t d> void add_walls (TiXmlElement *elem, Simulator<d> &sim);
template <> void add_walls (TiXmlElement *elem, Simulator<2> &sim) {
	if (!elem) return;
	float dx = sim.sand.dx;
	float x1 = sim.sand.m*dx, y1 = sim.sand.n*dx;
	if (elem->FirstChildElement("left"))   sim.bt->addRigidBody(new_plane(Vec3f(dx,0,0), Vec3f(1,0,0)));
	if (elem->FirstChildElement("right"))  sim.bt->addRigidBody(new_plane(Vec3f(x1-dx,0,0), Vec3f(-1,0,0)));
	if (elem->FirstChildElement("bottom")) sim.bt->addRigidBody(new_plane(Vec3f(0,dx,0), Vec3f(0,1,0)));
	if (elem->FirstChildElement("top"))    sim.bt->addRigidBody(new_plane(Vec3f(0,y1-dx,0), Vec3f(0,-1,0)));
}
template <> void add_walls (TiXmlElement *elem, Simulator<3> &sim) {
	if (!elem) return;
	float dx = sim.sand.dx;
	float x1 = sim.sand.m*dx, y1 = sim.sand.n*dx, z1 = sim.sand.o*dx;
	if (elem->FirstChildElement("left"))   sim.bt->addRigidBody(new_plane(Vec3f(dx,0,0), Vec3f(1,0,0)));
	if (elem->FirstChildElement("right"))  sim.bt->addRigidBody(new_plane(Vec3f(x1-dx,0,0), Vec3f(-1,0,0)));
	if (elem->FirstChildElement("bottom")) sim.bt->addRigidBody(new_plane(Vec3f(0,dx,0), Vec3f(0,1,0)));
	if (elem->FirstChildElement("top"))    sim.bt->addRigidBody(new_plane(Vec3f(0,y1-dx,0), Vec3f(0,-1,0)));
	if (elem->FirstChildElement("back"))   sim.bt->addRigidBody(new_plane(Vec3f(0,0,dx), Vec3f(0,0,1)));
	if (elem->FirstChildElement("front"))  sim.bt->addRigidBody(new_plane(Vec3f(0,0,z1-dx), Vec3f(0,0,-1)));
}

// Loading simulation setup

template <size_t d> bool load_setup (const string &filename, const string &modeldir, Simulator<d> &sim) {
	try {
		TiXmlDocument doc(filename);
		if (!doc.LoadFile()) cout << "Couldn't load setup file '" << filename << "'" << endl << yield;
		TiXmlElement *pRoot = doc.RootElement();
		if (!pRoot) {cout << "No root element" << endl; return false;}
		const char *dim = dim_string<d>();
		if (!pRoot->Attribute("type")) cout << "Type (2D or 3D) not specified" << endl << yield;
		if (string(pRoot->Attribute("type")).compare(dim) != 0) cout << "Type is not " << dim << endl << yield;
		sim.frame = 0;
		btDefaultCollisionConfiguration *cc = new btDefaultCollisionConfiguration;
		sim.bt = new btDiscreteDynamicsWorld(new btCollisionDispatcher(cc), new btDbvtBroadphase, new btSequentialImpulseConstraintSolver, cc);
		sim.bt->setGravity(btVector3(0,0,0));
		sim.sand.zhubridson = pRoot->Attribute("zhu-bridson") ? true : false;
		float dx;
		{
			// load params
			TiXmlElement *pParams = pRoot->FirstChildElement("params");
			if (!pParams) cout << "No params" << endl << yield;
			if (pParams->QueryFloatAttribute("dx", &dx) != TIXML_SUCCESS) cout << "Error reading params/dx" << endl << yield;
			if (pParams->QueryFloatAttribute("dt", &sim.dt) != TIXML_SUCCESS) cout << "Error reading params/dt" << endl << yield;
			if (pParams->QueryFloatAttribute("rho_max", &sim.sand.rho_max) != TIXML_SUCCESS) cout << "Error reading params/rho_max" << endl << yield;
			if (pParams->QueryFloatAttribute("r", &sim.sand.r) != TIXML_SUCCESS) cout << "Error reading params/r" << endl << yield;
			if (pParams->QueryFloatAttribute("mu", &sim.sand.mu) != TIXML_SUCCESS) cout << "Error reading params/mu" << endl << yield;
			load_vector_xml(pParams->FirstChildElement("g"), "%c", sim.g, "params/g");
		} {
			// load domain
			Vector<float,d> x1;
			load_vector_xml(pRoot->FirstChildElement("domain"), "%c1", x1, "domain");
			resize(sim.sand, lroundf(Vector<float,d>(x1/dx)), dx);
		} {
			// load sand sources
			TiXmlElement *pSource = pRoot->FirstChildElement("source");
			while (pSource) {
				float rho;
				Vector<float,d> v;
				if (pSource->QueryFloatAttribute("rho", &rho) != TIXML_SUCCESS) cout << "Error reading source/rho" << endl << yield;
				load_vector_xml(pSource->FirstChildElement("v"), "%c", v, "source/v");
				TiXmlElement *pShape = pSource->FirstChildElement();
				while (pShape) {
					if (pShape->ValueStr().compare("box") == 0) {
						Vector<float,d> x0, x1;
						load_vector_xml(pShape, "%c0", x0, "source/box");
						load_vector_xml(pShape, "%c1", x1, "source/box");
						add_particles_box<d>(sim.sand.particles, sim.sand.r, x0,x1, rho, v);
					} else if (pShape->ValueStr().compare("sphere") == 0) {
						Vector<float,d> c;
						float r;
						load_vector_xml(pShape, "c%c", c, "source/sphere");
						if (pShape->QueryFloatAttribute("r", &r) != TIXML_SUCCESS) cout << "Error reading source/sphere/r" << endl << yield;
						add_particles_sphere<d>(sim.sand.particles, sim.sand.r, c,r, rho, v);
					}
					pShape = pShape->NextSiblingElement();
				}
				pSource = pSource->NextSiblingElement("source");
			}
		} {
			// load subtraction for Zhu and Bridson
			TiXmlElement *pSubtract = pRoot->FirstChildElement("subtract");
			while (pSubtract) {
				if (!pSubtract->Attribute("type")) cout << "Error reading subtract/type" << endl << yield;
				string type = pSubtract->Attribute("type");
				if (type.compare("box") == 0) {
					Vector<float,d> x0, x1;
					load_vector_xml(pSubtract, "%c0", x0, "subtract/box");
					load_vector_xml(pSubtract, "%c1", x1, "subtract/box");
					remove_particles_box<d>(sim.sand.particles, sim.sand.r, x0,x1);
				} else if (type.compare("sphere") == 0) {
					Vector<float,d> c;
					float r;
					load_vector_xml(pSubtract, "c%c", c, "subtract/sphere");
					if (pSubtract->QueryFloatAttribute("r", &r) != TIXML_SUCCESS) cout << "Error reading subtract/sphere/r" << endl << yield;
					remove_particles_sphere<d>(sim.sand.particles, sim.sand.r, c,r);
				}
				pSubtract = pSubtract->NextSiblingElement("subtract");
			}
		} {
			// load rigid bodies
			TiXmlElement *pBody = pRoot->FirstChildElement("body");
			while (pBody) {
				RigidBody<d> body;
				body.dynamic = (pBody->Attribute("m") != NULL);
				body.inout = (pBody->Attribute("container") != NULL) ? -1 : 1;
				if (body.dynamic)
					load_body_xml(pBody, body);
				else {
					load_pose_xml(pBody, body.pose, "body");
					body.twist = Screw<d>(0.f);
					body.aux.dp = body.aux.ds = 0.f;
				}
				TiXmlElement *pDist = pBody->FirstChildElement("dist");
				if (!pDist) cout << "No dist" << endl << yield;
				RigidBodyInfo info;
				info.src = pDist->Attribute("src");
				if (pDist->QueryFloatAttribute("scale", &info.scale) != TIXML_SUCCESS) cout << "Error reading body/scale" << endl << yield;
				if (pDist->QueryFloatAttribute("offset", &info.offset) != TIXML_SUCCESS) cout << "Error reading body/offset" << endl << yield;
				if (!load_distance_field(info, modeldir, body.dist)) cout << "Couldn't load body " << info.src << endl << yield;
				dist_updated(body);
				if (body.dynamic) {
					sim.bt->addRigidBody(body.bt);
					bt_updated(body);
				}
				sim.bodies.push_back(body);
				sim.body_info.push_back(info);
				pBody = pBody->NextSiblingElement("body");
			}
		} {
			// load scripted actions
			TiXmlElement *pScripted = pRoot->FirstChildElement("scripted");
			while (pScripted) {
				string type = pScripted->Attribute("type");
				if (type.compare("explosion") == 0) {
					Explosion<d> *explo = new Explosion<d>;
					if (pScripted->QueryFloatAttribute("t", &explo->T) != TIXML_SUCCESS) cout << "Error reading explosion/t" << endl << yield;
					if (pScripted->QueryFloatAttribute("r", &explo->R) != TIXML_SUCCESS) cout << "Error reading explosion/r" << endl << yield;
					if (pScripted->QueryFloatAttribute("dVdt", &explo->dVdt) != TIXML_SUCCESS) cout << "Error reading explosion/dVdt" << endl << yield;
					load_vector_xml(pScripted->FirstChildElement("center"), "%c", explo->x, "explosion/center");
					explo->init(sim);
					sim.scripted.push_back(explo);
				} else if (type.compare("holdbody") == 0) {
					HoldBody<d> *hb = new HoldBody<d>;
					int b;
					if (pScripted->QueryIntAttribute("body", &b) != TIXML_SUCCESS) cout << "Error reading holdbody/b" << endl << yield;
					hb->body = &sim.bodies[b];
					if (pScripted->QueryFloatAttribute("until", &hb->T) != TIXML_SUCCESS) cout << "Error reading holdbody/b" << endl << yield;
					if (pScripted->FirstChildElement("x"))
						load_pose_xml(pScripted, hb->holdpose, "holdbody");
					else
						hb->holdpose = hb->body->pose;
					hb->init(sim);
					sim.scripted.push_back(hb);
				} else if (type.compare("source") == 0) {
					Source<d> *src = new Source<d>;
					load_vector_xml(pScripted->FirstChildElement("area"), "%c0", src->x0, "source/area");
					load_vector_xml(pScripted->FirstChildElement("area"), "%c1", src->x1, "source/area");
					load_vector_xml(pScripted->FirstChildElement("v"), "%c", src->v, "source/v");
					if (pScripted->QueryFloatAttribute("rho", &src->rho) != TIXML_SUCCESS) cout << "Error reading source/rho" << endl << yield;
					src->init(sim);
					sim.scripted.push_back(src);
				} else if (type.compare("scriptedbody") == 0) {
					ScriptedBody<d> *sb = new ScriptedBody<d>;
					int b;
					if (pScripted->QueryIntAttribute("body", &b) != TIXML_SUCCESS) cout << "Error reading scriptedbody/b" << endl << yield;
					sb->body = &sim.bodies[b];
					load_twist_xml(pScripted, sb->twist, "scriptedbody");
					sb->init(sim);
					sim.scripted.push_back(sb);
				} else cout << "Unknown scripted type " << type << endl;
				pScripted = pScripted->NextSiblingElement("scripted");
			}
		} {
			// add walls
			add_walls(pRoot->FirstChildElement("walls"), sim);
		} {
			// hacks
			memset(&sim.sand.hacks, 0, sizeof(sim.sand.hacks));
			TiXmlElement *pHacks = pRoot->FirstChildElement("hacks");
			if (pHacks) {
				if (pHacks->FirstChildElement("dont-regularize")) sim.sand.hacks.dont_regularize = true;
				if (pHacks->FirstChildElement("sparse-own-velocity")) sim.sand.hacks.sparse_own_velocity = true;
				if (pHacks->FirstChildElement("flip-not-pic")) sim.sand.hacks.flip_not_pic = true;
				if (pHacks->FirstChildElement("clamp-velocity")) pHacks->FirstChildElement("clamp-velocity")->QueryFloatAttribute("max", &sim.sand.hacks.clamp_velocity);
				if (pHacks->FirstChildElement("air")) {
					pHacks->FirstChildElement("air")->QueryFloatAttribute("drag", &sim.sand.hacks.air_drag);
					load_vector_xml(pHacks->FirstChildElement("air"), "v%c", sim.sand.hacks.air_vel, "hacks/air/v");
				}
				if (pHacks->FirstChildElement("zhu-bridson")) pHacks->FirstChildElement("zhu-bridson")->QueryFloatAttribute("cohesion", &sim.sand.hacks.cohesion);
				if (pHacks->FirstChildElement("cfl")) pHacks->FirstChildElement("cfl")->QueryFloatAttribute("factor", &sim.sand.hacks.cfl);
			}
		}
	} catch (Yield y) {return false;}
	splat(sim.sand);
	return true;
}
template bool load_setup (const string &filename, const string &modeldir, Simulator<2> &sim);
template bool load_setup (const string &filename, const string &modeldir, Simulator<3> &sim);

// ---------------- //
//  Saving  binary  //
// ---------------- //

template <size_t d> void save_obstacle_binary (const RigidBody<d> &body, fstream &file);
template <> void save_obstacle_binary (const RigidBody<2> &body, fstream &file) {
	const size_t d = 2;
	file.write((char*)body.pose.x.data(), d*4);
	file.write((char*)&body.pose.q, 4);
}
template <> void save_obstacle_binary (const RigidBody<3> &body, fstream &file) {
	const size_t d = 3;
	file.write((char*)body.pose.x.data(), d*4);
	file.write((char*)&body.pose.q.s, 4);
	file.write((char*)body.pose.q.v.data(), d*4);
}

template <size_t d> void save_debug (const Simulator<d> &sim, fstream &file);
template <> void save_debug (const Simulator<2> &sim, fstream &file) {
	file.write((char*)&sim.sand.aux.p.s[0], sim.sand.aux.p.s.size()*4);
	file.write((char*)&sim.sand.aux.s.xx.s[0], sim.sand.aux.s.xx.s.size()*4);
	file.write((char*)&sim.sand.aux.s.xy.s[0], sim.sand.aux.s.xy.s.size()*4);
	file.write((char*)&sim.sand.aux.dp.x.s[0], sim.sand.aux.dp.x.s.size()*4);
	file.write((char*)&sim.sand.aux.dp.y.s[0], sim.sand.aux.dp.y.s.size()*4);
	file.write((char*)&sim.sand.aux.ds.x.s[0], sim.sand.aux.ds.x.s.size()*4);
	file.write((char*)&sim.sand.aux.ds.y.s[0], sim.sand.aux.ds.y.s.size()*4);
}
template <> void save_debug (const Simulator<3> &sim, fstream &file) {
	// file.write((char*)&sim.sand.aux.p.s[0], sim.sand.aux.p.s.size()*4);
}

template <size_t d> bool save_state (const Simulator<d> &sim, const std::string &prefix) {
	// save sand state
	fstream file(stringf("%s.sand",prefix.c_str()).c_str(), ios::out|ios::binary);
	if (!file) return false;
	int32_t frame = (int32_t) sim.frame;
	file.write((char*)&frame, 4);
	file.write((char*)&sim.dt, 4);
	file.write((char*)&sim.simtime, 4);
	file.write((char*)sim.g.data(), d*4);
	Size<d> s = size<d>(sim.sand);
	file.write((char*)&s, d*4);
	file.write((char*)&sim.sand.dx, 4);
	file.write((char*)&sim.sand.rho_max, 4);
	file.write((char*)&sim.sand.r, 4);
	file.write((char*)&sim.sand.mu, 4);
	int32_t zb = sim.sand.zhubridson ? 1 : 0;
	file.write((char*)&zb, 4);
	int32_t np = (int32_t)sim.sand.particles.size();
	file.write((char*)&np, 4);
	for (int i = 0; i < np; i++) {
		const Particle<d> &p = sim.sand.particles[i];
		file.write((char*)p.x.data(), d*4);
		file.write((char*)p.v.data(), d*4);
		file.write((char*)&p.m, 4);
		file.write((char*)p.r2.data(), d*d*4);
	}
	int32_t nb = (int32_t)sim.bodies.size();
	file.write((char*)&nb, 4);
	for (int i = 0; i < nb; i++) {
		const RigidBody<d> &body = sim.bodies[i];
		int32_t dyn = body.dynamic ? 1 : 0;
		file.write((char*)&dyn, 4);
		file.write((char*)&body.inout, 4);
		if (!body.dynamic) {
			save_obstacle_binary(body, file);
		}
		int32_t len = (int32_t)sim.body_info[i].src.length();
		file.write((char*)&len, 4);
		file.write(sim.body_info[i].src.c_str(), len);
		file.write((char*)&sim.body_info[i].scale, 4);
		file.write((char*)&sim.body_info[i].offset, 4);
	}
	int32_t nsm = (int32_t)sim.sand.smmap.size();
	file.write((char*)&nsm, 4);
	for (int i = 0; i < nsm; i++) {
		const SplitMerge &sm = sim.sand.smmap[i];
		file.write((char*)&sm, sizeof(SplitMerge));
	}
	Sand<3> &sand = (Sand<3> &)sim.sand;
	sand.smmap.clear();
	save_debug(sim, file);
	// save Bullet state
	const int maxSerializeBufferSize = 5*1024*1024; // 5 MB?
	btDefaultSerializer *serializer = new btDefaultSerializer(maxSerializeBufferSize);
	sim.bt->serialize(serializer);
	fstream bullet(stringf("%s.bullet",prefix.c_str()).c_str(), ios::out|ios::binary);
	bullet.write((char*)serializer->getBufferPointer(), serializer->getCurrentBufferSize());
	delete serializer;
	return (file && bullet);
}
template bool save_state (const Simulator<2> &sim, const std::string &prefix);
template bool save_state (const Simulator<3> &sim, const std::string &prefix);

// ---------------- //
//  Loading binary  //
// ---------------- //

template <size_t d> void load_obstacle_binary (fstream &file, RigidBody<d> &body);
template <> void load_obstacle_binary (fstream &file, RigidBody<2> &body) {
	const size_t d = 2;
	file.read((char*)body.pose.x.data(), d*4);
	file.read((char*)&body.pose.q, 4);
}
template <> void load_obstacle_binary (fstream &file, RigidBody<3> &body) {
	const size_t d = 3;
	file.read((char*)body.pose.x.data(), d*4);
	file.read((char*)&body.pose.q.s, 4);
	file.read((char*)body.pose.q.v.data(), d*4);
}

template <size_t d> void load_debug (fstream &file, Simulator<d> &sim);
template <> void load_debug (fstream &file, Simulator<2> &sim) {
	file.read((char*)&sim.sand.aux.p.s[0], sim.sand.aux.p.s.size()*4);
	file.read((char*)&sim.sand.aux.s.xx.s[0], sim.sand.aux.s.xx.s.size()*4);
	file.read((char*)&sim.sand.aux.s.xy.s[0], sim.sand.aux.s.xy.s.size()*4);
	file.read((char*)&sim.sand.aux.dp.x.s[0], sim.sand.aux.dp.x.s.size()*4);
	file.read((char*)&sim.sand.aux.dp.y.s[0], sim.sand.aux.dp.y.s.size()*4);
	file.read((char*)&sim.sand.aux.ds.x.s[0], sim.sand.aux.ds.x.s.size()*4);
	file.read((char*)&sim.sand.aux.ds.y.s[0], sim.sand.aux.ds.y.s.size()*4);
}
template <> void load_debug (fstream &file, Simulator<3> &sim) {
	// file.read((char*)&sim.sand.aux.p.s[0], sim.sand.aux.p.s.size()*4);
}

template <size_t d> bool load_state (const std::string &prefix, const std::string &modeldir, Simulator<d> &sim) {
	// load sand state
	string simfile = stringf("%s.sand",prefix.c_str());
	fstream file(simfile.c_str(), ios::in|ios::binary);
	if (!file) {
		cout << "Error: unable to open file '" << simfile << "'" << endl;
		return false;
	}
	int32_t frame;
	file.read((char*)&frame, 4);
	sim.frame = (int)frame;
	file.read((char*)&sim.dt, 4);
	file.read((char*)&sim.simtime, 4);
	file.read((char*)sim.g.data(), d*4);
	Size<d> s;
	file.read((char*)&s, d*4);
	file.read((char*)&sim.sand.dx, 4);
	file.read((char*)&sim.sand.rho_max, 4);
	file.read((char*)&sim.sand.r, 4);
	file.read((char*)&sim.sand.mu, 4);
	int32_t zb;
	file.read((char*)&zb, 4);
	sim.sand.zhubridson = (zb != 0);
	resize(sim.sand, s, sim.sand.dx);
	int32_t np;
	file.read((char*)&np, 4);
	sim.sand.particles.resize((int)np);
	for (int i = 0; i < np; i++) {
		Particle<d> &p = sim.sand.particles[i];
		file.read((char*)p.x.data(), d*4);
		file.read((char*)p.v.data(), d*4);
		file.read((char*)&p.m, 4);
		file.read((char*)p.r2.data(), d*d*4);
	}
	int32_t nb;
	file.read((char*)&nb, 4);
	sim.bodies.resize((int)nb);
	sim.body_info.resize((int)nb);
	for (int i = 0; i < nb; i++) {
		RigidBody<d> &body = sim.bodies[i];
		int32_t dyn;
		file.read((char*)&dyn, 4);
		body.dynamic = (dyn != 0);
		file.read((char*)&body.inout, 4);
		if (!body.dynamic) {
			load_obstacle_binary(file, body);
		}
		int32_t len;
		char src[256];
		file.read((char*)&len, 4);
		file.read(src, len);
		src[len] = '\0';
		sim.body_info[i].src = src;
		file.read((char*)&sim.body_info[i].scale, 4);
		file.read((char*)&sim.body_info[i].offset, 4);
		if (body.dist.s.size() == 0) {
			if (!load_distance_field(sim.body_info[i], modeldir, body.dist)) { 
				cout << "Error: Unable to load model distance field" << endl;
				return false;
			}
			dist_updated(body);
		}
	}
	int32_t nsm;
	file.read((char*)&nsm, 4);
	sim.sand.smmap.resize((int)nsm);
	for (int i = 0; i < nsm; i++) {
		file.read((char*)&(sim.sand.smmap[i]), sizeof(SplitMerge));
	}
	load_debug(file, sim);
	splat(sim.sand);
	// load Bullet state
	static btDefaultCollisionConfiguration *cc = NULL;
	static btCollisionDispatcher *dispatcher = NULL;
	static btDbvtBroadphase *broadphase = NULL;
	static btSequentialImpulseConstraintSolver *solver = NULL;
	if (sim.bt) delete sim.bt;
	if (dispatcher) delete dispatcher;
	if (broadphase) delete broadphase;
	if (solver) delete solver;
	if (cc) delete cc;
	cc = new btDefaultCollisionConfiguration;
	dispatcher = new btCollisionDispatcher(cc);
	broadphase = new btDbvtBroadphase;
	solver = new btSequentialImpulseConstraintSolver;
	sim.bt = new btDiscreteDynamicsWorld(dispatcher, broadphase, solver, cc);
	sim.bt->setGravity(btVector3(0,0,0));
	btBulletWorldImporter *importer = new btBulletWorldImporter(sim.bt);
	string btfile = stringf("%s.bullet",prefix.c_str());
	if (!importer->loadFile(btfile.c_str())) {
		cout << "Error: unable to load file '" << btfile << "'" << endl;
		return false;
	}
	int nc = sim.bt->getCollisionObjectArray().size();
	for (int b = 0, c = 0; b < nb && c < nc; b++) {
		RigidBody<d> &body = sim.bodies[b];
		if (body.dynamic) {
			body.bt = (btRigidBody*)sim.bt->getCollisionObjectArray()[c];
			body.bt->setCollisionFlags(body.bt->getCollisionFlags() & (~btCollisionObject::CF_STATIC_OBJECT));
			bt_updated(body);
			c++;
		}
	}
	delete importer;
	return (bool)file;
}
template bool load_state (const std::string &prefix, const std::string &modeldir, Simulator<2> &sim);
template bool load_state (const std::string &prefix, const std::string &modeldir, Simulator<3> &sim);

// ------------------ //
//  Load/save scalars //
// ------------------ //

template <size_t d> bool load_scalar_field (const string &filename, ScalarField<d> &dist) {
	fstream file(filename.c_str(), ios::in|ios::binary);
	if (!file) return false;
	Size<d> s;
	file.read((char*)&s, d*4);
	file.read((char*)&dist.dx, 4);
	resize(dist, s, dist.dx);
	file.read((char*)&dist.s[0], dist.s.size()*4);
	return (bool)file;
}
template bool load_scalar_field (const string &filename, ScalarField<2> &dist);
template bool load_scalar_field (const string &filename, ScalarField<3> &dist);

template <size_t d> bool save_scalar_field (const ScalarField<d> &dist, const string &filename) {
	fstream file(filename.c_str(), ios::out|ios::binary);
	if (!file) return false;
	Size<d> s = size<d>(dist);
	file.write((char*)&s, d*4);
	file.write((char*)&dist.dx, 4);
	file.write((char*)&dist.s[0], dist.s.size()*4);
	return (bool)file;
}
template bool save_scalar_field (const ScalarField<2> &dist, const string &filename);
template bool save_scalar_field (const ScalarField<3> &dist, const string &filename);
