/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <iostream>
#include <fstream>
#include "../common.hpp"
#include "../sim/simulator.hpp"
#include "../model/primitive.hpp"
#include "../model/voxelizer.hpp"
#include "../model/distance.hpp"
#include "io.hpp"
#include "iohelpers.hpp"
#include "renderer.hpp"

using namespace std;

template <size_t d> int main_simulate (vector<char*> args) {
	Simulator<d> sim;
	/// Just a safety, it should be automatically initialized to 0
	sim.simtime = 0;

	string modeldir = get_opt(args, "--models");
	if (modeldir.empty()) modeldir = "models";
	{
		char final = modeldir[modeldir.length()-1];
		if (final != '/' && final != '\\')
			modeldir.push_back('/');
	}

	string setupfile = get_opt(args, "--setup");
	string outfile = get_opt(args, "--out");
	int inframe = get_opti(args, "--frame");

	if (outfile.empty()) cout << "No output file template (--out, with %d)" << endl << yield;

	if (!setupfile.empty()) {
		bool success = load_setup(setupfile, modeldir, sim);
		if (!success) cout << "Error loading setup " << setupfile << " (with models from " << modeldir << ")" << endl << yield;
		sim.frame = 0;
	}
	else if (inframe >= 0) {
		sim.bt = NULL;
		bool success = load_state(stringf(outfile, inframe), modeldir, sim);
		if (!success) cout << "Error loading state " << stringf(outfile,inframe) << " (with models from " << modeldir << ")" << endl << yield;
		sim.frame = inframe;
	}
	else cout << "Neither setup file (--setup) nor initial frame (--frame)" << endl << yield;

	float t1 = get_optf(args, "--t1");
	if (isnanl(t1)) {
		cout << "No end time (--t1), assuming infinity" << endl;
		t1 = infinity;
	}

	do {
		save_state(sim, stringf(outfile, sim.frame));
		sim.sand.smmap.clear();
		step(sim);
	} while (sim.simtime < t1);

	return 0;
}
#if 0
template <size_t d> int main_render (vector<char*> args) {
	Renderer<d> ren;

	string modeldir = get_opt(args, "--models");
	if (modeldir.empty()) modeldir = "models";
	{
		char final = modeldir[modeldir.length()-1];
		if (final != '/' && final != '\\')
			modeldir.push_back('/');
	}
	ren.modeldir = modeldir;

	ren.infile = get_opt(args, "--in");
	if (ren.infile.empty()) cout << "No input file template (--in, with %d)" << endl << yield;
	
	ren.frame = get_opti(args, "--frame");
	if (ren.frame < 0) {
		cout << "No start frame (--frame), assuming 0" << endl;
		ren.frame = 0;
	}
	
	render(ren);

	return 0;
}
#endif
int main_prep_model (vector<char*> args) {
	string infile = get_opt(args, "--in");
	string outfile = get_opt(args, "--out");
	string rasterfile = get_opt(args, "--raster");

	bool rot = get_optb(args, "--rotate-z");

	if (infile.empty()) cout << "No input file (--in)" << endl << yield;
	if (outfile.empty()) cout << "No output file (--out)" << endl << yield;

	ScalarField<3> voxels;
	load_voxels(infile, voxels);
	if (!rasterfile.empty()) save_scalar_field(voxels, rasterfile);
	cout << "Loaded " << voxels.m << "x" << voxels.n << "x" << voxels.o << " voxels" << endl;
	ScalarField<3> dist;
	dist.resize(voxels.m,voxels.n,voxels.o, voxels.dx);
	compute_distance_field(voxels, dist);
	cout << "Distance field computed" << endl;

	if (rot) {
		int m = dist.m, n = dist.n, o = dist.o;
		cout << m << " " << n << " " << o << endl;
		ScalarField<3> dist_new;
		dist_new.resize(m,n,o, dist.dx);
		for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
			dist_new(i,j,k) = min(min(dist(i,j,k), dist(m-i-1,n-j-1,k)), min(dist(j,i,k), dist(n-j-1,m-i-1,k)));
		}
		dist_new.s.swap(dist.s);
	}

	save_scalar_field(dist, outfile);
	
	return 0;
}

template <size_t d> inline Vector<float,d> get_optv (const std::vector<char*> &args, const std::string &opt) {
	int nargs = (int) args.size();
	for (int i = 0; i < nargs; i++)
		if (opt.compare(args[i]) == 0) {
			Vector<float,d> v;
			for (int j = 0; j < d; j++) v[j] = atof(args[i+1+j]);
			return v;
		}
	return Vector<float,d>(not_a_number);
}

template <size_t d> int main_primitive (vector<char*> args) {
	string type = get_opt(args, "--type");
	string outfile = get_opt(args, "--out");

	if (type.empty()) cout << "No primitive type (--type)" << endl << yield;
	if (outfile.empty()) cout << "No output file (--out)" << endl << yield;

	float dx = get_optf(args, "--dx");
	if (isnanl(dx)) cout << "No grid size (--dx)" << endl << yield;
	ScalarField<d> dist;
	if (type.compare("sphere") == 0) {
		float r = get_optf(args, "--r");
		if (isnanl(r)) cout << "No radius (--r)" << endl << yield;
		sphere_dist(r,dx, dist);
	} else if (type.compare("box") == 0) {
		Vector<float,d> size = get_optv<d>(args, "--size");
		if (isnanl(size[0])) cout << "No size (--size)" << endl << yield;
		box_dist(size,dx, dist);
	}
	save_scalar_field(dist, outfile);
	return 0;
}
#if 0
int main_renderman(vector<char*> args) {
	bool stopcout = get_optb(args, "--nocout");
	string snormals = get_opt(args, "--normals");
	NormalRenderingMode normals;
	if (snormals == "none") normals = NONE;
	else if (snormals == "nojitter") normals = NOJITTER;
	else normals = FULL;
	bool wireframe = get_optb(args, "--wireframe");
	streambuf *psbuf, *backup, *backup2, *psbuf2;
	ofstream filestr, filestr2;

	Renderer<3> ren;
	string modeldir = get_opt(args, "--models");
	if (modeldir.empty()) modeldir = "models";
	{
		char final = modeldir[modeldir.length()-1];
		if (final != '/' && final != '\\')
			modeldir.push_back('/');
	}
	ren.modeldir = modeldir;

	if (stopcout) {
		filestr.open("cout.txt");
		backup=cout.rdbuf();
		psbuf=filestr.rdbuf();
		cout.rdbuf(psbuf);

		/*filestr2.open("cerr.txt");
		backup2=cerr.rdbuf();
		psbuf2=filestr2.rdbuf();
		cerr.rdbuf(psbuf2);*/
	}
	ren.infile = get_opt(args, "--in");
	if (ren.infile.empty()) cout << "No input file template (--in, with %d)" << endl << yield;
	string outfile = get_opt(args, "--out");
	if (outfile.empty()) cout << "No output file template (--out with %d)" << endl << yield;
	//string initfile = get_opt(args, "--init");
	//string worldfile = get_opt(args, "--world");

	string setupfile = get_opt(args, "--setup");
	if (setupfile.empty()) cout << "No setup file template (--setup)" << endl << yield;
	ren.frame = get_opti(args, "--frame");
	int fps = get_opti(args, "--fps");
	if (fps <= 0) {
		cout << "No target frame rate specified, assuming 30" << endl;
		fps = 30;
	}
	if (ren.frame < 0) {
		cout << "No start frame (--frame), assuming 0" << endl;
		ren.frame = 0;
	}

	bool useglut = get_optb(args, "--useglut");
	int lastframe = get_opti(args, "--lastframe");
	if (lastframe <= 0) lastframe = 1000;

	render3r(ren, outfile, setupfile, useglut, (int)normals, wireframe, fps, lastframe);

	if (stopcout) {
		cout.rdbuf(backup);
		//cerr.rdbuf(backup2);
		filestr.close();
		//filestr2.close();
	}
	return 0;
}
#endif
ostream &operator<< (ostream &out, const ScalarField<2> &field) {
	int p = out.precision(4);
	for (int y = 0; y < field.n; y++) {
		for (int x = 0; x < field.m; x++) {
			out << "\t" << field(x,y);
		}
		out << endl;
	}
	out.precision(p);
	return out;
}

template <size_t d> Matrix<float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2, Vector<float,d> &eigenVals);
template <size_t d> Matrix<float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2, Vector<float,d> &eigenVals, Vector<float,d> eigenVecs[d]);
int main_debug (vector<char*> args) {
	Mat2f m;
	m = 1, 0.5,
		0.5, 1;
	Vec2f eval;
	decomposeMatrix(m, eval);
	cout << eval << endl;
	Vec2f evec[2];
	decomposeMatrix(m, eval, evec);
	cout << eval << endl;
	cout << evec[0] << ", " << evec[1] << endl;
	return 0;
}

int main (int argc, char **argv) {
	// srand(time(NULL));
	vector<char*> args(argv, &argv[argc]);
	int d;
	if (get_optb(args, "--prep-model"))
		return main_prep_model(args);
	else if (get_optb(args, "--debug"))
		return main_debug(args);
	if (get_optb(args, "--2d"))
		d = 2;
	else if (get_optb(args, "--3d"))
		d = 3;
	else cout << "Neither --2d nor --3d specified" << endl << yield;
	if (get_optb(args, "--simulate"))
		return d==2 ? main_simulate<2>(args) : main_simulate<3>(args);
	// else if (get_optb(args, "--render"))
	// 	return d==2 ? main_render<2>(args) : main_render<3>(args);
	else if (get_optb(args, "--primitive"))
		return d==2 ? main_primitive<2>(args) : main_primitive<3>(args);
	// else if (get_optb(args, "--renderman")) {
	// 	if (d == 2) cout << "Renderman rendering only allowed for 3D fields" << endl << yield;
	// 	else return main_renderman(args);
	// }
	else cout << "Neither --simulate, --render, --prep-model, --primitive, --renderman, nor --debug specified" << endl << yield;
}
