/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <GL/glut.h>
#include <ri.h>
#include "io.hpp"
#include "iohelpers.hpp"
#include <cstring>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include "../model/distance.hpp"
#include "render3r.hpp"
//#include "../model/scenegraph.hpp"
using namespace std;

#ifndef __GNUC__
using namespace stdext;
#endif

#define FRAND ((float)rand()/RAND_MAX)
#define FSRAND (2*FRAND - 1)

RendermanRenderer rren;
Renderer<3> *ren;
void glutrender();
void convert(float pos[3], int ipos[3], BucketField<3> &bf);
void convert(Vec3f pos, int ipos[3], BucketField<3> &bf);
int spillcount = 0;
bool renderSimParticles = false;
static int dl_id = -1;

/// Remove specified delimiter from the beginning and end of a string
void removeDelims(string &s, char delim) {
	if (s[0] == delim) s = s.substr(1);
	if (s[s.size()-1] == delim) s = s.substr(0, s.size() - 1);
}

inline Vec2f rotate (float a, const Vec2f &v) {return Vec2f(v[0]*cos(a) - v[1]*sin(a), v[0]*sin(a) + v[1]*cos(a));}
inline Vec2f irotate (float a, const Vec2f &v) {return Vec2f(v[0]*cos(a) + v[1]*sin(a), -v[0]*sin(a) + v[1]*cos(a));}
inline Vec3f rotate (const Quatf &q, const Vec3f &v) {return (q*Quatf(0,v)*inv(q)).v;}
inline Vec3f irotate (const Quatf &q, const Vec3f &v) {return (inv(q)*Quatf(0,v)*q).v;}

template <size_t d> inline Vector<float,d> world_to_body (const RigidBody<d> &body, const Vector<float,d> &x) {
	return irotate(body.pose.q, Vector<float,d>(x - body.pose.x));
}
template <size_t d> inline Vector<float,d> body_to_world (const RigidBody<d> &body, const Vector<float,d> &x) {
	return Vector<float,d>(rotate(body.pose.q, x) + body.pose.x);
}

Vec3f clamp_to_grid2 (const Vec3f &x, const ScalarField<3> &field) {
	int m = field.m, n = field.n, o = field.o;
	float dx = field.dx;
	return Vec3f(clampf(x[0], 0.5f*dx,(m-0.5f)*dx), clampf(x[1], 0.5f*dx,(n-0.5f)*dx), clampf(x[2], 0.5f*dx,(o-0.5f)*dx));
}

/// Project any render particles intersecting a rigid body out of it
void project_particles (const RigidBody<3> &body, vector< RenderParticle > &ps) {
	int np = ps.size();
	int ncount = 0;
	/// Each particle does not need to be processed sequentially for projection
#pragma omp parallel for
	for (int i = 0; i < np; i++) {
		RenderParticle &p = ps[i];
		Vec3f pos(ps[i].pos[0], ps[i].pos[1], ps[i].pos[2]);
		Vec3f p_body(world_to_body(body, pos) + body.origin);
		Vec3f dp_grid(p_body - clamp_to_grid2(p_body, body.dist));
		float dist_grid = lerp(p_body, body.dist);
		float dist_out = tvmet::norm2(dp_grid)/body.dist.dx;
		float dist = (dist_grid + dist_out)*body.inout;
		if (dist > 0) continue;
		//Vec3f n(gradient(p_body, body.dist)*body.dist.dx);
		Vec3f n(body.inout*gradient(p_body, body.dist)*body.dist.dx);
		if (dist_out > 0)
			n = normalize(dist_grid*n + dp_grid);
		else n = normalize(n);
		//Vec3f p_body_new(p_body - dist*n/tvmet::norm2(n));
		Vec3f p_body_new(p_body - 2*dist*body.dist.dx*n/dot(n,n));
		pos = body_to_world(body, Vec3f(p_body_new - body.origin));
		ps[i].pos[0] = pos[0];  ps[i].pos[1] = pos[1];  ps[i].pos[2] = pos[2];
		
		/// Verify the result
		p_body = Vec3f(world_to_body(body, pos) + body.origin);
		dist = lerp(p_body, body.dist);
#pragma omp atomic
		ncount += (dist < 0);
		//assert (dist > 0);
	}
	cout << "Project particle count: " << ncount << " of " << np << endl;
}

/// Tokenize a string about a default whitespace delimiter
void tokenize(string str, vector<string> &tokens) {
	stringstream ss(str);
	string buf;
	while(ss >> buf)
		tokens.push_back(buf);
}

/// Normalize a normal
bool normalize(RenderNormal &n) {
	float magn = sqrtf(sqf(n.pos[0]) + sqf(n.pos[1]) + sqf(n.pos[2]));
	bool valid = magn > 0.001f;
	for ( int i = 0; i < 3; i++ ) {
		n.pos[i] = (valid) ? n.pos[i] / magn : n.pos[i];
	}
	return valid;
}

bool normalize(Vec3f &n) {
	float magn = tvmet::norm2(n);
	bool valid = magn > epsilon;
	n = n / ( (valid) ? valid : valid + epsilon );
	return valid;
}

void RendermanRenderer::load_setup(const std::string &filename) {
	load_renderman_setup(filename, *this);
	sgraph.load(filename);
	wireframe = false;
	fps = 30;
}

RendermanRenderer::RendermanRenderer() {
	srand((unsigned int)time(NULL));
	tolerance = 1.2f;
	render_width = 0.01f;
	nJitter = 0.1f;
	rendertime = 0;
	distfield_supersampling = 1;
	fill_fraction = 0.7f;
	imageres[0] = 800; imageres[1] = 600;
	shadowmapres = 1024;
	shmapfov = 45;
	shmap_temp = "temp.depth";
	sandcolor[0] = 0.99f; sandcolor[1] = 0.9f; sandcolor[2] = 0.65f;
	sandopacity[0] = sandopacity[1] = sandopacity[2] = 0.7f;
	minRP = 0;
}

/// Cube root of a number
inline float cbrt(float x) {
	return powf(x, 1.0f/3);
}

/// Generates renderable points for a simulation particle
inline RenderParticle generate_render_particle(Vec3f eigenVals, Vec3f eigenVecs[3], Vec3f center) {
	//cout << "Test random numbers: " << FSRAND << ", " << FSRAND << ", " << FSRAND << endl;
	Vec3f renderPoint;

	/**
	 * New: Proper ellipsoid sampling
	 * From: http://en.wikipedia.org/wiki/Ellipsoid
	 */
	//bool placed = false;
	//int tries = 0;
	//while (!placed && tries < 1000) {
	float phi = FRAND * M_PI, theta = FRAND * 2 * M_PI, r = cbrt(FRAND);
	renderPoint = center + eigenVecs[0] * r * eigenVals[0] * sinf(phi) * cosf(theta) + eigenVecs[1] * r * eigenVals[1] * sinf(phi) * sinf(theta) + eigenVecs[2] * r * eigenVals[2] * cosf(phi);
		//renderPoint = center + eigenVecs[0] * eigenVals[0] * FSRAND + eigenVecs[1] * eigenVals[1] * FSRAND + eigenVecs[2] * eigenVals[2] * FSRAND ;
		/// TODO: do an intersection check and make sure that the distribution of points is non-intersecting

	//	placed = true;
	//	tries++;
	//}
	RenderParticle rp;
	rp.pos[0] = renderPoint[0]; rp.pos[1] = renderPoint[1]; rp.pos[2] = renderPoint[2];
	return rp;
}

/// Check whether a particle is in range of an ellipsoid
bool checkInRange(const Vec3f &relpos, const Mat3f& m, const Vec3f& eigenVals) {
	Vec3f teg(1/eigenVals[0],1/eigenVals[1],1/eigenVals[2]);
	Mat3f m2 = Mat3f(m * diagonal(teg) * tvmet::trans(m));

	return tvmet::dot(relpos, (m2 * relpos)) <= 1;
}

bool checkInRange(const float pos[3], const Vec3f& center, const Mat3f& m, const Vec3f& eigenVals) {
	Vec3f vp;
	vp[0] = pos[0]; vp[1] = pos[1]; vp[2] = pos[2];
	vp = vp - center;
	return checkInRange(vp, m, eigenVals);
}

/// Funcrion to check whether a generated render particle intersects with other render particles of the ellipsoid
bool RendermanRenderer::isNotIntersecting(RenderParticle &p, float rad, vector<int> &rpIndices) {
	float diameter2 = rad * rad * 4;
	for (int i = 0; i < (int)rpIndices.size(); i++) {
		RenderParticle &p2 = render_particles[rpIndices[i]];
		float dist2 = sqf(p.pos[0] - p2.pos[0]) + sqf(p.pos[1] - p2.pos[1]) + sqf(p.pos[2] - p2.pos[2]);
		if (dist2 < diameter2) return false;
	}
	return true;
}

/// Generate a new render particle, and a normal jitter set, also try to reuse 
RenderSet RendermanRenderer::generate_render_particles(Renderer<3> &ren, int sim_particle, Vec3f eigenVals, Vec3f eigenVecs[3], Mat3f evMat, Vec3f center, float rad, vector<RenderSet> &spill, bool usetolerance) {
	bool placed = false;
	/// Square of the diameter of render particle radius
	float rads = rad * rad * 4;
	RenderSet s;
	RenderParticle &p = s.first;
	RenderNormalJitter &jitter = s.second;
	int tries = 0;
	/// List of render particles we need to collision check against
	vector<int> &plist = sim_particles[sim_particle].renderParticleIndices;
	int n = (int)plist.size();
	int ipos[3];
	/// Get integer position of particle bucket we need to consider
	convert(center, ipos, render_particle_buckets);
	vector<int> &refs = render_particle_buckets(ipos);
	int m = (int)refs.size(), im = 0;
	Vec3f teigenVals(eigenVals * tolerance);

	while (!placed && tries < 10) {
		while ( im < m && ( refs[im] == -1 || !checkInRange(spill[refs[im]].first.pos, center, evMat, eigenVals) ) ) im++;
		if ( im < m && refs[im] != -1) {
			//p = spill[refs[im]];
			//jitter = njitters[refs[im]];
			s = spill[refs[im]];
			tries--;
		}
		else {
			if (usetolerance)
				p = generate_render_particle(teigenVals, eigenVecs, center);
			else
				p = generate_render_particle(eigenVals, eigenVecs, center);
			jitter = make_pair(nJitter * FRAND, (float)M_PI * FSRAND);
		}

		bool allpass = true;
		for ( int i = 0; i < n; i++ ) {
			RenderParticle &cp = render_particles[plist[i]];
			allpass = allpass && ( sqf(p.pos[0] - cp.pos[0]) + sqf(p.pos[1] - cp.pos[1]) + sqf(p.pos[2] - cp.pos[2]) > rads );
		}
		placed = placed || allpass;
		if ( im < m && placed) {
			/// If stuff is erased from spill, then the ref numbers in the buckets will become invalid, so DON'T tinker with them, they'll be automatically erased at the end of the frame
			//spill.erase(spill.begin() + refs[im]);
			//jitter = njitters[refs[im]];
			//refs.erase(refs.begin() + im);
			refs[im] = -1;
			spillcount++;
		}
		im++;
		tries++;
	}
	/*if (!checkInRange(p.pos, center, evMat, eigenVals) && p.pos[1] > 0.5) {
		cout << "Debug: generated particle (" << p.pos[0] << ", " << p.pos[1] << ", " << p.pos[2] << ") is outside the range of particle " << center << endl;
		cout << "pos: (" << p.pos[0] << ", " << p.pos[1] << ", " << p.pos[2] << ") center: " << center << " m: " << evMat << " evals: " << eigenVals << endl;
	}*/

	return s;
}

/// Get particles to stay inside the simulation grid
inline Vec3f reflectf (Vec3f &v, Sand<3> &sand) {
	return Vec3f(reflectf(v[0], 0,sand.m*sand.dx), reflectf(v[1], 0,sand.n*sand.dx), reflectf(v[2], 0,sand.o*sand.dx));
}

inline Vec3f clampf (Vec3f &v, Sand<3> &sand) {
	return Vec3f(clampf(v[0], 0,sand.m*sand.dx), clampf(v[1], 0,sand.n*sand.dx), clampf(v[2], 0,sand.o*sand.dx));
}

inline bool checkP(Vec3f &v, Sand<3> &sand) {
	bool retval = true;
	retval = retval && (v[0] >= 0) && (v[0] <= sand.m * sand.dx);
	retval = retval && (v[1] >= 0) && (v[1] <= sand.n * sand.dx);
	retval = retval && (v[2] >= 0) && (v[2] <= sand.o * sand.dx);
	return retval;
}

/// Find relevant bucket to look it for coordinate position
void convert(float pos[3], int ipos[3], BucketField<3> &bf) {
	for ( int coord = 0; coord < 3; coord++ ) {
		ipos[coord] = (int)(pos[coord] / bf.dx - 0.5f);
	}
	bf.bclamp(ipos);

	ipos[0] = (ipos[0] >= bf.m - 1) ? bf.m - 2 : ipos[0];
	ipos[1] = (ipos[1] >= bf.n - 1) ? bf.n - 2 : ipos[1];
	ipos[2] = (ipos[2] >= bf.o - 1) ? bf.o - 2 : ipos[2];
}

void convert(Vec3f pos, int ipos[3], BucketField<3> &bf) {
	for ( int coord = 0; coord < 3; coord++ ) {
		ipos[coord] = (int)(pos[coord] / bf.dx - 0.5f);
	}
	bf.bclamp(ipos);

	//ipos[0] = (ipos[0] >= bf.m - 1) ? bf.m - 2 : ipos[0];
	//ipos[1] = (ipos[1] >= bf.n - 1) ? bf.n - 2 : ipos[1];
	//ipos[2] = (ipos[2] >= bf.o - 1) ? bf.o - 2 : ipos[2];
}

/// Instead of looking in 8 buckets, just insert each particle into 8 buckets
void insertRPintoBucket(int rp, BucketField<3> &bf, float pos[3]) {
	int x[3];
	convert(pos, x, bf);

	int y[3];
	for ( int i = 0; i < 2; i++ ) {
		y[0] = x[0] + i;
		for ( int j = 0; j < 2; j++ ) {
			y[1] = x[1] + j;
			for ( int k = 0; k < 2; k++ ) {
				y[2] = x[2] + k;
				bf(y).push_back(rp);
			}
		}
	}
}

bool RendermanRenderer::consistency_check(Renderer<3> &ren, bool enable_mesgs, bool fix_errors) {
	/// First lets see if all the simulation particles have a match
	bool retval = true, warn = true;
	retval = retval && (sim_particles.size() == ren.sim.sand.particles.size());
	if (!retval) {
		if (enable_mesgs)
			cout << "Error: Simulation particles size mismatch, loaded " << ren.sim.sand.particles.size() << " expected " << sim_particles.size() << " particles" << endl;
		return false;
	}

	/// Now we check that every render particle that's expected by sim_particles
	int n = (int)sim_particles.size(), m = render_particles.size();
	for (int i = 0; i < n; i++) {
		SimParticle &s = sim_particles[i];
		for (int j = 0; j < (int)s.renderParticleIndices.size(); j++) {
			int rp = s.renderParticleIndices[j];
			retval = retval && (rp < m);
			bool check = render_sim_map[rp] == i;
			warn = warn && check;
			if (enable_mesgs && !check) {
				cout << "Warning: Inverse mapping for rp = " << rp << " is incorrect, should point to " << i << " instead, points to " << render_sim_map[rp] << endl;
			}
			if (fix_errors)
				render_sim_map[rp] = i;
		}
		if (!retval) {
			if (enable_mesgs) {
				cout << "Error: Simulation side expected particles which do not exist" << endl << "RP list (" << m << ": ";
				for (int j = 0; j < (int)s.renderParticleIndices.size(); j++) {
					int rp = s.renderParticleIndices[j];
					cout << rp << ", ";
				}
				cout << endl;
			}
			return false;
		}
	}

	/// Now that all that's done, check that each render particle has a parent
	for (int i = 0; i < m; i++) {
		int rp = i;
		RenderParticle &p = render_particles[rp];
		bool check = render_sim_map[rp] < n;
		bool validity = p.pos[0] == p.pos[0] && p.pos[1] == p.pos[1] && p.pos[2] == p.pos[2];
		warn = warn && check;
		retval = retval && validity;
		if (enable_mesgs && !check) {
			cout << "Warning: Inverse mapping for rp = " << rp << " is incorrect, maximum possible is " << n <<" points to " << render_sim_map[rp] << endl;
		}
		if (enable_mesgs && !validity) {
			cout << "Error: Particle positions are Nan! for rp = " << rp << endl;
			p.pos[0] = p.pos[1] = p.pos[2] = 0;
		}
		if (fix_errors) {
			/// Just fix it to a random value
			render_sim_map[rp] = 0;
		}
	}
	for (int i = 0; i < (int)render_normals.size(); i++) {
		RenderNormal &p = render_normals[i];
		bool validity = p.pos[0] == p.pos[0] && p.pos[1] == p.pos[1] && p.pos[2] == p.pos[2];
		if (enable_mesgs && !validity) {
			cout << "Error: Particle normals are Nan! for rp = " << i << endl;
			p.pos[0] = p.pos[1] = p.pos[2] = 0;
		}
	}
	if (!warn)
		cout << "Warnings present (with render_sim_map), enable debug messages to see them" << endl;
	return retval;
}

/// Vector form for isnan
int isnan(Vec3f &v) {
	return isnanl<float>(v[0]) || isnanl<float>(v[1]) || isnanl<float>(v[2]);
}

void RendermanRenderer::move_points(Renderer<3> &ren) {
	if (ren.sim.sand.m <= 0 || ren.sim.sand.n <= 0 || ren.sim.sand.o <=0 || ren.sim.sand.dx <= 0) {
		cout << "Error loading simulation data, one of m,n,o,dx is 0/negative!" << ren.sim.sand.m << " " << ren.sim.sand.n << " " << ren.sim.sand.o << " " << ren.sim.sand.dx << endl;
	}
	render_particle_buckets.resize(ren.sim.sand.m, ren.sim.sand.n, ren.sim.sand.o, ren.sim.sand.dx);
	build_distance_field(ren);

	clear_all(render_particle_buckets);
	vector <RenderSet> spill;
	gradient(distfield, grad);
	//gradient(ren.sim.sand.rho, grad);
	/// Apply split/merge mappings
	Sand<3> &sand = ren.sim.sand;
	float dx = sand.dx;
	Vec3f eigenVals, eigenVecs[3], teigenVals, eigenVals2, eigenVecs2[3];
	Mat3f evMatrix, evMatrix2;
	int i, j;
	int n = (int)sand.particles.size();
	int on = (int)sim_particles.size();
	cout << "Total render points: " << render_particles.size() << endl;
	int nRegen = 0, nNewgen = 0;
	float dt = ren.sim.simtime - rendertime;
	cout << "Simulation time updated from " << rendertime << " to " << ren.sim.simtime << endl;	
	rendertime = ren.sim.simtime;

	spillcount = 0;

	
	/*if (!consistency_check(ren, true, false)) {
		cout << "Point set consistency check failed before applying rigid" << endl;
	}
	else {
		cout << "Point set consistency check passed before applying rigid" << endl;
	}*/
	
	/*for (int i = 0; i < (int)render_particles.size(); i++) {
		RenderParticle &rp = render_particles[i];
		Vec3f actp(rp.pos[0], rp.pos[1], rp.pos[2]);
		if ( !checkP(actp, sand) ) {
			//cout << "Advection pushing particles out of bounds" << endl;
			cout << "Before was: actual position: " << actp << endl;
			//cout << "Limits: " << sand.m << ", " << sand.n << ", " << sand.o << ", " << sand.dx << endl;
			actp = clampf(actp, sand);
			cout << "Post clamping: " << actp << endl;
			rp.pos[0] = actp[0]; rp.pos[1] = actp[1]; rp.pos[2] = actp[2]; 
			//cout << "p: " << p.x << " actp: " << actp << endl;
		}
	}*/
	/// Push out all particles from the rigid bodies
	for (int i = 0; i < (int)ren.sim.bodies.size(); i++) {
		RigidBody<3> &body = ren.sim.bodies[i];
		project_particles(body, render_particles);
	}
	/*for (int i = 0; i < (int)render_particles.size(); i++) {
		RenderParticle &rp = render_particles[i];
		Vec3f actp(rp.pos[0], rp.pos[1], rp.pos[2]);
		if ( !checkP(actp, sand) ) {
			//cout << "Advection pushing particles out of bounds" << endl;
			cout << "PRGBP Before was: actual position: " << actp << endl;
			//cout << "Limits: " << sand.m << ", " << sand.n << ", " << sand.o << ", " << sand.dx << endl;
			actp = clampf(actp, sand);
			cout << "Post clamping: " << actp << endl;
			rp.pos[0] = actp[0]; rp.pos[1] = actp[1]; rp.pos[2] = actp[2]; 
			//cout << "p: " << p.x << " actp: " << actp << endl;
		}
	}*/
	
#pragma omp parallel for
	/// Move all render particles first
	for (int i = 0; i < (int)render_particles.size(); i++) {
		RenderParticle &rp = render_particles[i];
		RenderNormalJitter &nj = normal_jitters[i];
		Vec3f actp;
		actp[0] = rp.pos[0]; actp[1] = rp.pos[1]; actp[2] = rp.pos[2];
		Vec3f dx_sep = lerp_weighted(actp, sand.aux.dx_sep, sand.rho_s);
		Vec3f vel_perturb = lerp_weighted(actp, sand.v, sand.rho_s);
		if (isnan(dx_sep) || isnan(vel_perturb) || isnan(actp)) {
			cout << "Nan: dx_sep: " << isnan(dx_sep) << " vel_perturb: " << isnan(vel_perturb) << " actual position: " << isnan(actp) << endl;
			cout << "Data: dx_sep: " << dx_sep << " vel_perturb: " << vel_perturb << " actual position: " << actp << endl;
		}
		float vmod = tvmet::norm2(vel_perturb);
		nj.first += vmod * FSRAND * 0.001;
		nj.second += vmod * FSRAND * 0.001;
		nj.first = (nj.first < 0) ? nJitter + nj.first : nj.first;
		nj.first = (nj.first > nJitter) ? -2 * nJitter + nj.first : nj.first;
		nj.second = (nj.second > (float)M_PI) ? nj.second - M_PI : nj.second;
		nj.second = (nj.second < -(float)M_PI) ? nj.second + M_PI : nj.second;
		for ( int it = 0; it < 3; it++ )
			vel_perturb[it] = vel_perturb[it] * (1 + 0.01 * FSRAND);
		Vec3f befp(actp);
		actp += vel_perturb*dt + dx_sep;
		bool before = checkP(actp, sand);
		actp = reflectf(actp, sand);
		if ( !checkP(actp, sand) ) {
			cout << "Advection pushing particles out of bounds" << endl;
			cout << "velocity: " << vel_perturb << " dx_sep: " << dx_sep << " dt: " << dt << endl;
			cout << "Before was: " << before << " actual position: " << actp << " moved here from: " << befp << endl;
			cout << "Limits: " << sand.m*sand.dx << ", " << sand.n*sand.dx << ", " << sand.o*sand.dx << endl;
		}
		else {
			rp.pos[0] = actp[0]; rp.pos[1] = actp[1]; rp.pos[2] = actp[2];
		}
		//render_sim_map[s.renderParticleIndices[j]] = i;
		//RenderParticle &rp = render_particles[s.renderParticleIndices[j]];
		//pair<float, float> &nj = normal_jitters[s.renderParticleIndices[j]];
	}

	/*for (int i = 0; i < (int)render_particles.size(); i++) {
		RenderParticle &rp = render_particles[i];
		int simpID = render_sim_map[i];
		Particle<3> &p = ren.sim.sand.particles[simpID];
		evMatrix = decomposeMatrix(p.r2, eigenVals);
		for (int ik = 0; ik < 3; ik++) {
			eigenVals[ik] = sqrtf(eigenVals[ik]);
			teigenVals[ik] = tolerance * eigenVals[ik];
		}
		bool ok = checkInRange(rp.pos, p.x, evMatrix, teigenVals);
		if (!ok) {
			nRegen++;
			float rad = (rmax + rmin) / 2.0f;
			//float rad = rmin + FRAND * (rmax - rmin);
			spill.push_back(make_pair(render_particles[i], normal_jitters[i]));
			insertRPintoBucket((int)spill.size()-1, render_particle_buckets, render_particles[i].pos);
			RenderSet rs = generate_render_particles(ren, simpID, eigenVals, eigenVecs, evMatrix, p.x, rad, spill, samplingtol);

			Vec3f actp(rs.first.pos[0], rs.first.pos[1], rs.first.pos[2]);
			if (!checkP(actp, sand)) {
				//cout << "Generated particle is out of bounds" << endl;
				//cout << " pos 2 p: " << p.x << " actp: " << actp << endl;
				//cout << "Limits: " << sand.m << ", " << sand.n << ", " << sand.o << ", " << sand.dx << endl;
				actp = clampf(actp, sand);
				//cout << "Post clamping: " << actp << endl;
				rs.first.pos[0] = actp[0]; rs.first.pos[1] = actp[1]; rs.first.pos[2] = actp[2]; 
				//cout << "p: " << p.x << " actp: " << actp << endl;
			}
			render_particles[i] = rs.first;
			normal_jitters[i] = rs.second;
			render_sim_map[i] = simpID;
			//render_particles[s.renderParticleIndices[j]] = generate_render_particle(eigenVals, eigenVecs, p.x);
		}

	}*/
	for (int i = 0; i < (int)ren.sim.bodies.size(); i++) {
		RigidBody<3> &body = ren.sim.bodies[i];
		project_particles(body, render_particles);
	}


	/*if (!consistency_check(ren, true, false)) {
	  cout << "Point set consistency check failed before applying SMMAP" << endl;
	  }
	  else {
	  cout << "Point set consistency check passed before applying SMMAP" << endl;
	  }*/

	int nsm = (int)sand.smmap.size();
	int maxmove = 0;
	int nSp = 0, nMe = 0, nMo = 0;
	for (int i = 0; i < nsm; i++) {
		/**
		 * Split/Marge map - m decides which op
		 * split: pID[0] splits to pID[{0,1}]
		 * merge: pID[{0,1}] merges into pID[0]
		 * MOVE: pID[0] moved to pID[1]
		 */
		SplitMerge &sm = sand.smmap[i];
		switch(sm.m) {
			case SPLIT: 
				nSp++;
				/// Check each particle, move them to pID[1] if needed
				if (sm.pID[1] > (int)sim_particles.size() - 1) sim_particles.resize(sm.pID[1] + 1);
				{
					SimParticle &s = sim_particles[sm.pID[0]];
					SimParticle &s1 = sim_particles[sm.pID[1]];
					/// New, test, just split the render particles 50/50, hopefully they'll re-align after 1-2 steps
					int mid = (int)s.renderParticleIndices.size() / 2, nmoved = (int)s.renderParticleIndices.size() - mid;
					for (int j = mid; j < mid + nmoved; j++) {
						s1.renderParticleIndices.push_back(s.renderParticleIndices[j]);
						render_sim_map[s.renderParticleIndices[j]] = sm.pID[1];
					}
					//s.renderParticleIndices.resize(s.renderParticleIndices.size() - nmoved);
					s.renderParticleIndices.resize(mid);
				}
				break;
			case MERGE:
				nMe++;
				if (sm.pID[1] < (int)sim_particles.size() && sm.pID[0] < (int)sim_particles.size()) 
				{
					/// Move each particle from pID[1] to pID[0]
					for (int j = 0; j < (int)sim_particles[sm.pID[1]].renderParticleIndices.size(); j++) {
						sim_particles[sm.pID[0]].renderParticleIndices.push_back(sim_particles[sm.pID[1]].renderParticleIndices[j]);
						render_sim_map[sim_particles[sm.pID[1]].renderParticleIndices[j]] = sm.pID[0];
					}
					sim_particles[sm.pID[1]].renderParticleIndices.clear();
				}
				break;
			case MOVE:
				nMo++;
				//maxmove = (abs<int>(sm.pID[0] - sm.pID[1]) > maxmove) ? abs<int>(sm.pID[0] - sm.pID[1]) : maxmove;
				maxmove = 0;
				if (sm.pID[0] == sm.pID[1]) {
					cout << "Pointless move operation, moving from " << sm.pID[0] << " to " << sm.pID[1] << endl;
					break;
				}
				if (sm.pID[1] < (int)sim_particles.size() && sm.pID[0] < (int)sim_particles.size()) 
				{
					/// Moves are always the result of merges, hence the number of render particles across moves/merges/splits must be conserved
					/*{
						SimParticle &s = sim_particles[sm.pID[1]];
						int m = (int)s.renderParticleIndices.size();
						int nMoved = 0;
						for (int j = 0; j < m; j++) {
							int pos = s.renderParticleIndices[j];
							if (render_sim_map[pos] != sm.pID[1]) continue;
							int endp = (int)render_particles.size() - 1 - nMoved;
							if (pos != endp) nMoved++;
							render_particles[pos] = render_particles[endp];
							normal_jitters[pos] = normal_jitters[endp];
							int spID = render_sim_map[endp];
							render_sim_map[pos] = spID;
							if ( spID < sim_particles.size() ) {
								SimParticle &sp = sim_particles[spID];
								bool done = i == spID;
								for (int k = 0; !done && k < (int)sp.renderParticleIndices.size(); k++) {
									if (sp.renderParticleIndices[k] == endp) {
										sp.renderParticleIndices[k] = pos;
										done = true;
									}
								}
							}
						}
						render_particles.resize(render_particles.size() - nMoved);
						normal_jitters.resize(normal_jitters.size() - nMoved);
						s.renderParticleIndices.clear();
						assert(s.renderParticleIndices.size() == 0);
					}*/
					/// Move each particle from pID[0] to pID[1]
					sim_particles[sm.pID[1]] = sim_particles[sm.pID[0]];
					sim_particles[sm.pID[0]].renderParticleIndices.clear();
					if ( sm.pID[0] == (int)sim_particles.size() - 1 ) {
						sim_particles.resize(sim_particles.size() - 1);
					}
					else {
						cout << "Move from " << sm.pID[0] << " to " << sm.pID[1] << " did not resize, size = " << sim_particles.size() << endl;
					}
					for (int j = 0; j < (int)sim_particles[sm.pID[1]].renderParticleIndices.size(); j++) {
						render_sim_map[sim_particles[sm.pID[1]].renderParticleIndices[j]] = sm.pID[1];
					}
				}
				break;
			default:
				cout << "Error: Map type " << sm.m << " incorrect, for particle numbers (" << sm.pID[0] << ", " << sm.pID[1] << ")" << endl << yield;
		}
	}
	cout << "SplitMerge stats: (Splits, Merges, Moves) -> (" << nSp << ", " << nMe << ", " << nMo << ")" << endl;
	cout << "Largest move: " << maxmove << endl;
	if (maxmove > 0) {
		cout << "Resizing list from " << sim_particles.size() << " to " << sim_particles.size() - maxmove << endl;
		sim_particles.resize(sim_particles.size() - maxmove);
	}
	if ( sim_particles.size() > sand.particles.size() ) {
		cout << "SplitMerge mapping completed, sizes incompatible before resizing : sim_particles: " << sim_particles.size() << ", sand.particles: " << sand.particles.size() << endl;
	}

	sim_particles.resize(n);

	/// Check each particle that its within bounds of the simulation particle
	/// Temporary testing to see how things behave without this check
//#pragma omp parallel for
	//for (int i = 0; i < (int)render_particles.size(); i++) 
	if(0)
	{
		//render_sim_map[s.renderParticleIndices[j]] = i;
		int simpID = render_sim_map[i];
		RenderParticle &rp = render_particles[i];
		//pair<float, float> &nj = normal_jitters[s.renderParticleIndices[j]];
		
		/*if (simpID < 0 || simpID >= (int)ren.sim.sand.particles.size()) {
			cout << "Error: Particle " << i << " belongs to out of range simulation particle " << simpID << endl;
		}*/
		Particle<3> &p = ren.sim.sand.particles[simpID];
		evMatrix = decomposeMatrix(p.r2, eigenVals, eigenVecs);
		for ( int coord = 0; coord < 3; coord++ ) {
			eigenVals[coord] = sqrtf(fabsf(eigenVals[coord]));
			teigenVals[coord] = tolerance * eigenVals[coord];
		}
		bool ok = checkInRange(rp.pos, p.x, evMatrix, teigenVals);
		if (!ok) {
			nRegen++;
			float rad = (rmax + rmin) / 2.0f;
			//float rad = rmin + FRAND * (rmax - rmin);
			spill.push_back(make_pair(render_particles[i], normal_jitters[i]));
			insertRPintoBucket((int)spill.size()-1, render_particle_buckets, render_particles[i].pos);
			RenderSet rs = generate_render_particles(ren, simpID, eigenVals, eigenVecs, evMatrix, p.x, rad, spill, samplingtol);

			Vec3f actp(rs.first.pos[0], rs.first.pos[1], rs.first.pos[2]);
			if (!checkP(actp, sand)) {
				//cout << "Generated particle is out of bounds" << endl;
				//cout << " pos 2 p: " << p.x << " actp: " << actp << endl;
				//cout << "Limits: " << sand.m << ", " << sand.n << ", " << sand.o << ", " << sand.dx << endl;
				actp = clampf(actp, sand);
				//cout << "Post clamping: " << actp << endl;
				rs.first.pos[0] = actp[0]; rs.first.pos[1] = actp[1]; rs.first.pos[2] = actp[2]; 
				//cout << "p: " << p.x << " actp: " << actp << endl;
			}
			render_particles[i] = rs.first;
			normal_jitters[i] = rs.second;
			render_sim_map[i] = simpID;
			//render_particles[s.renderParticleIndices[j]] = generate_render_particle(eigenVals, eigenVecs, p.x);
		}

	}
	
	/*if (!consistency_check(ren, true, false)) {
		cout << "Point set consistency check failed after applying SMMAP" << endl;
	}
	else {
		cout << "Point set consistency check passed after applying SMMAP" << endl;
	}*/
	for ( i = 0; i < n; i++) {
		SimParticle &s = sim_particles[i];
		Particle<3> &p = ren.sim.sand.particles[i];
		int m = (int)s.renderParticleIndices.size();
		Vec3f eigenVals, teigenVals, eigenVecs[3];
		Mat3f evMatrix = decomposeMatrix(p.r2, eigenVals, eigenVecs);
		for ( int coord = 0; coord < 3; coord++ ) {
			eigenVals[coord] = sqrtf(fabsf(eigenVals[coord]));
			teigenVals[coord] = tolerance * eigenVals[coord];
			/*if (eigenVals[coord] <= 0 || eigenVals[coord] > ren.sim.sand.dx || eigenVals[coord] != eigenVals[coord] ) {
				cout << "Eigenvalues of particle " << i << " are out of range, they are: " << eigenVals;
				cout << " particle is at " << p.x << " and r2 is " << p.r2 << endl;
			}*/
		}
		float rad = (rmax + rmin) / 2.0f;
		//float rad = rmin + FRAND * (rmax - rmin);
		//int newm = p.m * (1 - fill_fraction * sqf(fill_fraction)) / sand.rho_max / (4/3 * (float)M_PI * rad * sqf(rad)) * tolerance + 5;
		int newm = p.m / rad * tolerance + minRP;
		Vec3f newpx(p.x + Vec3f(distfield.dx, distfield.dx, distfield.dx));
		float distf = fabsf (lerp(newpx, distfield) * distfield.dx);
		float dens = lerp(p.x, sand.rho);
		bool condition = (dens < sand.rho_max * cutoff_density) || (distf <= cutoff_distance);
		/// Adding the condition to not sample any spurious sim particles that may have a 0 eigenvalue
		float hackCutoff = 0.00001;
		if (minRP == 0)
			condition = condition && (eigenVals[0] > hackCutoff) && (eigenVals[1] > hackCutoff) && (eigenVals[2] > hackCutoff);
		//bool condition = distf >= cutoff_distance;
		if ((m == 0 || m < newm) && condition) {
			/// Generate a new m!
			//m = (int)( ( eigenVals[0] / rad ) * ( eigenVals[1] / rad ) * ( eigenVals[2] / rad ) ) + 5;
			int startm = m;
			m = newm;
			if ( isnanl( 1 / rad ) ) {
				cout << "Render particle radius range invalid: (" << rmin << ", " << rmax << ")" << " check setup XML file" << endl << yield;
			}
			nNewgen += (m - startm);

			//cout << "Generating " << m << " particles for simulation particle " << i;
			for ( int j = startm; j < m; j++ ) {
				//render_particles.push_back(generate_render_particle(eigenVals, eigenVecs, p.x));
				RenderSet rs = generate_render_particles(ren, i, eigenVals, eigenVecs, evMatrix, p.x, rad, spill, samplingtol);
				Vec3f actp(rs.first.pos[0], rs.first.pos[1], rs.first.pos[2]);
				if (!checkP(actp, sand)) {
					//cout << "Generated particle is out of bounds" << endl;
					//cout << "pos 1 p: " << p.x << " actp: " << actp << endl;
					//cout << "Limits: " << sand.m << ", " << sand.n << ", " << sand.o << ", " << sand.dx << endl;
					actp = clampf(actp, sand);
					//cout << "Post clamping: " << actp << endl;
					rs.first.pos[0] = actp[0]; rs.first.pos[1] = actp[1]; rs.first.pos[2] = actp[2]; 
					//cout << "p: " << p.x << " actp: " << actp << endl;
				}
				render_particles.push_back(rs.first);
				/// Note FSRAND generates in range -1, 1
				normal_jitters.push_back(rs.second);
				s.renderParticleIndices.push_back(render_particles.size() - 1);
				render_sim_map[render_particles.size() - 1] = i;
			}
			//cout << " done" << endl;
			//continue;
		}
		else if (!condition) {
			//if ( lerp(p.x, sand.phi) < 0.99f ) {
			//	cout << "Debug: distance/cutoff= " << distf << "/" << cutoff_distance << " for particle at: " << p.x << endl;
			//}
			//cout << "Debug: distance: " << distf << " cutoff: " << cutoff_distance << " position: " << Vec3f(p.x / sand.dx) << endl;
			/// Delete all these render particles
			int nMoved = 0;
			for (int j = 0; j < m; j++) {
				/**
				 * Steps:
				 * Move render_particles[size-1] to current pos, same for normal_jitter, and resize both by -1
				 * update sim_particles pointer for the relevant particle
				 */
				int pos = s.renderParticleIndices[j];
				int endp = (int)render_particles.size() - 1 - nMoved;
				if (pos != endp) nMoved++;
				render_particles[pos] = render_particles[endp];
				normal_jitters[pos] = normal_jitters[endp];
				int spID = render_sim_map[endp];
				render_sim_map[pos] = spID;
				if ( spID < sim_particles.size() ) {
					SimParticle &sp = sim_particles[spID];
					bool done = i == spID;
					for (int k = 0; !done && k < (int)sp.renderParticleIndices.size(); k++) {
						if (sp.renderParticleIndices[k] == endp) {
							sp.renderParticleIndices[k] = pos;
							done = true;
						}
					}
				}
				//if (!done) cout << "Did not find a replacement for particle " << endp << " while placing it at position " << pos << endl;
			}
			render_particles.resize(render_particles.size() - nMoved);
			normal_jitters.resize(normal_jitters.size() - nMoved);
			s.renderParticleIndices.clear();
			assert(s.renderParticleIndices.size() == 0);
			continue;
		}
		m = (int)s.renderParticleIndices.size();

		//if ( i == 0 ) {
		//	cout << "p.x = (" << p.x[0] << ", " << p.x[1] << ", " << p.x[2] << ") " << endl;
		//}
	}


	/*if (!consistency_check(ren, true, false)) {
	  cout << "Point set consistency check failed after doing book-keeping" << endl;
	  }
	  else {
	  cout << "Point set consistency check passed after doing book-keeping" << endl;
	  }*/

	/// HACK, remove later - hmm, scratch that, we need to redefine normals at each step
	render_normals.resize(render_particles.size());

	if (usenormals != NONE) {
#pragma omp parallel for
		for ( int i = 0; i < (int)render_particles.size(); i++ ) {
			Vec3f normal, pos, perp1, perp2;
			/// Changing it to e2 + e1, to create fewer cases with 0 cross product
			Vec3f e1; e1[0] = 1; e1[1] = 1; e1[2] = 0;
			Vec3f color; color[0] = 194.0f/255; color[1] = 178.0f/255; color[2] = 128.0f/255;
			RenderParticle &rp = render_particles[i];
			RenderNormal &np = render_normals[i];
			pos[0] = rp.pos[0];  pos[1] = rp.pos[1];  pos[2] = rp.pos[2];
			//pos = pos * (1 + 0.01f * FSRAND);
			//rp.pos[0] = pos[0]; rp.pos[1] = pos[1]; rp.pos[2] = pos[2];
			//normal = (Vec3f)tvmet::normalize(lerp(pos, grad)) * -1;
			//normal = (Vec3f)lerp(pos, grad) * -1;
			normal = gradient(pos, ren.sim.sand.rho) * -1;
			for (int nj = 0; nj < nGradientSamples; nj++) {
				normal += gradient(Vec3f(pos + Vec3f(FSRAND * dx, FSRAND * dx, FSRAND * dx)), ren.sim.sand.rho) * -1;
			}

			if (isnan(normal) ) {
				cout << "Error: Generated normal is useless, setting to 0" << endl;
				normal = Vec3f(1, 0, 0);
			}
			normal = normal / (tvmet::norm2(normal) + epsilon);
			if (tvmet::norm2(normal) < 0.1f) {
				normal = gradient(Vec3f(pos + Vec3f(distfield.dx, distfield.dx, distfield.dx)), distfield) * -1;
				normal = normal / (tvmet::norm2(normal) + epsilon);
				if (tvmet::norm2(normal) < 0.1f || isnan(normal)) {
					//cout << "Error: Gradient produced useless normal, generating random normal" << endl;
					normal = Vec3f(1, 0, 0);
				}
				//else
				//	normal = normal / (tvmet::norm2(normal) + 0.001f) * -1;
			}
			if (isnan(normal)) {
				cout << "Error: Generated normal is useless, setting to 0" << endl;
				normal = Vec3f(1, 0, 0);
			}
			if (usenormals == FULL) {
				//cout << "Normal: (" << normal[0] << ", " << normal[1] << ", " << normal[2] << ")" << endl;
				perp1 = (Vec3f)(tvmet::cross(normal, e1));
				perp1 = perp1 / (tvmet::norm2(perp1) + epsilon);
				if (tvmet::norm2(perp1) < 0.9f) {
					perp1 = tvmet::cross(normal, Vec3f(1, 1, 1));
					perp1 = perp1 / (tvmet::norm2(perp1) + epsilon);
				}
				perp2 = (Vec3f)(tvmet::cross(normal, perp1));
				perp2 = perp2 / (tvmet::norm2(perp2) + epsilon);
				if (isnan(perp1) || isnan(perp2)) {
					cout << "Basis for tangent plane is Nan" << endl;
				}
				/// Now we have a basis of the normal space
				// Generate 2 angles, theta and phi, which are the angles from the normal var and the perp1 vector
				float theta = normal_jitters[i].first, phi = normal_jitters[i].second;
				//float theta = nJitter * FRAND, phi = 180.0f * FSRAND;

				normal = (Vec3f)tvmet::normalize(normal * cosf(theta) + (perp1 * cosf(phi) + perp2 * sinf(phi)) * sinf(theta));
			}

			if (isnan(normal)) {
				cout << "Generated normal is Nan" << endl;
			}

			//normal = (Vec3f)tvmet::normalize(normal + ( perp1 * FSRAND + perp2 * FSRAND ) * nJitter );
			//normal = color * tvmet::dot(normal, aim);
			np.pos[0] = normal[0]; np.pos[1] = normal[1]; np.pos[2] = normal[2];
			//np.pos[0] = 0; np.pos[1] = 1; np.pos[2] = 0;
			//if ( isnormal(normal[0]) == 0 || isnormal(normal[1]) || isnormal(normal[2]) ) {
			//	cout << "Invalid normal: " << normal << " for particle at: " << pos << endl;
			//}
		}
	}
	cout << nRegen << "/" << render_particles.size() << " regenerated, " << nNewgen << " new particles, " << spillcount << " reused old particles" << endl;
	cout << spill.size() << " particles spilled over, " << spill.size() << " jitters left over" << endl;
}

void UnitCube(bool offcenter = false) {
	//static RtColor Color = {(RtFloat).2, (RtFloat).4, (RtFloat).6};
	//RiColor(Color);
#define L -.5
#define R  .5
#define D   L
#define U   R
#define F   R
#define N   L
	static RtPoint Cube[6][4] = {
		{ {L,D,F}, {R,D,F}, {R,D,N}, {L,D,N} },
		{ {L,D,F}, {L,D,N}, {L,U,N}, {L,U,F} },
		{ {R,U,N}, {R,U,F}, {L,U,F}, {L,U,N} },
		{ {R,U,N}, {R,D,N}, {R,D,F}, {R,U,F} },
		{ {R,D,F}, {L,D,F}, {L,U,F}, {R,U,F} },
		{ {L,U,N}, {L,D,N}, {R,D,N}, {R,U,N} }
	};
	if (offcenter)
		RiTranslate(0.5f, 0.5f, 0.5f);
	for ( int i = 0; i < 6; i++ )
		RiPolygon(4, RI_P, (RtPointer) Cube[i], RI_NULL);
}

void RendermanRenderer::build_distance_field(Renderer<3> &ren) {
	Sand<3> &sand = ren.sim.sand;
	int dfs = distfield_supersampling;
	int m = dfs * sand.m, n = dfs * sand.n, o = dfs * sand.o;
	float dx = sand.dx / dfs;
	distfield.resize(m, n, o, dx);
	rasterfield.resize(m, n, o, dx);
	float eps = 0.01f * sand.rho_max;

#pragma omp parallel for
	for (int i = 0; i < m; i++) {
		for (int j = 0; j < n; j++) {
			for (int k = 0; k < o; k++) {
				Vec3f pos(i * dx, j * dx, k * dx);
				rasterfield(i, j, k) = (lerp(pos, sand.rho) > eps) ? 1 : 0;
			}
		}
	}

	compute_distance_field(rasterfield, distfield);

	save_scalar_field(rasterfield, "raster.dat");
	save_scalar_field(distfield, "dist.dat");
	//for (int i = 0; i < m; i++) {
	//	for (int j = 0; j < n; j++) {
	//		cout << rasterfield(i, j, 4) << " ";
	//	}
	//	cout << endl;
	//}
	//for (int i = 0; i < m; i++) {
	//	for (int j = 0; j < n; j++) {
	//		//for (int k = 0; k < o; k++) {
	//			//cout << "raster = " << rasterfield(i, j, k) << " dist = " << distfield(i, j, k) << " ";
	//		//}
	//		cout << distfield(i, j, 4) << " ";
	//	}
	//	cout << endl;
	//}
	//cout << endl;
}

void RendermanRenderer::render_frames(Renderer<3> &ren, int start_frame, int end_frame) {
	bool success = true;
	Sand<3> &sand = ren.sim.sand;
	Timer perfTimer;
	perfTimer.tick();
	float rentime = 0, progtime = 0;
	//float diff_needed = 1.0f/fps;
	//float curr_rendertime = -diff_needed;
	int curr_renderframe = 0;
	string command = "launch:prman? -t -progress";

	for ( int i = start_frame; success && (i < end_frame); i++ ) {
		string outribfile = stringf(outfile, curr_renderframe);
		outribfile = outribfile + ".rib";
		int renderiter = 0;//(doshadows) ? 0 : (int)shmaplist.size();
		rentime = perfTimer.tick();
		cout << "Rendering frame " << i << " Timing (renderman, program) = (" << rentime << ", " << progtime << ")" << endl;
		string filename = stringf(ren.infile, ren.frame);
		success = load_state<3>(filename, ren.modeldir, ren.sim);
		//cout << "Moving points... ";
		move_points(ren);
		//cout << "done" << endl;
		progtime = perfTimer.tick();
		/// Only render the relevant frames
		//if (rendertime - curr_rendertime < diff_needed) {
		//	cout << "FPS needed: " << fps << " rendered last frame at sim time: " << curr_rendertime << " seconds so skipping this frame as current sim time is " << rendertime << " seconds" << endl;
		//	continue;
		//}
		//else {
			//curr_rendertime = rendertime;
			curr_renderframe++;
		//}
		RiBegin((RtToken)command.c_str());
		for (renderiter = 0; renderiter <= (int)shmaplist.size(); renderiter++) 
		{
			//RiFrameBegin(i * ((int)shmaplist.size() + 1) + renderiter);
			RiFrameBegin(renderiter);
			if ( !ribinitfile.empty()) {
				//cout << "Using RIB init file: " << ribinitfile << endl;
				RiReadArchive((RtString)ribinitfile.c_str(), (RtArchiveCallback)NULL, RI_NULL);
			}
			filename = stringf(outfile, i);
			if (renderiter == (int)shmaplist.size()) {
				char filename_const_kludge[256];
				strcpy(filename_const_kludge, filename.c_str());
				RiFormat(imageres[0], imageres[1], 1);
				RiPixelSamples(4, 4);
				RiPixelFilter(RiGaussianFilter, 2, 2);
				RiDisplay(filename_const_kludge, "file", "rgba", RI_NULL);
				//RiDisplay("+testview", "framebuffer", "rgb", RI_NULL);
				RiProjection("perspective", "fov" , &fov,  RI_NULL);
				cout << "Rendering image..." << endl;
			}
			else {
				ShadowMap &s = shmaplist[renderiter].second;
				RiFormat(shadowmapres, shadowmapres, 1);
				RiPixelSamples(1, 1);
				RiPixelFilter(RiBoxFilter, 1, 1);
				//RiDisplay((char*)s.outputfile.c_str(), "zfile", "z", RI_NULL);
				RiDisplay((char*)shmap_temp.c_str(), "zfile", "z", RI_NULL);
				RiDisplay((char*)("+" + s.outputfile).c_str(), "deepshad", "deepopacity", RI_NULL);
				RiProjection("perspective", "fov", &shmapfov, RI_NULL);
				//RiDisplay("+testview", "framebuffer", "rgb", RI_NULL);
				cout << "Generating shadow map " << renderiter + 1 << " of " << shmaplist.size() << endl;
			}
			if (renderiter < (int)shmaplist.size()) {
				ShadowMap &s = shmaplist[renderiter].second;
				PlaceCameraPts(s.from, s.to, 0);
				cout << "Placing camera from: [" << s.from[0] << " " << s.from[1] << " " << s.from[2] << "], to: [" << s.to[0] << " " << s.to[1] << " " << s.to[2] << "]" << endl;
			}
			else {
				PlaceCameraPts(camposition, camto, 0);
				cout << "Placing camera from: [" << camposition[0] << " " << camposition[1] << " " << camposition[2] << "], to: [" << camto[0] << " " << camto[1] << " " << camto[2] << "]" << endl;
			}
			//RiTranslate(camposition[0], camposition[1], camposition[2]);
			//RiRotate(-viewpoint.lat, 1,0,0);
			//RiRotate(-viewpoint.lon, 0,1,0);
			/// World begins
			//for (int ptccount = 0; ptccount < 2; ptccount++) 
			{
				//if (ptccount == 0) {
				//	RiSurface("ptc_write", RI_NULL);
				//	RiShadingRate(4);
				//	RiAttribute("cull", "hidden", "[0]", "backfacing", "[0]", RI_NULL);
				//	RiAttribute("dice", "rasterorient", "[0]", RI_NULL);
				//}
				//else {
				//	RiSurface("ptc_occlusion", RI_NULL);
				//}
				RiWorldBegin();
				RiAttributeBegin();
				RiTransformBegin();
				RtColor c = {0.9, 0.1, 0.1}, co;
				RiColor(c);

				if ( !ribworldfile.empty()) {
					//cout << "Using RIB world file: " << ribworldfile << endl;
					string modifile = ribworldfile + "m";
					if (renderiter < (int)shmaplist.size())
						RiReadArchive((RtString)modifile.c_str(), (RtArchiveCallback)NULL, RI_NULL);
					else
						RiReadArchive((RtString)ribworldfile.c_str(), (RtArchiveCallback)NULL, RI_NULL);
				}

				//render_points(ren);
				sgraph.render();

				RiTransformEnd();
				RiAttributeEnd();
				RiWorldEnd();
				if (renderiter < (int)shmaplist.size()) {
					ShadowMap &s = shmaplist[renderiter].second;
					stringstream out;
					out << i;
					//RiMakeShadow("temp.deep", (char*)s.outputfile.c_str(), RI_NULL);
					//RiMakeShadow("temp.depth", (char*)s.outputfile.c_str(), RI_NULL);
					//RiMakeShadow("temp.depth", (char*)(s.outputfile + out.str()).c_str(), RI_NULL);
				}
			}

			/// World ends
			RiFrameEnd();

		}

		/// Frame ends

		ren.frame++;
		RiEnd();
	}
}

void getRotation(Quatf &q, Vec3f &axis, float &angle) {
	/// Assuming quat to be of form q = cos(angle/2) + axis*sin(angle/2)
	angle = 2 * acosf(q.s);
	if (angle != 0)
		axis = q.v / (sinf(angle/2) + epsilon);
	angle = angle * 180.0f/(float)M_PI;
}

void RendermanSceneGraph::render() {
	int n = (int)scene.size();
	cout << "Rendering all " << n << " objects in scene...";
	for (int i = 0; i < n; i++) {
		Node &n = scene[i];
		RiAttributeBegin();
		RiTransformBegin();
		if (n.type == GRID) {
			Sand<3> &sand = (*ren).sim.sand;
			n.bound[3] = sand.dx * sand.m / 2;
			n.bound[4] = sand.dx * sand.n / 2;
			n.bound[5] = sand.dx * sand.o / 2;

			n.bound[0] = -n.bound[3];
			n.bound[1] = -n.bound[4];
			n.bound[2] = -n.bound[5];
		}
		if (n.align != -1 && n.align <= (int)scene.size()) {
			Node &n2 = scene[n.align];
			float min[3], max[3];
			min[0] = n2.bound[0] * n2.scale[0] + n2.location[0];
			min[1] = n2.bound[1] * n2.scale[1] + n2.location[1];
			min[2] = n2.bound[2] * n2.scale[2] + n2.location[2];
			max[0] = n2.bound[3] * n2.scale[0] + n2.location[0];
			max[1] = n2.bound[4] * n2.scale[1] + n2.location[1];
			max[2] = n2.bound[5] * n2.scale[2] + n2.location[2];

			float scale[3] = {(max[0] - min[0]) / (n.bound[3] - n.bound[0]), (max[1] - min[1]) / (n.bound[4] - n.bound[1]), (max[2] - min[2]) / (n.bound[5] - n.bound[2])};
			n.scale[0] = scale[0];  n.scale[1] = scale[1];  n.scale[2] = scale[2];

			n.location[0] = min[0] - scale[0] * n.bound[0];
			n.location[1] = min[0] - scale[1] * n.bound[1];
			n.location[2] = min[0] - scale[2] * n.bound[2];

			cout << "Debug: To align body " << i << " with body " << n.align << " info: (loc, bound, scale) : ([" << n2.location[0] << " " << n2.location[1] << " " << n2.location[2] << "], [" << n2.bound[0] << " " << n2.bound[1] << " " << n2.bound[2] << " " << n2.bound[3] << " " << n2.bound[4] << " " << n2.bound[5] << "], [" << n2.scale[0] << " " << n2.scale[1] << " " << n2.scale[2] << "]) we found: (loc, scale, bound) : ([" << n.location[0] << " " << n.location[1] << " " << n.location[2] << "], [" << n.scale[0] << " " << n.scale[1] << " " << n.scale[2] << "], [" << n.bound[0] << " " << n.bound[1] << " " << n.bound[2] << " " << n.bound[3] << " " << n.bound[4] << " " << n.bound[5] << "])" << endl;
		}
		if (n.tcouple != -1 && n.type == RIGIDBODY && n.tcouple <= (*ren).sim.bodies.size()) {
			RigidBody<3> &body = (*ren).sim.bodies[n.tcouple];
			RigidBodyInfo &body_info = (*ren).sim.body_info[n.tcouple];
			n.location[0] = body.pose.x[0];  n.location[1] = body.pose.x[1];  n.location[2] = body.pose.x[2];  
			n.scale[0] = n.scale[1] = n.scale[2] = body_info.scale;

			/// See whether this body is a scripted object and modify the position appropriately
			if (n.hold.active) {
				Vec3f g((*ren).sim.g * -1);
				float deltat = n.hold.endtime - ren->sim.simtime;
				//float deltat = n.hold.endtime - rren.rendertime;
				if (deltat > 0) {
					/// Now use simply s = ut + 1/2 a t^2, u = 0
					Vec3f newx(body.pose.x + g * 0.5f * sqf(deltat) - n.hold.v * deltat);
					n.location[0] = newx[0];  n.location[1] = newx[1];  n.location[2] = newx[2];  
				}
			}
		}
		if (n.couple != -1 && n.type == RIGIDBODY && n.couple <= (*ren).sim.bodies.size()) {
			RigidBody<3> &body = (*ren).sim.bodies[n.couple];
			RigidBodyInfo &body_info = (*ren).sim.body_info[n.couple];
			//n.location[0] = body.pose.x[0];  n.location[1] = body.pose.x[1];  n.location[2] = body.pose.x[2];  
			Vec3f axis;
			getRotation(body.pose.q, axis, n.angle);
			n.axis[0] = axis[0];  n.axis[1] = axis[1];  n.axis[2] = axis[2];
			//n.scale[0] = n.scale[1] = n.scale[2] = body_info.scale;
		}

		RiTranslate(n.location[0], n.location[1], n.location[2]);
		if (n.angle != 0)
			RiRotate(n.angle, n.axis[0], n.axis[1], n.axis[2]);
		if (n.scale[0] != 0 && n.scale[1] != 0 && n.scale != 0)
			RiScale(n.scale[0], n.scale[1], n.scale[2]);

		switch(n.type) {
			case RIGIDBODY:
				if (n.objdata != "")
					RiReadArchive((RtString)(n.objdata + ".rib").c_str(), (RtArchiveCallback)NULL, RI_NULL);
				break;
			case GRID:
				RiAttributeBegin();
				RiSides(2);
				rren.render_points(*ren);
				RiAttributeEnd();
				break;
			default:
				break;
		}

		RiTransformEnd();
		RiAttributeEnd();
	}
	cout << "done" << endl;
}

void RendermanRenderer::render_points(Renderer<3> &ren) {
	Sand<3> &sand = ren.sim.sand;
	//float scale = fmaxf(fmaxf(sand.m, sand.n), sand.o)*sand.dx;
	/**
	 * Logic: Scale entire world by 1/(size of grid)
	 *        So then, any transformations can be done at grid size resolution, and they'll be 
	 *	  actually at world res i.e. between -1,1
	 */
	RiTransformBegin();
	//RiScale(1/scale, 1/scale, 1/scale);
	RiTranslate(-sand.m*sand.dx/2, -sand.n*sand.dx/2, -sand.o*sand.dx/2);
	/// This rotation needs to be done to align the OpenGL view of coordinate axes, and the Renderman view
	//RiRotate(90, 0, 1, 0);

	//RtColor c, co;
	//c[0] = 194.0f/255; c[1] = 1.0f/255; c[2] = 1.0f/255;
	//c[0] = 237.0f/255; c[1] = 201.0f/255; c[2] = 175.0f/255;
	//co[0] = co[1] = co[2] = 0.7f;
	RiAttributeBegin();
	RiColor(sandcolor);
	RiOpacity(sandopacity);

	//float width = 0.01f;

	RiSurface("plastic", "float Ks", "0.7", "float roughness", "0.001", RI_NULL);
	//RiSurface("plastic", "float Ks", "0.7", "float Ka", "0.05", "float Kd", "0.7", "float roughness", "0.001", RI_NULL);
	if (!wireframe) {
		if (usenormals != NONE) {
			RiPoints((int)render_particles.size(), RI_P, &(render_particles[0]), RI_N, &(render_normals[0]), RI_CONSTANTWIDTH, &render_width, RI_NULL);
		}
		else {
			RiPoints((int)render_particles.size(), RI_P, &(render_particles[0]), RI_CONSTANTWIDTH, &render_width, RI_NULL);
		}
	}
	else {
		//RiScale (sand.m * sand.dx, sand.n * sand.dx, sand.o * sand.dx);
		//UnitCube(true);

		/// This wireframe mode should help us align models, so we render the cells which are covered by solids
		float dx = sand.dx;
		for (int i = 0; i < sand.m; i++) {
			for (int j = 0; j < sand.n; j++) {
				for (int k = 0; k < sand.o; k++) {
					Vec3f min(i*dx, j*dx, k*dx), max(min+Vec3f(dx,dx,dx));
					float cov = 1;
					for (int ri = 0; ri < (int)ren.sim.bodies.size(); ri++) {
						RigidBody<3> &body = ren.sim.bodies[ri];
						cov = cov * coverage(body, min, max);
					}
					if (cov < 1) {
						RiTransformBegin();
						RiTranslate(min[0], min[1], min[2]);
						RiScale(dx, dx, dx);
						UnitCube(true);
						RiTransformEnd();
					}
				}
			}
		}
	}
	RiAttributeEnd();
	RiTransformEnd();
}

string generateShadowMapList(string filename, vector < pair<int, ShadowMap> > &list) {
	string modifile = filename + "m";
	fstream file(filename.c_str(), ios::in);
	fstream file2(modifile.c_str(), ios::out);
	if (!file.is_open() || file.eof()) {
		cout << "Error: Parse, could not open file: " << filename << endl << yield;
	}

	char buffer[512];
	vector<string> tokens;
	ShadowMap s;
	while (!file.eof()) {
		file.getline(buffer, 512);
		/// empty or commented line
		if (buffer[0] == ' ' || buffer[0] == '#' || strlen(buffer) == 0)
			continue;
		//else cout << "Considering line: " << buffer << endl;
		string buf(buffer);
		/// not a light source declaration
		if (buf.find("LightSource") == string::npos) {
			file2 << buf << endl;
			//cout << "Not a light source" << endl;
			continue;
		}
		tokens.clear();
		tokenize(buf, tokens);

		int lno = atoi(tokens[2].c_str());
		string shstr("shadowname\""), shstr2("shadowmap\"");
		int loc = (find<string>(shstr, tokens) >= 0) ? find<string>(shstr, tokens) : find<string>(shstr2, tokens);
		if (loc > 0) {
			/// Truncate and remove '"' and '[]'
			s.outputfile = tokens[loc + 1].substr(1, tokens[loc + 1].size() - 2);
			removeDelims(s.outputfile, '\"');
			if (s.outputfile == "raytrace")
				continue;
			loc = find<string>(string("\"from\""), tokens);
			if (loc > 0) {
				//cout << "Found 'from' token at " << loc << endl;
				string from = tokens[loc + 1] + " " + tokens[loc + 2] + " " + tokens[loc + 3];
				from = from.substr(1, from.size() - 1);
				stringstream ss(from);
				ss >> s.from[0] >> s.from[1] >> s.from[2];
			}
			else cout << "No 'from' token for light, assuming [0 0 0]" << endl;
			loc = find<string>(string("\"to\""), tokens);
			if (loc > 0) {
				//cout << "Found 'to' token at " << loc << endl;
				string to = tokens[loc + 1] + " " + tokens[loc + 2] + " " + tokens[loc + 3];
				to = to.substr(1, to.size() - 1);
				stringstream ss(to);
				ss >> s.to[0] >> s.to[1] >> s.to[2];
			}
			else cout << "No 'to' token for light, assuming [0 0 0]" << endl;
			list.push_back(make_pair(lno, s));
		}
		//else cout << "No shadow map commands found" << endl;
	}
	return modifile;
}

void render3r(Renderer<3> &ren, string outfile, string setupfile, bool useglut, int normals, bool wireframe, int fps, int lastframe) {
	//RendermanRenderer r;
	::ren = &ren;
	rren.outfile = outfile;
	rren.load_setup(setupfile);
	rren.wireframe = wireframe;
	rren.fps = fps;
	cout << "Read setup file: " << setupfile << ", read RIBs(" << rren.ribinitfile << " , " << rren.ribworldfile << ")";
	cout << " tolerance: " << rren.tolerance << " radius range: (" << rren.rmin << ", " << rren.rmax << ")" << " cutoff: " << rren.cutoff_distance << endl;
	generateShadowMapList(rren.ribworldfile, rren.shmaplist);
	rren.usenormals = (NormalRenderingMode)normals;
	cout << "Need to generate " << rren.shmaplist.size() << " shadow maps: " << endl;
	for (int i = 0; i < (int)rren.shmaplist.size(); i++) {
		ShadowMap &s = rren.shmaplist[i].second;
		cout << "Light Source: " << rren.shmaplist[i].first << " from: [" << s.from[0] << " " << s.from[1] << " " << s.from[2] << "] to: [" << s.to[0] << " " << s.to[1] << " " << s.to[2] << "] output file: " << s.outputfile << endl;
	}
	switch (NormalRenderingMode(normals)) {
		case NONE: cout << "Not rendering normals" << endl; break;
		case NOJITTER: cout << "Not rendering normal jitter" << endl; break;
		case FULL: cout << "Rendering normals with jitter" << endl; break;
		default: cout << "Unknown normal rendering mode requested" << endl;
	}
	if (useglut)
		glutrender();
	else
		rren.render_frames(ren, 0, lastframe);
}


static Viewpoint viewpoint;
static struct Drag {bool down; int xprev, yprev;} drag;

static bool reload () {
	string filename = stringf(ren->infile, ren->frame);
	bool success = load_state(filename, ren->modeldir, ren->sim);
	cout << "Checking consistency of simulation particles..." << endl;
	sim_consistency_check(ren->sim.sand, 2);
	rren.move_points(*ren);
	if (success) glutSetWindowTitle(filename.c_str());
	else cout << "Failed to load file: " << filename << endl;
	return success;
}

static void colormap (float y) {
	glColor3f(0.5f+y*1.f,0.3f+y*0.6f,0.1f+y*0.2f);
}

static void draw_particle (const Particle<3> &p, float y1) {
	glPushMatrix();
	//colormap(p.x[1]/y1);
	float r_rms = sqrt((p.r2(0,0) + p.r2(1,1) + p.r2(2,2))/3.f);
	glTranslatef(p.x[0], p.x[1], p.x[2]);
	// Vec3f eigenVals, eigenVecs[3];
	// Mat3f m = decomposeMatrix(p.r2, eigenVals, eigenVecs);
	// eigenVals[0] = sqrtf(eigenVals[0]);  eigenVals[1] = sqrtf(eigenVals[1]);  eigenVals[2] = sqrtf(eigenVals[2]);
	// Mat3f mn = Mat3f(2 * m * diagonal(eigenVals) * trans(m));
	// float mat[16] = {mn(0,0), mn(1,0), mn(2,0), 0,   mn(0,1), mn(1,1), mn(2,1), 0,   mn(0,2), mn(1,2), mn(2,2), 0,   0, 0, 0, 1};
	// glMultMatrixf(mat);
	glScalef(r_rms*2, r_rms*2, r_rms*2);
	// glutSolidSphere(1, 8, 4);
	glCallList(dl_id);
	glPopMatrix();
}

static void draw_particle (const RenderParticle &p, float y1) {
	glBegin(GL_POINTS);
	colormap(p.pos[1]/y1);
	glVertex3f(p.pos[0],p.pos[1],p.pos[2]);
	glEnd();
}

static void display () {
	Sand<3> &sand = ren->sim.sand;

	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);
	//glDepthFunc(GL_LEQUAL);

	if(renderSimParticles) {
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_NORMALIZE);
	}
	else {
		glDisable(GL_LIGHTING);
	}
	//glPointSize(2);

	glLoadIdentity();
	glTranslatef(0,0,-viewpoint.dist);
	glRotatef(viewpoint.lat, 1,0,0);
	glRotatef(viewpoint.lon, 0,1,0);

	float scale = fmaxf(fmaxf(sand.m, sand.n), sand.o)*sand.dx;
	glScalef(1/scale, 1/scale, 1/scale);
	glTranslatef(-sand.m*sand.dx/2, -sand.n*sand.dx/2, -sand.o*sand.dx/2);

	float ambient[] = {0.6,0.6,0.6,1};
	float diffuse0[] = {0.4,0.4,0.4,1};
	float diffuse1[] = {0.2,0.2,0.2,1};
	float specular[] = {0,0,0,1};
	float pos0[] = {3,4,5,0};
	float pos1[] = {-3,-4,-5,0};
	if (renderSimParticles) {
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
		glEnable(GL_LIGHT0);
		glLightfv(GL_LIGHT0, GL_POSITION, pos0);
		glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse0);
		glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
		glEnable(GL_LIGHT1);
		glLightfv(GL_LIGHT1, GL_POSITION, pos1);
		glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse1);
		glLightfv(GL_LIGHT1, GL_SPECULAR, specular);
	}
	else {
		glDisable(GL_LIGHT0);
		glDisable(GL_LIGHT1);
	}

	glPushMatrix();
	glScalef(sand.m * sand.dx, sand.n * sand.dx, sand.o * sand.dx);
	glColor3f(0,0,0);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0,0,0);
	glVertex3f(1,0,0);
	glVertex3f(1,1,0);
	glVertex3f(0,1,0);
	glVertex3f(0,0,0);
	glVertex3f(0,0,1);
	glVertex3f(1,0,1);
	glVertex3f(1,1,1);
	glVertex3f(0,1,1);
	glVertex3f(0,0,1);
	glEnd();
	glBegin(GL_LINES);
	glVertex3f(1,0,0);
	glVertex3f(1,0,1);
	glVertex3f(1,1,0);
	glVertex3f(1,1,1);
	glVertex3f(0,1,0);
	glVertex3f(0,1,1);
	glEnd();
	glPopMatrix();

	if (renderSimParticles) {
		vector<Particle<3> > &particles = sand.particles;
		int np = particles.size();
		int h = glutGet(GLUT_WINDOW_HEIGHT);
		//glPointSize(2*sand.r/scale * h/(3*tan(M_PI/8)));
		for (int i = 0; i < np; i++) {
			Particle<3> &p = particles[i];
			Vec3f newpx(p.x + Vec3f(rren.distfield.dx, rren.distfield.dx, rren.distfield.dx));
			float distf = fabsf (lerp(newpx, rren.distfield) * rren.distfield.dx);
			float dens = lerp(p.x, sand.rho);
			bool condition = (dens < sand.rho_max * rren.cutoff_density) || (distf <= rren.cutoff_distance);
			if (condition)
				glColor3f(0.9f, 0.1f, 0.1f);
			else
				glColor3f(0.1f, 0.1f, 0.9f);
			draw_particle(p, sand.n*sand.dx);
		}
	}
	else {
		const vector<RenderParticle > &particles = rren.render_particles;
		int np = particles.size();
		int h = glutGet(GLUT_WINDOW_HEIGHT);
		//glPointSize(2*rren.render_width/scale * h/(3*tan(M_PI/8)));
		//glPointSize(2*sand.r/scale * h/(3*tan(M_PI/8)));
		glPointSize(rren.render_width);
		for (int p = 0; p < np; p++) {
			draw_particle(particles[p], sand.n*sand.dx);
		}
	}

	glutSwapBuffers();
}

static void reshape (int w, int h) {
	glViewport(0,0, w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	float aspect = (float) w/h;
	gluPerspective(45, aspect, 1,10);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void keyboard (unsigned char key, int x, int y) {
	switch (key) {
		case 'q':
		case 'Q':
			exit(0);
			break;
		case 'p':
		case 'P':
			cout << "(" << viewpoint.lat << " , " << viewpoint.lon << ")" << endl;
			break;
		case 's':
		case 'S':
			renderSimParticles = !renderSimParticles;
	}
	glutPostRedisplay();
}

static void special (int key, int x, int y) {
	switch (key) {
		case GLUT_KEY_LEFT:
			ren->frame--;
			if (!reload()) ren->frame++;
			break;
		case GLUT_KEY_RIGHT:
			ren->frame++;
			if (!reload()) ren->frame--;
			break;
		case GLUT_KEY_HOME:
			ren->frame = 0;
			reload();
			break;
	}
	glutPostRedisplay();
}

static void mouse (int button, int state, int x, int y) {
	if (button == GLUT_LEFT_BUTTON) {
		drag.down = (state==GLUT_DOWN);
		drag.xprev = x;
		drag.yprev = y;
	}
}

static void motion (int x, int y) {
	if (drag.down) {
		float dx = x - drag.xprev;
		float dy = y - drag.yprev;
		viewpoint.lat += dy*0.2f; // degrees per pixel
		viewpoint.lon += dx*0.2f; // degrees per pixel
		drag.xprev = x;
		drag.yprev = y;
	}
	glutPostRedisplay();
}

void glutrender () {
	//::ren = &ren;
	viewpoint.lat = viewpoint.lon = 0;
	viewpoint.dist = 2;
	int argc = 1;
	char argv0[] = "STFU, GLUT";
	char *argv[] = {argv0};
	glutInit(&argc, argv);
	glutInitWindowSize(800, 800);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow("");
	dl_id = glGenLists(1);
	glNewList(dl_id, GL_COMPILE);
	glutSolidSphere(1, 8, 4);
	glEndList();


	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutReshapeFunc(reshape);
	glutSpecialFunc(special);
	if (!reload()) cout << "Couldn't load frame " << ::ren->frame << endl << yield;
	glutMainLoop();
}
