/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef __RENDER3R_H__
#define __RENDER3R_H__

#include <vector>
#include <string>
#include <list>
#include "renderer.hpp"
#include "rihelpers.hpp"

#ifdef __GNUC__
#include <ext/hash_map>
namespace std
{
	using namespace __gnu_cxx;
}
using __gnu_cxx::hash_map;
#else
#include <hash_map>
using stdext::hash_map;
#endif

/**
 * How this renderer works, a short story:
 *
 * Utilizes existing Renderer<3> to load simulation data files
 * A list of simulation particles is kept, each simulation particle has it's own list of render particles (indices)
 * This allows book-keeping so spurious render particles can be removed
 * 
 * Render particles are kept in their own vector, with 3D coordinates only - so this can be passed directly to
 * Renderman to render
 *
 *
 * Functions needed
 * - Sample render particles to each simulation particles
 * - One function to iterate over all sim particles which will
 * -- Advect all render particles
 * -- Make sure particle indices remain consistent over merges/splits
 * -- Remove and regenerate any spurious render particles
 * -- Future: Take care of multiple LOD rendering
 * - One function to issue actual Renderman calls
 *
 * Data needed
 * - Renderer<3> &ren - for data i/o
 * - vector<RenderParticles> - List of 3D coordinates of render particles
 * - vector<SimParticles> - List of simulation particles - indices must be matched to Renderer<3>'s particle list
 */

struct RenderParticle {
	float pos[3];
};

typedef std::pair <float, float> RenderNormalJitter;
typedef std::pair<RenderParticle, RenderNormalJitter> RenderSet;

typedef RenderParticle RenderNormal;
bool normalize(RenderNormal &n);

struct SimParticle {
	std::vector<int> renderParticleIndices;
};

struct Viewpoint {
	float lat, lon, dist;
}; 

struct ShadowMap {
	float from[3], to[3];
	std::string outputfile;
};

enum NormalRenderingMode {FULL, NONE, NOJITTER};

class RendermanRenderer {
	public:
		/// Variables
		/// RIB files for external initialization - will be specified only through XML setup file
		std::string						ribinitfile, ribworldfile, ribdsmworldfile, ribdsminitfile;
		int							imageres[2], shadowmapres;
		/// Output file template
		std::string						outfile;
		/// Render particle buckets
		BucketField<3>						render_particle_buckets;
		/// List of render particles
		std::vector<RenderParticle>				render_particles;
		std::vector<RenderNormal>				render_normals;
		std::vector< RenderNormalJitter >			normal_jitters;
		/// List of simulation particles
		std::vector<SimParticle>				sim_particles;
		/// Map from render particles to simulation particles
		hash_map<int, int>					render_sim_map;
		/// Tolerance - specifies grace distance for which render particle is valid while outside sim particle
		float							tolerance;
		/// Camera Viewpoint
		//Viewpoint						viewpoint;
		/// Camera Position
		float							camposition[3], camto[3];
		float							fov, shmapfov;
		std::string							shmap_temp;
		int							fps;
		RtColor							sandcolor, sandopacity;
		/// Shadow map light list
		std::vector< std::pair<int, ShadowMap> >		shmaplist;
		NormalRenderingMode					usenormals;
		std::string						envmapname;
		float							envmappos[3];
		int							envmapres[2];
		/// Render particle radii range
		float							rmin, rmax;
		int							minRP;
		/// Render particle render radius
		float							render_width;
		/// Gradient field
		VectorField<3>						grad;
		/// Normal jitter scaling
		float							nJitter;
		/// Number of gradient samples
		int							nGradientSamples;
		/// Current render time
		float							rendertime;
		/// Signed distance field representing approximate surface
		ScalarField<3>						distfield, rasterfield;
		/// Cutoff distance from surface for sampling
		float							cutoff_distance, cutoff_density;
		int							distfield_supersampling;
		float							fill_fraction;
		RendermanSceneGraph					sgraph;
		bool							wireframe;
		bool							samplingtol;

	public:
		RendermanRenderer();
		/// Functions
		/// Load XML config file for render
		void							load_setup(const std::string &filename);
		/// Render frames in range start_frame to end_frame, calls below 2 functions - main function to call
		void							render_frames(Renderer<3> &ren, int start_frame, int end_frame);
		void							render_points(Renderer<3> &ren);
		/// Generate and move points as needed
		void							move_points(Renderer<3> &ren);
		RenderSet						generate_render_particles(Renderer<3> &ren, int sim_particle, Vec3f eigenVals, Vec3f eigenVecs[3], Mat3f evMat, Vec3f center, float rad, std::vector<RenderSet> &spill, bool usetolerance = false);
		bool							isNotIntersecting(RenderParticle &p, float rad, std::vector<int> &rpIndices);
		void							set_viewpoint(Viewpoint v);
		void							set_tolerance(float tol);
		/// Build distance field - from density, to find approximate surface definition
		void							build_distance_field(Renderer<3> &ren);
		bool							consistency_check(Renderer<3> &ren, bool enable_mesgs = false, bool fix_errors = false);
};

#endif
