/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "renderer.hpp"
#include <GL/glut.h>
#include <vector>
#include "io.hpp"
#include "iohelpers.hpp"
#include "marching.hpp"
using namespace std;

static Renderer<2> *ren;
static struct Visibility {bool particles, density, grid, surface, bodies;} visible;

RigidBody<2> body; // DEBUG

float zoom;

static bool reload () {
	string filename = stringf(ren->infile, ren->frame);
	bool success = load_state(filename, ren->modeldir, ren->sim);
	if (success) glutSetWindowTitle(filename.c_str());
	return success;
}

static void draw_particle (const Particle<2> &p, float dt = 0) {
	static float x[12] = {1, sqrtf(3)/2, 0.5f, 0, -0.5f, -sqrtf(3)/2, -1, -sqrtf(3)/2, -0.5f, 0, 0.5f, sqrtf(3)/2};
	static float y[12] = {0, 0.5f, sqrtf(3)/2, 1, sqrtf(3)/2, 0.5f, 0, -0.5f, -sqrtf(3)/2, -1, -sqrtf(3)/2, -0.5f};
	Vec2f evs;
	Mat2f m = decomposeMatrix(p.r2,evs);
	evs[0] = sqrtf(evs[0]);		evs[1] = sqrtf(evs[1]);
	//Mat2f(m*diagonal(Vec2f(sqrtf(e1),sqrtf(e2)))*trans(m))
	Mat2f r = Mat2f(m * diagonal(evs) * trans(m));
	//Mat2f r = radius(p.r2);
	glColor4f(0.5,0.3,0.1,0.5);
	glBegin(GL_LINE_LOOP);
	for (int i = 0; i < 12; i++) {
		Vec2f v(p.x + r*Vec2f(x[i],y[i]));
		glVertex2f(v[0],v[1]);
	}
	glEnd();
	glBegin(GL_LINES);
	glColor4f(0.5,0.3,0.1,0);
	glVertex2f(p.x[0]-p.v[0]*dt,p.x[1]-p.v[1]*dt);
	glColor4f(0.5,0.3,0.1,1);
	glVertex2f(p.x[0],p.x[1]);
	glEnd();
}

static void draw_arrow (const Vec2f &v) {
	float s = norm2(v);
	float t = atan2(v[1],v[0]);
	glPushMatrix();
	glRotatef(t*180/M_PI, 0,0,1);
	glScalef(s,s,1);
	glBegin(GL_QUADS);
	glVertex2f(-0.5f, -0.1f);
	glVertex2f(0.f, -0.1f);
	glVertex2f(0.f, 0.1f);
	glVertex2f(-0.5f, 0.1f);
	glEnd();
	glBegin(GL_TRIANGLES);
	glVertex2f(0.5f, 0.f);
	glVertex2f(0.f, 0.3f);
	glVertex2f(0.f, -0.3f);
	glEnd();
	glPopMatrix();
}

static void colormap (float x) {
	if (x > 0)      glColor3f(clampf(1-x*0.5f,0.5f,1), 1-x, 1-x*2);
	else if (x < 0) glColor3f(1+x*2, 1+x, clampf(1+x*0.5f,0.5f,1));
	else            glColor3f(1,1,1);
}

extern void pressure_J (const Sand<2> &sand, const RigidBody<2> &body, vector< Screw<2> > &J);

static void display () {
	Sand<2> &sand = ren->sim.sand;

	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	//glLineWidth(2);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScalef(zoom, zoom, 1);
	glTranslatef(-sand.m*sand.dx/2, -sand.n*sand.dx/2, 0);

	// debug pressure
	if (true) {
		glBegin(GL_QUADS);
		for (int j = 0; j < sand.n; j++) for (int i = 0; i < sand.m; i++) {
			float s = sand.aux.p(i+1,j+1);
			s /= sand.rho_max*norm2(ren->sim.g)*sand.m*sand.dx;
			if (s > 0) glColor3f(1-0.25*s, 1-s, 1-4*s);
			else glColor3f(1+4*s, 1+s, 1+0.25*s);
			glVertex2f(i*sand.dx,     j*sand.dx);
			glVertex2f((i+1)*sand.dx, j*sand.dx);
			glVertex2f((i+1)*sand.dx, (j+1)*sand.dx);
			glVertex2f(i*sand.dx,     (j+1)*sand.dx);
		}
		glEnd();
	}

	// debug stress
	if (false) {
		glBegin(GL_QUADS);
		for (int j = 0; j < sand.n; j++) for (int i = 0; i < sand.m; i++) {
			float s = sand.aux.s.xx(i,j);
			s /= sand.rho_max*norm2(ren->sim.g)*sand.m*sand.dx;
			if (s > 0) glColor3f(1-0.25*s, 1-s, 1-4*s);
			else glColor3f(1+4*s, 1+s, 1+0.25*s);
			glVertex2f(i*sand.dx,     j*sand.dx);
			glVertex2f((i+1)*sand.dx, j*sand.dx);
			glVertex2f((i+1)*sand.dx, (j+1)*sand.dx);
			glVertex2f(i*sand.dx,     (j+1)*sand.dx);
		}
		glEnd();
		glTranslatef(-sand.dx/2,-sand.dx/2,0);
		glBegin(GL_QUADS);
		for (int j = 0; j <= sand.n; j++) for (int i = 0; i <= sand.m; i++) {
			float s = sand.aux.s.xy(i,j);
			s /= 10000;
			if (s > 0) glColor4f(1-2*s, 1-0.25*s, 1-2*s, 0.5);
			else glColor4f(1+0.5*s, 1+4*s, 1+0.5*s, 0.5);
			glVertex2f(i*sand.dx,     j*sand.dx);
			glVertex2f((i+1)*sand.dx, j*sand.dx);
			glVertex2f((i+1)*sand.dx, (j+1)*sand.dx);
			glVertex2f(i*sand.dx,     (j+1)*sand.dx);
		}
		glEnd();
		glTranslatef(sand.dx/2,sand.dx/2,0);
	}

	// debug coupling
	if (false) {
		if (ren->sim.bodies.size() > 0) {
			glColor3f(0,0,0);
			const RigidBody<2> &body = ren->sim.bodies[0];
			int pm = sand.m + 2, pn = sand.n + 2;
			vector< Screw<2> > J(pm*pn);
			pressure_J (sand, body, J);
			for (int j = 0; j < sand.n; j++) for (int i = 0; i < sand.m; i++) {
				Screw<2> s = J[(i+1)+pm*(j+1)];
				Vec2f v(lin(s)*sand.dx);
				glPushMatrix();
				glTranslatef((i+0.5f)*sand.dx, (j+0.5f)*sand.dx, 0);
				//draw_arrow(v);
				glutSolidCube(ang(s)*sand.dx);
				glPopMatrix();
			}
		}
	}

	if (visible.density) {
		glBegin(GL_QUADS);
		for (int j = 0; j < sand.n; j++) for (int i = 0; i < sand.m; i++) {
			float rr = sand.rho(i,j)/sand.rho_max;
			/*if (rr > 1) glColor3f(1-rr*0.2, 0.8-rr*0.4, 0);
			  else*/ glColor3f(1-rr*0.2, 1-rr*0.4, 1-rr*0.6);
			glVertex2f(i*sand.dx,     j*sand.dx);
			glVertex2f((i+1)*sand.dx, j*sand.dx);
			glVertex2f((i+1)*sand.dx, (j+1)*sand.dx);
			glVertex2f(i*sand.dx,     (j+1)*sand.dx);
		}
		glEnd();
	}

	if (visible.bodies) {
		for (int b = 0; b < ren->sim.bodies.size(); b++) {
			const RigidBody<2> &body = ren->sim.bodies[b];
			glPushMatrix();
			const Vec2f &com = body.pose.x;
			glTranslatef(com[0], com[1], 0);
			glRotatef(body.pose.q*180/M_PI, 0,0,1);
			glColor3f(0.6,0.6,0.6);
			glTranslatef(-body.origin[0], -body.origin[1], 0);
			// glBegin(GL_LINE_LOOP);
			// glVertex3f(0,0,0);
			// glVertex3f(body.dist.m*body.dist.dx,0,0);
			// glVertex3f(body.dist.m*body.dist.dx,body.dist.n*body.dist.dx,0);
			// glVertex3f(0,body.dist.n*body.dist.dx,0);
			// glEnd();
			glPopMatrix();
			glBegin(GL_QUADS);
			for (int j = 0; j < sand.n; j++) for (int i = 0; i < sand.m; i++) {
				Vec2f x0(i*sand.dx,j*sand.dx), x1((i+1)*sand.dx, (j+1)*sand.dx);
				float f = coverage(body, x0,x1);
				glColor4f(0.6,0.6,0.6, f);
				glVertex2f(i*sand.dx,     j*sand.dx);
				glVertex2f((i+1)*sand.dx, j*sand.dx);
				glVertex2f((i+1)*sand.dx, (j+1)*sand.dx);
				glVertex2f(i*sand.dx,     (j+1)*sand.dx);
			}
			glEnd();
		}
	}

	// vector body
	if (false) {
		for (int b = 0; b < ren->sim.bodies.size(); b++) {
			const RigidBody<2> &body = ren->sim.bodies[b];
			glPushMatrix();
			const Vec2f &com = body.pose.x;
			glTranslatef(com[0], com[1], 0);
			glRotatef(body.pose.q*180/M_PI, 0,0,1);
			glColor3f(0.6,0.6,0.6);
			glLineWidth(3);
			glBegin(GL_LINE_LOOP);
			float r = 0.1;
			for (int i = 0; i < 50; i++) {
				glVertex2f(r*cos(2*M_PI*i/50), r*sin(2*M_PI*i/50));
			}
			glEnd();
			glBegin(GL_LINES);
			glVertex2f(0,0);
			glColor4f(0.6,0.6,0.6,0);
			Vec2f v(0.0055f,-0.010f);
			cout << com << endl;
			glVertex2f(-v[0], -v[1]);
			glEnd();
			glLineWidth(1.5);
			glPopMatrix();
		}
	}

	// cute arrows
	if (true) {
		glEnable(GL_POLYGON_SMOOTH);
		for (int j = 0; j <= sand.n; j++) for (int i = 0; i <= sand.m; i++) {
			Vec2f x(i*sand.dx,j*sand.dx);
			Vec2f v(lerp_weighted(x, /*sand.v*/sand.aux.dp, sand.rho_s)*5);
			//v += Vec2f(0.f,-9.81f)*0.0334;
			//v += 0.5f*(lerp_weighted(x, sand.aux.dp, sand.rho_s) + lerp_weighted(x, sand.aux.ds, sand.rho_s));
			v *= ren->sim.dt;
			v *= sqrt(norm2(v)*sand.dx)/norm2(v);
			glColor4f(0,0,0, lerp(x,sand.rho)/sand.rho_max);
			glPushMatrix();
			glTranslatef(i*sand.dx, j*sand.dx, 0);
			draw_arrow(v);
			glPopMatrix();
		}
		glDisable(GL_POLYGON_SMOOTH);
	}

	if (visible.grid) {
		glColor3f(0.9,0.9,0.9);
		glBegin(GL_LINES);
		for (int i = 0; i <= sand.m; i++) {
			glVertex2f(i*sand.dx, 0);
			glVertex2f(i*sand.dx, sand.n*sand.dx);
		}
		for (int j = 0; j <= sand.n; j++) {
			glVertex2f(0,              j*sand.dx);
			glVertex2f(sand.m*sand.dx, j*sand.dx);
		}
		glEnd();
		glColor3f(0.5,0.5,0.5);
		glBegin(GL_LINE_LOOP);
		glVertex2f(0, 0);
		glVertex2f(0, sand.n*sand.dx);
		glVertex2f(sand.m*sand.dx, sand.n*sand.dx);
		glVertex2f(sand.m*sand.dx, 0);
		glEnd();
	}

	if (visible.particles) {
		const vector<Particle<2> > &particles = sand.particles;
		int np = particles.size();
		for (int p = 0; p < np; p++) {
			draw_particle(particles[p], ren->sim.dt);
		}
	}
	
	if (visible.surface) {
		vector<Edge2D> edges = extract_surface(sand.rho, sand.rho_max, sand.aux.lambda_sep);
		int ne = edges.size();
		glColor3f(0,0,0);
		glTranslatef(sand.dx/2,sand.dx/2,0);
		glBegin(GL_LINES);
		for (int e = 0; e < ne; e++) {
			glVertex2f(edges[e][0][0], edges[e][0][1]);
			glVertex2f(edges[e][1][0], edges[e][1][1]);
		}
		glEnd();
		glTranslatef(-sand.dx/2,-sand.dx/2,0);
	}

	glutSwapBuffers();

	GLenum error = glGetError();
	if (error != GL_NO_ERROR)
		cout << "OpenGL error: " << gluErrorString(error) << endl;

}

static void reshape (int w, int h) {
	glViewport(0,0, w,h);
	float aspect = (float) w/h;
	float x1 = ren->sim.sand.m*ren->sim.sand.dx, y1 = ren->sim.sand.n*ren->sim.sand.dx;
	if (aspect > (float)x1/y1) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(-aspect,aspect, -1,1, -1,1);
		glScalef(2/y1, 2/y1, 2/y1);
	} else {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(-1,1, -1/aspect,1/aspect, -1,1);
		glScalef(2/x1, 2/x1, 2/x1);
	}
}

static void keyboard (unsigned char key, int x, int y) {
	switch (key) {
	case 'q':
	case 'Q':
		exit(0);
	break;
	case 'r':
	case 'R':
		reload();
	break;
	case 'p':
	case 'P':
		visible.particles ^= true;
	break;
	case 'g':
	case 'G':
		visible.grid ^= true;
	break;
	case 'd':
	case 'D':
		visible.density ^= true;
	break;
	case 's':
	case 'S':
		visible.surface ^= true;
	break;
	case 'b':
	case 'B':
		visible.bodies ^= true;
	break;
	}
	glutPostRedisplay();
}

static void special (int key, int x, int y) {
	switch (key) {
	case GLUT_KEY_LEFT:
		ren->frame--;
		if (!reload()) ren->frame++;
		break;
	case GLUT_KEY_RIGHT:
		ren->frame++;
		if (!reload()) ren->frame--;
		break;
	case GLUT_KEY_HOME:
		ren->frame = 0;
		reload();
		break;
	}
	glutPostRedisplay();
}

static void mouse (int button, int state, int x, int y) {
	if (button == /*GLUT_WHEEL_UP*/ 3 && state == GLUT_UP) {
		zoom /= 0.8;
		glutPostRedisplay();
	} else if (button == /*GLUT_WHEEL_DOWN*/ 4 && state == GLUT_UP) {
		zoom *= 0.8;
		glutPostRedisplay();
	}
}

static void debug () {
}

template <> void render (Renderer<2> &ren) {
	::ren = &ren;
	ren.sim.bt = NULL;
	visible.particles = visible.density = visible.grid = visible.surface = visible.bodies = true;
	zoom = 1.f;
	int argc = 1;
	char argv0[] = "STFU, GLUT";
	char *argv[] = {argv0};
	glutInit(&argc, argv);
	glutInitWindowSize(800, 800);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutCreateWindow("");
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutReshapeFunc(reshape);
	glutSpecialFunc(special);
	if (!reload()) cout << "Couldn't load frame " << ::ren->frame << endl << yield;
	glutMainLoop();
}
