/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "renderer.hpp"
#include <GL/glut.h>
#include "io.hpp"
#include "iohelpers.hpp"
using namespace std;

static Renderer<3> *ren;
static Viewpoint viewpoint;
static struct Drag {bool down; int xprev, yprev;} drag;
static struct Visibility {bool particles, density, bodies, antibodies;} visible;

int debug_i = 0, debug_j = 0, debug_k = 0;

int dl_id;

#define RENDER_HACK 0

bool screenshooting = false;

#include <png.h>
void png_write (const char *filename, int width, int height, const unsigned char *pixels, bool has_alpha = false) {
   FILE* file = fopen(filename, "wb");
   if (!file) {
      printf("Couldn't open file %s for writing.\n", filename);
      return;
   }
   png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
   if (!png_ptr) {
      printf("Couldn't create a PNG write structure.\n");
      fclose(file);
      return;
   }
   png_infop info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr) {
      printf("Couldn't create a PNG info structure.\n");
      png_destroy_write_struct(&png_ptr, NULL);
      fclose(file);
      return;
   }
   if (setjmp(png_jmpbuf(png_ptr))) {
      printf("Had a problem writing %s.\n", filename);
      png_destroy_write_struct(&png_ptr, &info_ptr);
      fclose(file);
      return;
   }
   png_init_io(png_ptr, file);
   png_set_filter(png_ptr, 0, PNG_ALL_FILTERS);
   png_set_compression_level(png_ptr, Z_BEST_SPEED);
   png_set_IHDR(png_ptr, info_ptr, width, height, 8, has_alpha ? PNG_COLOR_TYPE_RGBA : PNG_COLOR_TYPE_RGB,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
   int channels = has_alpha ? 4 : 3;
   png_bytep* row_pointers = (png_bytep*) new unsigned char*[height];
   for (int y = 0; y < height; y++)
      row_pointers[y] = (png_bytep) &pixels[(height-1-y)*width*channels];
   png_set_rows(png_ptr, info_ptr, row_pointers);
   png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);
   delete[] row_pointers;
   png_destroy_write_struct(&png_ptr, &info_ptr);
   fclose(file);
}
bool SaveScreenGrabPng(const char* filename, unsigned sw, unsigned sh) {

	// get some info about the screen size
	//unsigned sw      =  800;//glutGet(GLUT_WINDOW_WIDTH);
	//unsigned sh      =  600;//glutGet(GLUT_WINDOW_HEIGHT);
	unsigned bpp     =  3;//glutGet(GLUT_WINDOW_RGBA) ? 4 : 3;
	GLenum   format  =  (bpp==4) ? GL_RGBA : GL_RGB;


	// allocate memory to store image data
	unsigned char* pdata = new unsigned char[sw*sh*bpp];

	// read from front buffer
	glReadBuffer(GL_FRONT);

	// read pixel data
	glReadPixels(0,0,sw,sh,format,GL_UNSIGNED_BYTE,pdata);

	// write data as a tga file
	bool ret = true;
	png_write(filename,sw,sh,pdata);

	// clean up
	delete [] pdata;

	// done
	return ret;
}

static bool reload () {
	string filename = stringf(ren->infile, ren->frame);
	bool success = load_state(filename, ren->modeldir, ren->sim);
	if (success) glutSetWindowTitle(filename.c_str());
	return success;
}

static void colormap (float y) {
	glColor3f(0.5f+y*1.f,0.3f+y*0.6f,0.1f+y*0.2f);
	//glColor3f(0.9f,0.65f,0.2f);
}

static void draw_particle (const Particle<3> &p, float y1, float rho_max) {
	glPushMatrix();
	colormap(p.x[1]/y1);
	//float r_rms = sqrt((p.r2(0,0) + p.r2(1,1) + p.r2(2,2))/3.f);
	float r = cbrtf(p.m/rho_max)*2/3.f;
	//float r = 0.01;
	r = max(r, ren->sim.sand.dx/10);
	glTranslatef(p.x[0], p.x[1], p.x[2]);
	// Vec3f eigenVals, eigenVecs[3];
	// Mat3f m = decomposeMatrix(p.r2, eigenVals, eigenVecs);
	// eigenVals[0] = sqrtf(eigenVals[0]);  eigenVals[1] = sqrtf(eigenVals[1]);  eigenVals[2] = sqrtf(eigenVals[2]);
	// Mat3f mn = Mat3f(2 * m * diagonal(eigenVals) * trans(m));
	// float mat[16] = {mn(0,0), mn(1,0), mn(2,0), 0,   mn(0,1), mn(1,1), mn(2,1), 0,   mn(0,2), mn(1,2), mn(2,2), 0,   0, 0, 0, 1};
	// glMultMatrixf(mat);
	glScalef(r, r, r);
	// glutSolidSphere(1, 8, 4);
	glCallList(dl_id);
	glPopMatrix();
}

static void display () {
	Sand<3> &sand = ren->sim.sand;

	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);
	//glDepthFunc(GL_LEQUAL);

	glEnable(GL_LIGHTING);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_NORMALIZE);

	//glPointSize(2);

	glLoadIdentity();
	glTranslatef(0,0,-viewpoint.dist);
	glRotatef(viewpoint.lat, 1,0,0);
	glRotatef(viewpoint.lon, 0,1,0);

	float scale = fmaxf(fmaxf(sand.m, sand.n), sand.o)*sand.dx;
	glScalef(1/scale, 1/scale, 1/scale);
	glTranslatef(-sand.m*sand.dx/2, -sand.n*sand.dx/2, -sand.o*sand.dx/2);

	float ambient[] = {0.6,0.6,0.6,1};
	float diffuse0[] = {0.4,0.4,0.4,1};
	float diffuse1[] = {0.2,0.2,0.2,1};
	float specular[] = {0,0,0,1};
	float pos0[] = {3,4,5,0};
	float pos1[] = {-3,-4,-5,0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, pos0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse0);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
	glEnable(GL_LIGHT1);
	glLightfv(GL_LIGHT1, GL_POSITION, pos1);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse1);
	glLightfv(GL_LIGHT1, GL_SPECULAR, specular);

	float x1 = sand.m*sand.dx, y1 = sand.n*sand.dx, z1 = sand.o*sand.dx;
	glColor3f(0,0,0);
	glBegin(GL_LINE_STRIP);
	glVertex3f(0, 0, 0);
	glVertex3f(x1,0, 0);
	glVertex3f(x1,y1,0);
	glVertex3f(0, y1,0);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 0, z1);
	glVertex3f(x1,0, z1);
	glVertex3f(x1,y1,z1);
	glVertex3f(0, y1,z1);
	glVertex3f(0, 0, z1);
	glEnd();
	glBegin(GL_LINES);
	glVertex3f(x1,0, 0);
	glVertex3f(x1,0, z1);
	glVertex3f(x1,y1,0);
	glVertex3f(x1,y1,z1);
	glVertex3f(0, y1,0);
	glVertex3f(0, y1,z1);
	glEnd();

	if (visible.particles) {
		const vector<Particle<3> > &particles = sand.particles;
		int np = particles.size();
		int h = glutGet(GLUT_WINDOW_HEIGHT);
		//glPointSize(2*sand.r/scale * h/(3*tan(M_PI/8)));
		for (int p = 0; p < np; p++) {
			draw_particle(particles[p], sand.n*sand.dx, sand.rho_max);
		}
	}

	// debug stress
	if (false) {
		glDisable(GL_LIGHTING);
		glBegin(GL_QUADS);
		float scale = 1.f/(sand.rho_max*norm2(ren->sim.g)*sand.m*sand.dx); // maximum pressure = rho g h
		int i = debug_i;
		for (int k = 0; k < sand.o; k++) for (int j = 0; j < sand.n; j++) {
			float s = sand.aux.p(i+1,j+1,k+1)*scale;
			if (s > 0) glColor3f(1-s*0.25, 1-s, 1-s*4);
			else glColor3f(1+s*4, 1+s, 1+s*0.25);
			glVertex3f((i+0.5f)*sand.dx,  j   *sand.dx,  k   *sand.dx);
			glVertex3f((i+0.5f)*sand.dx, (j+1)*sand.dx,  k   *sand.dx);
			glVertex3f((i+0.5f)*sand.dx, (j+1)*sand.dx, (k+1)*sand.dx);
			glVertex3f((i+0.5f)*sand.dx,  j   *sand.dx, (k+1)*sand.dx);
		}
		int j = debug_j;
		for (int k = 0; k < sand.o; k++) for (int i = 0; i < sand.m; i++) {
			float s = sand.aux.p(i+1,j+1,k+1)*scale;
			if (s > 0) glColor3f(1-s*0.25, 1-s, 1-s*4);
			else glColor3f(1+s*4, 1+s, 1+s*0.25);
			glVertex3f( i   *sand.dx, (j+0.5f)*sand.dx,  k   *sand.dx);
			glVertex3f( i   *sand.dx, (j+0.5f)*sand.dx, (k+1)*sand.dx);
			glVertex3f((i+1)*sand.dx, (j+0.5f)*sand.dx, (k+1)*sand.dx);
			glVertex3f((i+1)*sand.dx, (j+0.5f)*sand.dx,  k   *sand.dx);
		}
		int k = debug_k;
		for (int j = 0; j < sand.n; j++) for (int i = 0; i < sand.m; i++) {
			float s = sand.aux.p(i+1,j+1,k+1)*scale;
			if (s > 0) glColor3f(1-s*0.25, 1-s, 1-s*4);
			else glColor3f(1+s*4, 1+s, 1+s*0.25);
			glVertex3f( i   *sand.dx,  j   *sand.dx, (k+0.5f)*sand.dx);
			glVertex3f((i+1)*sand.dx,  j   *sand.dx, (k+0.5f)*sand.dx);
			glVertex3f((i+1)*sand.dx, (j+1)*sand.dx, (k+0.5f)*sand.dx);
			glVertex3f( i   *sand.dx, (j+1)*sand.dx, (k+0.5f)*sand.dx);
		}
		glEnd();
		glEnable(GL_LIGHTING);
	}

	if (visible.bodies) {
		glColor3f(0.6,0.6,0.6);
		for (int b = 0; b < ren->sim.bodies.size(); b++) {
			const RigidBody<3> &body = ren->sim.bodies[b];
			const Vec3f &com = body.pose.x;
			const Quatf &q = body.pose.q;
			glPushMatrix();
			glTranslatef(com[0], com[1], com[2]);
			float a = 2*atan2((float)norm2(q.v), q.s);
			if (a > epsilon) {
				Vec3f axis(normalize(q.v));
				glRotatef(a*180/M_PI, axis[0],axis[1],axis[2]);
			}
			glScalef(body.dist.m*body.dist.dx, body.dist.n*body.dist.dx, body.dist.o*body.dist.dx);
			glutWireCube(1);
			glPopMatrix();
#if RENDER_HACK
			glPushMatrix();
			glTranslatef(com[0],com[1],com[2]);
			glScalef(0.1,0.1,0.1);
			glutSolidSphere(1, 16, 16);
			glPopMatrix();
#else
			float x0 = com[0] - body.r, x1 = com[0] + body.r;
			float y0 = com[1] - body.r, y1 = com[1] + body.r;
			float z0 = com[2] - body.r, z1 = com[2] + body.r;
			int i0 = floor(x0/sand.dx), i1 = ceil(x1/sand.dx);
			int j0 = floor(y0/sand.dx), j1 = ceil(y1/sand.dx);
			int k0 = floor(z0/sand.dx), k1 = ceil(z1/sand.dx);
			i0 = min(max(i0, 0), sand.m);
			i1 = min(max(i1, 0), sand.m);
			j0 = min(max(j0, 0), sand.n);
			j1 = min(max(j1, 0), sand.n);
			k0 = min(max(k0, 0), sand.o);
			k1 = min(max(k1, 0), sand.o);
			for (int k = k0; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
				Vec3f x0(i*sand.dx, j*sand.dx, k*sand.dx);
				Vec3f x1(x0 + Vec3f(sand.dx));
				float f = coverage(body, x0,x1);
				if (visible.antibodies) f = 1 - f;
				if (f > 0) {
					glPushMatrix();
					glTranslatef((i+0.5f)*sand.dx, (j+0.5f)*sand.dx, (k+0.5f)*sand.dx);
					glutSolidCube(sand.dx*sqrtf(f));
					glPopMatrix();
				}
			}
#endif
		}
	}

	// debug velocity
	if (false) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_LIGHTING);
		int k = debug_k;
		for (int j = 0; j <= sand.n; j++) for (int i = 0; i <= sand.m; i++) {
			Vec3f x(i*sand.dx,j*sand.dx,(k+0.5f)*sand.dx);
			Vec3f v = lerp_weighted(x, sand.v, sand.rho_s);
			//v += Vec2f(0.f,-9.81f)*0.0334;
			//v += 0.5f*(lerp_weighted(x, sand.aux.dp, sand.rho_s) + lerp_weighted(x, sand.aux.ds, sand.rho_s));
			glColor4f(0,0,0, lerp(x,sand.rho)/sand.rho_max);
			glPushMatrix();
			glTranslatef(i*sand.dx, j*sand.dx, (k+0.5f)*sand.dx);
			float s = norm2(v)*ren->sim.dt;
			float t = atan2(v[1],v[0]);
			glRotatef(t*180/M_PI, 0,0,1);
			glScalef(s,s,1);
			glBegin(GL_QUADS);
			glVertex3f(-0.5f, -0.1f, 0.f);
			glVertex3f(0.f, -0.1f, 0.f);
			glVertex3f(0.f, 0.1f, 0.f);
			glVertex3f(-0.5f, 0.1f, 0.f);
			glEnd();
			glBegin(GL_TRIANGLES);
			glVertex3f(0.5f, 0.f, 0.f);
			glVertex3f(0.f, 0.3f, 0.f);
			glVertex3f(0.f, -0.3f, 0.f);
			glEnd();
			glPopMatrix();
		}
		glEnable(GL_LIGHTING);
		glDisable(GL_BLEND);
	}

	glutSwapBuffers();

	if (screenshooting) {
		string filename = stringf("local/sshots/sshot%04d.png", ren->frame);
		SaveScreenGrabPng(filename.c_str(), glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT));
		ren->frame += 1;
		if (!reload()) {
			ren->frame -= 1;
			screenshooting = false;
		}
		glutPostRedisplay();
	}
}

static void reshape (int w, int h) {
	glViewport(0,0, w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	float aspect = (float) w/h;
	if (aspect > 1)
		gluPerspective(45, aspect, 0.1,10);
	else
		gluPerspective(45*aspect, aspect, 0.1,10);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void keyboard (unsigned char key, int x, int y) {
	switch (key) {
	case 'q':
	case 'Q':
		exit(0);
	break;
	case 'p':
	case 'P':
		visible.particles ^= true;
	break;
	case 'd':
	case 'D':
		visible.density ^= true;
	break;
	case 'b':
		visible.bodies ^= true;
	break;
	case 'B':
		visible.antibodies ^= true;
	break;
	case 'i':
		debug_i = min(max(debug_i + 1, 0), ren->sim.sand.m-1);
		break;
	case 'I':
		debug_i = min(max(debug_i - 1, 0), ren->sim.sand.m-1);
		break;
	case 'j':
		debug_j = min(max(debug_j + 1, 0), ren->sim.sand.n-1);
		break;
	case 'J':
		debug_j = min(max(debug_j - 1, 0), ren->sim.sand.n-1);
		break;
	case 'k':
		debug_k = min(max(debug_k + 1, 0), ren->sim.sand.o-1);
		break;
	case 'K':
		debug_k = min(max(debug_k - 1, 0), ren->sim.sand.o-1);
		break;
	case 's':
	case 'S':
		screenshooting = !screenshooting;
	break;
	}
	glutPostRedisplay();
}

static void special (int key, int x, int y) {
	switch (key) {
	case GLUT_KEY_LEFT:
		ren->frame--;
		if (!reload()) ren->frame++;
		break;
	case GLUT_KEY_RIGHT:
		ren->frame++;
		if (!reload()) ren->frame--;
		break;
	case GLUT_KEY_HOME:
		ren->frame = 0;
		reload();
		break;
	}
	glutPostRedisplay();
}

static void mouse (int button, int state, int x, int y) {
	if (button == GLUT_LEFT_BUTTON) {
		drag.down = (state==GLUT_DOWN);
		drag.xprev = x;
		drag.yprev = y;
	} else if (button == /*GLUT_WHEEL_UP*/ 3 && state == GLUT_UP) {
		viewpoint.dist *= 0.8;
		glutPostRedisplay();
	} else if (button == /*GLUT_WHEEL_DOWN*/ 4 && state == GLUT_UP) {
		viewpoint.dist /= 0.8;
		glutPostRedisplay();
	}
}

static void motion (int x, int y) {
	if (drag.down) {
		float dx = x - drag.xprev;
		float dy = y - drag.yprev;
		viewpoint.lat += dy*0.2f; // degrees per pixel
		viewpoint.lon += dx*0.2f; // degrees per pixel
		viewpoint.lat = clampf(viewpoint.lat, -90,90);
		drag.xprev = x;
		drag.yprev = y;
	}
	glutPostRedisplay();
}

template <> void render (Renderer<3> &ren) {
	::ren = &ren;
	ren.sim.bt = NULL;
	viewpoint.lat = viewpoint.lon = 0;
	viewpoint.dist = 2;
	visible.particles = visible.density = visible.bodies = true;
	int argc = 1;
	char argv0[] = "STFU, GLUT";
	char *argv[] = {argv0};
	glutInit(&argc, argv);
	glutInitWindowSize(800, 800);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow("");

	dl_id = glGenLists(1);
	glNewList(dl_id, GL_COMPILE);
	glutSolidSphere(1, 8, 4);
	glEndList();

	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutReshapeFunc(reshape);
	glutSpecialFunc(special);
	if (!reload()) cout << "Couldn't load frame " << ::ren->frame << endl << yield;
	glutMainLoop();
}
