#ifndef BLAS_WRAPPER_H
#define BLAS_WRAPPER_H

// Simplified BLAS wrapper (overloaded readable names)
// with stride==1 assumed, and for std::vector's.
// For the moment, no complex number support, and many routines have been dropped.

#include <vector>
#include <cblas.h>

namespace BLAS{

// dot products ==============================================================

inline float dotf(int n, const float *x, const float *y)
{ return cblas_sdot(n, x, 1, y, 1); }
//{
//	float val=0;
//	for(int i=0;i<n;i++)
//		val+=x[i]*y[i];
//	return val;
//}

inline float dotf(const std::vector<float> &x, const std::vector<float> &y)
{ return cblas_sdot((int)x.size(), &x[0], 1, &y[0], 1); }
//{
//	float val=0;
//	for(int i=0;i<(int)x.size();i++)
//		val+=x[i]*y[i];
//	return val;
//}

inline double dot(int n, const float *x, const float *y)
{ return cblas_dsdot(n, x, 1, y, 1); } // double precision calculation
//{
//	double val=0;
//	for(int i=0;i<n;i++)
//		val+=x[i]*y[i];
//	return val;
//}

inline double dot(const std::vector<float> &x, const std::vector<float> &y)
{ return cblas_dsdot((int)x.size(), &x[0], 1, &y[0], 1); } // double precision calculation
//{
//	double val=0;
//	for(int i=0;i<(int)x.size();i++)
//		val+=x[i]*y[i];
//	return val;
//}

inline double dot(int n, const double *x, const double *y)
{ return cblas_ddot(n, x, 1, y, 1); }
//{
//	double val=0;
//	for(int i=0;i<n;i++)
//		val+=x[i]*y[i];
//	return val;
//}

inline double dot(const std::vector<double> &x, const std::vector<double> &y)
{ return cblas_ddot((int)x.size(), &x[0], 1, &y[0], 1); }
//{
//	double val=0;
//	for(int i=0;i<(int)x.size();i++)
//		val+=x[i]*y[i];
//	return val;
//}

// 2-norm ====================================================================

template <class T>
T nm2(int n, const T *x) {
	T val=x[0]*x[0];
	for(int i=1;i<n;i++)
		val+=x[i]*x[i];
	return sqrtf((float)val);
}

template <class T>
T nm2(std::vector<T>x) {
	int n=(int)x.size();
	T val=x[0]*x[0];
	for(int i=1;i<n;i++)
		val+=x[i]*x[i];
	return sqrtf((float)val);
}

inline float norm2(int n, const float *x)
{ return cblas_snrm2(n, x, 1); }
//{ return nm2<float>(n,x); }

inline float norm2(const std::vector<float> &x)
{ return cblas_snrm2((int)x.size(), &x[0], 1); }
//{ return nm2<float>(x); }
//{
//	float norm=0;
//	for(int i=0;i<(int)x.size();i++)
//		norm+=x[i]*x[i];
//	return sqrtf(norm);
//}

inline double norm2(int n, const double *x)
{ return cblas_dnrm2(n, x, 1); }
//{ return nm2<double>(n,x); }

inline double norm2(const std::vector<double> &x)
{ return cblas_dnrm2((int)x.size(), &x[0], 1); }
//{ return nm2<double>(x); }
//{
//	float norm=0;
//	for(int i=0;i<(int)x.size();i++)
//		norm+=(float)(x[i]*x[i]);
//	return sqrt(norm);
//}

// 1-norm (sum of absolute values) ===========================================


template <class T>
T as(int n, const T *x)
{
	T val=fabs(x[0]);
	for(int i=1;i<n;i++)
		val+=fabs(x[i]);
	return val;
}
inline float abs_sum(int n, const float *x)
{ return cblas_sasum(n, x, 1); }

inline float abs_sum(const std::vector<float> &x)
{ return cblas_sasum((int)x.size(), &x[0], 1); }

inline double abs_sum(int n, const double *x)
{ return cblas_dasum(n, x, 1); }

inline double abs_sum(const std::vector<double> &x)
{ return cblas_dasum((int)x.size(), &x[0], 1); }

// inf-norm (maximum absolute value: index of max returned) ==================

template <class T>
int iom(int n, const T *x) {
	//T max=x[0];
	int i=0;
	for(int j=1;j<n;j++) {
		if(x[j]>x[i]) {
			i=j;
		}
	}
	return i;
}

template <class T>
int iom(const std::vector<T>x) {
	//T max=x[0];
	int n=(int)x.size();
	int i=0;
	for(int j=1;j<n;j++) {
		if(x[j]>x[i]) {
			i=j;
		}
	}
	return i;
}

inline int index_abs_max(int n, const float *x)
{ return (int)cblas_isamax(n, x, 1); }
//{ return iom<float>(n,x); }

inline int index_abs_max(const std::vector<float> &x)
{ return (int)cblas_isamax((int)x.size(), &x[0], 1); }
//{ return iom<float>(x); }
//{
//	int maxi=0;
//	for(int i=1;i<(int)x.size();i++) {
//		if(x[i]>x[maxi])
//			maxi=i;
//	}
//	return maxi;
//}

inline int index_abs_max(int n, const double *x)
{ return (int)cblas_idamax(n, x, 1); }
//{ return iom<double>(n,x); }

inline int index_abs_max(const std::vector<double> &x)
{ return (int)cblas_idamax((int)x.size(), &x[0], 1); }
//{ return iom<double>(x); }
//{
//	int maxi=0;
//	for(int i=1;i<(int)x.size();i++) {
//		if(x[i]>x[maxi])
//			maxi=i;
//	}
//	return maxi;
//}

// inf-norm (maximum absolute value) =========================================
// technically not part of BLAS, but useful

inline float abs_max(int n, const float *x)
{ return std::fabs(x[index_abs_max(n, x)]); }

inline float abs_max(const std::vector<float> &x)
{ return std::fabs(x[index_abs_max(x)]); }

inline double abs_max(int n, const double *x)
{ return std::fabs(x[index_abs_max(n, x)]); }

inline double abs_max(const std::vector<double> &x)
{ return std::fabs(x[index_abs_max(x)]); }

// saxpy (y=alpha*x+y) =======================================================

template <class T>
void ads(int n,T alpha, const T *x, T *y) {
	for(int i=0;i<n;i++)
		y[i]+=alpha*x[i];
}

template <class T>
void ads(T alpha, const std::vector<T>x, std::vector<T>y) {
	int n=(int)x.size();
	for(int i=0;i<n;i++)
		y[i]+=alpha*x[i];
}

inline void add_scaled(int n, float alpha, const float *x, float *y)
{ cblas_saxpy(n, alpha, x, 1, y, 1); }
//{ ads<float>(n,alpha,x,y); }

inline void add_scaled(float alpha, const std::vector<float> &x, std::vector<float> &y)
{ cblas_saxpy((int)x.size(), alpha, &x[0], 1, &y[0], 1); }
//{ ads<float>(alpha,x,y); }
//{
//	if(x.size()!=y.size()) {
//		fprintf(stderr,"Size mismatch.\n");
//		return;
//	}
//	for(int i=0;i<(int)x.size();i++) {
//		y[i]+=alpha*x[i];
//	}
//}

inline void add_scaled(int n, double alpha, const double *x, double *y)
{ cblas_daxpy(n, alpha, x, 1, y, 1); }
//{ ads<double>(n,alpha,x,y); }

inline void add_scaled(double alpha, const std::vector<double> &x, std::vector<double> &y)
{ cblas_daxpy((int)x.size(), alpha, &x[0], 1, &y[0], 1); }
//{ ads<double>(alpha,x,y); }
//{
//	if(x.size()!=y.size()) {
//		fprintf(stderr,"Size mismatch.\n");
//		return;
//	}
//	for(int i=0;i<(int)x.size();i++) {
//		y[i]+=alpha*x[i];
//	}
//}

// scale (x=alpha*x) =========================================================


inline void scale(int n, float alpha, float *x)
{ cblas_sscal(n, alpha, x, 1); }

inline void scale(float alpha, std::vector<float> &x)
{ cblas_sscal((int)x.size(), alpha, &x[0], 1); }

inline void scale(int n, double alpha, double *x)
{ cblas_dscal(n, alpha, x, 1); }

inline void scale(double alpha, std::vector<double> &x)
{ cblas_dscal((int)x.size(), alpha, &x[0], 1); }

};

#endif
