/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "dostal05.h"

using namespace std;

float infinity_norm (const SparseMatrix<float>& A) {
	float norm = 0;
	for (int i = A.n-1; i >= 0; i--) {
		float row_sum = 0;
		for (int j = A.value[i].size()-1; j >= 0; j--) {
			row_sum += fabsf(A.value[i][j]);
		}
		norm = fmaxf(norm, row_sum);
	}
	return norm;
}

#define ACTIVE(i) (fabs(ap[i]))
#define FREE(i) (1-ACTIVE(i))
#define IFACTIVE(i, ifactive, iffree) (ACTIVE(i)*(ifactive) + FREE(i)*(iffree))
#define POS(x) (fmaxf(x,0.f))
#define NEG(x) (fminf(x,-0.f))
#define CLAMPF(x,a,b) (fminf(fmaxf(x,a),b))
#define IN(v,i) (ap[i]*POS(ap[i]*v[i]))
#define OUT(v,i) (ap[i]*NEG(ap[i]*v[i]))

vector<float> var;

inline float projectedgmax (const vector<float> &r, const vector<float> &a) {
	float pgmax = 0;
	const float *rp = &r[0], *ap = &a[0];
	for (int i = (int)r.size()-1; i >= 0; i--) {
		float pg = fabsf(IFACTIVE(i, OUT(rp,i), rp[i]));
		pg *= var[i];
		pgmax = max(pgmax, pg);
	}
	return pgmax;
}

inline float choppedgnorm2 (const vector<float> &r, const vector<float> &a) {
	float cgnorm2 = 0;
	const float *rp = &r[0], *ap = &a[0];
	for (int i = (int)r.size()-1; i >= 0; i--) {
		float c = IFACTIVE(i, OUT(rp,i), 0);
		c *= var[i];
		cgnorm2 += c*c;
	}
	return cgnorm2;
}

inline float freedotrfreeg (const vector<float> &r, const vector<float> &x, float alpha_bar, const vector<float> &a, const vector<float> &xmin, const vector<float> &xmax) {
	float fdrf = 0;
	const float *rp = &r[0], *xp = &x[0], *ap = &a[0], *xminp = &xmin[0], *xmaxp = &xmax[0];
	for (int i = (int)r.size()-1; i >= 0; i--) {
		float f = IFACTIVE(i, 0, rp[i]);
		float rf = CLAMPF(f, (xp[i]-xmaxp[i])/alpha_bar, (xp[i]-xminp[i])/alpha_bar);
		fdrf += f*rf;
	}
	return fdrf;
}

inline void choppedg (const vector<float> &r, const vector<float> &a, vector<float> &c) {
	const float *rp = &r[0], *ap = &a[0];
	float *cp = &c[0];
	for (int i = (int)r.size()-1; i >= 0; i--) {
		cp[i] = IFACTIVE(i, OUT(rp,i), 0);
		cp[i] *= var[i];
	}
}

inline float max_steplength (const vector<float> &x, const vector<float> &p, const vector<float> &a, const vector<float> &xmin, const vector<float> &xmax) {
	float alpha = numeric_limits<float>::infinity();
	const float *xp = &x[0], *pp = &p[0], *ap = &a[0], *xminp = &xmin[0], *xmaxp = &xmax[0];
	debugPrintf("computing max steplength: ");
	for (int i = (int)x.size()-1; i >= 0; i--) {
		//debugPrintf("%f = (%f+%f)/max(%f,(float)0)  \t", ((1-fp[i])+xp[i])/max(pp[i],(float)0),(1-fp[i]),xp[i],pp[i]);
		float alpha_down = fabsf((ACTIVE(i)+(xp[i]-xminp[i]))/POS(pp[i]));
		float alpha_up = fabsf((ACTIVE(i)+(xmaxp[i]-xp[i]))/POS(-pp[i]));
		if (var[i]) alpha = fminf(alpha, fminf(alpha_down,alpha_up));
		if (alpha == -infinity) {
			cout << i << endl;
			cout << "alpha_down = ("<<ACTIVE(i)<<"+(" << xp[i]<<"-"<<xminp[i]<<"))/"<<POS(pp[i]) << endl;
			cout << "alpha_up = ("<<ACTIVE(i)<<"+("<<xmaxp[i]<<"-"<<xp[i]<<"))/"<<POS(-pp[i]) << endl;
			exit(1);
		}
	}
	return alpha;
}

void printVector (const char *name, const vector<float> &v) {
	mexPrintf("%s = [", name);
	for (int i = 0; i < (int) v.size(); i++) {
		mexPrintf("%0.3f\t", v[i]);
	}
	mexPrintf("]\n");
}

void printVector (const char *name, const vector<double> &v) {
	mexPrintf("%s = [", name);
	for (int i = 0; i < (int) v.size(); i++) {
		mexPrintf("%0.3f\t", v[i]);
	}
	mexPrintf("]\n");
}

inline void add_scaled_free(float alpha, const vector<float> &x, const vector<float> &a, const vector<float> &ymin, const vector<float> &ymax, vector<float> &y) {
	const float *xp = &x[0], *ap = &a[0], *yminp = &ymin[0], *ymaxp = &ymax[0];
	float *yp = &y[0];
	for (int i = (int)x.size()-1; i >= 0; i--) {
		yp[i] = CLAMPF(yp[i] + IFACTIVE(i, 0, alpha*xp[i]), yminp[i],ymaxp[i]);
	}
}

inline void project_to_bounds(const vector<float> &xmin, const vector<float> &xmax, vector<float> &x) {
	const float *xminp = &xmin[0], *xmaxp = &xmax[0];
	float *xp = &x[0];
	for (int i = (int)x.size()-1; i >= 0; i--) {
		xp[i] = CLAMPF(xp[i], xminp[i],xmaxp[i]);
	}
}

inline void recompute_active(const vector<float> &x, const vector<float> &xmin, const vector<float> &xmax, vector<float> &a) {
	const float *xp = &x[0], *xminp = &xmin[0], *xmaxp = &xmax[0];
	float *ap = &a[0];
	for (int i = (int)x.size()-1; i >= 0; i--) {
		ap[i] = xp[i]==xminp[i] ? 1 : xp[i]==xmaxp[i] ? -1 : 0;
	}
}

// perform result=matrix_aa*x_a
void multiply_free(const FixedSparseMatrix<float> &matrix, const vector<float> &x, const vector<float> &a, vector<float> &result)
{
	assert(matrix.n==x.size());
	result.resize(matrix.n);
	const float *ap = &a[0];
	for(unsigned int i=0; i<matrix.n; ++i){
		result[i]=0;
		for(unsigned int j=matrix.rowstart[i]; j<matrix.rowstart[i+1]; ++j){
			result[i] += FREE(i)*FREE(matrix.colindex[j]) * matrix.value[j]*x[matrix.colindex[j]];
		}
	}
}

// solve L*result=rhs
void solve_lower_free(const SparseColumnLowerFactor<float> &factor, const vector<float> &rhs, const vector<float> &a, vector<float> &result)
{
	assert(factor.n==rhs.size());
	assert(factor.n==result.size());
	result=rhs;
	const float *ap = &a[0];
	for(unsigned int i=0; i<factor.n; ++i){
		result[i] *= FREE(i)*factor.invdiag[i];
		for(unsigned int j=factor.colstart[i]; j<factor.colstart[i+1]; ++j){
			result[factor.rowindex[j]] -= FREE(i)*FREE(factor.rowindex[j]) * factor.value[j]*result[i];
		}
	}
}

// solve L^float*result=rhs
void solve_lower_transpose_in_place_free(const SparseColumnLowerFactor<float> &factor, vector<float> &x, const vector<float> &a)
{
	assert(factor.n==x.size());
	assert(factor.n>0);
	unsigned int i=factor.n;
	const float *ap = &a[0];
	do{
		--i;
		for(unsigned int j=factor.colstart[i]; j<factor.colstart[i+1]; ++j){
			x[i] -= FREE(i)*FREE(factor.rowindex[j]) * factor.value[j]*x[factor.rowindex[j]];
		}
		x[i] *= FREE(i)*factor.invdiag[i];
	}while(i!=0);
}

void form_preconditioner(const SparseMatrix<float>& A, SparseColumnLowerFactor<float> &M)
{
	factor_modified_incomplete_cholesky0(A, M);
}

void apply_preconditioner(const SparseColumnLowerFactor<float> &M, const vector<float> &x, vector<float> &y)
{
	solve_lower(M, x, y);
	solve_lower_transpose_in_place(M,y);
}

void apply_preconditioner_free(const SparseColumnLowerFactor<float> &M, const vector<float> &x, const vector<float> &a, vector<float> &y)
{
	solve_lower_free(M, x, a, y);
	solve_lower_transpose_in_place_free(M, y, a);
}

Dostal05Solver::Dostal05Solver(void)
{
	set_solver_parameters(1e-5, 100, 1.0, 0.97, 0.25);
}

void Dostal05Solver::set_solver_parameters(float tolerance_, int max_iterations_, float proportioning_factor_, float modified_incomplete_cholesky_parameter_, float min_diagonal_ratio_)
{
	tolerance=tolerance_;
	if(tolerance<1e-30) tolerance=1e-30;
	max_iterations=max_iterations_;
	proportioning_factor=proportioning_factor_;
	modified_incomplete_cholesky_parameter=modified_incomplete_cholesky_parameter_;
	min_diagonal_ratio=min_diagonal_ratio_;
}

void Dostal05Solver::set_matrix(const SparseMatrix<float> &A) {
	this->A = A;
	form_preconditioner(A, M);
	A_fixed.construct_from_matrix(A);
	alpha_bar = 1./infinity_norm(A);
}

void Dostal05Solver::set_rhs(const vector<float> &b) {
	this->b = b;
}

void Dostal05Solver::set_initial(const vector<float> &x) {
	this->x = x;
}

void Dostal05Solver::set_lower(const vector<float> &xmin) {
	this->xmin = xmin;
}

void Dostal05Solver::set_upper(const vector<float> &xmax) {
	this->xmax = xmax;
}

// void sanity_check (const vector<float> &xmin, const vector<float> &xmax) {
// 	for (int i = 0; i < xmin.size(); i++) {
// 		if (xmin[i] > 0) {
// 			cout << "xmin["<<i<<"] = " << xmin[i] << endl;
// 			exit(1);
// 		}
// 	}
// 	for (int i = 0; i < xmin.size(); i++) {
// 		if (xmax[i] < 0) {
// 			cout << "xmax["<<i<<"] = " << xmax[i] << endl;
// 			exit(1);
// 		}
// 	}
// }

bool Dostal05Solver::solve(float &residual_initial_out, float &residual_out, int &iterations_out, vector<float> *stats) 
{
	unsigned int n=A.n;

	if (x.size() != n) {x.assign(n, 0);}
	if (xmin.size() != n) {xmin.assign(n, 0);}
	if (xmax.size() != n) {xmax.assign(n, infinity);}
	project_to_bounds(xmin,xmax, x);

	var.assign(n,1);
	for (int i = 0; i < n; i++) if (xmin[i] == xmax[i]) var[i] = 0;

	r.resize(n);
	p.resize(n);
	z.resize(n);
	a.resize(n);
	pfg.resize(n);

	recompute_active(x,xmin,xmax,a);
	multiply(A_fixed, x, r);
	BLAS::add_scaled(-1, b, r);						// r = Ax - b
	apply_preconditioner_free(M, r, a, p);
	residual_initial_out = projectedgmax(r,a);

	int iteration;
	for(iteration=0; iteration<max_iterations; ++iteration){
		debugPrintf("iteration %d\n", iteration);
		debugPrintVector("x", x);
		debugPrintVector("r", r);
		debugPrintVector("p", p);
		residual_out = projectedgmax(r,a);
		debugStats(stats, RESIDUAL, residual_out);
		debugStats(stats, QUADRATIC, 0.5*(BLAS::dot(x,r)-BLAS::dot(x,b)));
		debugStats(stats, ACTIVE, BLAS::abs_sum(a));
		//printf("residual = %f\n", residual_out);
		if (residual_out <= tolerance) {
			iterations_out = iteration;
			return true;
		}
		float c2 = choppedgnorm2(r,a), frf = freedotrfreeg(r,x,alpha_bar,a,xmin,xmax);
		debugPrintf("choppedgnorm2 %f freedotrfreeg %f\n", c2, frf);
		if (c2 <= proportioning_factor*frf) {
			// proportional
			debugPrintf("proportional\n");
			multiply(A_fixed, p, z);							// z = Ap
			debugPrintVector("z", z);
			float rho = BLAS::dot(p,z);
			float alpha_cg = BLAS::dot(r,p) / rho;				// alpha_cg = r'p / p'Ap
			float alpha_f = max_steplength(x,p,a,xmin,xmax);	// alpha_f = min(x/p)
			debugPrintf("alpha_cg = %f, alpha_f = %f\n", alpha_cg, alpha_f);
			if (alpha_cg < alpha_f) {
				// conjugate gradient
				debugPrintf("conjugate gradient\n");
				debugStats(stats, STEP, 0);
				BLAS::add_scaled(-alpha_cg, p, x);				// x -= alpha_cg p
				BLAS::add_scaled(-alpha_cg, z, r);				// r -= alpha_cg Ap
				apply_preconditioner_free(M, r, a, pfg);
				float gamma = BLAS::dot(pfg,z) / rho;			// gamma = (Wf)'(Ap) / p'Ap
				debugPrintf("gamma = %f\n", gamma);
				BLAS::add_scaled(-gamma, p, pfg); p.swap(pfg);	// p = Wf - gamma p
			} else {
				// expansion
				debugPrintf("expansion\n");
				debugStats(stats, STEP, 1);
				BLAS::add_scaled(-alpha_f, p, x);				// x -= alpha_f p
				BLAS::add_scaled(-alpha_f, z, r);				// r -= alpha_f z
				add_scaled_free(-alpha_bar,r,a,xmin,xmax,x);	// x = P(x - alpha_bar f)
				debugPrintVector("x", x);
				recompute_active(x,xmin,xmax, a);
				debugPrintVector("a", a);
				multiply(A_fixed, x, r);
				BLAS::add_scaled(-1, b, r);						// r = Ax - b
				apply_preconditioner_free(M, r, a, p);			// p = Wf
			}
			/*{ // sanity check
				for (int i = 0; i < n; i++) {
					bool activex = !f[i] && x[i]!=0;
					bool negativex = x[i]<0;
					bool zerox = f[i] && x[i]==0;
					bool activep = !f[i] && p[i]!=0;
					if (activex || negativex || zerox || activep) {
						printf("HOLY SHIP HOLY SHIP HOLY SHIP\n");
						A.write_matlab(fstream("matrix.mat",ios::out),"A");
						exit(1);
					}
				}
			}*/
		} else {
			// proportioning
			debugPrintf("proportioning\n");
			debugStats(stats, STEP, 2);
			debugPrintVector("r", r);
			choppedg(r, a, p);
			debugPrintVector("p", p);
			multiply(A_fixed, p, z);							// z = Ap
			debugPrintVector("z", z);
			float alpha_cg = BLAS::dot(r,p) / BLAS::dot(p,z);	// alpha_cg = r'p / p'Ap
			debugPrintf("alpha_cg = %f\n", alpha_cg);
			BLAS::add_scaled(-alpha_cg, p, x);					// x -= alpha_cg p
			// BLAS::add_scaled(-alpha_cg, z, r);					// r -= alpha_cg z
			// recompute_active(x,xmin,xmax, a);
			// debugPrintVector("a", a);
			// apply_preconditioner_free(M, r, a, p);				// p = Wf
			project_to_bounds(xmin,xmax, x);
			debugPrintVector("x", x);
			recompute_active(x,xmin,xmax, a);
			debugPrintVector("a", a);
			multiply(A_fixed, x, r);
			BLAS::add_scaled(-1, b, r);							// r = Ax - b
			apply_preconditioner_free(M, r, a, p);				// p = Wf
		}
	}
	iterations_out=iteration;
	return false;
}

