/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "fields.hpp"

using namespace tvmet;

void ScalarField<2>::resize (int m, int n, float dx) {
	this->m = m;
	this->n = n;
	this->dx = dx;
	s.resize(m*n);
}
void ScalarField<3>::resize (int m, int n, int o, float dx) {
	this->m = m;
	this->n = n;
	this->o = o;
	this->dx = dx;
	s.resize(m*n*o);
}

void BucketField<2>::resize (int m, int n, float dx) {
	this->m = m;
	this->n = n;
	this->dx = dx;
	s.resize(m*n);
}
void BucketField<3>::resize (int m, int n, int o, float dx) {
	this->m = m;
	this->n = n;
	this->o = o;
	this->dx = dx;
	s.resize(m*n*o);
}

void VectorField<2>::resize (int m, int n, float dx) {
	this->m = m;
	this->n = n;
	this->dx = dx;
	x.resize(m+1,n,dx);
	y.resize(m,n+1,dx);
}
void VectorField<3>::resize (int m, int n, int o, float dx) {
	this->m = m;
	this->n = n;
	this->o = o;
	this->dx = dx;
	x.resize(m+1,n,o,dx);
	y.resize(m,n+1,o,dx);
	z.resize(m,n,o+1,dx);
}

void TensorField<2>::resize (int m, int n, float dx) {
	this->m = m;
	this->n = n;
	this->dx = dx;
	xx.resize(m,n,dx);
	yy.resize(m,n,dx);
	xy.resize(m+1,n+1,dx);
}
void TensorField<3>::resize (int m, int n, int o, float dx) {
	this->m = m;
	this->n = n;
	this->o = o;
	this->dx = dx;
	xx.resize(m,n,o,dx);
	yy.resize(m,n,o,dx);
	zz.resize(m,n,o,dx);
	xy.resize(m+1,n+1,o,dx);
	xz.resize(m+1,n,o+1,dx);
	yz.resize(m,n+1,o+1,dx);
}

template <size_t d> void clear_all(BucketField<d> &field) {
	for ( int i = 0; i < (int)field.s.size(); i++ ) {
		field.s[i].clear();
	}
}
template void clear_all (BucketField<2> &field);
template void clear_all (BucketField<3> &field);

template <size_t d> void set_all (ScalarField<d> &field, float s) {
	field.s.assign(field.s.size(), s);
}
template void set_all (ScalarField<2> &field, float s);
template void set_all (ScalarField<3> &field, float s);

template <> void set_all (VectorField<2> &field, const Vec2f &v) {
	set_all(field.x, v[0]);
	set_all(field.y, v[1]);
}
template <> void set_all (VectorField<3> &field, const Vec3f &v) {
	set_all(field.x, v[0]);
	set_all(field.y, v[1]);
	set_all(field.z, v[2]);
}

template <> void set_all (TensorField<2> &field, const Mat2f &t) {
	set_all(field.xx, t(0,0));
	set_all(field.yy, t(1,1));
	set_all(field.xy, t(0,1));
}
template <> void set_all (TensorField<3> &field, const Mat3f &t) {
	set_all(field.xx, t(0,0));
	set_all(field.yy, t(1,1));
	set_all(field.zz, t(2,2));
	set_all(field.xy, t(0,1));
	set_all(field.xz, t(0,2));
	set_all(field.yz, t(1,2));
}

template <size_t d> void add_into (ScalarField<d> &aug, const ScalarField<d> &add) {
	for (int i = 0; i < (int)aug.s.size(); i++) {
		aug.s[i] += add.s[i];
	}
}

template <> void add_into (VectorField<2> &aug, const VectorField<2> &add) {
	add_into(aug.x, add.x);
	add_into(aug.y, add.y);
}
template <> void add_into (VectorField<3> &aug, const VectorField<3> &add) {
	add_into(aug.x, add.x);
	add_into(aug.y, add.y);
	add_into(aug.z, add.z);
}

inline void lerp_weight (float x, int m, int *x0, float *f) {
	*x0 = (int)floor(x);
	*f = x - *x0;
	if (*x0 < 0) {*x0 = 0; *f = 0;}
	if (*x0 >= m-1) {*x0 = m-2; *f = 1;}
}

template <> float lerp (const Vec2f &x, const ScalarField<2> &field) {
	int x0, y0;
	float fx, fy;
	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
	return lerp(fy,
	            lerp(fx, field(x0,y0),   field(x0+1,y0  )),
	            lerp(fx, field(x0,y0+1), field(x0+1,y0+1)));
}
template <> float lerp (const Vec3f &x, const ScalarField<3> &field) {
	int x0, y0, z0;
	float fx, fy, fz;
	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
	lerp_weight(x[2]/field.dx-0.5f, field.o, &z0, &fz);
	return lerp(fz,
	            lerp(fy,
	                 lerp(fx, field(x0,y0,z0),   field(x0+1,y0,z0  )),
	                 lerp(fx, field(x0,y0+1,z0), field(x0+1,y0+1,z0))),
	            lerp(fy,
	                 lerp(fx, field(x0,y0,z0+1),   field(x0+1,y0,z0+1  )),
	                 lerp(fx, field(x0,y0+1,z0+1), field(x0+1,y0+1,z0+1))));
}

template <> float lerp_weighted (const Vec2f &x, const ScalarField<2> &field, const ScalarField<2> &weight) {
	int x0, y0;
	float fx, fy;
	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
	return lerp(fy,
	            lerp(fx, field(x0,y0) * weight(x0,y0),   field(x0+1,y0) * weight(x0+1,y0)),
	            lerp(fx, field(x0,y0+1)*weight(x0,y0+1), field(x0+1,y0+1)*weight(x0+1,y0+1)))
		/ lerp(fy,
		       lerp(fx, weight(x0,y0),   weight(x0+1,y0)),
		       lerp(fx, weight(x0,y0+1), weight(x0+1,y0+1)));
}
template <> float lerp_weighted (const Vec3f &x, const ScalarField<3> &field, const ScalarField<3> &weight) {
	int x0, y0, z0;
	float fx, fy, fz;
	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
	lerp_weight(x[2]/field.dx-0.5f, field.o, &z0, &fz);
	float divisor = lerp(fz,
		       lerp(fy,
		            lerp(fx, weight(x0,y0,z0),   weight(x0+1,y0,z0)),
		            lerp(fx, weight(x0,y0+1,z0), weight(x0+1,y0+1,z0))),
		       lerp(fy,
		            lerp(fx, weight(x0,y0,z0+1),   weight(x0+1,y0,z0+1)),
		            lerp(fx, weight(x0,y0+1,z0+1), weight(x0+1,y0+1,z0+1))));
	divisor = (divisor < epsilon) ? epsilon : divisor;
	return lerp(fz,
	            lerp(fy,
	                 lerp(fx, field(x0,y0,z0) * weight(x0,y0,z0),   field(x0+1,y0,z0) * weight(x0+1,y0,z0)),
	                 lerp(fx, field(x0,y0+1,z0)*weight(x0,y0+1,z0), field(x0+1,y0+1,z0)*weight(x0+1,y0+1,z0))),
	            lerp(fy,
	                 lerp(fx, field(x0,y0,z0+1) * weight(x0,y0,z0+1),   field(x0+1,y0,z0+1) * weight(x0+1,y0,z0+1)),
	                 lerp(fx, field(x0,y0+1,z0+1)*weight(x0,y0+1,z0+1), field(x0+1,y0+1,z0+1)*weight(x0+1,y0+1,z0+1))))
		/ divisor;
}

template <> Vec2f lerp (const Vec2f &x, const VectorField<2> &field) {
	Vec2f xx = x; xx[0] += field.dx*0.5f;
	Vec2f xy = x; xy[1] += field.dx*0.5f;
	return Vec2f(lerp(xx, field.x), lerp(xy, field.y));
}
template <> Vec3f lerp (const Vec3f &x, const VectorField<3> &field) {
	Vec3f xx = x; xx[0] += field.dx*0.5f;
	Vec3f xy = x; xy[1] += field.dx*0.5f;
	Vec3f xz = x; xz[2] += field.dx*0.5f;
	return Vec3f(lerp(xx, field.x), lerp(xy, field.y), lerp(xz, field.z));
}

template <> Vec2f lerp_weighted (const Vec2f &x, const VectorField<2> &field, const VectorField<2> &weight) {
	Vec2f xx = x; xx[0] += field.dx*0.5f;
	Vec2f xy = x; xy[1] += field.dx*0.5f;
	return Vec2f(lerp_weighted(xx, field.x, weight.x), lerp_weighted(xy, field.y, weight.y));
}
template <> Vec3f lerp_weighted (const Vec3f &x, const VectorField<3> &field, const VectorField<3> &weight) {
	Vec3f xx = x; xx[0] += field.dx*0.5f;
	Vec3f xy = x; xy[1] += field.dx*0.5f;
	Vec3f xz = x; xz[2] += field.dx*0.5f;
	return Vec3f(lerp_weighted(xx, field.x, weight.x), lerp_weighted(xy, field.y, weight.y), lerp_weighted(xz, field.z, weight.z));
}

template <> void add_lerped (float s, const Vec2f &x, ScalarField<2> &field) {
	int x0, y0;
	float fx, fy;
	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
	field(x0,  y0)   += (1-fx)*(1-fy)*s;
	field(x0+1,y0)   += (  fx)*(1-fy)*s;
	field(x0,  y0+1) += (1-fx)*(  fy)*s;
	field(x0+1,y0+1) += (  fx)*(  fy)*s;
}
template <> void add_lerped (float s, const Vec3f &x, ScalarField<3> &field) {
	int x0, y0, z0;
	float fx, fy, fz;
	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
	lerp_weight(x[2]/field.dx-0.5f, field.o, &z0, &fz);
	field(x0  ,y0  ,z0)   += (1-fx)*(1-fy)*(1-fz)*s;
	field(x0+1,y0  ,z0)   += (  fx)*(1-fy)*(1-fz)*s;
	field(x0  ,y0+1,z0)   += (1-fx)*(  fy)*(1-fz)*s;
	field(x0+1,y0+1,z0)   += (  fx)*(  fy)*(1-fz)*s;
	field(x0  ,y0  ,z0+1) += (1-fx)*(1-fy)*(  fz)*s;
	field(x0+1,y0  ,z0+1) += (  fx)*(1-fy)*(  fz)*s;
	field(x0  ,y0+1,z0+1) += (1-fx)*(  fy)*(  fz)*s;
	field(x0+1,y0+1,z0+1) += (  fx)*(  fy)*(  fz)*s;
}

template <> void add_lerped (const Vec2f &v, const Vec2f &x, VectorField<2> &field) {
	Vec2f xx = x; xx[0] += field.dx*0.5f;
	Vec2f xy = x; xy[1] += field.dx*0.5f;
	add_lerped(v[0], xx, field.x);
	add_lerped(v[1], xy, field.y);
}
template <> void add_lerped (const Vec3f &v, const Vec3f &x, VectorField<3> &field) {
	Vec3f xx = x; xx[0] += field.dx*0.5f;
	Vec3f xy = x; xy[1] += field.dx*0.5f;
	Vec3f xz = x; xz[2] += field.dx*0.5f;
	add_lerped(v[0], xx, field.x);
	add_lerped(v[1], xy, field.y);
	add_lerped(v[2], xz, field.z);
}

template <> Vec2f gradient (const Vec2f &x, const ScalarField<2> &field) {
	Vec2f v;
	Vec2f xx0 = x, xx1 = x; xx0[0] -= field.dx*0.5f; xx1[0] += field.dx*0.5f;
	Vec2f xy0 = x, xy1 = x; xy0[1] -= field.dx*0.5f; xy1[1] += field.dx*0.5f;
	float dsx((lerp(xx1, field) - lerp(xx0, field))/field.dx);
	float dsy((lerp(xy1, field) - lerp(xy0, field))/field.dx);
	return Vec2f(dsx, dsy);
}
template <> Vec3f gradient (const Vec3f &x, const ScalarField<3> &field) {
	Vec3f v;
	Vec3f xx0 = x, xx1 = x; xx0[0] -= field.dx*0.5f; xx1[0] += field.dx*0.5f;
	Vec3f xy0 = x, xy1 = x; xy0[1] -= field.dx*0.5f; xy1[1] += field.dx*0.5f;
	Vec3f xz0 = x, xz1 = x; xz0[2] -= field.dx*0.5f; xz1[2] += field.dx*0.5f;
	float dsx((lerp(xx1, field) - lerp(xx0, field))/field.dx);
	float dsy((lerp(xy1, field) - lerp(xy0, field))/field.dx);
	float dsz((lerp(xz1, field) - lerp(xz0, field))/field.dx);
	return Vec3f(dsx, dsy, dsz);
}

static const float dj = 0.1;

template <> Mat2f jacobian (const Vec2f &x, const VectorField<2> &field) {
	Mat2f m;
	Vec2f xx0 = x, xx1 = x; xx0[0] -= field.dx*dj; xx1[0] += field.dx*dj;
	Vec2f xy0 = x, xy1 = x; xy0[1] -= field.dx*dj; xy1[1] += field.dx*dj;
	Vec2f dvx((lerp(xx1, field) - lerp(xx0, field))/(2*field.dx*dj));
	Vec2f dvy((lerp(xy1, field) - lerp(xy0, field))/(2*field.dx*dj));
	m = dvx[0], dvy[0],
		dvx[1], dvy[1];
	return m;
}
template <> Mat3f jacobian (const Vec3f &x, const VectorField<3> &field) {
	Mat3f m;
	Vec3f xx0 = x, xx1 = x; xx0[0] -= field.dx*dj; xx1[0] += field.dx*dj;
	Vec3f xy0 = x, xy1 = x; xy0[1] -= field.dx*dj; xy1[1] += field.dx*dj;
	Vec3f xz0 = x, xz1 = x; xz0[2] -= field.dx*dj; xz1[2] += field.dx*dj;
	Vec3f dvx((lerp(xx1, field) - lerp(xx0, field))/(2*field.dx*dj));
	Vec3f dvy((lerp(xy1, field) - lerp(xy0, field))/(2*field.dx*dj));
	Vec3f dvz((lerp(xz1, field) - lerp(xz0, field))/(2*field.dx*dj));
	m = dvx[0], dvy[0], dvz[0],
		dvx[1], dvy[1], dvz[1],
		dvx[2], dvy[2], dvz[2];
	return m;
}

template <> Mat2f jacobian_weighted (const Vec2f &x, const VectorField<2> &field, const VectorField<2> &weight){
	Mat2f m;
	Vec2f xx0 = x, xx1 = x; xx0[0] -= field.dx*dj; xx1[0] += field.dx*dj;
	Vec2f xy0 = x, xy1 = x; xy0[1] -= field.dx*dj; xy1[1] += field.dx*dj;
	Vec2f dvx((lerp_weighted(xx1, field, weight) - lerp_weighted(xx0, field, weight))/(2*field.dx*dj));
	Vec2f dvy((lerp_weighted(xy1, field, weight) - lerp_weighted(xy0, field, weight))/(2*field.dx*dj));
	m = dvx[0], dvy[0],
		dvx[1], dvy[1];
	return m;
}
template <> Mat3f jacobian_weighted (const Vec3f &x, const VectorField<3> &field, const VectorField<3> &weight){
	Mat3f m;
	Vec3f xx0 = x, xx1 = x; xx0[0] -= field.dx*dj; xx1[0] += field.dx*dj;
	Vec3f xy0 = x, xy1 = x; xy0[1] -= field.dx*dj; xy1[1] += field.dx*dj;
	Vec3f xz0 = x, xz1 = x; xz0[2] -= field.dx*dj; xz1[2] += field.dx*dj;
	Vec3f dvx((lerp_weighted(xx1, field, weight) - lerp_weighted(xx0, field, weight))/(2*field.dx*dj));
	Vec3f dvy((lerp_weighted(xy1, field, weight) - lerp_weighted(xy0, field, weight))/(2*field.dx*dj));
	Vec3f dvz((lerp_weighted(xz1, field, weight) - lerp_weighted(xz0, field, weight))/(2*field.dx*dj));
	m = dvx[0], dvy[0], dvz[0],
		dvx[1], dvy[1], dvz[1],
		dvx[2], dvy[2], dvz[2];
	return m;
}

// Vec2f gradient_unsmoothed (const Vec2f &x, const ScalarField<2> &field, const ScalarField<2> &weight) {
// 	// del (<fw>/<w>) = del<fw>/<w> - <fw>del<w>/<w>^2
// 	// del<fw>/<w> - <f|w>del<w>/<w>
// 	int x0, y0;
// 	float fx, fy;
// 	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
// 	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
// 	Vec2f grad_w(lerp(fy, weight(x0+1,y0) - weight(x0,y0), weight(x0+1,y0+1) - weight(x0,y0+1)),
// 	             lerp(fx, weight(x0,y0+1) - weight(x0,y0), weight(x0+1,y0+1) - weight(x0+1,y0)));
// 	Vec2f grad_fw(lerp(fy, field(x0+1,y0)*weight(x0+1,y0) - field(x0,y0)*weight(x0,y0), field(x0+1,y0+1)*weight(x0+1,y0+1) - field(x0,y0+1)*weight(x0,y0+1)),
// 	              lerp(fx, field(x0,y0+1)*weight(x0,y0+1) - field(x0,y0)*weight(x0,y0), field(x0+1,y0+1)*weight(x0+1,y0+1) - field(x0+1,y0)*weight(x0+1,y0)));
// 	return Vec2f(grad_fw/lerp(x,weight) - lerp_weighted(x,field,weight)*grad_w/lerp(x,weight));
// }
// Vec3f gradient_unsmoothed (const Vec3f &x, const ScalarField<3> &field, const ScalarField<3> &weight) {
// 	// del (<fw>/<w>) = del<fw>/<w> - <fw>del<w>/<w>^2
// 	// del<fw>/<w> - <f|w>del<w>/<w>
// 	int x0, y0, z0;
// 	float fx, fy, fz;
// 	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
// 	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
// 	lerp_weight(x[2]/field.dx-0.5f, field.o, &z0, &fz);
// 	Vec3f grad_w(lerp(fy, weight(x0+1,y0,z0) - weight(x0,y0), weight(x0+1,y0+1) - weight(x0,y0+1)),
// 	             lerp(fx, weight(x0,y0+1) - weight(x0,y0), weight(x0+1,y0+1) - weight(x0+1,y0)));
// 	Vec3f grad_fw(lerp(fy, field(x0+1,y0)*weight(x0+1,y0) - field(x0,y0)*weight(x0,y0), field(x0+1,y0+1)*weight(x0+1,y0+1) - field(x0,y0+1)*weight(x0,y0+1)),
// 	              lerp(fx, field(x0,y0+1)*weight(x0,y0+1) - field(x0,y0)*weight(x0,y0), field(x0+1,y0+1)*weight(x0+1,y0+1) - field(x0+1,y0)*weight(x0+1,y0)));
// 	return Vec2f(grad_fw/lerp(x,weight) - lerp_weighted(x,field,weight)*grad_w/lerp(x,weight));

// 	int x0, y0, z0;
// 	float fx, fy, fz;
// 	lerp_weight(x[0]/field.dx-0.5f, field.m, &x0, &fx);
// 	lerp_weight(x[1]/field.dx-0.5f, field.n, &y0, &fy);
// 	lerp_weight(x[2]/field.dx-0.5f, field.o, &z0, &fz);
// 	float divisor = lerp(fz,
// 		       lerp(fy,
// 		            lerp(fx, weight(x0,y0,z0),   weight(x0+1,y0,z0)),
// 		            lerp(fx, weight(x0,y0+1,z0), weight(x0+1,y0+1,z0))),
// 		       lerp(fy,
// 		            lerp(fx, weight(x0,y0,z0+1),   weight(x0+1,y0,z0+1)),
// 		            lerp(fx, weight(x0,y0+1,z0+1), weight(x0+1,y0+1,z0+1))));
// 	divisor = (divisor < epsilon) ? epsilon : divisor;
// 	return lerp(fz,
// 	            lerp(fy,
// 	                 lerp(fx, field(x0,y0,z0) * weight(x0,y0,z0),   field(x0+1,y0,z0) * weight(x0+1,y0,z0)),
// 	                 lerp(fx, field(x0,y0+1,z0)*weight(x0,y0+1,z0), field(x0+1,y0+1,z0)*weight(x0+1,y0+1,z0))),
// 	            lerp(fy,
// 	                 lerp(fx, field(x0,y0,z0+1) * weight(x0,y0,z0+1),   field(x0+1,y0,z0+1) * weight(x0+1,y0,z0+1)),
// 	                 lerp(fx, field(x0,y0+1,z0+1)*weight(x0,y0+1,z0+1), field(x0+1,y0+1,z0+1)*weight(x0+1,y0+1,z0+1))))
// 		/ divisor;
// }

// template <> Mat2f jacobian_unsmoothed (const Vec2f &x, const VectorField<2> &field, const VectorField<2> &weight){
// 	Mat2f m;
// 	Vec2f xx = x; xx[0] += field.dx*0.5f;
// 	Vec2f xy = x; xy[1] += field.dx*0.5f;
// 	Vec2f dvx = gradient_unsmoothed(xx, field.x, weight.x);
// 	Vec2f dvy = gradient_unsmoothed(xy, field.y, weight.y);
// 	m = dvx[0], dvx[1],
// 		dvy[0], dvy[1];
// 	return m;
// }
// template <> Mat3f jacobian_unsmoothed (const Vec3f &x, const VectorField<3> &field, const VectorField<3> &weight){
// 	Mat3f m;
// 	Vec3f xx = x; xx[0] += field.dx*0.5f;
// 	Vec3f xy = x; xy[1] += field.dx*0.5f;
// 	Vec3f xz = x; xz[2] += field.dx*0.5f;
// 	Vec3f dvx = gradient_unsmoothed(xx, field.x, weight.x);
// 	Vec3f dvy = gradient_unsmoothed(xy, field.y, weight.y);
// 	Vec3f dvz = gradient_unsmoothed(xz, field.z, weight.z);
// 	m = dvx[0], dvx[1], dvx[2],
// 		dvy[0], dvy[1], dvy[2],
// 		dvz[0], dvz[1], dvz[2];
// 	return m;
// }

/// Note: This is not an accurate gradient field on the boundaries
template <> void gradient (const ScalarField<3> &sfield, VectorField<3> &grad) {
	int m = sfield.m, n = sfield.n, o = sfield.o;
	float dx = sfield.dx;
	grad.resize(m, n, o, sfield.dx);
	for ( int i = 1; i < m; i++ ) {
		for ( int j = 0; j < n; j++ ) {
			for ( int k = 0; k < o; k++ ) {
				grad.x(i, j, k) = ( sfield(i, j, k) - sfield(i-1, j, k) ) / dx;
				//std::cout << "Debug: grad= " << grad.x(i, j, k) << " dx= " << dx << "sfield = (" << sfield(i, j, k) << ", " << sfield(i-1, j, k) << ")" << std::endl;
			}
		}
	}
	for ( int i = 0; i < m; i++ ) {
		for ( int j = 1; j < n; j++ ) {
			for ( int k = 0; k < o; k++ ) {
				grad.y(i, j, k) = ( sfield(i, j, k) - sfield(i, j-1, k) ) / dx;
			}
		}
	}
	for ( int i = 1; i < m; i++ ) {
		for ( int j = 0; j < n; j++ ) {
			for ( int k = 1; k < o; k++ ) {
				grad.z(i, j, k) = ( sfield(i, j, k) - sfield(i, j, k-1) ) / dx;
			}
		}
	}
	for ( int j = 0; j < n; j++ ) for ( int k = 0; k < o; k++ ) { {grad.x(0, j, k) = -1; grad.x(m, j, k) = 1;} }
	for ( int i = 0; i < m; i++ ) for ( int k = 0; k < o; k++ ) { {grad.y(i, 0, k) = -1; grad.y(i, n, k) = 1;} }
	for ( int i = 0; i < m; i++ ) for ( int j = 0; j < n; j++ ) { {grad.z(i, j, 0) = -1; grad.z(i, j, o) = 1;} }
}
