/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef FIELDS_HPP
#define FIELDS_HPP

#include <vector>
#include "../common.hpp"

template <size_t d> class ScalarField;
template <> class ScalarField<2> {
public:
	int m, n;
	float dx;
	std::vector<float> s;   // x-major order
	void resize (int m, int n, float dx);
	inline float operator() (int i, int j) const {return s[i+j*m];}
	inline float &operator() (int i, int j) {return s[i+j*m];}
	inline float &operator() (int x[]) {return operator()(x[0], x[1]);}
};
template <> class ScalarField<3> {
public:
	int m, n, o;
	float dx;
	std::vector<float> s;   // x-major order
	void resize (int m, int n, int o, float dx);
	inline float operator() (int i, int j, int k) const {return s[i+m*(j+n*k)];}
	inline float &operator() (int i, int j, int k) {return s[i+m*(j+n*k)];}
	inline float &operator() (int x[]) {return operator()(x[0], x[1], x[2]);}
};

template <size_t d> class BucketField;
template <> class BucketField<2> {
public:
	int m,n;
	float dx;
	std::vector< std::vector<int> > s; // x-major order
	inline std::vector<int> &operator() (int i, int j) {return s[i+j*m];}
	inline std::vector<int> &operator() (int x[]) {return s[x[0]+x[1]*m];}
	inline void bclamp(int x[]) { x[0] = clamp<int>(x[0], 0, m - 1); x[1] = clamp<int>(x[1], 0, n - 1); }
	void resize (int m, int n, float dx);
};
template <> class BucketField<3> {
public:
	int m,n,o;
	float dx;
	std::vector< std::vector<int> > s; // x-major order
	inline std::vector<int> &operator() (int i, int j, int k) {return s[i+m*(j+n*k)];}
	inline std::vector<int> &operator() (int x[]) {return s[x[0]+m*(x[1]+n*x[2])];}
	inline void bclamp(int x[]) { x[0] = clamp<int>(x[0], 0, m - 1); x[1] = clamp<int>(x[1], 0, n - 1); x[2] = clamp<int>(x[2], 0, o - 1); }
	void resize (int m, int n, int o, float dx);
};

template <size_t d> class VectorField;
template <> class VectorField<2> {
public:
	int m, n;
	float dx;
	ScalarField<2> x, y;   // staggered
	void resize (int m, int n, float dx);
};
template <> class VectorField<3> {
public:
	int m, n, o;
	float dx;
	ScalarField<3> x, y, z;   // staggered
	void resize (int m, int n, int o, float dx);
};

// Only for *symmetric* tensor fields
template <size_t d> class TensorField;
template <> class TensorField<2> {
public:
	int m, n;
	float dx;
	ScalarField<2> xx, xy, yy;   // diagonal elements cell-centered, off-diagonal elements at vertices
	void resize (int m, int n, float dx);
};
template <> class TensorField<3> {
public:
	int m, n, o;
	float dx;
	ScalarField<3> xx, yy, zz, xy, xz, yz;   // diagonal elements cell-centered, off-diagonal elements at edges
	void resize (int m, int n, int o, float dx);
};

template <size_t d> void set_all (ScalarField<d> &field, float s);
template <size_t d> void set_all (VectorField<d> &field, const Vector<float,(unsigned)d> &v);
template <size_t d> void set_all (TensorField<d> &field, const Matrix<float,(unsigned)d,(unsigned)d> &t);

template <size_t d> void clear_all (BucketField<d> &field);

template <size_t d> void add_into (ScalarField<d> &aug, const ScalarField<d> &add);
template <size_t d> void add_into (VectorField<d> &aug, const VectorField<d> &add);

// origin at vertex; cell centers at half-integer coordinates
template <size_t d> float lerp (const Vector<float,d> &x, const ScalarField<d> &field);
template <size_t d> Vector<float,d> lerp (const Vector<float,d> &x, const VectorField<d> &field);
template <size_t d> float lerp_weighted (const Vector<float,d> &x, const ScalarField<d> &field, const ScalarField<d> &weight);
template <size_t d> Vector<float,d> lerp_weighted (const Vector<float,d> &x, const VectorField<d> &field, const VectorField<d> &weight);

template <size_t d> void add_lerped (float s, const Vector<float,d> &x, ScalarField<d> &field);
template <size_t d> void add_lerped (const Vector<float,d> &v, const Vector<float,d> &x, VectorField<d> &field);

template <size_t d> Vector<float,d> gradient (const Vector<float,d> &x, const ScalarField<d> &field);

template <size_t d> Matrix<float,d,d> jacobian (const Vector<float,d> &x, const VectorField<d> &field);
template <size_t d> Matrix<float,d,d> jacobian_weighted (const Vector<float,d> &x, const VectorField<d> &field, const VectorField<d> &weight);

template <size_t d> void gradient(const ScalarField<d> &sfield, VectorField<d> &grad);

#endif
