#include "scenegraph.hpp"
#include <cstdlib>
#include <fstream>
#include <stdint.h>
#include "../common.hpp"
#define TIXML_USE_STL
#include <tinyxml.h>
#include "../exec/iohelpers.hpp"
#include <sstream>

using namespace std;

HoldBodyRender::HoldBodyRender() : v(0,0,0) {
	active = false;
	endtime = 0;
}

/// Remove specified delimiter from the beginning and end of a string
void removeDelims(string &s, string &delims) {
	int pos;
	while ((pos = s.find_first_of(delims)) != string::npos) {
		s.replace(pos, 1, "");
	}
}

Node::Node() {
	angle = 0; axis[0] = axis[1] = axis[2] = 0;
	scale[0] = scale[1] = scale[2] = 1;
	location[0] = location[1] = location[2] = 0;
	for (int i = 0; i < 6; i++) bound[i] = 0;
	type = RIGIDBODY;
	objdata = string("");
	align = -1;
}

Node::Node(float _location[], float _bound[], NodeType _type) {
	location[0] = _location[0];		location[1] = _location[1];		location[2] = _location[2];
	for (int i = 0; i < 6; i++) bound[i] = _bound[i];
	type = _type;
	angle = 0; axis[0] = axis[1] = axis[2] = 0;
	scale[0] = scale[1] = scale[2] = 1;
	align = -1;
}

void Node::setLocation(float _location[]) {
	location[0] = _location[0];		location[1] = _location[1];		location[2] = _location[2];
}

void Node::setRotation(float _axis[], float _angle) {
	axis[0] = _axis[0];		axis[1] = _axis[1];		axis[2] = _axis[2];
	angle = _angle;
}

string &Node::objectdata() {
	return objdata;
}

void SceneGraph::clearall() {
	scene.clear();
}

/// Load the entire setup XML
bool SceneGraph::load(std::string filename) {
	cout << "Loading objects into scene graph" << endl;
	clearall();
	string mydelims("[]");
	try {
		TiXmlDocument doc(filename);
		if (!doc.LoadFile()) cout << "Couldn't load setup file '" << filename << "'" << endl << yield;
		TiXmlElement *pRoot = doc.RootElement();
		if (!pRoot) {cout << "No root element" << endl; return false;}

		TiXmlElement *pEntry = pRoot->FirstChildElement("node");
		while (pEntry) {
			Node n;
			if (pEntry->Attribute("type")) {
				string type = string(pEntry->Attribute("type"));
				if (type == "camera") n.type = CAMERA;
				else if (type == "grid") n.type = GRID;
				else if (type == "rigidbody") n.type = RIGIDBODY;
				else if (type == "light") n.type = LIGHT;
				else cout << "Unknown node type: " << type << endl << yield;
			}
			else {
				cout << "Must specify node type" << endl << yield;
			}
			if (pEntry->Attribute("location")) {
				string loc = string(pEntry->Attribute("location"));
				removeDelims(loc, mydelims);
				stringstream s(loc);
				s >> n.location[0] >> n.location[1] >> n.location[2];
			}
			else {
				cout << "Assuming location to be [0 0 0]" << endl;
			}
			if (pEntry->Attribute("min")) {
				string boundmin = string(pEntry->Attribute("min"));
				string boundmax = string(pEntry->Attribute("max"));
				removeDelims(boundmin, mydelims);
				removeDelims(boundmax, mydelims);
				stringstream s2(boundmin), s3(boundmax);
				s2 >> n.bound[0] >> n.bound[1] >> n.bound[2];
				s3 >> n.bound[3] >> n.bound[4] >> n.bound[5];
			}
			else if (n.type != GRID) {
				cout << "Bounds must be specified for non grid elements" << endl << yield;
			}

			if (pEntry->Attribute("axis") && pEntry->Attribute("angle")) {
				string axis = string(pEntry->Attribute("axis"));
				string angle = string(pEntry->Attribute("angle"));
				removeDelims(axis, mydelims);
				stringstream s6(axis), s4(angle);
				s6 >> n.axis[0] >> n.axis[1] >> n.axis[2];
				s4 >> n.angle;
			}
			else {
				//cout << "No rotations specified" << endl;
				/// Don't want the axis to be 0, else the entire matrix would become 0
				n.axis[0] = 1;
			}

			if (pEntry->Attribute("scale")) {
				string scale = string(pEntry->Attribute("scale"));
				removeDelims(scale, mydelims);
				stringstream s5(scale);
				s5 >> n.scale[0] >> n.scale[1] >> n.scale[2];
			}

			if (pEntry->Attribute("data"))
				n.objdata = string(pEntry->Attribute("data"));

			if (pEntry->Attribute("align")) {
				string align = string(pEntry->Attribute("align"));
				stringstream s7(align);
				s7 >> n.align;
			}

			if (pEntry->QueryIntAttribute("couple", &(n.couple)) != TIXML_SUCCESS) n.couple = -1;
			if (pEntry->QueryIntAttribute("tcouple", &(n.tcouple)) != TIXML_SUCCESS) n.tcouple = -1;
			if (n.couple > -1) n.tcouple = n.couple;

			if (pEntry->Attribute("hold") && !strcmp(pEntry->Attribute("hold"), "yes")) {
				HoldBodyRender &h = n.hold;
				h.active = true;
				if (pEntry->QueryFloatAttribute("until", &(h.endtime)) != TIXML_SUCCESS) cout << "Error: End time not specified for node/hold" << endl << yield;
				if ( pEntry->Attribute("v") == NULL) cout << "Error: velocity not specified for node/hold, assuming 0" << endl;
				else {
					string vel(pEntry->Attribute("v"));
					removeDelims(vel, mydelims);
					stringstream vs(vel);
					vs >> h.v[0] >> h.v[1] >> h.v[2];
				}
				cout << "Holding body till t = " << h.endtime << " with target vel = [" << h.v[0] << ", " << h.v[1] << ", " << h.v[2] << "]" << endl;
			}

			pEntry = pEntry->NextSiblingElement("node");
			scene.push_back(n);
		}
	}
	catch (Yield y) {return false;}
	return true;
}

/// Dump this config to the setup XML
bool SceneGraph::save(std::string filename) {
	cout << "Sorry, save is not implemented yet!" << endl;
	return true;
}

/// Update body objID
void SceneGraph::update(std::string filename, int objID) {
	if (objID >= (int)scene.size()) {
		cout << "Error (Update): Requested object does not exist. Call information follows: (filename: " << filename << " , objID: " << objID << ") nObjectsInSceneGraph: " << scene.size() << endl;
		return;
	}
	Node &n = scene[objID];
	switch (n.type) {
		case CAMERA: break;
		case GRID: break;
		case RIGIDBODY: break;
		case LIGHT: break;
		default: cout << "Error (Update): Unknown object type: " << n.type << endl;
	}
}
