/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "coupling.hpp"

using namespace std;

Screw<2> divide_by_mass (const Screw<2> &wrench, const RigidBody<2> &body) {
	return screw(Vec2f(m_inv(body)*lin(wrench)), world_I_inv(body)*ang(wrench));
}
Screw<3> divide_by_mass (const Screw<3> &wrench, const RigidBody<3> &body) {
	return screw(Vec3f(m_inv(body)*lin(wrench)), Vec3f(world_I_inv(body)*ang(wrench)));
}

template <typename T> void pack (vector<T> &J, vector<int> &idx) {
	int n = J.size();
	idx.clear();
	for (int i = 0; i < n; i++) {
		if (dot(J[i],J[i]) < epsilon) continue;
		J[idx.size()] = J[i];
		idx.push_back(i);
	}
	J.resize(idx.size());
}

template <typename T> void project (vector<T> &J, const vector<int> &idx) {
	int n = idx.size();
	for (int i = 0; i < n; i++) {
		J[i] = J[idx[i]];
	}
	J.resize(idx.size());
}

// ---------------- //
//        2D        //
// ---------------- //

typedef vector< Screw<2> > Coupling2;

void pressure_J (const Sand<2> &sand, const RigidBody<2> &body, vector< Screw<2> > &J) {
	int m = sand.m, n = sand.n;
	int pm = m + 2, pn = n + 2;
	float dx = sand.dx;
	J.assign(pm*pn, Screw<2>(0));
	const Vec2f &com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	// x faces
	for (int j = j0; j < j1; j++) for (int i = i0; i <= i1; i++) {
		int a = i+(j+1)*pm, b = (i+1)+(j+1)*pm;
		Vec2f cx0((i-0.5f)*dx, j*dx), cx1((i+0.5f)*dx, (j+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<2> s(1, 0, com[1] - (j+0.5f)*dx);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// y faces
	for (int j = j0; j <= j1; j++) for (int i = i0; i < i1; i++) {
		int a = (i+1)+j*pm, b = (i+1)+(j+1)*pm;
		Vec2f cx0(i*dx, (j-0.5f)*dx), cx1((i+1)*dx, (j+0.5f)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<2> s(0, 1, (i+0.5f)*dx - com[0]);
		J[a] += f*s;
		J[b] -= f*s;
	}
}

template <> void pressure_coupling (const Sand<2> &sand, const RigidBody<2> &body, float dt, const IntArray<2> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n;
	int pm = m + 2, pn = n + 2;
	float dx = sand.dx;
	// static vector< Screw<2> > J;
	// pressure_J(sand, body, J);
	const Coupling2 &J = body.aux.coupling.p_J;
	// A += dt^2 J' M_inv J
	// b += dt J' v
	float mi = m_inv(body);
	float Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt;
	Screw<2> t(get_twist(body) + divide_by_mass(body.aux.ds, body));
	for (int a = 0; a < pm*pn; a++) {
		if (idx.a[a] < 0) continue;
		if (all_elements(abs(J[a]) <= epsilon)) continue;
		for (int b = 0; b < pm*pn; b++) {
			if (idx.a[b] < 0) continue;
			if (all_elements(abs(J[b]) <= epsilon)) continue;
			float Aab = dot(lin(J[a]),lin(J[b]))*mi + ang(J[a])*Ii*ang(J[b]);
			if (Aab > epsilon)
				A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
		}
		// int i = a%pm - 1, j = a/pm - 1;
		// Vec2f x0(i*dx,j*dx), x1((i+1)*dx, (j+1)*dx);
		// float f = coverage(body, x0,x1);
		// if (i >= 0 && i < m && j >= 0 && j < n) B[idx.a[a]] += f;
		B[idx.a[a]] -= dot(J[a],t)*coeffr;
	}
}

template <> Screw<2> pressure_force (const Sand<2> &sand, const RigidBody<2> &body) {
	int m = sand.m, n = sand.n;
	int pm = m + 2, pn = n + 2;
	float dx = sand.dx;
	// static vector< Screw<2> > J;
	// pressure_J(sand, body, J);
	const Coupling2 &J = body.aux.coupling.p_J;
	Screw<2> f(0);
	for (int a = 0; a < pm*pn; a++) {
		f += J[a]*sand.aux.p.s[a];
	}
	return Screw<2>(dx*f);
}

void friction_xx_yy_J (const Sand<2> &sand, const RigidBody<2> &body, vector< Screw<2> > &J) {
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	J.assign(m*n, Screw<2>(0));
	const Vec2f &com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	// x faces
	for (int j = j0; j < j1; j++) for (int i = i0+1; i < i1; i++) {
		int a = (i-1)+j*m, b = i+j*m;
		Vec2f cx0((i-0.5f)*dx, j*dx), cx1((i+0.5f)*dx, (j+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<2> s(1, 0, com[1] - (j+0.5f)*dx);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// y faces
	for (int j = j0+1; j < j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+(j-1)*m, b = i+j*m;
		Vec2f cx0(i*dx, (j-0.5f)*dx), cx1((i+1)*dx, (j+0.5f)*dx);
		float f = coverage(body, cx0,cx1) ;
		Screw<2> s(0, 1, (i+0.5f)*dx - com[0]);
		J[a] -= f*s;
		J[b] += f*s;
	}
}

void friction_xy_J (const Sand<2> &sand, const RigidBody<2> &body, vector< Screw<2> > &J) {
	int m = sand.m, n = sand.n;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	J.assign(sm*sn, Screw<2>(0));
	const Vec2f &com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	// x faces
	for (int j = j0; j < j1; j++) for (int i = i0; i <= i1; i++) {
		int a = i+j*sm, b = i+(j+1)*sm;
		Vec2f cx0((i-0.5f)*dx, j*dx), cx1((i+0.5f)*dx, (j+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<2> s(1, 0, com[1] - (j+0.5f)*dx);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// y faces
	for (int j = j0; j <= j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+j*sm, b = (i+1)+j*sm;
		Vec2f cx0(i*dx, (j-0.5f)*dx), cx1((i+1)*dx, (j+0.5f)*dx);
		float f = coverage(body, cx0,cx1) ;
		Screw<2> s(0, 1, (i+0.5f)*dx - com[0]);
		J[a] += f*s;
		J[b] -= f*s;
	}
}

void friction_xx_yy_coupling (const Sand<2> &sand, const RigidBody<2> &body, float dt, const IntArray<2> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	// static vector< Screw<2> > J;
	// friction_xx_yy_J(sand, body, J);
	const Coupling2 &J = body.aux.coupling.f_xx_yy_J;
	// A += dt^2 J' M_inv J
	// b += dt J' v
	float mi = m_inv(body);
	float Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt;
	Screw<2> t(get_twist(body) + divide_by_mass(body.aux.dp + body.aux.ds, body));
	for (int a = 0; a < m*n; a++) {
		if (idx.a[a] < 0) continue;
		if (all_elements(abs(J[a]) <= epsilon)) continue;
		for (int b = 0; b < m*n; b++) {
			if (idx.a[b] < 0) continue;
			if (all_elements(abs(J[b]) <= epsilon)) continue;
			float Aab = dot(lin(J[a]),lin(J[b]))*mi + ang(J[a])*Ii*ang(J[b]);
			if (Aab > epsilon)
				A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
		}
		B[idx.a[a]] -= dot(J[a],t)*coeffr;
	}
}

void friction_xy_coupling (const Sand<2> &sand, const RigidBody<2> &body, float dt, const IntArray<2> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	// static vector< Screw<2> > J;
	// friction_xy_J(sand, body, J);
	const Coupling2 &J = body.aux.coupling.f_xy_J;
	// A += dt^2 J' M_inv J
	// b += dt J' v
	float mi = m_inv(body);
	float Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt;
	Screw<2> t(get_twist(body) + divide_by_mass(body.aux.dp + body.aux.ds, body));
	for (int a = 0; a < sm*sn; a++) {
		if (idx.a[a] < 0) continue;
		if (all_elements(abs(J[a]) <= epsilon)) continue;
		for (int b = 0; b < sm*sn; b++) {
			if (idx.a[b] < 0) continue;
			if (all_elements(abs(J[b]) <= epsilon)) continue;
			float Aab = dot(lin(J[a]),lin(J[b]))*mi + ang(J[a])*Ii*ang(J[b]);
			if (Aab > epsilon)
				A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
		}
		B[idx.a[a]] -= dot(J[a],t)*coeffr;
	}
}

Screw<2> friction_xx_yy_force (const Sand<2> &sand, const RigidBody<2> &body) {
	//return Screw<2>(0);
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	// static vector< Screw<2> > J;
	// friction_xx_yy_J(sand, body, J);
	const Coupling2 &J = body.aux.coupling.f_xx_yy_J;
	Screw<2> f(0);
	for (int a = 0; a < m*n; a++) {
		f += J[a]*sand.aux.s.xx.s[a];
	}
	//cout << "friction xx/yy force " << f[0] << " " << f[1] << " " << f[2] << endl;
	return Screw<2>(dx*f);
}

Screw<2> friction_xy_force (const Sand<2> &sand, const RigidBody<2> &body) {
	//return Screw<2>(0);
	int m = sand.m, n = sand.n;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	// static vector< Screw<2> > J;
	// friction_xy_J(sand, body, J);
	const Coupling2 &J = body.aux.coupling.f_xy_J;
	Screw<2> f(0);
	for (int a = 0; a < sm*sn; a++) {
		f += J[a]*sand.aux.s.xy.s[a];
	}
	//cout << "friction xy force " << f[0] << " " << f[1] << " " << f[2] << endl;
	return Screw<2>(dx*f);
}

template <> void precompute_coupling_matrices (const Sand<2> &sand, RigidBody<2> &body) {
	pressure_J(sand, body, body.aux.coupling.p_J);
	friction_xx_yy_J(sand, body, body.aux.coupling.f_xx_yy_J);
	friction_xy_J(sand, body, body.aux.coupling.f_xy_J);
}

// ---------------- //
//        3D        //
// ---------------- //

typedef vector< Screw<3> > Coupling3;

void pressure_J (const Sand<3> &sand, RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int pm = m + 2, pn = n + 2, po = o + 2;
	float dx = sand.dx;
	Coupling3 &J = body.aux.coupling.p_J;
	J.assign(pm*pn*po, Screw<3>(0));
	Vec3f com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	float z0 = com[2] - body.r, z1 = com[2] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	int k0 = floor(z0/dx), k1 = ceil(z1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	k0 = min(max(k0, 0), o);
	k1 = min(max(k1, 0), o);
	// x faces
	for (int k = k0; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i <= i1; i++) {
		int a = i+pm*((j+1)+pn*(k+1)), b = (i+1)+pm*((j+1)+pn*(k+1));
		Vec3f cx0((i-0.5f)*dx, j*dx, k*dx), cx1((i+0.5f)*dx, (j+1)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(1,0,0, 0, (k+0.5f)*dx - com[2], com[1] - (j+0.5f)*dx);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// y faces
	for (int k = k0; k < k1; k++) for (int j = j0; j <= j1; j++) for (int i = i0; i < i1; i++) {
		int a = (i+1)+pm*(j+pn*(k+1)), b = (i+1)+pm*((j+1)+pn*(k+1));
		Vec3f cx0(i*dx, (j-0.5f)*dx, k*dx), cx1((i+1)*dx, (j+0.5f)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,1,0, com[2] - (k+0.5f)*dx, 0, (i+0.5f)*dx - com[0]);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// z faces
	for (int k = k0; k <= k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
		int a = (i+1)+pm*((j+1)+pn*k), b = (i+1)+pm*((j+1)+pn*(k+1));
		Vec3f cx0(i*dx, j*dx, (k-0.5f)*dx), cx1((i+1)*dx, (j+1)*dx, (k+0.5f)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,0,1, (j+0.5f)*dx - com[1], com[0] - (i+0.5f)*dx, 0);
		J[a] += f*s;
		J[b] -= f*s;
	}
	pack(J, body.aux.coupling.p_i);
}

template <> void pressure_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n, o = sand.o;
	int pm = m + 2, pn = n + 2, po = o + 2;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// pressure_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.p_i;
	const Coupling3 &J = body.aux.coupling.p_J;
	// A += dt^2 J' M_inv J
	// b += dt J' v
	float mi = m_inv(body);
	Mat3f Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt*dx;
	Screw<3> t(get_twist(body) + divide_by_mass(body.aux.ds, body));
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		if (idx.a[a] < 0) continue;
		if (body.dynamic) {
			for (int j = 0; j < Jidx.size(); j++) {
				int b = Jidx[j];
				if (idx.a[b] < 0) continue;
				float Aab = dot(lin(J[i]),lin(J[j]))*mi + dot(ang(J[i]), Ii*ang(J[j]));
				if (Aab > epsilon)
					A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
			}
		}
		// int i = a%pm - 1, j = (a/pm)%pn - 1, k = (a/pm)/pn - 1;
		// Vec3f x0(i*dx,j*dx,k*dx), x1((i+1)*dx,(j+1)*dx,(k+1)*dx);
		// float f = coverage(body, x0,x1);
		// if (i >= 0 && i < m && j >= 0 && j < n && k >= 0 && k < o) B[idx.a[a]] += f;
		B[idx.a[a]] -= dot(J[i],t)*coeffr;
	}
}

template <> Screw<3> pressure_force (const Sand<3> &sand, const RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int pm = m + 2, pn = n + 2, po = o + 2;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// pressure_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.p_i;
	const Coupling3 &J = body.aux.coupling.p_J;
	Screw<3> f(0);
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		f += J[i]*sand.aux.p.s[a];
	}
	return Screw<3>(dx*dx*f);
}

void friction_xx_zz_J (const Sand<3> &sand, RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	Coupling3 &J = body.aux.coupling.f_xx_zz_J, &Jzz = body.aux.coupling.f_zz_J;
	J.assign(m*n*o, Screw<3>(0));
	Jzz.assign(m*n*o, Screw<3>(0));
	Vec3f com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	float z0 = com[2] - body.r, z1 = com[2] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	int k0 = floor(z0/dx), k1 = ceil(z1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	k0 = min(max(k0, 0), o);
	k1 = min(max(k1, 0), o);
	// x faces
	for (int k = k0; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0+1; i < i1; i++) {
		int a = (i-1)+m*(j+n*k), b = i+m*(j+n*k);
		Vec3f cx0((i-0.5f)*dx, j*dx, k*dx), cx1((i+0.5f)*dx, (j+1)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(1,0,0, 0, (k+0.5f)*dx - com[2], com[1] - (j+0.5f)*dx);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// z faces
	for (int k = k0+1; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+m*(j+n*(k-1)), b = i+m*(j+n*k);
		Vec3f cx0(i*dx, j*dx, (k-0.5f)*dx), cx1((i+1)*dx, (j+1)*dx, (k+0.5f)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,0,1, (j+0.5f)*dx - com[1], com[0] - (i+0.5f)*dx, 0);
		Jzz[a] += f*s;
		Jzz[b] -= f*s;
		J[a] -= Jzz[a];
		J[b] -= Jzz[b];
	}
	vector<int> &J_idx = body.aux.coupling.f_xx_zz_i, Jzz_idx = body.aux.coupling.f_zz_i;
	pack(J, J_idx);
	Jzz_idx = J_idx;
	project(Jzz, Jzz_idx);
}

void friction_yy_J (const Sand<3> &sand, RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	Coupling3 &J = body.aux.coupling.f_yy_J;
	J.assign(m*n*o, Screw<3>(0));
	Vec3f com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	float z0 = com[2] - body.r, z1 = com[2] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	int k0 = floor(z0/dx), k1 = ceil(z1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	k0 = min(max(k0, 0), o);
	k1 = min(max(k1, 0), o);
	// y faces
	for (int k = k0; k < k1; k++) for (int j = j0+1; j < j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+m*((j-1)+n*k), b = i+m*(j+n*k);
		Vec3f cx0(i*dx, (j-0.5f)*dx, k*dx), cx1((i+1)*dx, (j+0.5f)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,1,0, com[2] - (k+0.5f)*dx, 0, (i+0.5f)*dx - com[0]);
		J[a] += f*s;
		J[b] -= f*s;
	}
	pack(J, body.aux.coupling.f_yy_i);
}

void friction_xy_J (const Sand<3> &sand, RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	Coupling3 &J = body.aux.coupling.f_xy_J;
	J.assign(sm*sn*o, Screw<3>(0));
	Vec3f com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	float z0 = com[2] - body.r, z1 = com[2] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	int k0 = floor(z0/dx), k1 = ceil(z1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	k0 = min(max(k0, 0), o);
	k1 = min(max(k1, 0), o);
	// x faces
	for (int k = k0; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i <= i1; i++) {
		int a = i+sm*(j+sn*k), b = i+sm*((j+1)+sn*k);
		Vec3f cx0((i-0.5f)*dx, j*dx, k*dx), cx1((i+0.5f)*dx, (j+1)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(1,0,0, 0, (k+0.5f)*dx - com[2], com[1] - (j+0.5f)*dx);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// y faces
	for (int k = k0; k < k1; k++) for (int j = j0; j <= j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+sm*(j+sn*k), b = (i+1)+sm*(j+sn*k);
		Vec3f cx0(i*dx, (j-0.5f)*dx, k*dx), cx1((i+1)*dx, (j+0.5f)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,1,0, com[2] - (k+0.5f)*dx, 0, (i+0.5f)*dx - com[0]);
		J[a] += f*s;
		J[b] -= f*s;
	}
	pack(J, body.aux.coupling.f_xy_i);
}

void friction_xz_J (const Sand<3> &sand, RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, so = o+1;
	float dx = sand.dx;
	Coupling3 &J = body.aux.coupling.f_xz_J;
	J.assign(sm*n*so, Screw<3>(0));
	Vec3f com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	float z0 = com[2] - body.r, z1 = com[2] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	int k0 = floor(z0/dx), k1 = ceil(z1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	k0 = min(max(k0, 0), o);
	k1 = min(max(k1, 0), o);
	// x faces
	for (int k = k0; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i <= i1; i++) {
		int a = i+sm*(j+n*k), b = i+sm*(j+n*(k+1));
		Vec3f cx0((i-0.5f)*dx, j*dx, k*dx), cx1((i+0.5f)*dx, (j+1)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(1,0,0, 0, (k+0.5f)*dx - com[2], com[1] - (j+0.5f)*dx);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// z faces
	for (int k = k0; k <= k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+sm*(j+n*k), b = (i+1)+sm*(j+n*k);
		Vec3f cx0(i*dx, j*dx, (k-0.5f)*dx), cx1((i+1)*dx, (j+1)*dx, (k+0.5f)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,0,1, (j+0.5f)*dx - com[1], com[0] - (i+0.5f)*dx, 0);
		J[a] += f*s;
		J[b] -= f*s;
	}
	pack(J, body.aux.coupling.f_xz_i);
}

void friction_yz_J (const Sand<3> &sand, RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sn = n+1, so = o+1;
	float dx = sand.dx;
	Coupling3 &J = body.aux.coupling.f_yz_J;
	J.assign(m*sn*so, Screw<3>(0));
	Vec3f com = body.pose.x;
	float x0 = com[0] - body.r, x1 = com[0] + body.r;
	float y0 = com[1] - body.r, y1 = com[1] + body.r;
	float z0 = com[2] - body.r, z1 = com[2] + body.r;
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	int k0 = floor(z0/dx), k1 = ceil(z1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	k0 = min(max(k0, 0), o);
	k1 = min(max(k1, 0), o);
	// y faces
	for (int k = k0; k < k1; k++) for (int j = j0; j <= j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+m*(j+sn*k), b = i+m*(j+sn*(k+1));
		Vec3f cx0(i*dx, (j-0.5f)*dx, k*dx), cx1((i+1)*dx, (j+0.5f)*dx, (k+1)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,1,0, com[2] - (k+0.5f)*dx, 0, (i+0.5f)*dx - com[0]);
		J[a] += f*s;
		J[b] -= f*s;
	}
	// z faces
	for (int k = k0; k <= k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
		int a = i+m*(j+sn*k), b = i+m*((j+1)+sn*k);
		Vec3f cx0(i*dx, j*dx, (k-0.5f)*dx), cx1((i+1)*dx, (j+1)*dx, (k+0.5f)*dx);
		float f = coverage(body, cx0,cx1);
		Screw<3> s(0,0,1, (j+0.5f)*dx - com[1], com[0] - (i+0.5f)*dx, 0);
		J[a] += f*s;
		J[b] -= f*s;
	}
	pack(J, body.aux.coupling.f_yz_i);
}

void friction_xx_zz_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	// static vector< Screw<3> > J, Jzz;
	// friction_xx_zz_J(sand, body, J, Jzz);
	const vector<int> &Jidx = body.aux.coupling.f_xx_zz_i;
	const Coupling3 &J = body.aux.coupling.f_xx_zz_J, &Jzz = body.aux.coupling.f_zz_J;
	// A += dt^2 dx J' M_inv J
	// b += dt/dx J' v
	float mi = m_inv(body);
	Mat3f Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt*dx;
	Screw<3> t(get_twist(body) + divide_by_mass(body.aux.dp + body.aux.ds, body));
	// i, j: index into J
	// a, b: index into full m*n*o field
	// idx.a[a,b,c]: index into A, b
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		if (idx.a[a] < 0) continue;
		if (body.dynamic) {
			for (int j = 0; j < Jidx.size(); j++) {
				int b = Jidx[j];
				if (idx.a[b] < 0) continue;
				float Aab = dot(lin(J[i]),lin(J[j]))*mi + dot(ang(J[i]), Ii*ang(J[j]));
				if (Aab > epsilon)
					A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
				float syy = sand.aux.s.yy.s[b];
				float Bab = dot(lin(J[i]),lin(Jzz[j])*syy)*mi + dot(ang(J[i]), Ii*ang(Jzz[j])*syy);
				B[idx.a[a]] -= Bab*coeffl;
			}
		}
		B[idx.a[a]] -= dot(J[i],t)*coeffr;
	}
}

void friction_yy_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_yy_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_yy_i;
	const Coupling3 &J = body.aux.coupling.f_yy_J;
	// A += dt^2 dx J' M_inv J
	// b += dt/dx J' v
	float mi = m_inv(body);
	Mat3f Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt*dx;
	Screw<3> t(get_twist(body) + divide_by_mass(body.aux.dp + body.aux.ds, body));
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		if (idx.a[a] < 0) continue;
		if (body.dynamic) {
			for (int j = 0; j < Jidx.size(); j++) {
				int b = Jidx[j];
				if (idx.a[b] < 0) continue;
				float Aab = dot(lin(J[i]),lin(J[j]))*mi + dot(ang(J[i]), Ii*ang(J[j]));
				if (Aab > epsilon)
					A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
			}
		}
		B[idx.a[a]] -= dot(J[i],t)*coeffr;
	}
}

void friction_xy_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_xy_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_xy_i;
	const Coupling3 &J = body.aux.coupling.f_xy_J;
	// A += dt^2 dx J' M_inv J
	// b += dt/dx J' v
	float mi = m_inv(body);
	Mat3f Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt*dx;
	Screw<3> t(get_twist(body) + divide_by_mass(body.aux.dp + body.aux.ds, body));
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		if (idx.a[a] < 0) continue;
		if (body.dynamic) {
			for (int j = 0; j < Jidx.size(); j++) {
				int b = Jidx[j];
				if (idx.a[b] < 0) continue;
				float Aab = dot(lin(J[i]),lin(J[j]))*mi + dot(ang(J[i]), Ii*ang(J[j]));
				if (Aab > epsilon)
					A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
			}
		}
		B[idx.a[a]] -= dot(J[i],t)*coeffr;
	}
}

void friction_xz_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, so = o+1;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_xz_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_xz_i;
	const Coupling3 &J = body.aux.coupling.f_xz_J;
	// A += dt^2 dx J' M_inv J
	// b += dt/dx J' v
	float mi = m_inv(body);
	Mat3f Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt*dx;
	Screw<3> t(get_twist(body) + divide_by_mass(body.aux.dp + body.aux.ds, body));
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		if (idx.a[a] < 0) continue;
		if (body.dynamic) {
			for (int j = 0; j < Jidx.size(); j++) {
				int b = Jidx[j];
				if (idx.a[b] < 0) continue;
				float Aab = dot(lin(J[i]),lin(J[j]))*mi + dot(ang(J[i]), Ii*ang(J[j]));
				if (Aab > epsilon)
					A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
			}
		}
		B[idx.a[a]] -= dot(J[i],t)*coeffr;
	}
}

void friction_yz_coupling (const Sand<3> &sand, const RigidBody<3> &body, float dt, const IntArray<3> &idx, SparseMatrixf &A, vector<float> &B) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sn = n+1, so = o+1;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_yz_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_yz_i;
	const Coupling3 &J = body.aux.coupling.f_yz_J;
	// A += dt^2 dx J' M_inv J
	// b += dt/dx J' v
	float mi = m_inv(body);
	Mat3f Ii = world_I_inv(body);
	float coeffr = dt/dx, coeffl = dt*dt*dx;
	Screw<3> t(get_twist(body) + divide_by_mass(body.aux.dp + body.aux.ds, body));
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		if (idx.a[a] < 0) continue;
		if (body.dynamic) {
			for (int j = 0; j < Jidx.size(); j++) {
				int b = Jidx[j];
				if (idx.a[b] < 0) continue;
				float Aab = dot(lin(J[i]),lin(J[j]))*mi + dot(ang(J[i]), Ii*ang(J[j]));
				if (Aab > epsilon)
					A.add_to_element(idx.a[a],idx.a[b], Aab*coeffl);
			}
		}
		B[idx.a[a]] -= dot(J[i],t)*coeffr;
	}
}

Screw<3> friction_xx_zz_force (const Sand<3> &sand, const RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	// static vector< Screw<3> > J, Jzz;
	// friction_xx_zz_J(sand, body, J, Jzz);
	const vector<int> &Jidx = body.aux.coupling.f_xx_zz_i;
	const Coupling3 &J = body.aux.coupling.f_xx_zz_J, &Jzz = body.aux.coupling.f_zz_J;
	Screw<3> f(0);
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		f += J[i]*sand.aux.s.xx.s[a];
		f -= Jzz[i]*sand.aux.s.yy.s[a];
	}
	return Screw<3>(dx*dx*f);
}

Screw<3> friction_yy_force (const Sand<3> &sand, const RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_yy_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_yy_i;
	const Coupling3 &J = body.aux.coupling.f_yy_J;
	Screw<3> f(0);
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		f += J[i]*sand.aux.s.yy.s[a];
	}
	return Screw<3>(dx*dx*f);
}

Screw<3> friction_xy_force (const Sand<3> &sand, const RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_xy_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_xy_i;
	const Coupling3 &J = body.aux.coupling.f_xy_J;
	Screw<3> f(0);
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		f += J[i]*sand.aux.s.xy.s[a];
	}
	return Screw<3>(dx*dx*f);
}

Screw<3> friction_xz_force (const Sand<3> &sand, const RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, so = o+1;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_xz_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_xz_i;
	const Coupling3 &J = body.aux.coupling.f_xz_J;
	Screw<3> f(0);
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		f += J[i]*sand.aux.s.xz.s[a];
	}
	return Screw<3>(dx*dx*f);
}

Screw<3> friction_yz_force (const Sand<3> &sand, const RigidBody<3> &body) {
	int m = sand.m, n = sand.n, o = sand.o;
	int sn = n+1, so = o+1;
	float dx = sand.dx;
	// static vector< Screw<3> > J;
	// friction_yz_J(sand, body, J);
	const vector<int> &Jidx = body.aux.coupling.f_yz_i;
	const Coupling3 &J = body.aux.coupling.f_yz_J;
	Screw<3> f(0);
	for (int i = 0; i < Jidx.size(); i++) {
		int a = Jidx[i];
		f += J[i]*sand.aux.s.yz.s[a];
	}
	return Screw<3>(dx*dx*f);
}

template <> void precompute_coupling_matrices (const Sand<3> &sand, RigidBody<3> &body) {
	pressure_J(sand, body);
	friction_xx_zz_J(sand, body);
	friction_yy_J(sand, body);
	friction_xy_J(sand, body);
	friction_yz_J(sand, body);
	friction_xz_J(sand, body);
}
