/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "friction.hpp"

#include "../math/qp.hpp"
#include "coupling.hpp"
#include "simhelpers.hpp"
#include <fstream>

using namespace std;

#define dropdim(i,j,m,n) ((i) +(j)*(m))

template <class T>
void writeMatlab(std::vector<T> &v, char *filename, char *variable) {
	FILE *file = fopen(filename, "w");
	fprintf(file,"%s = [",variable);
	for ( int x = 0; x < (int)v.size(); x++ ) {
		fprintf(file, "%f ",v[x]);
	}
	fprintf(file,"];\n");
	fclose(file);
}

// ---------------- //
//        2D        //
// ---------------- //

float friction_xx_yy_solve (Sand<2> &sand, vector< RigidBody<2> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static vector<float> smin, smax;
	static IntArray<2> idx;
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	idx.initialize(m,n, -1);
	int nactive = 0;
	// enumerate active cells
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		bool active = (sand.rho(i,j) > 0);
		if (active)
			idx(i,j) = nactive++;
	}
	if (nactive == 0) {
		set_all(sand.aux.s.xx, 0);
		set_all(sand.aux.s.yy, 0);
		return 0;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	smin.resize(nactive, 0);
	smax.resize(nactive, 0);
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = i==0 ? -1 : idx(i-1,j), b = i==m ? -1 : idx(i,j);
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float vx = sand.v.x(i,j) + sand.aux.dp.x(i,j) + sand.aux.ds.x(i,j);
		float f_obs = sand.aux.f_obs_s.x(i,j);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*(vx*mass + f_obs*sand.aux.v_obs.x(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = j==0 ? -1 : idx(i,j-1), b = j==n ? -1 : idx(i,j);
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float vy = sand.v.y(i,j) + sand.aux.dp.y(i,j) + sand.aux.ds.y(i,j);
		float f_obs = sand.aux.f_obs_s.y(i,j);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*(vy*mass + f_obs*sand.aux.v_obs.y(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] += fmultiplier; // because s_yy = -s_xx
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] -= fmultiplier; // because s_yy = -s_xx
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j);
		if (a >= 0) {
			x[a] = sand.aux.s.xx(i,j);
			float sabs = sand.mu * sand.aux.p(i+1,j+1);
			smin[a] = -sabs;
			smax[a] = +sabs;
		}
	}
	for (int b = 0; b < bodies.size(); b++) {
		friction_xx_yy_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, smin, smax, x, report);
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j);
		sand.aux.s.xx(i,j) = a<0 ? 0 : x[a];
		sand.aux.s.yy(i,j) = -sand.aux.s.xx(i,j);
	}
	return report.residual_initial;
}

float friction_xy_solve (Sand<2> &sand, vector< RigidBody<2> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static vector<float> smin, smax;
	static IntArray<2> idx;
	int m = sand.m, n = sand.n;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	idx.initialize(sm,sn, -1);
	int nactive = 0;
	// enumerate active cells
	for (int j = 0; j <= n; j++) for (int i = 0; i <= m; i++) {
		bool active = true;
		if (i > 0 && j > 0 && sand.rho(i-1,j-1)==0) active = false;
		if (i > 0 && j < n && sand.rho(i-1,j)==0)   active = false;
		if (i < m && j > 0 && sand.rho(i,j-1)==0)   active = false;
		if (i < m && j < n && sand.rho(i,j)==0)     active = false;
		if (active)
			idx(i,j) = nactive++;
	}
	if (nactive == 0) {
		set_all(sand.aux.s.xy, 0);
		return 0;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	smin.assign(nactive, 0);
	smax.assign(nactive, 0);
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j), b = idx(i,j+1);
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float vx = sand.v.x(i,j) + sand.aux.dp.x(i,j) + sand.aux.ds.x(i,j);
		float f_obs = sand.aux.f_obs_s.x(i,j);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*(vx*mass + f_obs*sand.aux.v_obs.x(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j), b = idx(i+1,j);
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float vy = sand.v.y(i,j) + sand.aux.dp.y(i,j) + sand.aux.ds.y(i,j);
		float f_obs = sand.aux.f_obs_s.y(i,j);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*(vy*mass + f_obs*sand.aux.v_obs.y(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	for (int j = 0; j <= n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j);
		if (a >= 0) {
			x[a] = sand.aux.s.xy(i,j);
			float sabs = sand.mu * (sand.aux.p(i,j) + sand.aux.p(i+1,j) + sand.aux.p(i,j+1) + sand.aux.p(i+1,j+1))*0.25f;
			smin[a] = -sabs;
			smax[a] = +sabs;
		}
	}
	for (int b = 0; b < bodies.size(); b++) {
		friction_xy_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, smin, smax, x, report);
	for (int j = 0; j <= n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j);
		sand.aux.s.xy(i,j) = a<0 ? 0 : x[a];
	}
	return report.residual_initial;
}

void friction_xx_yy_apply (Sand<2> &sand, std::vector< RigidBody<2> > &bodies, float dt) {
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	float coeff = dt/(dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 1; i < m; i++) {
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.x(i,j);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.x(i,j) -= (mass<epsilon) ? 0 : (sand.aux.s.xx(i,j) - sand.aux.s.xx(i-1,j))*multiplier;
	}
	// y faces
	for (int j = 1; j < n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.y(i,j);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.y(i,j) -= (mass<epsilon) ? 0 : (sand.aux.s.yy(i,j) - sand.aux.s.yy(i,j-1))*multiplier;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds += dt*friction_xx_yy_force(sand, bodies[b]);
	}
}

void friction_xy_apply (Sand<2> &sand, std::vector< RigidBody<2> > &bodies, float dt) {
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	float coeff = dt/(dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.x(i,j);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.x(i,j) -= (mass<epsilon) ? 0 : (sand.aux.s.xy(i,j+1) - sand.aux.s.xy(i,j))*multiplier;
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.y(i,j);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.y(i,j) -= (mass<epsilon) ? 0 : (sand.aux.s.xy(i+1,j) - sand.aux.s.xy(i,j))*multiplier;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds += dt*friction_xy_force(sand, bodies[b]);
	}
}

template <> void friction_apply (Sand<2> &sand, vector< RigidBody<2> >& bodies, float dt) {
	set_all(sand.aux.ds, Vec2f(0));
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds = Screw<2>(0);
	}
	friction_xx_yy_apply(sand, bodies, dt);
	friction_xy_apply(sand, bodies, dt);
}

template <> float friction_solve (Sand<2> &sand, vector< RigidBody<2> > &bodies, float dt) {
	float r = 0;
	friction_xx_yy_apply(sand, bodies, -dt);
	r += friction_xx_yy_solve(sand, bodies, dt);
	friction_xx_yy_apply(sand, bodies, dt);
	friction_xy_apply(sand, bodies, -dt);
	r += friction_xy_solve(sand, bodies, dt);
	friction_xy_apply(sand, bodies, dt);
	return r;
}

// ---------------- //
//        3D        //
// ---------------- //

float friction_yy_solve (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static vector<float> smin, smax;
	static IntArray<3> idx;
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	idx.initialize(m,n,o, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		bool active = (sand.rho(i,j,k) > 0);
		if (active)
			idx(i,j,k) = nactive++;
	}
	if (nactive == 0) {
		set_all(sand.aux.s.yy, 0);
		return 0;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	smin.resize(nactive, 0);
	smax.resize(nactive, 0);
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = j==0 ? -1 : idx(i,j-1,k), b = j==n ? -1 : idx(i,j,k);
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float vy = sand.v.y(i,j,k) + sand.aux.dp.y(i,j,k) + sand.aux.ds.y(i,j,k);
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vy*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k);
		if (a >= 0) {
			x[a] = sand.aux.s.yy(i,j,k);
			float sabs = sand.mu * sand.aux.p(i+1,j+1,k+1);
			smin[a] = -sabs;
			smax[a] = +sabs;
		}
	}
	for (int b = 0; b < bodies.size(); b++) {
		friction_yy_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, smin, smax, x, report);
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k);
		sand.aux.s.yy(i,j,k) = a<0 ? 0 : x[a];
	}
	return report.residual_initial;
}

float friction_xx_zz_solve (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static vector<float> smin, smax;
	static IntArray<3> idx;
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	idx.initialize(m,n,o, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		bool active = (sand.rho(i,j,k) > 0);
		if (active)
			idx(i,j,k) = nactive++;
	}
	if (nactive == 0) {
		set_all(sand.aux.s.xx, 0);
		set_all(sand.aux.s.zz, 0);
		return 0;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	smin.resize(nactive, 0);
	smax.resize(nactive, 0);
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = i==0 ? -1 : idx(i-1,j,k), b = i==m ? -1 : idx(i,j,k);
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float vx = sand.v.x(i,j,k) + sand.aux.dp.x(i,j,k) + sand.aux.ds.x(i,j,k);
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vx*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = k==0 ? -1 : idx(i,j,k-1), b = k==o ? -1 : idx(i,j,k);
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float vz = sand.v.z(i,j,k) + sand.aux.dp.z(i,j,k) + sand.aux.ds.z(i,j,k);
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vz*mass;
		if (k > 0 && k < o) {
			// because szz = -syy - sxx
			// vz += (Delta_z s_yy) * coeffr*(1-f_obs)/rho_z
			// => fmultiplier += (Delta_z s_yy) * coeffr*(1-f_obs)/rho_z * coeffr*(1-f_obs)*mass
			// => fmultiplier += (Delta_z s_yy) * coeffr*coeffr*mass/rho_z * sqr(1-f_obs)
			fmultiplier += (sand.aux.s.yy(i,j,k)-sand.aux.s.yy(i,j,k-1))*multiplier;
		}
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] += fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] -= fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k);
		if (a >= 0) {
			x[a] = (sand.aux.s.xx(i,j,k) - sand.aux.s.yy(i,j,k) - sand.aux.s.zz(i,j,k))/2;
			float syy = sand.aux.s.yy(i,j,k);
			float sabs = sand.mu * sand.aux.p(i+1,j+1,k+1);
			smin[a] = max(-sabs, -sabs-syy);  // szz = -syy - sxx <= sabs
			smax[a] = min(+sabs, +sabs-syy);  // szz = -syy - sxx >= -sabs
		}
	}
	for (int b = 0; b < bodies.size(); b++) {
		friction_xx_zz_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, smin, smax, x, report);
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k);
		float syy = sand.aux.s.yy(i,j,k);
		sand.aux.s.xx(i,j,k) = a<0 ? 0 : x[a];
		sand.aux.s.zz(i,j,k) = -syy - sand.aux.s.xx(i,j,k);
	}
	return report.residual_initial;
}

float friction_xz_solve (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static vector<float> smin, smax;
	static IntArray<3> idx;
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, so = o+1;
	float dx = sand.dx;
	idx.initialize(sm,n,so, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		bool active = true;
		if (i > 0 && k > 0 && sand.rho(i-1,j,k-1)==0) active = false;
		if (i > 0 && k < o && sand.rho(i-1,j,k)==0)   active = false;
		if (i < m && k > 0 && sand.rho(i,j,k-1)==0)   active = false;
		if (i < m && k < o && sand.rho(i,j,k)==0)     active = false;
		if (active)
			idx(i,j,k) = nactive++;
	}
	if (nactive == 0) {
		set_all(sand.aux.s.xz, 0);
		return 0;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	smin.assign(nactive, 0);
	smax.assign(nactive, 0);
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j,k), b = idx(i,j,k+1);
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float vx = sand.v.x(i,j,k) + sand.aux.dp.x(i,j,k) + sand.aux.ds.x(i,j,k);
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vx*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k), b = idx(i+1,j,k);
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float vz = sand.v.z(i,j,k) + sand.aux.dp.z(i,j,k) + sand.aux.ds.z(i,j,k);
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vz*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j,k);
		if (a >= 0) {
			x[a] = sand.aux.s.xz(i,j,k);
			float sabs = sand.mu * (sand.aux.p(i,j+1,k) + sand.aux.p(i+1,j+1,k) + sand.aux.p(i,j+1,k+1) + sand.aux.p(i+1,j+1,k+1))*0.25f;
			smin[a] = -sabs;
			smax[a] = +sabs;
		}
	}
	for (int b = 0; b < bodies.size(); b++) {
		friction_xz_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, smin, smax, x, report);
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j,k);
		sand.aux.s.xz(i,j,k) = a<0 ? 0 : x[a];
	}
	return report.residual_initial;
}

float friction_xy_solve (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static vector<float> smin, smax;
	static IntArray<3> idx;
	int m = sand.m, n = sand.n, o = sand.o;
	int sm = m+1, sn = n+1;
	float dx = sand.dx;
	idx.initialize(sm,sn,o, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i <= m; i++) {
		bool active = true;
		if (i > 0 && j > 0 && sand.rho(i-1,j-1,k)==0) active = false;
		if (i > 0 && j < n && sand.rho(i-1,j,k)==0)   active = false;
		if (i < m && j > 0 && sand.rho(i,j-1,k)==0)   active = false;
		if (i < m && j < n && sand.rho(i,j,k)==0)     active = false;
		if (active)
			idx(i,j,k) = nactive++;
	}
	if (nactive == 0) {
		set_all(sand.aux.s.xy, 0);
		return 0;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	smin.assign(nactive, 0);
	smax.assign(nactive, 0);
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j,k), b = idx(i,j+1,k);
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float vx = sand.v.x(i,j,k) + sand.aux.dp.x(i,j,k) + sand.aux.ds.x(i,j,k);
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vx*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k), b = idx(i+1,j,k);
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float vy = sand.v.y(i,j,k) + sand.aux.dp.y(i,j,k) + sand.aux.ds.y(i,j,k);
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vy*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j,k);
		if (a >= 0) {
			x[a] = sand.aux.s.xy(i,j,k);
			float sabs = sand.mu * (sand.aux.p(i,j,k+1) + sand.aux.p(i+1,j,k+1) + sand.aux.p(i,j+1,k+1) + sand.aux.p(i+1,j+1,k+1))*0.25f;
			smin[a] = -sabs;
			smax[a] = +sabs;
		}
	}
	for (int b = 0; b < bodies.size(); b++) {
		friction_xy_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, smin, smax, x, report);
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j,k);
		sand.aux.s.xy(i,j,k) = a<0 ? 0 : x[a];
	}
	return report.residual_initial;
}

float friction_yz_solve (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static vector<float> smin, smax;
	static IntArray<3> idx;
	int m = sand.m, n = sand.n, o = sand.o;
	int sn = n+1, so = o+1;
	float dx = sand.dx;
	idx.initialize(m,sn,so, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = 0; k <= o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		bool active = true;
		if (j > 0 && k > 0 && sand.rho(i,j-1,k-1)==0) active = false;
		if (j > 0 && k < o && sand.rho(i,j-1,k)==0)   active = false;
		if (j < n && k > 0 && sand.rho(i,j,k-1)==0)   active = false;
		if (j < n && k < o && sand.rho(i,j,k)==0)     active = false;
		if (active)
			idx(i,j,k) = nactive++;
	}
	if (nactive == 0) {
		set_all(sand.aux.s.yz, 0);
		return 0;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	smin.assign(nactive, 0);
	smax.assign(nactive, 0);
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k), b = idx(i,j,k+1);
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float vy = sand.v.y(i,j,k) + sand.aux.dp.y(i,j,k) + sand.aux.ds.y(i,j,k);
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vy*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k), b = idx(i,j+1,k);
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float vz = sand.v.z(i,j,k) + sand.aux.dp.z(i,j,k) + sand.aux.ds.z(i,j,k);
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float multiplier = coeffl*sqr(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vz*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	for (int k = 0; k <= o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k);
		if (a >= 0) {
			x[a] = sand.aux.s.yz(i,j,k);
			float sabs = sand.mu * (sand.aux.p(i+1,j,k) + sand.aux.p(i+1,j+1,k) + sand.aux.p(i+1,j,k+1) + sand.aux.p(i+1,j+1,k+1))*0.25f;
			smin[a] = -sabs;
			smax[a] = +sabs;
		}
	}
	for (int b = 0; b < bodies.size(); b++) {
		friction_yz_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, smin, smax, x, report);
	for (int k = 0; k <= o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i,j,k);
		sand.aux.s.yz(i,j,k) = a<0 ? 0 : x[a];
	}
	return report.residual_initial;
}

void friction_yy_apply (Sand<3> &sand, std::vector< RigidBody<3> > &bodies, float dt) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	float coeff = dt/(dx*sand.rho_max);
	// y faces
	for (int k = 0; k < o; k++) for (int j = 1; j < n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.y(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.yy(i,j,k) - sand.aux.s.yy(i,j-1,k))*multiplier;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds += dt*friction_yy_force(sand, bodies[b]);
	}
}

void friction_xx_zz_apply (Sand<3> &sand, std::vector< RigidBody<3> > &bodies, float dt) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	float coeff = dt/(dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 1; i < m; i++) {
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.x(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.xx(i,j,k) - sand.aux.s.xx(i-1,j,k))*multiplier;
	}
	// z faces
	for (int k = 1; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.z(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.zz(i,j,k) - sand.aux.s.zz(i,j,k-1))*multiplier;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds += dt*friction_xx_zz_force(sand, bodies[b]);
	}
}

void friction_xy_apply (Sand<3> &sand, std::vector< RigidBody<3> > &bodies, float dt) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	float coeff = dt/(dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.x(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.xy(i,j+1,k) - sand.aux.s.xy(i,j,k))*multiplier;
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.y(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.xy(i+1,j,k) - sand.aux.s.xy(i,j,k))*multiplier;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds += dt*friction_xy_force(sand, bodies[b]);
	}
}

void friction_xz_apply (Sand<3> &sand, std::vector< RigidBody<3> > &bodies, float dt) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	float coeff = dt/(dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.x(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.xz(i,j,k+1) - sand.aux.s.xz(i,j,k))*multiplier;
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.z(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.xz(i+1,j,k) - sand.aux.s.xz(i,j,k))*multiplier;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds += dt*friction_xz_force(sand, bodies[b]);
	}
}

void friction_yz_apply (Sand<3> &sand, std::vector< RigidBody<3> > &bodies, float dt) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	float coeff = dt/(dx*sand.rho_max);
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.y(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.yz(i,j,k+1) - sand.aux.s.yz(i,j,k))*multiplier;
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float multiplier = coeff*(1-f_obs)/mass;
		sand.aux.ds.z(i,j,k) -= (mass<epsilon) ? 0 : (sand.aux.s.yz(i,j+1,k) - sand.aux.s.yz(i,j,k))*multiplier;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds += dt*friction_yz_force(sand, bodies[b]);
	}
}

template <> void friction_apply (Sand<3> &sand, vector< RigidBody<3> >& bodies, float dt) {
	set_all(sand.aux.ds, Vec3f(0));
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.ds = Screw<3>(0);
	}
	friction_yy_apply(sand, bodies, dt);
	friction_xx_zz_apply(sand, bodies, dt);
	friction_xy_apply(sand, bodies, dt);
	friction_xz_apply(sand, bodies, dt);
	friction_yz_apply(sand, bodies, dt);
}

template <> float friction_solve (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	float r = 0;

	friction_yy_apply(sand, bodies, -dt);
	r += friction_yy_solve(sand, bodies, dt);
	friction_yy_apply(sand, bodies, dt);

	friction_xx_zz_apply(sand, bodies, -dt);
	r += friction_xx_zz_solve(sand, bodies, dt);
	friction_xx_zz_apply(sand, bodies, dt);

	friction_xy_apply(sand, bodies, -dt);
	r += friction_xy_solve(sand, bodies, dt);
	friction_xy_apply(sand, bodies, dt);

	friction_yz_apply(sand, bodies, -dt);
	r += friction_yz_solve(sand, bodies, dt);
	friction_yz_apply(sand, bodies, dt);

	friction_xz_apply(sand, bodies, -dt);
	r += friction_xz_solve(sand, bodies, dt);
	friction_xz_apply(sand, bodies, dt);

	return r;
}
