/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "particle.hpp"
#include <algorithm>
#include <vector>
#include <cstdlib>
#include <assert.h>
#include <functional>
#include "../math/util.h"
#include <lapacke.h>
using namespace tvmet;
using namespace std;

#define USE_LAPACK_EIGEN_DECOMP

template <size_t d> Matrix<float,d,d> advect_radii (const Matrix<float,d,d> &r2, const Matrix<float,d,d> &J) {
	Matrix<float,d,d> J1(identity<float,d,d>() + J);
	return Matrix<float,d,d>(J1*r2*trans(J1));
}
template Mat2f advect_radii (const Mat2f &r2, const Mat2f &J);
template Mat3f advect_radii (const Mat3f &r2, const Mat3f &J);

template <size_t d> Matrix<float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2) {
	Vector<float,d> eigenVals;
	return decomposeMatrix<d>(r2, eigenVals);
}

template Mat2f decomposeMatrix(const Mat2f &r2);
template Mat3f decomposeMatrix(const Mat3f &r2);

// template <> Mat2f decomposeMatrix(const Mat2f &r2, Vec2f &eigenVals) {
// 	float a = r2(0,0), b = r2(0,1), c = r2(1,1);
// 	float tr = a + c;
// 	float det = a*c - b*b;
// 	float e1, e2;
// 	Vec2f v1, v2;
// 	e1 = (tr + sqrtf(tr*tr - 4*det))*0.5f;
// 	e2 = (tr - sqrtf(tr*tr - 4*det))*0.5f;
// 	v1 = Vec2f(normalize(fabs(e1-a)>fabs(e1-c) ? Vec2f(b,e1-a) : Vec2f(e1-c,b)));
// 	v2 = Vec2f(normalize(fabs(e2-a)>fabs(e2-c) ? Vec2f(b,e2-a) : Vec2f(e2-c,b)));
// 	Mat2f m;
// 	m = v1[0], v2[0],
// 		v1[1], v2[1];
// 	eigenVals[0] = e1;
// 	eigenVals[1] = e2;
// 	//return Mat2f(m*diagonal(Vec2f(sqrtf(e1),sqrtf(e2)))*trans(m));
// 	return m;
// }

bool UDgreater ( float elem1, float elem2 ) {
	return fabsf(elem1) > fabsf(elem2);
}

template <class T> bool tolEqual ( T elem1, T elem2, T tol ) {
	assert ( tol >= 0 );
	return abs<T>(elem1 - elem2)/(abs<T>(elem1) + abs<T>(elem2)) < tol;
}


#ifdef USE_LAPACK_EIGEN_DECOMP
template <> Mat2f decomposeMatrix(const Mat2f &r2, Vec2f &eigenVals) {
	float a = r2(0,0), b = r2(0,1), c = r2(1,1);
	float tr = a + c;
	float det = a*c - b*b;
	float e1, e2;
	Vec2f v1, v2;
	e1 = (tr + sqrtf(tr*tr - 4*det))*0.5f;
	e2 = (tr - sqrtf(tr*tr - 4*det))*0.5f;
	v1 = Vec2f(normalize(fabs(e1-a)>fabs(e1-c) ? Vec2f(b,e1-a) : Vec2f(e1-c,b)));
	v2 = Vec2f(normalize(fabs(e2-a)>fabs(e2-c) ? Vec2f(b,e2-a) : Vec2f(e2-c,b)));
	Mat2f m;
	m = v1[0], v2[0],
		v1[1], v2[1];
	eigenVals[0] = e1;
	eigenVals[1] = e2;
	//return Mat2f(m*diagonal(Vec2f(sqrtf(e1),sqrtf(e2)))*trans(m));
	return m;
}

template <> Mat3f decomposeMatrix(const Mat3f &r2, Vec3f &eigenVals) {
	float a = r2(0,0), b = r2(1,1), c = r2(2,2), d = r2(0,1), e = r2(0,2), f = r2(1,2);
	float m = (a + b + c) / 3;
	Mat3f K = (Mat3f)(r2 - tvmet::identity<float, 3, 3>() * m);
	float ka = a - m, kb = b - m, kc = c - m;
	float q = (ka * (kb * kc - sqf(f)) + d * (f * e - d * kc) + e * (d * f - e * kb))/2;
	float p = ( sqf(ka) + sqf(kb) + sqf(kc) + 2 * (sqf(d) + sqf(e) + sqf(f)) ) / 6;

	float phi = 1/3 * atanf(sqrtf(fmaxf(p * sqf(p) - sqf(q), 0))/(q+0.000001f));
	if ( phi < 0 )
		phi += (float)M_PI / 3;

	float sp = sqrtf(p), cphi = cosf(phi), sphi = sinf(phi), st = sqrtf(3);
	vector<float> ev(3);
	ev[0] = m + 2 * sp * cphi;
	ev[1] = m - sp * (cphi + st * sphi);
	ev[2] = m - sp * (cphi - st * sphi);
	/// Sort the eigenvalues in order of magnitude - assuming no negative eigenvalues, else resort to UDgreater
	//sort(ev.begin(), ev.end(), greater<float>());
	sort(ev.begin(), ev.end(), UDgreater);
	eigenVals(0) = ev[0]; eigenVals(1) = ev[1]; eigenVals(2) = ev[2];

	return Mat3f(tvmet::identity<float, 3, 3>());
}
/*
template <> Mat3f decomposeMatrix(const Mat3f &r2, Vec3f &eigenVals) {
	int n = 3, lda = 3, info, lwork = 37, liwork = 18;
	int iwkopt = 18;
	//int* iwork;
	int iwork[18];
	float wkopt;
	//float* work;
	float work[37];
	float w[3], a[9];
	for ( int i = 0; i < 3; i++ ) {
		for ( int j = 0; j <= i; j++ ) {
			a[i * 3 + j] = r2(i, j);
		}
	}
	//lwork = liwork = -1;
	//LAPACK_ssyevd("V", "U", &n, a, &lda, w, &wkopt, &lwork, &iwkopt, &liwork, &info );
	//lwork = (int)wkopt;
	//work = new float[lwork];
	//liwork = iwkopt;
	//iwork = new int[liwork];
	//cout << "Debug: lwork: " << lwork << " liwork: " << liwork << " iwkopt: " << iwkopt << endl;

	LAPACK_ssyevd("N", "U", &n, a, &lda, w, work, &lwork, iwork, &liwork, &info);
	if ( info > 0 ) {
		cout << "Error: Eigenvalue computation failed for matrix: " << r2 << endl;
	}
	eigenVals[0] = w[2]; eigenVals[1] = w[1]; eigenVals[2] = w[0];
	//cout << "Debug: eigenvalues: " << eigenVals << endl;
	Mat3f m = Mat3f(tvmet::identity<float, 3, 3>());
	//m = a[6], a[3], a[0],
	//	a[7], a[4], a[1],
	//	a[8], a[5], a[2];

	//delete work;
	//delete iwork;
	return m;
}*/

template <> Mat2f decomposeMatrix(const Mat2f &r2, Vec2f &eigenVals, Vec2f eigenVecs[2]) {
	Mat2f m;
	int n = 2, lda = 2, info, lwork = 21, liwork = 13;
	int iwkopt = 13;
	int iwork[13];
	float wkopt;
	float work[21];
	float w[2], a[4];
	for ( int i = 0; i < 2; i++ ) {
		for ( int j = 0; j <= i; j++ ) {
			a[i * 2 + j] = r2(i, j);
		}
	}

        char U[]="U",V[]="V";
	LAPACK_ssyevd(V, U, &n, a, &lda, w, work, &lwork, iwork, &liwork, &info);
	if ( info > 0 ) {
		cout << "Error: Eigenvalue computation failed for matrix: " << r2 << endl;
	}
	eigenVals[0] = w[1]; eigenVals[1] = w[0];
	m = a[2], a[0],
		a[3], a[1];
	
	for ( int i = 0; i < 2; i++ ) {
		for ( int j = 0; j < 2; j++ ) {
			eigenVecs[j][i] = m(i, j);
		}
	}
	return m;
}

template <> Mat3f decomposeMatrix(const Mat3f &r2, Vec3f &eigenVals, Vec3f eigenVecs[3]) {
	Mat3f m;
	int n = 3, lda = 3, info, lwork = 37, liwork = 18;
	int iwkopt = 18;
	int iwork[18];
	float wkopt;
	float work[37];
	float w[3], a[9];
	for ( int i = 0; i < 3; i++ ) {
		for ( int j = 0; j <= i; j++ ) {
			a[i * 3 + j] = r2(i, j);
		}
	}

        char U[]="U",V[]="V";
	LAPACK_ssyevd(V, U, &n, a, &lda, w, work, &lwork, iwork, &liwork, &info);
	if ( info > 0 ) {
		cout << "Error: Eigenvalue computation failed for matrix: " << r2 << endl;
	}
	eigenVals[0] = w[2]; eigenVals[1] = w[1]; eigenVals[2] = w[0];
	m = a[6], a[3], a[0],
		a[7], a[4], a[1],
		a[8], a[5], a[2];
	
	for ( int i = 0; i < 3; i++ ) {
		for ( int j = 0; j < 3; j++ ) {
			eigenVecs[j][i] = m(i, j);
		}
	}
	return m;
}

#else
template <> Mat3f decomposeMatrix(const Mat3f &r2, Vec3f &eigenVals) {
/*
	% Given symmetric 3x3 matrix M, compute the eigenvalues
		m = trace(M)/3;
	K = M-m*eye(3);
	q = det(K)/2;

	p = 0
		for i=1:3
			for j=1:3
				p = p + K(i,j)^2;
	end
		end
		p = p/6;

	phi = 1/3*atan(sqrt(p^3-q^2)/q);
	if(phi<0)
		phi=phi+pi/3;
	end

		eig1 = m + 2*sqrt(p)*cos(phi)
		eig2 = m - sqrt(p)*(cos(phi) + sqrt(3)*sin(phi))
		eig3 = m - sqrt(p)*(cos(phi) - sqrt(3)*sin(phi))
*/
	float a = r2(0,0), b = r2(1,1), c = r2(2,2), d = r2(0,1), e = r2(0,2), f = r2(1,2);
	float m = (a + b + c) / 3;
	Mat3f K = (Mat3f)(r2 - tvmet::identity<float, 3, 3>() * m);
	float ka = a - m, kb = b - m, kc = c - m;
	float q = (ka * (kb * kc - sqf(f)) + d * (f * e - d * kc) + e * (d * f - e * kb))/2;
	float p = ( sqf(ka) + sqf(kb) + sqf(kc) + 2 * (sqf(d) + sqf(e) + sqf(f)) ) / 6;

	float phi = 1/3 * atanf(sqrtf(fmaxf(p * sqf(p) - sqf(q), 0))/(q+0.000001f));
	if ( phi < 0 )
		phi += (float)M_PI / 3;

	float sp = sqrtf(p), cphi = cosf(phi), sphi = sinf(phi), st = sqrtf(3);
	vector<float> ev(3);
	ev[0] = m + 2 * sp * cphi;
	ev[1] = m - sp * (cphi + st * sphi);
	ev[2] = m - sp * (cphi - st * sphi);
	/// Sort the eigenvalues in order of magnitude - assuming no negative eigenvalues, else resort to UDgreater
	//sort(ev.begin(), ev.end(), greater<float>());
	sort(ev.begin(), ev.end(), UDgreater<float>);
	eigenVals(0) = ev[0]; eigenVals(1) = ev[1]; eigenVals(2) = ev[2];
	//assert ( ( ev[0] >= ev[1] ) && ( ev[1] >= ev[2] ) );
	//assert ( (ev[0] >= 0) && (ev[1] >= 0) && (ev[2] >= 0) );

	/**
	 * Ok, we found the eigenvalues, now find the eigenvectors
	 *
	 * There are 3 cases for the number of eigenvalues:
	 * 1) All eigenvalues are the same - sphere case, we can return an identity matrix
	 * 2) 2 eigenvalues are the same - the repeat eigenvalue will have an eigenbasis for a 2D space
	 *	  Generate a random vector, cross with the 1 unique row vector to get 1st eigenbasis, 
	 *	  cross that vector with the row vector to get the second basis
	 * 3) Different eigenvalues - get the 2 independent row vectors, cross them to get the eigen vector
	 */

	Mat3f m2;
	Vec3f eigenVectors[3], maxv;
	int wcase;
	for ( int i = 0; i < 3; i++ ) {
		wcase = 0;

		float eval = ev[i];
		Mat3f A = (Mat3f)(r2 - tvmet::identity<float, 3, 3>() * eval);
		Vec3f v[3];
		for ( int j = 0; j < 3; j++ ) v[j] = (Vec3f)tvmet::normalize(tvmet::row(A, j));

		while ( i + 1 < 3 && tolEqual<float>( ev[i], ev[i+1], 0.0001f ) ) {
			i++;
			wcase++;
		}

		int vmax = 0;
		//unsigned int t = (unsigned int)time(NULL);
		switch(wcase) {
			case 2: /// All eigenvalues are the same
				m2 = tvmet::identity<float, 3, 3>();
				break;
			case 1: /// 2 eigenvalues same i & i-1
				/// Now all row vectors are the same, so we can choose any of them :)
#define FSRAND ((float)rand()/RAND_MAX)
				//eigenVectors[i - 1] = randhashf(t), randhashf(t+1), randhashf(t+2);
				if (tvmet::norm2(v[1]) > tvmet::norm2(v[vmax])) vmax = 1;
				if (tvmet::norm2(v[2]) > tvmet::norm2(v[vmax])) vmax = 2;
				eigenVectors[i - 1] = FSRAND, FSRAND, FSRAND;
				eigenVectors[i - 1] = tvmet::normalize(eigenVectors[i - 1]);
				v[vmax] = tvmet::normalize(v[vmax]);
				eigenVectors[i] = tvmet::normalize(tvmet::cross(v[vmax], eigenVectors[i - 1]));
				eigenVectors[i - 1] = tvmet::normalize(tvmet::cross(v[vmax], eigenVectors[i]));
#undef FSRAND
				break;
			case 0:
			default: /// 1 eigenvalue -> 1 eigenvector
				Vec3f evecs[3];
				for ( int k = 0; k < 3; k++ )
					v[k] = tvmet::normalize(v[k]);
				for ( int k = 0; k < 3; k++ ) {
					evecs[k] = tvmet::cross(v[k], v[(k+1)%3]);
				}
				bool b1 = tvmet::norm2(evecs[0]) > tvmet::norm2(evecs[1]);
				bool b2 = tvmet::norm2(evecs[1]) > tvmet::norm2(evecs[2]);
				bool b3 = tvmet::norm2(evecs[2]) > tvmet::norm2(evecs[0]);
				eigenVectors[i] = (b1) ? ((b3) ? evecs[2] : evecs[0]) : ((b2) ? evecs[1] : evecs[2]);
				eigenVectors[i] = tvmet::normalize(eigenVectors[i]);
		}
		//if ( ( tvmet::norm2(eigenVectors[i]) < 0.2 && wcase < 2 ) || ( tvmet::norm2(eigenVectors[i-1]) < 0.2 && wcase == 1)) {
		//	cout << "Error: eigenvector is too small: " << eigenVectors[i] << " case used: " << wcase << " relevant matrices r2= " << r2 << "A = " << A << " for eigenvalue " << eval << endl;
		//}
	}

	if ( wcase != 2 ) {
		m2 = eigenVectors[0][0], eigenVectors[1][0], eigenVectors[2][0],
			 eigenVectors[0][1], eigenVectors[1][1], eigenVectors[2][1],
			 eigenVectors[0][2], eigenVectors[1][2], eigenVectors[2][2];
	}
	return m2;
}
#endif

template <size_t d> Matrix<float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2, Vector<float,d> &eigenVals, Vector<float,d> eigenVecs[d]) {
	Matrix <float,d,d> m = decomposeMatrix(r2, eigenVals);
	for ( int i = 0; i < d; i++ ) {
		for ( int j = 0; j < d; j++ ) {
			eigenVecs[j][i] = m(i, j);
		}
	}
	return m;
}

template Mat2f decomposeMatrix(const Mat2f &r2, Vec2f &eigenVals, Vec2f eigenVecs[2]);
template Mat3f decomposeMatrix(const Mat3f &r2, Vec3f &eigenVals, Vec3f eigenVecs[3]);

template <size_t d> Matrix<float,d,d> decomposeMatrix(const Matrix<float,d,d> &r2, vector<float> &eigenVals, Vector<float,d> eigenVecs[d]) {
	Vector<float,d> ev;
	Matrix <float,d,d> m= decomposeMatrix(r2, ev, eigenVecs);
	eigenVals = vector<float>(ev);
	return m;
}

template Mat2f decomposeMatrix(const Mat2f &r2, Vec2f &eigenVals);
template Mat3f decomposeMatrix(const Mat3f &r2, Vec3f &eigenVals);

/// Checks whether a particle needs to be split or merge, return value is true if needs merging
template <size_t d> bool needSplitMerge(Particle<d> &p, bool &split, float r, float tol) {
	bool merge = needSplitOrMergeApprox<d>(p, split, r, tol);
	if (split || merge) {
		return merge;
	} else {
		Vector<float,d> eigenVal;
		//Vector<float,d> eigenVec[d];
		decomposeMatrix(p.r2, eigenVal);
		for ( int i = 0; i < d; i++ ) eigenVal[i] = sqrtf(fabsf(eigenVal[i]));

		float maxeval = tvmet::max<float,d>(eigenVal);
		float mineval = tvmet::min<float,d>(eigenVal);

		split = (maxeval > ( (float)M_SQRT2 + tol ) * r);
		bool merge = (mineval < ( (float)M_SQRT1_2 - tol ) * r);

		return merge;
	}
}

template bool needSplitMerge<2>(Particle<2> &p, bool &split, float r, float tol);
template bool needSplitMerge<3>(Particle<3> &p, bool &split, float r, float tol);

template <> bool needSplitOrMergeApprox<2> (Particle<2> &p, bool &split, float r, float tol) {
	float sum = trace(p.r2);
	float prod = det(p.r2);
	// prod <= eval*(sum - eval)
	float maxeval = sqrtf(sum - prod/sum);
	float mineval = sqrtf(prod/sum);
	split = (maxeval > ( (float)M_SQRT2 + tol ) * r);
	bool merge = (mineval < ( (float)M_SQRT1_2 - tol ) * r);
	return merge;
}
template <> bool needSplitOrMergeApprox<3> (Particle<3> &p, bool &split, float r, float tol) {
	float sum = trace(p.r2);
	float prod = det(p.r2);
	// prod <= eval*((sum - eval)/2)^2
	float maxeval = sqrtf(sum - 16*prod/(sum*sum));
	float mineval = sqrtf(4*prod/(sum*sum));
	split = (maxeval > ( (float)M_SQRT2 + tol ) * r);
	bool merge = (mineval < ( (float)M_SQRT1_2 - tol ) * r);
	return merge;
}

template <size_t d> void splitParticle(const Particle<d> &src, Particle<d> &dest1, Particle<d> &dest2) {
	/// We need to define m, x, v, and r2 for both particles
	dest1.m = dest2.m = src.m / 2;
	Vector<float,d> eigenVal;
	Vector<float,d> eigenVec[d], shift;
	/// The eigenvalues obtained are actually eval squares
	//Mat2f newM = decomposeMatrix(src.r2, eigenVal[0], eigenVal[1], eigenVec[0], eigenVec[1]);
	Matrix<float,d,d> newM = decomposeMatrix(src.r2, eigenVal, eigenVec);

	/// So we divide by 4, to get the actual evals/2
	int modifyIndex = 0;//(eigenVal[0] > eigenVal[1] ) ? 0 : 1;
	for ( int i = 1; i < d; i++ ) {
		modifyIndex = ( fabsf(eigenVal[i]) > fabsf(eigenVal[modifyIndex]) ) ? i : modifyIndex;
	}
	eigenVal[modifyIndex] /= 4;
	shift = eigenVec[modifyIndex] * sqrtf(eigenVal[modifyIndex]);

	Matrix<float,d,d> newR2(newM*diagonal(Vector<float,d>(eigenVal))*trans(newM));
	dest1.r2 = dest2.r2 = newR2;
	/// Would ruin rotational aspect of velocity - but no possible solution seems likely
	dest1.v = dest2.v = src.v;

	dest1.x = src.x + shift;
	dest2.x = src.x - shift;

	//cout << "Splitting particle src= " << src.r2 << " into dest1= " << dest1.r2 << " and dest2= " << dest2.r2 << endl;
}

template void splitParticle<2>(const Particle<2> &src, Particle<2> &dest1, Particle<2> &dest2);
template void splitParticle<3>(const Particle<3> &src, Particle<3> &dest1, Particle<3> &dest2);

// Already implemented in common.hpp: Matrix outer(Vector, Vector) returns outer product M = u v^T

// template <size_t d> Matrix<float, d, d> vvtrans(Vector<float, d> &v) {
// 	Matrix<float, d, d> m;
// 	for ( int i = 0; i < d; i++ ) {
// 		for ( int j = 0; j < d; j++ ) {
// 			m(i,j) = v(i) * v(j);
// 		}
// 	}
// 	return m;
// }

template <size_t d> Particle<d> mergeParticles(Particle<d> &p1, Particle<d> &p2) {
	//cout <<"Merging particles p1= " << p1.r2 << " and p2= " << p2.r2;
	Particle<d> p;
	p.m = p1.m + p2.m;
	p.x = (p1.x * p1.m + p2.x * p2.m) / p.m;
	p.v = (p1.v * p1.m + p2.v * p2.m) / p.m;
	/// For calculating Moment of Inertia for new particle, we rely on closed form given at http://scienceworld.wolfram.com/physics/MomentofInertiaEllipse.html
	/// Then, conservation of moment of inertia is used
	// Actually, let's conserve the covariance tensor
	// (http://en.wikipedia.org/wiki/Moment_of_inertia#Comparison_with_covariance_matrix)
	Vector<float, d> r1(p1.x - p.x), r2(p2.x - p.x);
	/// Calc translation matrices for MOIs = r'rI - rr' = [r_y^2,  -r_x*r_y \\ -r_x*r_y, r_x^2] 
	Matrix<float,d,d> M1(outer(r1,r1)), M2(outer(r2,r2));
	Matrix<float,d,d> id(identity<float,d,d>());

	float fudge = 1;
	float mult = (d + 2)*fudge;
	p.r2 = ( p1.m*p1.r2 + p1.m*mult*M1 + p2.m*p2.r2 + p2.m*mult*M2 ) / p.m;

	//cout << " into particle p= " << p.r2 << endl;
	return p;
}

template Particle<2> mergeParticles<2>(Particle<2> &p1, Particle<2> &p2);
template Particle<3> mergeParticles<3>(Particle<3> &p1, Particle<3> &p2);

//Particle<3> mergeParticles(Particle<3> &p1, Particle<3> &p2) {
//	Particle<3> p;
//	return p;
//}

//Particle2D mergeParticles(Particle2D &p1, Particle2D &p2) {
//	Particle2D result;
//	/// Conservation of mass
//	result.m = p1.m + p2.m;
//	/// Conservation of momentum
//	result.v = (p1.v * p1.m + p2.v * p2.m) / result.m;
//	/// Conserving center of mass
//	result.x = (p1.x * p1.m + p2.x * p2.m) / result.m;
//
//	/// Now, only r2 is left
//	/**
//	 * What all we wanna do:
//	 * 
//	 * 1) Make sure that the product of eigenvalues is conserved: pe1*pe2 = p1e1*p1e2 + p2e1*p2e2 -> conserves volume
//	 * 2) Preserve general orientation of eigenvalues as far as possible
//	 * 3) The orientation of major axes should be similar for the 2 ellipses
//	 */
//	Vec2f evec1[2], evec2[2];
//	float eval1[2], eval2[2];
//	Mat2f m1, m2;
//	m1 = decomposeMatrix(p1.r2, eval1, evec1);
//	m2 = decomposeMatrix(p2.r2, eval2, evec2);
//	/// Assumed that eigenvalues are in order
//	assert (eval1[0] >= eval1[1]);
//	assert (eval2[0] >= eval2[1]);
//	eval1[0] = sqrtf(eval1[0]); eval1[1] = sqrtf(eval1[1]);
//	eval2[0] = sqrtf(eval2[0]); eval2[1] = sqrtf(eval2[1]);
//
//	/// We will use 45 degrees as a threshold, if the major axes of both ellipses are within 45 degrees, we can merge
//	assert (dot(evec1[0], evec2[0]) >= (float)M_SQRT1_2);
//	Vec2f evec[2];
//	float eval[2];
//	/// Since we're assuming the 2 ellipses to have close to parallel major axes, this isn't that bad
//	evec[0] = (evec1[0] + evec2[0])/2;
//	eval[0] = (eval1[0] + eval2[0])/2;
//
//	/// Conserving volume
//	assert (eval[0] > 0);
//	eval[1] = ( eval1[0] * eval1[1] + eval2[0] * eval2[1] ) / (eval[0]);
//	evec[1] = Vec2f(-evec[0][1], evec[0][0]);
//	Mat2f m;
//	m = evec[0][0], evec[1][0],
//		evec[0][1], evec[1][1];
//	result.r2 = Mat2f(m*diagonal(Vec2f(sqf(eval[0]),sqf(eval[1])))*trans(m));
//	return result;
//}
//bool isMergeAble(const Particle<2> &p1, const Particle<2> &p2, float r) {
//	if ( !p1.merge || !p2.merge )
//		return false;
//	float evals[2];
//	Vec2f evec1[2], evec2[2];
//	decomposeMatrix(p1.r2, evals, evec1);
//	decomposeMatrix(p2.r2, evals, evec2);
//	/// Check that both particles need to be merged, and that the major axes of both are aligned
//	return ( (dot(evec1[0], evec2[0]) >= (float)M_SQRT1_2) );
//}
