/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "pressure.hpp"
#include "../math/qp.hpp"
#include "coupling.hpp"
#include "simhelpers.hpp"

#include <fstream>
using namespace std;

template <> void regularize_pressure (Sand<2> &sand) {
	if (sand.hacks.dont_regularize) return;
	int m = sand.m, n = sand.n;
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float f_obs = min(min(sand.aux.f_obs_s.x(i,j), sand.aux.f_obs_s.x(i+1,j)),
		                  min(sand.aux.f_obs_s.y(i,j), sand.aux.f_obs_s.y(i,j+1)));
		sand.aux.p(i+1,j+1) *= (1 - f_obs);
	}
}

template <> void regularize_pressure (Sand<3> &sand) {
	if (sand.hacks.dont_regularize) return;
	int m = sand.m, n = sand.n, o = sand.o;
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float f_obs = min(min(sand.aux.f_obs_s.x(i,j,k), sand.aux.f_obs_s.x(i+1,j,k)),
		                  min(min(sand.aux.f_obs_s.y(i,j,k), sand.aux.f_obs_s.y(i,j+1,k)),
		                      min(sand.aux.f_obs_s.z(i,j,k), sand.aux.f_obs_s.z(i,j,k+1))));
		sand.aux.p(i+1,j+1,k+1) *= (1 - f_obs);
	}
}

// ---------------- //
//        2D        //
// ---------------- //

template <> void density_correction (Sand<2> &sand, const vector< RigidBody<2> > &bodies) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static IntArray<2> idx;
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n;
	int pm = m + 2, pn = n + 2;
	float dx = sand.dx;
	idx.initialize(pm,pn, -1);
	int nactive = 0;
	// enumerate active cells
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		bool active = true;
		if (i >= 0 && i < m && j >= 0 && j < n) active = (sand.rho(i,j) > 0);
		if ((i == -1 || i == m) && (j == -1 || j == n)) active = false;
		if (active)
			idx(i+1,j+1) = nactive++;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	// for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
	// 	int a = idx(i+1,j+1);
	// 	if (a >= 0) x[a] = sand.aux.lambda_sep(i+1,j+1);
	// }
	float coeffr = 1/dx, coeffl = 1/(dx*dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j+1), b = idx(i+1,j+1);
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float multiplier = coeffl*(1 - sand.aux.f_obs_s.x(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j), b = idx(i+1,j+1);
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float multiplier = coeffl*(1 - sand.aux.f_obs_s.y(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// cells
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1);
		if (a < 0) continue;
		f[a] += (i < 0 || i >= m || j < 0 || j >= n) ? 0 : sand.rho(i,j)/sand.rho_max + sand.aux.f_obs(i,j) - 1.f;
	}
	// for (int b = 0; b < bodies.size(); b++) {
	// 	pressure_coupling(sand, bodies[b], dt, idx, H, f);
	// }
	QpReport report;
	qp_solve_dostal(H, f, x, report);
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1);
		if (a >= 0) sand.aux.lambda_sep(i+1,j+1) = x[a];
		else sand.aux.lambda_sep(i+1,j+1) = 0;
	}
	float multiplier = 1/(dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.x(i,j);
		sand.aux.dx_sep.x(i,j) = (mass<epsilon) ? 0 : (sand.aux.lambda_sep(i,j+1) - sand.aux.lambda_sep(i+1,j+1))*multiplier*(1-f_obs)/mass;
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.y(i,j);
		sand.aux.dx_sep.y(i,j) = (mass<epsilon) ? 0 : (sand.aux.lambda_sep(i+1,j) - sand.aux.lambda_sep(i+1,j+1))*multiplier*(1-f_obs)/mass;
	}
	float multiplier2 = 1/dx;
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		sand.rho(i,j) += (sand.aux.dx_sep.x(i,j)*rho_x(sand,i,j) - sand.aux.dx_sep.x(i+1,j)*rho_x(sand,i+1,j))*multiplier2;
		sand.rho(i,j) += (sand.aux.dx_sep.y(i,j)*rho_y(sand,i,j) - sand.aux.dx_sep.y(i,j+1)*rho_y(sand,i,j+1))*multiplier2;
	}
}

template <> void pressure_apply (Sand<2> &sand, vector< RigidBody<2> > &bodies, float dt) {
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	float multiplier = dt/(dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.x(i,j);
		sand.aux.dp.x(i,j) = (mass<epsilon) ? 0 : (sand.aux.p(i,j+1) - sand.aux.p(i+1,j+1))*multiplier*(1-f_obs)/mass;
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float f_obs = sand.aux.f_obs_s.y(i,j);
		sand.aux.dp.y(i,j) = (mass<epsilon) ? 0 : (sand.aux.p(i+1,j) - sand.aux.p(i+1,j+1))*multiplier*(1-f_obs)/mass;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.dp = dt*pressure_force(sand, bodies[b]);
	}
}

template <> float pressure_solve (Sand<2> &sand, vector< RigidBody<2> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static IntArray<2> idx;
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n;
	int pm = m + 2, pn = n + 2;
	float dx = sand.dx;
	idx.initialize(pm,pn, -1);
	int nactive = 0;
	// enumerate active cells
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		bool active = false;
		if (i >= 0 && j >= 0 && j < n && rho_x(sand,i,j) > epsilon) active = true;
		if (i < m && j >= 0 && j < n && rho_x(sand,i+1,j) > epsilon) active = true;
		if (i >= 0 && i < m && j >= 0 && rho_y(sand,i,j) > epsilon) active = true;
		if (i >= 0 && i < m && j < n && rho_y(sand,i,j+1) > epsilon) active = true;
		//if (i >= 0 && i < m && j >= 0 && j < n) active = (sand.rho(i,j) > epsilon);
		if ((i == -1 || i == m) && (j == -1 || j == n)) active = false;
		if (active)
			idx(i+1,j+1) = nactive++;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1);
		if (a >= 0)
			x[a] = sand.aux.p(i+1,j+1);
	}
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j+1), b = idx(i+1,j+1);
		float mass = rho_x(sand,i,j)/sand.rho_max;
		float vx = sand.v.x(i,j) + sand.aux.ds.x(i,j);
		float f_obs = sand.aux.f_obs_s.x(i,j);
		float multiplier = coeffl*sqf(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*(vx*mass + f_obs*sand.aux.v_obs.x(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j), b = idx(i+1,j+1);
		float mass = rho_y(sand,i,j)/sand.rho_max;
		float vy = sand.v.y(i,j) + sand.aux.ds.y(i,j);
		float f_obs = sand.aux.f_obs_s.y(i,j);
		float multiplier = coeffl*sqf(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*(vy*mass + f_obs*sand.aux.v_obs.y(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// cells
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1);
		if (a < 0) continue;
		f[a] += (i < 0 || i >= m || j < 0 || j >= n) ? 0 : sand.rho(i,j)/sand.rho_max + sand.aux.f_obs(i,j) - 1.f;
	}
	for (int b = 0; b < bodies.size(); b++) {
		pressure_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, x, report);
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1);
		if (a >= 0) sand.aux.p(i+1,j+1) = x[a];
		else sand.aux.p(i+1,j+1) = 0;
	}
	pressure_apply(sand, bodies, dt);
	return report.residual_initial;
}

// ---------------- //
//        3D        //
// ---------------- //

template <> void density_correction (Sand<3> &sand, const vector< RigidBody<3> > &bodies) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static IntArray<3> idx;
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n, o = sand.o;
	int pm = m + 2, pn = n + 2, po = o + 2;
	float dx = sand.dx;
	idx.initialize(pm,pn,po, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		bool active = true;
		if (i >= 0 && i < m && j >= 0 && j < n && k >= 0 && k < o) active = (sand.rho(i,j,k) > 0);
	    if ((i == -1 || i == m) && (j == -1 || j == n) && (k == -1 || k == o)) active = false;
		if (active)
			idx(i+1,j+1,k+1) = nactive++;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	// for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
	// 	int a = idx(i+1,j+1,k+1);
	// 	if (a >= 0) x[a] = sand.aux.lambda_sep(i+1,j+1,k+1);
	// }
	float coeffr = 1/dx, coeffl = 1/(dx*dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j+1,k+1), b = idx(i+1,j+1,k+1);
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float multiplier = coeffl*(1 - sand.aux.f_obs_s.x(i,j,k));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j,k+1), b = idx(i+1,j+1,k+1);
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float multiplier = coeffl*(1 - sand.aux.f_obs_s.y(i,j,k));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j+1,k), b = idx(i+1,j+1,k+1);
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float multiplier = coeffl*(1 - sand.aux.f_obs_s.z(i,j,k));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// cells
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1,k+1);
		if (a < 0) continue;
		f[a] += (i < 0 || i >= m || j < 0 || j >= n || k < 0 || k >= o) ? 0 : sand.rho(i,j,k)/sand.rho_max + sand.aux.f_obs(i,j,k) - 1.f;
	}
	// for (int b = 0; b < bodies.size(); b++) {
	// 	pressure_coupling(sand, bodies[b], dt, idx, H, f);
	// }
	QpReport report;
	qp_solve_dostal(H, f, x, report);
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1,k+1);
		if (a >= 0) sand.aux.lambda_sep(i+1,j+1,k+1) = x[a];
		else sand.aux.lambda_sep(i+1,j+1,k+1) = 0;
	}
	float multiplier = 1/dx;
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float mass = rho_x(sand,i,j,k);
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float dx_sep = (mass<epsilon) ? 0 : (sand.aux.lambda_sep(i,j+1,k+1) - sand.aux.lambda_sep(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
		sand.aux.dx_sep.x(i,j,k) = dx_sep;
		if (i > 0 && i < m) {
			sand.rho(i,j,k) += mass*dx_sep/dx;
			sand.rho(i-1,j,k) -= mass*dx_sep/dx;
		}
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j,k);
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float dx_sep = (mass<epsilon) ? 0 : (sand.aux.lambda_sep(i+1,j,k+1) - sand.aux.lambda_sep(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
		sand.aux.dx_sep.y(i,j,k) = dx_sep;
		if (j > 0 && j < n) {
			sand.rho(i,j,k) += mass*dx_sep/dx;
			sand.rho(i,j-1,k) -= mass*dx_sep/dx;
		}
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_z(sand,i,j,k);
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float dx_sep = (mass<epsilon) ? 0 : (sand.aux.lambda_sep(i+1,j+1,k) - sand.aux.lambda_sep(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
		sand.aux.dx_sep.z(i,j,k) = dx_sep;
		if (k > 0 && k < o) {
			sand.rho(i,j,k) += mass*dx_sep/dx;
			sand.rho(i,j,k-1) -= mass*dx_sep/dx;
		}
	}
}

template <> void pressure_apply (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	float multiplier = dt/dx;
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float mass = rho_x(sand,i,j,k);
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		sand.aux.dp.x(i,j,k) = (mass<epsilon) ? 0 : (sand.aux.p(i,j+1,k+1) - sand.aux.p(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_y(sand,i,j,k);
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		sand.aux.dp.y(i,j,k) = (mass<epsilon) ? 0 : (sand.aux.p(i+1,j,k+1) - sand.aux.p(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float mass = rho_z(sand,i,j,k);
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		sand.aux.dp.z(i,j,k) = (mass<epsilon) ? 0 : (sand.aux.p(i+1,j+1,k) - sand.aux.p(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
	}
	for (int b = 0; b < bodies.size(); b++) {
		bodies[b].aux.dp = dt*pressure_force(sand, bodies[b]);
	}
}

template <> float pressure_solve (Sand<3> &sand, vector< RigidBody<3> > &bodies, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static IntArray<3> idx;
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n, o = sand.o;
	int pm = m + 2, pn = n + 2, po = o + 2;
	float dx = sand.dx;
	idx.initialize(pm,pn,po, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		bool active = false;
		if (i >= 0 && j >= 0 && j < n  && k >= 0 && k < o  && rho_x(sand,i,j,k)   > epsilon) active = true;
		if (i < m  && j >= 0 && j < n  && k >= 0 && k < o  && rho_x(sand,i+1,j,k) > epsilon) active = true;
		if (i >= 0 && i < m  && j >= 0 && k >= 0 && k < o  && rho_y(sand,i,j,k)   > epsilon) active = true;
		if (i >= 0 && i < m  && j < n  && k >= 0 && k < o  && rho_y(sand,i,j+1,k) > epsilon) active = true;
		if (i >= 0 && i < m  && j >= 0 && j < n  && k >= 0 && rho_z(sand,i,j,k)   > epsilon) active = true;
		if (i >= 0 && i < m  && j >= 0 && j < n  && k < o  && rho_z(sand,i,j,k+1) > epsilon) active = true;
	    if ((i == -1 || i == m) && (j == -1 || j == n) && (k == -1 || k == o)) active = false;
		if (active)
			idx(i+1,j+1,k+1) = nactive++;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1,k+1);
		if (a >= 0) x[a] = sand.aux.p(i+1,j+1,k+1);
	}
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		int a = idx(i,j+1,k+1), b = idx(i+1,j+1,k+1);
		float mass = rho_x(sand,i,j,k)/sand.rho_max;
		float vx = sand.v.x(i,j,k) + sand.aux.ds.x(i,j,k);
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float multiplier = coeffl*sqf(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vx*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j,k+1), b = idx(i+1,j+1,k+1);
		float mass = rho_y(sand,i,j,k)/sand.rho_max;
		float vy = sand.v.y(i,j,k) + sand.aux.ds.y(i,j,k);
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float multiplier = coeffl*sqf(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vy*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j+1,k), b = idx(i+1,j+1,k+1);
		float mass = rho_z(sand,i,j,k)/sand.rho_max;
		float vz = sand.v.z(i,j,k) + sand.aux.ds.z(i,j,k);
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float multiplier = coeffl*sqf(1 - f_obs);
		float fmultiplier = coeffr*(1 - f_obs)*vz*mass;
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// cells
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1,k+1);
		if (a < 0) continue;
		f[a] += (i < 0 || i >= m || j < 0 || j >= n || k < 0 || k >= o) ? 0 : sand.rho(i,j,k)/sand.rho_max + sand.aux.f_obs(i,j,k) - 1.f;
	}
	for (int b = 0; b < bodies.size(); b++) {
		pressure_coupling(sand, bodies[b], dt, idx, H, f);
	}
	QpReport report;
	qp_solve_dostal(H, f, x, report);
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1,k+1);
		if (a >= 0) sand.aux.p(i+1,j+1,k+1) = x[a];
		else sand.aux.p(i+1,j+1,k+1) = 0;
	}
	pressure_apply(sand, bodies, dt);
	return report.residual_initial;
}
