/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "rigid.hpp"
using namespace std;
#undef max

// Helper functions

inline Vec2f rotate (float a, const Vec2f &v) {return Vec2f(v[0]*cos(a) - v[1]*sin(a), v[0]*sin(a) + v[1]*cos(a));}
inline Vec2f irotate (float a, const Vec2f &v) {return Vec2f(v[0]*cos(a) + v[1]*sin(a), -v[0]*sin(a) + v[1]*cos(a));}
inline Vec3f rotate (const Quatf &q, const Vec3f &v) {return (q*Quatf(0,v)*inv(q)).v;}
inline Vec3f irotate (const Quatf &q, const Vec3f &v) {return (inv(q)*Quatf(0,v)*q).v;}

template <size_t d> inline Vector<float,d> world_to_body (const RigidBody<d> &body, const Vector<float,d> &x) {
	return irotate(body.pose.q, Vector<float,d>(x - body.pose.x));
}
template <size_t d> inline Vector<float,d> body_to_world (const RigidBody<d> &body, const Vector<float,d> &x) {
	return Vector<float,d>(rotate(body.pose.q, x) + body.pose.x);
}

Vec2f clamp_to_grid (const Vec2f &x, const ScalarField<2> &field) {
	int m = field.m, n = field.n;
	float dx = field.dx;
	return Vec2f(clampf(x[0], 0.5f*dx,(m-0.5f)*dx), clampf(x[1], 0.5f*dx,(n-0.5f)*dx));
}
Vec3f clamp_to_grid (const Vec3f &x, const ScalarField<3> &field) {
	int m = field.m, n = field.n, o = field.o;
	float dx = field.dx;
	return Vec3f(clampf(x[0], 0.5f*dx,(m-0.5f)*dx), clampf(x[1], 0.5f*dx,(n-0.5f)*dx), clampf(x[2], 0.5f*dx,(o-0.5f)*dx));
}

// End helper functions

template <> void dist_updated (RigidBody<2> &body) {
	int m = body.dist.m, n = body.dist.n;
	float dx = body.dist.dx;
	body.origin = Vec2f(m*dx,n*dx)/2.f;
	body.r = 0.f;
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		Vec2f x((i+0.5f)*dx, (j+0.5f)*dx);
		if (body.dist(i,j) < 0)
			body.r = max(body.r, (float)norm2(x - body.origin) + sqrt(2.f)*dx);
	}
}
template <> void dist_updated (RigidBody<3> &body) {
	int m = body.dist.m, n = body.dist.n, o = body.dist.o;
	float dx = body.dist.dx;
	body.origin = Vec3f(m*dx,n*dx,o*dx)/2.f;
	body.r = 0.f;
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		Vec3f x((i+0.5f)*dx, (j+0.5f)*dx, (k+0.5f)*dx);
		if (body.dist(i,j,k) < 0)
			body.r = max(body.r, (float)norm2(x - body.origin) + sqrt(3.f)*dx);
	}
	//cout << "body dx " << body.dist.dx << " origin " << body.origin[0] << " " << body.origin[1] << " " << body.origin[2] << " r " << body.r << endl;
}

template <> void set_pose (RigidBody<2> &body, const RigidPose<2> &pose) {
	body.pose = pose;
	if (body.dynamic) {
		btTransform xform(btQuaternion(btVector3(0,0,1),pose.q), to_bt(pose.x));
		body.bt->setWorldTransform(xform);
	}
}
template <> void set_pose (RigidBody<3> &body, const RigidPose<3> &pose) {
	body.pose = pose;
	if (body.dynamic) {
		body.bt->setWorldTransform(btTransform(to_bt(pose.q), to_bt(pose.x)));
	}
}

template <> Screw<2> get_twist (const RigidBody<2> &body) {
	if (!body.dynamic) return body.twist;
	Vec2f lin = from_bt<2>(body.bt->getLinearVelocity());
	float ang = body.bt->getAngularVelocity()[2];
	return screw(lin, ang);
}
template <> Screw<3> get_twist (const RigidBody<3> &body) {
	if (!body.dynamic) return body.twist;
	Vec3f lin = from_bt<3>(body.bt->getLinearVelocity());
	Vec3f ang = from_bt<3>(body.bt->getAngularVelocity());
	return screw(lin, ang);
}

template <> void set_twist (RigidBody<2> &body, const Screw<2> &twist) {
	if (body.dynamic) {
		body.bt->setLinearVelocity(to_bt(lin(twist)));
		body.bt->setAngularVelocity(btVector3(0,0,ang(twist)));
	} else {
		body.twist = twist;
	}
}
template <> void set_twist (RigidBody<3> &body, const Screw<3> &twist) {
	if (body.dynamic) {
		body.bt->setLinearVelocity(to_bt(lin(twist)));
		body.bt->setAngularVelocity(to_bt(ang(twist)));
	} else {
		body.twist = twist;
	}
}

template <size_t d> float distance (const RigidBody<d> &body, const Vector<float,d> &x) {
	Vector<float,d> x_body(world_to_body(body,x) + body.origin);
	return lerp(x_body, body.dist)*body.dist.dx + norm2(x_body - clamp_to_grid(x_body, body.dist));
}
template float distance (const RigidBody<2> &body, const Vec2f &v);
template float distance (const RigidBody<3> &body, const Vec3f &v);

template <size_t d> void apply_wrench (RigidBody<d> &body, const Screw<d> &w);
template <> void apply_wrench (RigidBody<2> &body, const Screw<2> &w) {
	if (!body.dynamic) return;
	body.bt->applyCentralImpulse(to_bt(lin(w)));
	body.bt->applyTorqueImpulse(btVector3(0,0,ang(w)));
}
template <> void apply_wrench (RigidBody<3> &body, const Screw<3> &w) {
	if (!body.dynamic) return;
	body.bt->applyCentralImpulse(to_bt(lin(w)));
	body.bt->applyTorqueImpulse(to_bt(ang(w)));
}

template <> void bt_updated (RigidBody<2> &body) {
	body.pose.x = from_bt<2>(body.bt->getWorldTransform().getOrigin());
	const btQuaternion &q = body.bt->getWorldTransform().getRotation();
	body.pose.q = q.getAngle()*q.getAxis()[2];
}
template <> void bt_updated (RigidBody<3> &body) {
	body.pose.x = from_bt<3>(body.bt->getWorldTransform().getOrigin());
	body.pose.q = from_bt(body.bt->getWorldTransform().getRotation());
}

Mat3f from_bt (const btMatrix3x3 &a) {
	Mat3f b;
	for (int j = 0; j < 3; j++) for (int i = 0; i < 3; i++)
		b(i,j) = a[i][j];
	return b;
}

template <size_t d> float m_inv (const RigidBody<d> &body) {
	return body.dynamic ? body.bt->getInvMass() : 0;
}
template float m_inv (const RigidBody<2> &body);
template float m_inv (const RigidBody<3> &body);
float world_I_inv (const RigidBody<2> &body) {
	return body.dynamic ? body.bt->getInvInertiaDiagLocal()[2] : 0;
}
Mat3f world_I_inv (const RigidBody<3> &body) {
	return body.dynamic ? from_bt(body.bt->getInvInertiaTensorWorld()) : Mat3f(0.f);
}

template <> float speed (const RigidBody<2> &body) {
	Screw<2> t = get_twist(body);
	return norm2(lin(t)) + body.r*ang(t);
}
template <> float speed (const RigidBody<3> &body) {
	Screw<3> t = get_twist(body);
	return norm2(lin(t)) + body.r*norm2(ang(t));
}

template <> float coverage (const RigidBody<2> &body, const Vec2f &x0, const Vec2f &x1) {
	const Vec2f &com = body.pose.x;
	if (x0[0] > com[0] + body.r || x1[0] < com[0] - body.r || x0[1] > com[1] + body.r || x1[1] < com[1] - body.r)
		return 0;
	Vec2f x00(x0), x01(x0[0],x1[1]), x10(x1[0],x0[1]), x11(x1);
	float d00 = distance(body, x00);
	float d01 = distance(body, x01);
	float d10 = distance(body, x10);
	float d11 = distance(body, x11);
	return 0.5f - body.inout * 0.5f*(d00 + d01 + d10 + d11) / (fabsf(d00) + fabsf(d01) + fabsf(d10) + fabsf(d11));
}
template <> float coverage (const RigidBody<3> &body, const Vec3f &x0, const Vec3f &x1) {
	const Vec3f &com = body.pose.x;
	if (x0[0] > com[0] + body.r || x1[0] < com[0] - body.r || x0[1] > com[1] + body.r || x1[1] < com[1] - body.r || x0[2] > com[2] + body.r || x1[2] < com[2] - body.r)
		return 0;
	Vec3f x000(x0), x001(x0[0],x0[1],x1[2]), x010(x0[0],x1[1],x0[2]), x011(x0[0],x1[1],x1[2]);
	Vec3f x100(x1[0],x0[1],x0[2]), x101(x1[0],x0[1],x1[2]), x110(x1[0],x1[1],x0[2]), x111(x1);
	float d000 = distance(body, x000);
	float d001 = distance(body, x001);
	float d010 = distance(body, x010);
	float d011 = distance(body, x011);
	float d100 = distance(body, x100);
	float d101 = distance(body, x101);
	float d110 = distance(body, x110);
	float d111 = distance(body, x111);
	return 0.5f - body.inout  *0.5f*(d000+d001+d010+d011+d100+d101+d110+d111)
	             / (fabsf(d000)+fabsf(d001)+fabsf(d010)+fabsf(d011)+fabsf(d100)+fabsf(d101)+fabsf(d110)+fabsf(d111));
}

Vec2f body_velocity_at (const RigidBody<2> &body, const Screw<2> &twist, const Vec2f &x) {
	return Vec2f(lin(twist) + ang(twist) * Vec2f(body.pose.x[1] - x[1], x[0] - body.pose.x[0]));
}
Vec3f body_velocity_at (const RigidBody<3> &body, const Screw<3> &twist, const Vec3f &x) {
	return Vec3f(lin(twist) + cross(ang(twist), x - body.pose.x));
}

static ostream& operator<< (ostream &out, const Quatf &q) {out << "[" << q.s << "," << q.v << "]"; return out;}

template <size_t d> void project_particles (const RigidBody<d> &body, vector< Particle<d> > &ps) {
	Screw<d> twist = get_twist(body);
	int np = ps.size();
	for (int i = 0; i < np; i++) {
		Particle<d> &p = ps[i];
		Vector<float,d> p_body(world_to_body(body,p.x) + body.origin);
		Vector<float,d> dp_grid(p_body - clamp_to_grid(p_body, body.dist));
		float dist_grid = lerp(p_body, body.dist);
		float dist_out = norm2(dp_grid)/body.dist.dx;
		float dist = (dist_grid + dist_out)*body.inout - sqrt(trace(p.r2)/d)/body.dist.dx;
		if (dist > 0) continue;
		Vector<float,d> n(body.inout*gradient(p_body, body.dist)*body.dist.dx);
		if (dist_out > 0)
			n = normalize(dist_grid*n + dp_grid);
		else n = normalize(n);
		Vector<float,d> p_body_new(p_body - 2*dist*body.dist.dx*n/dot(n,n));
		p.x = body_to_world(body, Vector<float,d>(p_body_new - body.origin));
		Vector<float,d> body_v = body_velocity_at(body, twist, p.x);
		p.v -= n*min(0.f, dot(n, p.v - body_v));
	}
}
template void project_particles (const RigidBody<2> &body, vector< Particle<2> > &ps);
template void project_particles (const RigidBody<3> &body, vector< Particle<3> > &ps);

template <> void add_obstacle (const RigidBody<2> &body, ScalarField<2> &f_obs, VectorField<2> &f_obs_s, VectorField<2> &v_obs) {
	int m = f_obs.m, n = f_obs.n;
	float dx = f_obs.dx;
	const Vec2f &pos = body.pose.x;
	Screw<2> twist = get_twist(body);
	float x0 = pos[0] - body.r, x1 = pos[0] + body.r;
	float y0 = pos[1] - body.r, y1 = pos[1] + body.r;
	// i0, i1, j0, j1 index vertices, not cells
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	// cells
	for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
		Vec2f cx0(i*dx, j*dx), cx1((i+1)*dx, (j+1)*dx);
		f_obs(i,j) += coverage(body, cx0,cx1);
	}
	// x faces
	for (int j = j0; j < j1; j++) for (int i = i0; i <= i1; i++) {
		Vec2f cx(i*dx, (j+0.5f)*dx);
		Vec2f cx0((i-0.5f)*dx, j*dx), cx1((i+0.5f)*dx, (j+1)*dx);
		f_obs_s.x(i,j) += coverage(body, cx0,cx1);
		if (!body.dynamic && f_obs_s.x(i,j) > epsilon) v_obs.x(i,j) += body_velocity_at(body, twist, cx)[0];
	}
	// y faces
	for (int j = j0; j <= j1; j++) for (int i = i0; i < i1; i++) {
		Vec2f cx((i+0.5f)*dx, j*dx);
		Vec2f cx0(i*dx, (j-0.5f)*dx), cx1((i+1)*dx, (j+0.5f)*dx);
		f_obs_s.y(i,j) += coverage(body, cx0,cx1);
		if (!body.dynamic && f_obs_s.y(i,j) > epsilon) v_obs.y(i,j) += body_velocity_at(body, twist, cx)[1];
	}
}
template <> void add_obstacle (const RigidBody<3> &body, ScalarField<3> &f_obs, VectorField<3> &f_obs_s, VectorField<3> &v_obs) {
	int m = f_obs.m, n = f_obs.n, o = f_obs.o;
	float dx = f_obs.dx;
	const Vec3f &pos = body.pose.x;
	Screw<3> twist = get_twist(body);
	float x0 = pos[0] - body.r, x1 = pos[0] + body.r;
	float y0 = pos[1] - body.r, y1 = pos[1] + body.r;
	float z0 = pos[2] - body.r, z1 = pos[2] + body.r;
	// i0, i1, j0, j1, k0, k1 index vertices, not cells
	int i0 = floor(x0/dx), i1 = ceil(x1/dx);
	int j0 = floor(y0/dx), j1 = ceil(y1/dx);
	int k0 = floor(z0/dx), k1 = ceil(z1/dx);
	i0 = min(max(i0, 0), m);
	i1 = min(max(i1, 0), m);
	j0 = min(max(j0, 0), n);
	j1 = min(max(j1, 0), n);
	k0 = min(max(k0, 0), o);
	k1 = min(max(k1, 0), o);
	// cells
	for (int k = k0; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
		Vec3f cx0(i*dx, j*dx, k*dx), cx1((i+1)*dx, (j+1)*dx, (k+1)*dx);
		f_obs(i,j,k) += coverage(body, cx0,cx1);
	}
	// x faces
	for (int k = k0; k < k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i <= i1; i++) {
		Vec3f cx(i*dx, (j+0.5f)*dx, (k+0.5f)*dx);
		Vec3f cx0((i-0.5f)*dx, j*dx, k*dx), cx1((i+0.5f)*dx, (j+1)*dx, (k+1)*dx);
		f_obs_s.x(i,j,k) += coverage(body, cx0,cx1);
		if (!body.dynamic && f_obs_s.x(i,j,k) > epsilon) v_obs.x(i,j,k) += body_velocity_at(body, twist, cx)[0];
	}
	// y faces
	for (int k = k0; k < k1; k++) for (int j = j0; j <= j1; j++) for (int i = i0; i < i1; i++) {
		Vec3f cx((i+0.5f)*dx, j*dx, (k+0.5f)*dx);
		Vec3f cx0(i*dx, (j-0.5f)*dx, k*dx), cx1((i+1)*dx, (j+0.5f)*dx, (k+1)*dx);
		f_obs_s.y(i,j,k) += coverage(body, cx0,cx1);
		if (!body.dynamic && f_obs_s.y(i,j,k) > epsilon) v_obs.y(i,j,k) += body_velocity_at(body, twist, cx)[1];
	}
	// z faces
	for (int k = k0; k <= k1; k++) for (int j = j0; j < j1; j++) for (int i = i0; i < i1; i++) {
		Vec3f cx((i+0.5f)*dx, (j+0.5f)*dx, k*dx);
		Vec3f cx0(i*dx, j*dx, (k-0.5f)*dx), cx1((i+1)*dx, (j+1)*dx, (k+0.5f)*dx);
		f_obs_s.z(i,j,k) += coverage(body, cx0,cx1);
		if (!body.dynamic && f_obs_s.z(i,j,k) > epsilon) v_obs.z(i,j,k) += body_velocity_at(body, twist, cx)[2];
	}
}
