/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef RIGID_HPP
#define RIGID_HPP

#include "../common.hpp"
#include "../math/fields.hpp"
#include "particle.hpp"

#include <btBulletDynamicsCommon.h>

// To/from Bullet conversion functions

struct Quatf {
	float s; Vec3f v;
	inline Quatf () : s(1), v(0) {}
	inline Quatf (float s, const Vec3f &v) : s(s), v(v) {}
};
inline Quatf operator* (const Quatf &p, const Quatf &q) {return Quatf(p.s*q.s - dot(p.v,q.v), Vec3f(p.s*q.v + p.v*q.s + cross(p.v,q.v)));}
inline Quatf normalize (const Quatf &q) {float len = sqrtf(sqf(q.s) + dot(q.v,q.v)); return Quatf(q.s/len, Vec3f(q.v/len));}
inline Quatf inv (const Quatf &q) {return Quatf(q.s,Vec3f(-q.v));}

template <size_t d> inline Vector<float,d> from_bt (const btVector3 &v) {return Vector<float,d>((const btScalar*)v,d);}
inline Quatf from_bt (const btQuaternion &q) {return Quatf(q[3],Vec3f(q[0],q[1],q[2]));}
template <size_t d> btVector3 to_bt (const Vector<float,d> &v);
template <> inline btVector3 to_bt (const Vec2f &v) {return btVector3(v[0],v[1],0);}
template <> inline btVector3 to_bt (const Vec3f &v) {return btVector3(v[0],v[1],v[2]);}
inline btQuaternion to_bt (const Quatf &q) {return btQuaternion(q.v[0],q.v[1],q.v[2],q.s);}

// Screws (linear+angular velocity ("twist"), linear+angular momentum ("wrench"))

template <size_t d> class Screw : public Vector<float,d*(d+1)/2> {
public:
	typedef Vector<float,d*(d+1)/2> Vecsf;
	inline Screw () : Vecsf() {}
	inline Screw (float x) : Vecsf(x) {}
	inline Screw (float x, float y, float w) : Vecsf(x,y,w) {}
	inline Screw (float x, float y, float z, float u, float v, float w) : Vecsf(x,y,z,u,v,w) {}
	template <typename E> inline Screw (const XprVector<E,d*(d+1)/2>& e) : Vecsf(e) {}
};
inline Screw<2> screw (const Vec2f &v, float w) {return Screw<2>(v[0],v[1],w);}
inline Screw<2> screw (const Vec2f &v) {return Screw<2>(v[0],v[1],0);}
inline Screw<3> screw (const Vec3f &v, const Vec3f &w) {return Screw<3>(v[0],v[1],v[2],w[0],w[1],w[2]);}
inline Screw<3> screw (const Vec3f &v) {return Screw<3>(v[0],v[1],v[2],0,0,0);}
inline Vec2f lin (const Screw<2> &s) {return Vec2f(s[0],s[1]);}
inline Vec3f lin (const Screw<3> &s) {return Vec3f(s[0],s[1],s[2]);}
inline float ang (const Screw<2> &s) {return s[2];}
inline Vec3f ang (const Screw<3> &s) {return Vec3f(s[3],s[4],s[5]);}

// Rigid bodies

template <size_t d> struct RigidPose;
template <> struct RigidPose<2> {
	Vec2f x;
	float q;
};
template <> struct RigidPose<3> {
	Vec3f x;
	Quatf q;
};

// sparse d-by-n matrices with k nonzero entries
// stored as index : k -> n and entries : d-by-k
template <size_t d> struct CouplingMatrices;
template <> struct CouplingMatrices<2> {
	std::vector<int>        p_i, f_xx_yy_i, f_xy_i;
	std::vector< Screw<2> > p_J, f_xx_yy_J, f_xy_J;
};
template <> struct CouplingMatrices<3> {
	std::vector<int>        p_i, f_yy_i, f_xx_zz_i, f_zz_i, f_xy_i, f_yz_i, f_xz_i;
	std::vector< Screw<3> > p_J, f_yy_J, f_xx_zz_J, f_zz_J, f_xy_J, f_yz_J, f_xz_J;
};

template <size_t d> class RigidBody {
public:
	bool dynamic;           // moving rigid body or static/scripted obstacle?
	float inout;            // either +1 (just another obstacle) or -1 (sand is *inside* the body)
	btRigidBody *bt;        // Bullet rigid body, NULL if !dynamic
	RigidPose<d> pose;      // position and orientation of body, updated from bt if dynamic
	Screw<d> twist;         // current twist if scripted obstacle, zero otherwise
	ScalarField<d> dist;    // distance field
	Vector<float,d> origin; // position of center of mass relative to distance field
	float r;                // radius of conservative bounding sphere around center of mass
	struct {
		Screw<d> dp, ds;
		CouplingMatrices<d> coupling;
	} aux;
};

template <size_t d> void dist_updated (RigidBody<d> &body);

template <size_t d> void bt_updated (RigidBody<d> &body);

template <size_t d> float m_inv (const RigidBody<d> &body);
float world_I_inv (const RigidBody<2> &body);
Mat3f world_I_inv (const RigidBody<3> &body);

template <size_t d> Vector<float,d> world_to_body (const RigidBody<d> &body, const Vector<float,d> &x);
template <size_t d> Vector<float,d> body_to_world (const RigidBody<d> &body, const Vector<float,d> &x);

template <size_t d> float distance (const RigidBody<d> &body, const Vector<float,d> &x);

template <size_t d> void set_pose (RigidBody<d> &body, const RigidPose<d> &pose);
template <size_t d> Screw<d> get_twist (const RigidBody<d> &body);
template <size_t d> void set_twist (RigidBody<d> &body, const Screw<d> &twist);
template <size_t d> void apply_wrench (RigidBody<d> &body, const Screw<d> &wrench);

template <size_t d> float speed (const RigidBody<d> &body); // maximum speed for timestep selection
template <size_t d> float coverage (const RigidBody<d> &body, const Vector<float,d> &x0, const Vector<float,d> &x1); // covering fraction in box

template <size_t d> void project_particles (const RigidBody<d> &body, std::vector< Particle<d> > &ps);

template <size_t d> void add_obstacle (const RigidBody<d> &body, ScalarField<d> &f_obs, VectorField<d> &f_obs_s, VectorField<d> &v_obs);

#endif
