/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "sand.hpp"
#include <queue>

using namespace std;

void Sand<2>::resize (int m, int n, float dx) {
	this->m = m;
	this->n = n;
	this->dx = dx;
	rho.resize(m,n, dx);
	rho_s.resize(m,n, dx);
	v.resize(m,n, dx);
	aux.lambda_sep.resize(m+2,n+2, dx);
	aux.dx_sep.resize(m,n, dx);
	aux.f_obs.resize(m,n, dx);
	aux.f_obs_s.resize(m,n, dx);
	aux.v_obs.resize(m,n, dx);
	aux.p.resize(m+2,n+2, dx);
	aux.s.resize(m,n, dx);
	aux.dp.resize(m,n, dx);
	aux.ds.resize(m,n, dx);
	set_all(aux.p, 0);
	set_all(aux.s, Mat2f(0));
	set_all(aux.dp, Vec2f(0));
	set_all(aux.ds, Vec2f(0));
	aux.pbin.resize(m,n, dx);
}
void Sand<3>::resize (int m, int n, int o, float dx) {
	this->m = m;
	this->n = n;
	this->o = o;
	this->dx = dx;
	rho.resize(m,n,o, dx);
	rho_s.resize(m,n,o, dx);
	v.resize(m,n,o, dx);
	aux.lambda_sep.resize(m+2,n+2,o+2, dx);
	aux.dx_sep.resize(m,n,o, dx);
	aux.f_obs.resize(m,n,o, dx);
	aux.f_obs_s.resize(m,n,o, dx);
	aux.v_obs.resize(m,n,o, dx);
	aux.p.resize(m+2,n+2,o+2, dx);
	aux.s.resize(m,n,o, dx);
	aux.dp.resize(m,n,o, dx);
	aux.ds.resize(m,n,o, dx);
	set_all(aux.p, 0);
	set_all(aux.s, Mat3f(0));
	set_all(aux.dp, Vec3f(0));
	set_all(aux.ds, Vec3f(0));
	aux.pbin.resize(m,n,o, dx);
}

inline void confine (Particle<2> &p, Sand<2> &sand) {
	if (p.x[0] < 0)              {p.x[0] = -p.x[0]; p.v[0] = 0;}
	if (p.x[0] > sand.m*sand.dx) {p.x[0] = 2*sand.m*sand.dx - p.x[0]; p.v[0] = 0;}
	if (p.x[1] < 0)              {p.x[1] = -p.x[1]; p.v[1] = 0;}
	if (p.x[1] > sand.n*sand.dx) {p.x[1] = 2*sand.n*sand.dx - p.x[1]; p.v[1] = 0;}
}
inline void confine (Particle<3> &p, Sand<3> &sand) {
	if (p.x[0] < 0)              {p.x[0] = -p.x[0]; p.v[0] = 0;}
	if (p.x[0] > sand.m*sand.dx) {p.x[0] = 2*sand.m*sand.dx - p.x[0]; p.v[0] = 0;}
	if (p.x[1] < 0)              {p.x[1] = -p.x[1]; p.v[1] = 0;}
	if (p.x[1] > sand.n*sand.dx) {p.x[1] = 2*sand.n*sand.dx - p.x[1]; p.v[1] = 0;}
	if (p.x[2] < 0)              {p.x[2] = -p.x[2]; p.v[2] = 0;}
	if (p.x[2] > sand.o*sand.dx) {p.x[2] = 2*sand.o*sand.dx - p.x[2]; p.v[2] = 0;}
}

template <typename T> static inline T clamp (const T &v, float a) {
	//return v;
	float vmax = max(abs(v));
	return vmax>a ? T(v*a/vmax) : v;
}
template <size_t d> void advect (Sand<d> &sand, float dt) {
	int np = (int) sand.particles.size();
	float vmax = 0;
	for (int p = 0; p < np; p++)
		if (max(abs(sand.particles[p].v)) > vmax)
			vmax = max(abs(sand.particles[p].v));
	for (int p = 0; p < np; p++) {
		Particle<d> &particle = sand.particles[p];
		// assuming PIC
		Vector<float,d> fluid_v(lerp_weighted(particle.x, sand.v, sand.rho_s));
		float f_obs = lerp(particle.x, sand.aux.f_obs);
		float mass = lerp(particle.x, sand.rho)/(sand.rho_max*(1 - f_obs));
		mass = clampf(mass, 0.05,1);
		if (sand.hacks.flip_not_pic) {
			Vector<float,d> dv(lerp_weighted(particle.x, sand.aux.dp, sand.rho_s) + lerp_weighted(particle.x, sand.aux.ds, sand.rho_s));
			particle.v += clamp(dv, sand.dx/dt);
			particle.v += 0.05*(fluid_v - particle.v);
		} else {
			particle.v = fluid_v;
		}
		if (sand.hacks.clamp_velocity > 0)
			particle.v = clamp(particle.v, sand.hacks.clamp_velocity);
		Vector<float,d> movement_v;
		if (sand.hacks.sparse_own_velocity)
			movement_v = (particle.v + mass*(fluid_v - particle.v));
		else movement_v = fluid_v;
		if (sand.hacks.air_drag > 0) {
			Vector<float,d> dv(sand.hacks.air_vel - particle.v);
			Vector<float,d> n(gradient(particle.x, sand.rho));
			if (norm2(n) > epsilon) dv -= mass*(n*dot(n,dv)/dot(n,n));
			particle.v += dt*sand.hacks.air_drag*(1 - mass)*dv;
		}
		Vector<float,d> dx_sep(lerp_weighted(particle.x, sand.aux.dx_sep, sand.rho_s));
		Matrix<float,d,d> J(0);
		J += dt*jacobian_weighted(particle.x, sand.v, sand.rho_s);
		J += jacobian_weighted(particle.x, sand.aux.dx_sep, sand.rho_s);
		particle.r2 = advect_radii(particle.r2, clamp(J, 1));
		particle.x += movement_v*dt + clamp(dx_sep, sand.dx);
		confine(particle, sand);
	}
	vmax = 0;
	for (int p = 0; p < np; p++)
		if (max(abs(sand.particles[p].v)) > vmax)
			vmax = max(abs(sand.particles[p].v));
}
template void advect (Sand<2> &sand, float dt);
template void advect (Sand<3> &sand, float dt);

template <size_t d> void force (Sand<d> &sand, const Vector<float,d> &gdt) {
	int np = (int) sand.particles.size();
	for (int p = 0; p < np; p++) {
		sand.particles[p].v += gdt;
	}
}
template void force (Sand<2> &sand, const Vec2f &gdt);
template void force (Sand<3> &sand, const Vec3f &gdt);

inline void divide (VectorField<2> &num, VectorField<2> &den) {
	for (int i = 0; i < (int)num.x.s.size(); i++) num.x.s[i] /= (den.x.s[i] + epsilon);
	for (int i = 0; i < (int)num.y.s.size(); i++) num.y.s[i] /= (den.y.s[i] + epsilon);
}
inline void divide (VectorField<3> &num, VectorField<3> &den) {
	for (int i = 0; i < (int)num.x.s.size(); i++) num.x.s[i] /= (den.x.s[i] + epsilon);
	for (int i = 0; i < (int)num.y.s.size(); i++) num.y.s[i] /= (den.y.s[i] + epsilon);
	for (int i = 0; i < (int)num.z.s.size(); i++) num.z.s[i] /= (den.z.s[i] + epsilon);
}

template <size_t d> void splat (Sand<d> &sand) {
	typedef Vector<float,d> Vecdf;
	int np = (int) sand.particles.size();
	float dx = sand.dx;
	set_all(sand.rho, 0);
	set_all(sand.rho_s, Vecdf(0));
	set_all(sand.v, Vecdf(0));
	float dvol = pow(dx, (int)d);
	for (int p = 0; p < np; p++) {
		Particle<d> &particle = sand.particles[p];
		add_lerped(particle.m/dvol, particle.x, sand.rho);
		add_lerped(Vecdf(particle.m*particle.v/dvol), particle.x, sand.v);
		add_lerped(Vecdf(particle.m*Vecdf(1)/dvol), particle.x, sand.rho_s);
	}
	divide(sand.v, sand.rho_s);
}
template void splat (Sand<2> &sand);
template void splat (Sand<3> &sand);

template <size_t d> void bClamp(int coord, int axis, BucketField<d> &b);
template <> void bClamp(int coord, int axis, BucketField<2> &b) {
	assert ( ( axis >= 0 ) && ( axis < 2 ) );
	switch(axis) {
		case 0: coord = clamp<int>(coord, 0, b.m - 1); break;
		case 1: coord = clamp<int>(coord, 0, b.n - 1); break;
	}
}

template <> void bClamp(int coord, int axis, BucketField<3> &b) {
	assert ( ( axis >= 0 ) && ( axis < 3 ) );
	switch(axis) {
		case 0: coord = clamp<int>(coord, 0, b.m - 1); break;
		case 1: coord = clamp<int>(coord, 0, b.n - 1); break;
		case 2: coord = clamp<int>(coord, 0, b.o - 1); break;
	}
}

template <size_t d> int fillBuckets (Sand<d> &sand, bool fillall) {
	int nSplits = 0;
	BucketField<d> &pbin = sand.aux.pbin;
	clear_all(pbin);
	bool split = false;
	int n = (int)sand.particles.size();
	for ( int i = 0; i < n; i++ ) {
		Particle<d> &p = sand.particles[i];
		p.merge = false;
		int x[d];
		for ( int coord = 0; coord < d; coord++ ) {
			x[coord] = (int)(p.x[coord] / sand.dx - 0.5f);
		}
		pbin.bclamp(x);
		// if ( /*sand.rho(x)*/lerp(p.x, sand.rho) >= sand.rho_max * 0.99f)
		// 	continue;
		bool merge = needSplitMerge<d>(p, split, sand.r);
		if ( split ) {
			Particle<d> p1, p2;
			splitParticle<d>(p, p1, p2);
			sand.particles[i] = p1;
			sand.particles.push_back(p2);
			sand.smmap.push_back(SplitMerge(SPLIT, i, (int)sand.particles.size() - 1));
			split = false;
			//sand.pbin(x, y).push_back(i);
			//sand.pbin(x, y).push_back(sand.particles.size()-1);
			pbin(x).push_back(i);
			pbin(x).push_back(sand.particles.size()-1);
			nSplits++;
		}
		else if ( merge || fillall ) {
			pbin(x).push_back(i);
		}
	}
	return nSplits;
}

template int fillBuckets(Sand<2> &sand, bool fillall);
template int fillBuckets(Sand<3> &sand, bool fillall);

template<class T, class V>
class Record {
public:
	T		key;
	V		data;
	Record(T key, V data) {this->key = key; this->data = data;}
	bool operator<(const Record<T,V> &right) const {
		/// We want lower key to mean higher priority
		return this->key > right.key;
	}
};

template <size_t d> bool within(Sand<d> &sand, Particle<d> &p);

template <> bool within(Sand<2> &sand, Particle<2> &p) {
	return (p.x[0] >= 0) && (p.x[1] >= 0) && (p.x[0] <= sand.dx * sand.m) && (p.x[1] <= sand.dx * sand.n);
}

template <> bool within(Sand<3> &sand, Particle<3> &p) {
	return (p.x[0] >= 0) && (p.x[1] >= 0) && (p.x[2] >= 0) && (p.x[0] <= sand.dx * sand.m) && (p.x[1] <= sand.dx * sand.n) && (p.x[2] <= sand.dx * sand.o);
}

template <size_t d> void sim_consistency_check(Sand<d> &sand, int level) {
	int n = (int)sand.particles.size(), nm;
	Matrix <float, d, d> m;
	Vector <float, d> evals;
	vector<int> invalid_particles;
	for (int i = 0; i < n; i++) {
		Particle<d> &p = sand.particles[i];
		m = decomposeMatrix(p.r2, evals);
		//m = decomposeMatrix(sand.particles[i].r2, evals, eigenVecs);
		bool ok = true;
		/// Check that all eigenvalues are fine
		for (int j = 0; j < (int)d; j++) {
			evals[j] = sqrtf(evals[j]);
			switch(level) {
				case 1: ok = ok && (evals[j] == evals[j]) && (evals[j] >= 0) && (evals[j] <= ((float)M_SQRT2 + 0.01f) * sand.r);
					break;
				case 0: ok = ok && (evals[j] == evals[j]) && (evals[j] >= ((float)M_SQRT1_2 - 0.01f) * sand.r) && (evals[j] <= ((float)M_SQRT2 + 0.01f) * sand.r) && (evals[j] != 0);
					break;
				case 2: ok = ok && (evals[j] == evals[j]) && (evals[j] >= 0);
					break;
				default: cout << "Invalid check level: " << level << endl;
			}
			ok = ok && within(sand, sand.particles[i]); 
		}
		if (!ok) {
			//cout << "pushing in particle " << i << endl;
			invalid_particles.push_back(i);
		}
	}

	if (invalid_particles.size() > 0) {
		nm = (int)invalid_particles.size();
		cout << "Found " << nm << " invalid particles out of " << n << " particles, positions and r2 matrices follow: " << endl;
		cout << "For reference, sand.r is " << sand.r << endl;
		if (level == 0) {
			for (int i = 0; i < nm; i++) {
				int place = invalid_particles[i];
				cout << "Particle " << place << " p.m: " << sand.particles[place].m << " p.x: " << sand.particles[place].x << " p.r2: " << sand.particles[place].r2;
				m = decomposeMatrix(sand.particles[place].r2, evals);
				//m = decomposeMatrix(sand.particles[place].r2, evals, eigenVecs);
				cout << " evals: " << evals << endl;
				cout << "Bounds: [" << ((float)M_SQRT1_2 - 0.01f) * sand.r << ", " << ((float)M_SQRT2 + 0.01f) * sand.r << "]" << endl;
			}
		}
	}
}

template void sim_consistency_check(Sand<2> &sand, int level);
template void sim_consistency_check(Sand<3> &sand, int level);

/// Main particle split and merge function
template <> void splitnmerge (Sand<2> &sand) {
	int nSplits = fillBuckets<2>(sand, true);
	//int nSplits = fillBuckets(sand, false);
	BucketField<2> &b = sand.aux.pbin;
	priority_queue<int> s;
	int nMerges = 0;

	for ( int x = 0; x < b.m; x++ ) {
		for ( int y = 0; y < b.n; y++ ) {
			/// Each bin looks at bins (x+{0,1}, y+{0,1}) for other particles
			vector<int> &bin = b(x,y);
			for ( int i = 0; i < (int) bin.size(); i++ ) {
				Particle<2> &p = sand.particles[bin[i]];
				/*if ( p.merge )*/ {
					priority_queue< Record<float, int> > candidates;
					int dxmin = (x > 0) ? -1 : 0, dxmax = (x < b.m - 1) ? 1 : 0;
					int dymin = (y > 0) ? -1 : 0, dymax = (y < b.n - 1) ? 1 : 0;
					/// Gather all candidates, sort by distance to target particle
					for ( int dx = dxmin; dx < dxmax; dx++ ) {
						for ( int dy = dymin; dy < dymax; dy++ ) {
							vector<int> &cand_bin = b(x+dx, y+dy);
							for ( int j = (dx == 0 && dy == 0) ? i + 1 : 0; j < (int)cand_bin.size(); j++ ) {
								Vec2f v(p.x - sand.particles[cand_bin[j]].x);
								candidates.push(Record<float, int>(tvmet::dot(v, v), cand_bin[j]));
							}
						}
					}

					/// All particles gathered, now merge
					bool done = false;
					while ( !done && !candidates.empty() ) {
						Record<float, int> r = candidates.top();
						candidates.pop();
						Particle<2> &p1 = sand.particles[r.data];
						if ( p1.merge ) continue;
						//if ( p1.merge && isMergeAble(p, p1, sand.r) ) {
						Particle<2> pnew = mergeParticles<2>(p, p1);
						bool split = false;
						needSplitMerge<2>(pnew, split, sand.r);
						//split = split || ( (pnew.m / ellipseVolume(pnew)) > sand.rho_max * 1.1 );
						if ( !split ) {
							p = pnew;
							p1 = pnew;
							p1.merge = true;
							/// Put extra location into stack to be cleared later
							s.push(r.data);
							done = true;
							nMerges++;
						}
					}
				}
			}
		}
	}
	while ( !s.empty() ) {
		int loc = s.top();
		s.pop();
		if ( loc != (int)sand.particles.size() - 1 ) {
			//fprintf(stderr,"Debug: swapping particle at %d with last particle at %d\n", loc, sand.particles.size() - 1);
			sand.particles[loc] = sand.particles[sand.particles.size() - 1];
		}
		sand.particles.resize(sand.particles.size() - 1);
	}
	//cout<<"nS = "<<nSplits<<" nM = "<<nMerges<<endl;
}

template <> void splitnmerge (Sand<3> &sand) {
	//sand.smmap.clear();
	Timer perfTimer, intTimer, pTimer;
	perfTimer.tick();
	int nSplits = fillBuckets<3>(sand, true);
	float perfFill = perfTimer.tick();
	//int nSplits = fillBuckets(sand, false);
	BucketField<3> &b = sand.aux.pbin;
	priority_queue<int> s;
	int nMerges = 0;

	float perfGather = 0, perfDoMerge = 0;

	float perfNSP = 0, perfpMerge = 0, perfRest = 0;
	for ( int x = 0; x < b.m; x++ ) {
		for ( int y = 0; y < b.n; y++ ) {
			for ( int z = 0; z < b.o; z++ ) {
				/// Each bin looks at bins (x+{0,1}, y+{0,1}, z+{0,1}) for other particles
				vector<int> &bin = b(x,y,z);
				for ( int i = 0; i < (int) bin.size(); i++ ) {
					intTimer.tick();
					Particle<3> &p = sand.particles[bin[i]];
					if (tvmet::all_elements(p.r2 != p.r2)) {
						/// Corruption in simulation -> one particle has gone nan, sweep it under the carpet
						s.push(i);
						cout << "Simulation Error: Particle " << i << " at: " << p.x << " has invalid radius matrix r2: " << p.r2 << " removing from simulation" << endl;
						continue;
					}
					/*if ( p.merge )*/ {
						priority_queue< Record<float, int> > candidates;
						int dxmin = (x > 0) ? -1 : 0, dxmax = (x < b.m - 1) ? 1 : 0;
						int dymin = (y > 0) ? -1 : 0, dymax = (y < b.n - 1) ? 1 : 0;
						int dzmin = (z > 0) ? -1 : 0, dzmax = (y < b.o - 1) ? 1 : 0;
						/// Gather all candidates, sort by distance to target particle
						for ( int dx = dxmin; dx < dxmax; dx++ ) {
							for ( int dy = dymin; dy < dymax; dy++ ) {
								for ( int dz = dzmin; dz < dzmax; dz++ ) {
									vector<int> &cand_bin = b(x+dx, y+dy, z+dz);
									for ( int j = (dx == 0 && dy == 0 && dz == 0) ? i + 1 : 0; j < (int)cand_bin.size(); j++ ) {
										Vec3f v(p.x - sand.particles[cand_bin[j]].x);
										if (dot(v,v) < sand.r*sand.r*4)
											candidates.push(Record<float, int>(tvmet::dot(v, v), cand_bin[j]));
									}
								}
							}
						}
						perfGather += intTimer.tick();

						/// All particles gathered, now merge
						pTimer.tick();
						bool done = false;
						/*Record<float, int> r = candidates.top();
						Particle<3> &closestp = sand.particles[r.data];
						while (closestp.merge && !candidates.empty()) {
							candidates.pop();
							r = candidates.top();
							closestp = sand.particles[r.data];
						}*/
						while ( !done && !candidates.empty() ) {
							Record<float, int> r = candidates.top();
							candidates.pop();
							Particle<3> &p1 = sand.particles[r.data];
							if ( p1.merge ) continue;
							//if ( p1.merge && isMergeAble(p, p1, sand.r) ) {
							perfRest += pTimer.tick();
							Particle<3> pnew = mergeParticles<3>(p, p1);
							perfpMerge += pTimer.tick();
							bool split = false;
							//needSplitMerge<3>(pnew, split, sand.r);
							float sum = trace(pnew.r2), prod = det(pnew.r2);
							float r2max = 0.5f*(sum - 9*prod/(sum*sum)); // evalmax is at least this much
							// because prod >= evalmax^2 (sum - 2*evalmax), and evalmax >= sum/3
							split = sqrtf(r2max) > M_SQRT2*sand.r;
							if (!split) // max(r^2) not obviously too big, may be non-obviously too big
								needSplitMerge<3>(pnew, split, sand.r);
							perfNSP += pTimer.tick();
							//split = split || ( (pnew.m / ellipseVolume(pnew)) > sand.rho_max * 1.1 );
							if ( !split ) {
								p = pnew;
								p1 = pnew;
								p1.merge = true;
								sand.smmap.push_back(SplitMerge(MERGE, bin[i], r.data));
								/// Put extra location into stack to be cleared later
								s.push(r.data);
								done = true;
								nMerges++;
							}
						}
						/*if (!done && p.m < 0.00001 && !closestp.merge) {
							/// Oh we couldn't find a particle to merge this particle with, and it's too small to care about
							/// so just push this mass into the nearest particle

							closestp.m += p.m;
							closestp.r2 += p.r2;
							sand.smmap.push_back(SplitMerge(MERGE, r.data, bin[i]));
							s.push(bin[i]);
							//Particle<3> pnew = mergeParticles<3>(p, closestp);
							//p = pnew;
							//closestp = pnew;
							//closestp.merge = true;
							//sand.smmap.push_back(SplitMerge(MERGE, bin[i], r.data));
							//s.push(r.data);
							nMerges++;
						}*/

						perfDoMerge += intTimer.tick();
					}
				}
			}
		}
	}
	float perfMerge = perfTimer.tick();
	int ndels = 0;
	while ( !s.empty() ) {
		int loc = s.top();
		s.pop();
		if ( loc < (int)sand.particles.size() - 1 - ndels ) {
			//fprintf(stderr,"Debug: swapping particle at %d with last particle at %d\n", loc, sand.particles.size() - 1);
			int tloc = (int)sand.particles.size() - 1 - ndels;
			sand.particles[loc] = sand.particles[tloc];
			//for (int tempi = 0; tempi < (int)sand.smmap.size(); tempi++) {
			//	SplitMerge &sm = sand.smmap[tempi];
			//	if ( sm.m != SPLIT ) continue;
			//	sm.pID[0] = (tloc == sm.pID[0]) ? loc : sm.pID[0];
			//	sm.pID[1] = (tloc == sm.pID[1]) ? loc : sm.pID[1];
			//}
			sand.smmap.push_back(SplitMerge(MOVE, (int)sand.particles.size() - 1 - ndels, loc));
		}
		ndels++;
	}
	sand.particles.resize(sand.particles.size() - ndels);
	float perfDelete = perfTimer.tick();

	//cout << "Performance: Fill: " << perfFill << " Merge: " << perfMerge << "(G:" << perfGather << ", DM:" << perfDoMerge << ")" << " Delete: " << perfDelete << " Precise: (Rest, NSP, Merge) = (" << perfRest << ", " << perfNSP << ", " << perfpMerge << ") " << endl;
	//cout<<"nS = "<<nSplits<<" nM = "<<nMerges<<endl;
}
