/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "scripted.hpp"
#include "rigid.hpp"
#include "simulator.hpp"
using namespace std;

template <size_t d> string Explosion<d>::type() {
	return string("explosion");
}

template <size_t d> string HoldBody<d>::type() {
	return string("holdbody");
}

template <size_t d> string Source<d>::type() {
	return string("source");
}

template <size_t d> string ScriptedBody<d>::type() {
	return string("scriptedbody");
}

template string Explosion<2>::type();
template string Explosion<3>::type();

template string HoldBody<2>::type();
template string HoldBody<3>::type();

template string Source<2>::type();
template string Source<3>::type();

template string ScriptedBody<2>::type();
template string ScriptedBody<3>::type();

template <size_t d> void Explosion<d>::init (const Simulator<d> &sim) {
	t = 0;
	blown = false;
}
template void Explosion<2>::init (const Simulator<2> &sim);
template void Explosion<3>::init (const Simulator<3> &sim);

template <size_t d> float explosion_speed (float r, float R, float dVdt);
template <> float explosion_speed<2> (float r, float R, float dVdt) {
	return dVdt*r/(2*M_PI*R*R);
}
template <> float explosion_speed<3> (float r, float R, float dVdt) {
	return dVdt*r/(4*M_PI*R*R*R);
}

template <size_t d> float Explosion<d>::speed (const Simulator<d> &sim) {
	if (t < T || blown) return 0;
	return explosion_speed<d>(2*R, R, dVdt);
}
template float Explosion<2>::speed (const Simulator<2> &sim);
template float Explosion<3>::speed (const Simulator<3> &sim);

template <size_t d> void Explosion<d>::run (Simulator<d> &sim, float dt) {
	if (t < T || blown) {
		t += dt;
		return;
	}
	t += dt;
	blown = true;
	cout << "KABOOM!" << endl;
	// explode!
	int np = sim.sand.particles.size();
	for (int p = 0; p < np; p++) {
		Particle<d> &particle = sim.sand.particles[p];
		Vector<float,d> px0(particle.x - x);
		float r0 = norm2(px0);
		float falloff =
			r0<R   ? 1 :
			r0<2*R ? 1.f/pow(r0/R,(int)d) :
			0;
		if (falloff > 0) {
			Vector<float,d> px1(px0);
			px1[1] += R;
			float r1 = norm2(px1);
			Vector<float,d> v(dVdt * px1 / (2*(d-1)*M_PI*powf(R,(int)d)));
			v[0] *= 0.5f;
			v[1] *= 2.f;
			v[2] *= 0.5f;
			particle.v += v*falloff;
		}
	}
}
template void Explosion<2>::run (Simulator<2> &sim, float dt);
template void Explosion<3>::run (Simulator<3> &sim, float dt);

template <size_t d> void HoldBody<d>::init (const Simulator<d> &sim) {
	t = 0;
	released = false;
	releasepose = body->pose;
	releasetwist = get_twist(*body);
	set_pose(*body, holdpose);
	set_twist(*body, Screw<d>(0.f));
}
template void HoldBody<2>::init (const Simulator<2> &sim);
template void HoldBody<3>::init (const Simulator<3> &sim);

template <> float HoldBody<2>::speed (const Simulator<2> &sim) {
	if (t < T || released) return 0;
	return norm2(lin(releasetwist)) + body->r*ang(releasetwist);

}
template <> float HoldBody<3>::speed (const Simulator<3> &sim) {
	if (t < T || released) return 0;
	return norm2(lin(releasetwist)) + body->r*norm2(ang(releasetwist));
}

template <size_t d> void HoldBody<d>::run (Simulator<d> &sim, float dt) {
	if (t < T) {
		set_twist(*body, Screw<d>(0.f));
	} else if (!released) {
		released = true;
		set_pose(*body, releasepose);
		set_twist(*body, releasetwist);
	}
	t += dt;
}
template void HoldBody<2>::run (Simulator<2> &sim, float dt);
template void HoldBody<3>::run (Simulator<3> &sim, float dt);

template <size_t d> void Source<d>::init (const Simulator<d> &sim) {
	t0 = t = 0;
}
template void Source<2>::init (const Simulator<2> &sim);
template void Source<3>::init (const Simulator<3> &sim);

template <size_t d> float Source<d>::speed (const Simulator<d> &sim) {
	return norm2(v);
}
template float Source<2>::speed (const Simulator<2> &sim);
template float Source<3>::speed (const Simulator<3> &sim);

// Referencing function in io.cpp. Bad style, FIXME.
template <size_t d> void add_particles_box (vector< Particle<d> > &particles, float r, Vector<float,d> x0, Vector<float,d> x1, float rho, Vector<float,d> v);

template <size_t d> void Source<d>::run (Simulator<d> &sim, float dt) {
	Vector<float,d> deltax(v*(t-t0) + sim.g*sqf(t-t0)/2);
	if (norm2(deltax) < sim.sand.dx) {
		t += dt;
		return;
	}
	add_particles_box(sim.sand.particles, sim.sand.r, x0, Vector<float,d>(x1 + deltax), rho, v);
	t0 = t;
	t = t + dt;
};
template void Source<2>::run (Simulator<2> &sim, float dt);
template void Source<3>::run (Simulator<3> &sim, float dt);

template <size_t d> void integrate_twist (RigidPose<d> &pose, const Screw<d> &twist, float dt);
template <> void integrate_twist (RigidPose<2> &pose, const Screw<2> &twist, float dt) {
	pose.x += lin(twist)*dt;
	pose.q += ang(twist)*dt;
}
template <> void integrate_twist (RigidPose<3> &pose, const Screw<3> &twist, float dt) {
	pose.x += lin(twist)*dt;
	float w = norm2(ang(twist));
	Quatf dq;
	if (w > epsilon) {
		dq.s = cos(w*dt/2);
		dq.v = sin(w*dt/2)*normalize(ang(twist));
	}
	pose.q = dq*pose.q;
}

template <size_t d> void ScriptedBody<d>::init (const Simulator<d> &sim) {
}
template void ScriptedBody<2>::init (const Simulator<2> &sim);
template void ScriptedBody<3>::init (const Simulator<3> &sim);

template <> float ScriptedBody<2>::speed (const Simulator<2> &sim) {
	return 0;
}
template <> float ScriptedBody<3>::speed (const Simulator<3> &sim) {
	return 0;
}

template <size_t d> void ScriptedBody<d>::run (Simulator<d> &sim, float dt) {
	set_twist(*body, twist);
	integrate_twist(body->pose, twist, dt);
}
template void ScriptedBody<2>::run (Simulator<2> &sim, float dt);
template void ScriptedBody<3>::run (Simulator<3> &sim, float dt);
