/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "simulator.hpp"
#include "pressure.hpp"
#include "friction.hpp"
#include "coupling.hpp"
#include "scripted.hpp"
#include "zhubridson.hpp"
//#include <tvmet/tvmet.h>

#include <iostream>
using namespace std;
using namespace tvmet;
#undef max

Timer timer;
struct {float force, splat, density, friction, pressure, advect, splitmerge;} times;

template <size_t d> float max_vel (const Sand<d> &sand) {
	float vmax = 0;
	int np = sand.particles.size();
	for (int p = 0; p < np; p++)
		if (max(abs(sand.particles[p].v)) > vmax)
			vmax = max(abs(sand.particles[p].v));
	return vmax;
}
template <size_t d> float max_coverage (const Sand<d> &sand) {
	float fmax = 0;
	int ncells = sand.rho.s.size();
	for (int i = 0; i < ncells; i++)
		fmax = max(fmax, sand.rho.s[i]/sand.rho_max + sand.aux.f_obs.s[i]);
	return fmax;
}

template <size_t d> void substep (Simulator<d> &sim, float dt) {
	int nb = sim.bodies.size();
	int p = cout.precision(4);
	cout << "dt " << dt << " t " << sim.simtime << endl;
	timer.tick();
#define REPORT_TIMING(step_name) //cout << step_name << " " << timer.tick() << "\n";
	// apply gravity
	force(sim.sand, Vector<float,d>(sim.g*dt));
	for (int i = 0; i < nb; i++) {
		RigidBody<d> &body = sim.bodies[i];
		if (m_inv(body) > 0) apply_wrench(body, screw(Vector<float,d>(sim.g*dt/m_inv(body))));
	}
	REPORT_TIMING("force");
	// run scripted events
	for (int i = 0; i < sim.scripted.size(); i++) {
		sim.scripted[i]->run(sim, dt);
	}
	// accumulate rigid body volume coverage
	set_all(sim.sand.aux.f_obs, 0);
	set_all(sim.sand.aux.f_obs_s, Vector<float,d>(0));
	set_all(sim.sand.aux.v_obs, Vector<float,d>(0));
	for (int i = 0; i < nb; i++) {
		RigidBody<d> &body = sim.bodies[i];
		add_obstacle(body, sim.sand.aux.f_obs, sim.sand.aux.f_obs_s, sim.sand.aux.v_obs);
	}
	if (sim.sand.particles.size() == 0) {
		sim.simtime += dt;
		cout << endl;
		return;
	}
	// splat particles onto grid
	splat(sim.sand);
	REPORT_TIMING("splat");
	// density correction
	density_correction(sim.sand, sim.bodies);
	REPORT_TIMING("density");
	// pressure and friction solves
	if (sim.sand.zhubridson) {
		pressure_solve_zb(sim.sand, dt);
		friction_solve_zb(sim.sand, dt);
	} else {
		int n_iters = 2;
		for (int i = 0; i < nb; i++) {
			precompute_coupling_matrices(sim.sand, sim.bodies[i]);
		}
		REPORT_TIMING("coupling");
		pressure_apply(sim.sand, sim.bodies, dt);
		friction_apply(sim.sand, sim.bodies, dt);
		REPORT_TIMING("apply");
		for (int i = 0; i < n_iters; i++) {
			float rf = friction_solve(sim.sand, sim.bodies, dt);
			REPORT_TIMING("friction");
			float rp = pressure_solve(sim.sand, sim.bodies, dt);
			REPORT_TIMING("pressure");
		}
		// apply pressure and friction impulses
		add_into(sim.sand.v, sim.sand.aux.dp);
		add_into(sim.sand.v, sim.sand.aux.ds);
		for (int i = 0; i < nb; i++) {
			RigidBody<d> &body = sim.bodies[i];
			apply_wrench(body, body.aux.dp);
			apply_wrench(body, body.aux.ds);
			// cout << "PRESSURE IMPULSE " << body.aux.dp << endl;
			// cout << "FRICTION IMPULSE " << body.aux.ds << endl;
		}
		if (!sim.sand.hacks.dont_regularize)
			regularize_pressure(sim.sand);
	}
	// advance simulation state
	//cout << "Checking consistency BEFORE advect" << endl;
	//sim_consistency_check(sim.sand);
	advect(sim.sand, dt);

	//cout << "Checking sim_consistency AFTER advect, BEFORE splitmerge" << endl;
	//sim_consistency_check(sim.sand);
	REPORT_TIMING("advect");
	splitnmerge(sim.sand);
	splitnmerge(sim.sand);
	//cout << "Checking sim_consistency AFTER splitmerge, BEFORE projectparticles" << endl;
	//sim_consistency_check(sim.sand);
	REPORT_TIMING("splitnmerge");
	sim.bt->stepSimulation(dt, 0);
	for (int i = 0; i < nb; i++) {
		if (sim.bodies[i].dynamic)
			bt_updated(sim.bodies[i]);
		project_particles<d>(sim.bodies[i], sim.sand.particles);
	}
	//cout << "Checking sim_consistency AFTER projectparticles" << endl;
	//sim_consistency_check(sim.sand);
	// done
	sim.simtime += dt;
	cout << endl;
	cout.precision(p);
}

template <size_t d> void step (Simulator<d> &sim) {
	float t = 0;
	do {
		int np = sim.sand.particles.size(), nb = sim.bodies.size();
		// find largest timestep allowed
		float dt = infinity;
		vector<float> dist_to_body(nb, infinity);
		float vmax = 0;
		int imax = 0;
		splat(sim.sand);
		for (int i = 0; i < np; i++) {
			Particle<d> &particle = sim.sand.particles[i];
			float v = tvmet::max(tvmet::abs(particle.v));
			if (v > vmax) imax = i;
			vmax = fmaxf(vmax, v);
			for (int b = 0; b < nb; b++)
				dist_to_body[b] = min(dist_to_body[b], distance(sim.bodies[b], particle.x));
		}
		dt = sim.sand.dx/(vmax + sqrtf(vmax*vmax + 2*norm2(sim.g)*sim.sand.dx)); // so dx/2 >= vmax*dt + g*dt*dt/2
		for (int i = 0; i < nb; i++) {
			RigidBody<d> &body = sim.bodies[i];
			float dist = max(dist_to_body[i], sim.sand.dx/2);
			float v = speed(body);
			float dt_body = 2*dist/(v + sqrtf(v*v + 4*norm2(sim.g)*dist));
			dt = min(dt, dt_body);
		}
		for (int i = 0; i < sim.scripted.size(); i++) {
			float v = sim.scripted[i]->speed(sim);
			float dt_scripted = sim.sand.dx/(v + sqrtf(v*v + 2*norm2(sim.g)*sim.sand.dx));
			dt = min(dt, dt_scripted);
		}
		if (sim.sand.hacks.cfl > 0)
			dt *= sim.sand.hacks.cfl;
		if (t + dt > sim.dt) {
			dt = sim.dt - t;
			t = sim.dt;
		} else if (t + dt*2 > sim.dt) {
			dt = (sim.dt - t)/2;
			t += dt;
		} else {
			t += dt;
		}
		Timer timer; timer.tick();
		substep(sim, dt);
		cout << "SUBSTEP: " << timer.tick() << endl;
	} while (t < sim.dt);
	sim.frame++;
}
template void step (Simulator<2> &sim);
template void step (Simulator<3> &sim);
