/*

Copyright (c) 2010, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "zhubridson.hpp"
#include "../math/pcg_solver.h"
#include "../math/qp.hpp"
#include "simhelpers.hpp"
#include "rigid.hpp"
using namespace std;

inline bool is_fluid (Sand<2> &sand, int i, int j) {return sand.rho(i,j)/sand.rho_max > epsilon || sand.aux.f_obs(i,j) > 0.5;}
inline bool is_fluid (Sand<3> &sand, int i, int j, int k) {return sand.rho(i,j,k)/sand.rho_max > epsilon || sand.aux.f_obs(i,j,k) > 0.5f;}

void splat_zb (Sand<2> &sand) {
	float dx = sand.dx;
	set_all(sand.rho, 0);
	float dvol = dx*dx;
	int np = sand.particles.size();
	for (int p = 0; p < np; p++) {
		Particle<2> &particle = sand.particles[p];
		int i = (int)floor(particle.x[0]/dx), j = (int)floor(particle.x[1]/dx);
		i = i<0 ? 0 : i>=sand.m ? sand.m-1 : i;
		j = j<0 ? 0 : j>=sand.n ? sand.n-1 : j;
		sand.rho(i,j) += particle.m/dvol;
	}
}
void splat_zb (Sand<3> &sand) {
	float dx = sand.dx;
	set_all(sand.rho, 0);
	float dvol = dx*dx*dx;
	int np = sand.particles.size();
	for (int p = 0; p < np; p++) {
		Particle<3> &particle = sand.particles[p];
		int i = (int)floor(particle.x[0]/dx), j = (int)floor(particle.x[1]/dx), k = (int)floor(particle.x[2]/dx);
		i = i<0 ? 0 : i>=sand.m ? sand.m-1 : i;
		j = j<0 ? 0 : j>=sand.n ? sand.n-1 : j;
		k = k<0 ? 0 : k>=sand.o ? sand.o-1 : k;
		sand.rho(i,j,k) += particle.m/dvol;
	}
}

inline void frictional_boundary (float &vt, float mu, float dvn, float vt_obs) {
	if (fabsf(dvn) < epsilon) return;
	float f = min(1.f, 5*mu*fabsf(dvn/(vt - vt_obs)));
	vt = lerp(f, vt, vt_obs);
}

template <> void pressure_solve_zb (Sand<2> &sand, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static IntArray<2> idx;
	splat_zb(sand);
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	int pm = m + 2, pn = n + 2;
	idx.initialize(pm,pn, -1);
	int nactive = 0;
	// enumerate active cells
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		bool active = true;
		//if (i > 0 && j >= 0 && j < n && sand.rho(i-1,j)+sand.aux.f_obs(i-1,j) < epsilon) active = false;
		//if (i < m-1 && j >= 0 && j < n && sand.rho(i+1,j)+sand.aux.f_obs(i+1,j) < epsilon) active = false;
		//if (i >= 0 && i < m && j > 0 && sand.rho(i,j-1)+sand.aux.f_obs(i,j-1) < epsilon) active = false;
		//if (i >= 0 && i < m && j < n-1 && sand.rho(i,j+1)+sand.aux.f_obs(i,j+1) < epsilon) active = false;
		if (i >= 0 && i < m && j >= 0 && j < n) active = is_fluid(sand, i,j);
		if ((i == -1 || i == m) && (j == -1 || j == n)) active = false;
		if (active)
			idx(i+1,j+1) = nactive++;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1);
		if (a >= 0)
			x[a] = sand.aux.p(i+1,j+1);
	}
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 1; i < m; i++) {
		int a = idx(i,j+1), b = idx(i+1,j+1);
		float f_obs = sand.aux.f_obs_s.x(i,j);
		float mass = 1 - f_obs;
		float vx = sand.v.x(i,j);
		float multiplier = coeffl*mass;
		float fmultiplier = coeffr*(mass*vx + f_obs*sand.aux.v_obs.x(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int j = 1; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j), b = idx(i+1,j+1);
		float f_obs = sand.aux.f_obs_s.y(i,j);
		float mass = 1 - f_obs;
		float vy = sand.v.y(i,j);
		float multiplier = coeffl*mass;
		float fmultiplier = coeffr*(mass*vy + f_obs*sand.aux.v_obs.y(i,j));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// // cells
	// for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
	// 	int a = idx(i+1,j+1);
	// 	if (a < 0) continue;
	// 	if (min(min(sand.rho_s.x(i,j), sand.rho_s.x(i+1,j)), min(sand.rho_s.y(i,j), sand.rho_s.y(i,j+1))) < epsilon) 
	// 		f[a] = 0;
	// }
	PCGSolver<float> solver;
	solver.set_solver_parameters(1e-3f, 100);
	QpReport report;
	solver.solve(H, f, x, report.residual, report.iterations);
	for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1);
		if (a >= 0) sand.aux.p(i+1,j+1) = x[a];
		else sand.aux.p(i+1,j+1) = 0;
	}
	float multiplier = dt/(dx*sand.rho_max);
	// x faces
	for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float f_obs = sand.aux.f_obs_s.x(i,j);
		float mass = 1 - f_obs;
		sand.aux.dp.x(i,j) = (mass<epsilon) ? 0 : (sand.aux.p(i,j+1) - sand.aux.p(i+1,j+1))*multiplier*(1-f_obs)/mass;
		if (i == 0 || i == m) sand.aux.dp.x(i,j) = -sand.v.x(i,j);
		sand.v.x(i,j) += sand.aux.dp.x(i,j);
		// wall friction
		float dvx = sand.aux.dp.x(i,j);
		if (i < m) if (i == 0 || sand.aux.f_obs(i-1,j) > 0.5f) {
			frictional_boundary(sand.v.y(i,j),   sand.mu, dvx, i==0 ? 0 : sand.aux.v_obs.y(i,j));
			frictional_boundary(sand.v.y(i,j+1), sand.mu, dvx, i==0 ? 0 : sand.aux.v_obs.y(i,j+1));
		}
		if (i > 0) if (i == m || sand.aux.f_obs(i,j) > 0.5f) {
			frictional_boundary(sand.v.y(i-1,j),   sand.mu, dvx, i==m ? 0 : sand.aux.v_obs.y(i-1,j));
			frictional_boundary(sand.v.y(i-1,j+1), sand.mu, dvx, i==m ? 0 : sand.aux.v_obs.y(i-1,j+1));
		}
	}
	// y faces
	for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float f_obs = sand.aux.f_obs_s.y(i,j);
		float mass = 1 - f_obs;
		sand.aux.dp.y(i,j) = (mass<epsilon) ? 0 : (sand.aux.p(i+1,j) - sand.aux.p(i+1,j+1))*multiplier*(1-f_obs)/mass;
		if (j == 0 || j == n) sand.aux.dp.y(i,j) = -sand.v.y(i,j);
		sand.v.y(i,j) += sand.aux.dp.y(i,j);
		// wall friction
		float dvy = sand.aux.dp.y(i,j);
		if (j < n) if (j == 0 || sand.aux.f_obs(i,j-1) > 0.5f) {
			frictional_boundary(sand.v.x(i,j),   sand.mu, dvy, j==0 ? 0 : sand.aux.v_obs.x(i,j));
			frictional_boundary(sand.v.x(i+1,j), sand.mu, dvy, j==0 ? 0 : sand.aux.v_obs.x(i+1,j));
		}
		if (j > 0) if (j == n || sand.aux.f_obs(i,j) > 0.5f) {
			frictional_boundary(sand.v.x(i,j-1),   sand.mu, dvy, j==n ? 0 : sand.aux.v_obs.x(i,j-1));
			frictional_boundary(sand.v.x(i+1,j-1), sand.mu, dvy, j==n ? 0 : sand.aux.v_obs.x(i+1,j-1));
		}
	}
}

template <> void pressure_solve_zb (Sand<3> &sand, float dt) {
	static SparseMatrixf H;
	static vector<float> f, x;
	static IntArray<3> idx;
	splat_zb(sand);
	// Timer timer;
	// timer.set();
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	int pm = m + 2, pn = n + 2, po = o + 2;
	idx.initialize(pm,pn,po, -1);
	int nactive = 0;
	// enumerate active cells
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		bool active = true;
		// if (i >= 0 && j >= 0 && j < n && rho_x(sand,i,j) > epsilon) active = true;
		// if (i < m && j >= 0 && j < n && rho_x(sand,i+1,j) > epsilon) active = true;
		// if (i >= 0 && i < m && j >= 0 && rho_y(sand,i,j) > epsilon) active = true;
		// if (i >= 0 && i < m && j < n && rho_y(sand,i,j+1) > epsilon) active = true;
		if (i >= 0 && i < m && j >= 0 && j < n && k >= 0 && k < o) active = is_fluid(sand, i,j,k);
		if ((i == -1 || i == m) && (j == -1 || j == n) && (k == -1 || k == o)) active = false;
		if (active)
			idx(i+1,j+1,k+1) = nactive++;
	}
	H.resize(nactive);
	H.zero();
	f.assign(nactive, 0);
	x.assign(nactive, 0);
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1,k+1);
		if (a >= 0)
			x[a] = sand.aux.p(i+1,j+1,k+1);
	}
	float coeffr = dt/dx, coeffl = dt*dt/(dx*dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 1; i < m; i++) {
		int a = idx(i,j+1,k+1), b = idx(i+1,j+1,k+1);
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float mass = 1 - f_obs;
		float vx = sand.v.x(i,j,k);
		float multiplier = coeffl*mass;
		float fmultiplier = coeffr*(mass*vx + f_obs*sand.aux.v_obs.x(i,j,k));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 1; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j,k+1), b = idx(i+1,j+1,k+1);
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float mass = 1 - f_obs;
		float vy = sand.v.y(i,j,k);
		float multiplier = coeffl*mass;
		float fmultiplier = coeffr*(mass*vy + f_obs*sand.aux.v_obs.y(i,j,k));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// z faces
	for (int k = 1; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		int a = idx(i+1,j+1,k), b = idx(i+1,j+1,k+1);
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float mass = 1 - f_obs;
		float vz = sand.v.z(i,j,k);
		float multiplier = coeffl*mass;
		float fmultiplier = coeffr*(mass*vz + f_obs*sand.aux.v_obs.z(i,j,k));
		if (a >= 0) {
			H.add_to_element(a, a, multiplier);
			f[a] -= fmultiplier;
		}
		if (b >= 0) {
			H.add_to_element(b, b, multiplier);
			f[b] += fmultiplier;
		}
		if (a >= 0 && b >= 0) {
			H.add_to_element(a, b, -multiplier);
			H.add_to_element(b, a, -multiplier);
		}
	}
	// // cells
	// for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
	// 	int a = idx(i+1,j+1);
	// 	if (a < 0) continue;
	// 	if (min(min(sand.rho_s.x(i,j), sand.rho_s.x(i+1,j)), min(sand.rho_s.y(i,j), sand.rho_s.y(i,j+1))) < epsilon) 
	// 		f[a] = 0;
	// }
	PCGSolver<float> solver;
	solver.set_solver_parameters(1e-3f, 100);
	QpReport report;
	solver.solve(H, f, x, report.residual, report.iterations);
	for (int k = -1; k <= o; k++) for (int j = -1; j <= n; j++) for (int i = -1; i <= m; i++) {
		int a = idx(i+1,j+1,k+1);
		if (a >= 0) sand.aux.p(i+1,j+1,k+1) = x[a];
		else sand.aux.p(i+1,j+1,k+1) = 0;
	}
	float multiplier = dt/(dx*sand.rho_max);
	// x faces
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i <= m; i++) {
		float f_obs = sand.aux.f_obs_s.x(i,j,k);
		float mass = 1 - f_obs;
		float dvx = (mass<epsilon) ? 0 : (sand.aux.p(i,j+1,k+1) - sand.aux.p(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
		if (i == 0 || i == m) dvx = -sand.v.x(i,j,k);
		sand.aux.dp.x(i,j,k) = dvx;
		sand.v.x(i,j,k) += dvx;
		// wall friction
		if (i < m) if (i == 0 || sand.aux.f_obs(i-1,j,k) > 0.5f) {
			frictional_boundary(sand.v.y(i,j,  k), sand.mu, dvx, i==0 ? 0 : sand.aux.v_obs.y(i,j,  k));
			frictional_boundary(sand.v.y(i,j+1,k), sand.mu, dvx, i==0 ? 0 : sand.aux.v_obs.y(i,j+1,k));
			frictional_boundary(sand.v.z(i,j,k),   sand.mu, dvx, i==0 ? 0 : sand.aux.v_obs.z(i,j,k));
			frictional_boundary(sand.v.z(i,j,k+1), sand.mu, dvx, i==0 ? 0 : sand.aux.v_obs.z(i,j,k+1));
		}
		if (i > 0) if (i == m || sand.aux.f_obs(i,j,k) > 0.5f) {
			frictional_boundary(sand.v.y(i-1,j,  k), sand.mu, dvx, i==m ? 0 : sand.aux.v_obs.y(i-1,j,  k));
			frictional_boundary(sand.v.y(i-1,j+1,k), sand.mu, dvx, i==m ? 0 : sand.aux.v_obs.y(i-1,j+1,k));
			frictional_boundary(sand.v.z(i-1,j,k),   sand.mu, dvx, i==m ? 0 : sand.aux.v_obs.z(i-1,j,k));
			frictional_boundary(sand.v.z(i-1,j,k+1), sand.mu, dvx, i==m ? 0 : sand.aux.v_obs.z(i-1,j,k+1));
		}
	}
	// y faces
	for (int k = 0; k < o; k++) for (int j = 0; j <= n; j++) for (int i = 0; i < m; i++) {
		float f_obs = sand.aux.f_obs_s.y(i,j,k);
		float mass = 1 - f_obs;
		float dvy = (mass<epsilon) ? 0 : (sand.aux.p(i+1,j,k+1) - sand.aux.p(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
		if (j == 0 || j == n) dvy = -sand.v.y(i,j,k);
		sand.aux.dp.y(i,j,k) = dvy;
		sand.v.y(i,j,k) += dvy;
		// wall friction
		if (j < n) if (j == 0 || sand.aux.f_obs(i,j-1,k) > 0.5f) {
			frictional_boundary(sand.v.x(i,  j,k), sand.mu, dvy, j==0 ? 0 : sand.aux.v_obs.x(i,  j,k));
			frictional_boundary(sand.v.x(i+1,j,k), sand.mu, dvy, j==0 ? 0 : sand.aux.v_obs.x(i+1,j,k));
			frictional_boundary(sand.v.z(i,j,k),   sand.mu, dvy, j==0 ? 0 : sand.aux.v_obs.z(i,j,k));
			frictional_boundary(sand.v.z(i,j,k+1), sand.mu, dvy, j==0 ? 0 : sand.aux.v_obs.z(i,j,k+1));
		}
		if (j > 0) if (j == n || sand.aux.f_obs(i,j,k) > 0.5f) {
			frictional_boundary(sand.v.x(i,  j-1,k), sand.mu, dvy, j==n ? 0 : sand.aux.v_obs.x(i,  j-1,k));
			frictional_boundary(sand.v.x(i+1,j-1,k), sand.mu, dvy, j==n ? 0 : sand.aux.v_obs.x(i+1,j-1,k));
			frictional_boundary(sand.v.z(i,j-1,k),   sand.mu, dvy, j==n ? 0 : sand.aux.v_obs.z(i,j-1,k));
			frictional_boundary(sand.v.z(i,j-1,k+1), sand.mu, dvy, j==n ? 0 : sand.aux.v_obs.z(i,j-1,k+1));
		}
	}
	// z faces
	for (int k = 0; k <= o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		float f_obs = sand.aux.f_obs_s.z(i,j,k);
		float mass = 1 - f_obs;
		float dvz = (mass<epsilon) ? 0 : (sand.aux.p(i+1,j+1,k) - sand.aux.p(i+1,j+1,k+1))*multiplier*(1-f_obs)/mass;
		if (k == 0 || k == o) dvz = -sand.v.z(i,j,k);
		sand.aux.dp.z(i,j,k) = dvz;
		sand.v.z(i,j,k) += dvz;
		// wall friction
		if (k < o) if (k == 0 || sand.aux.f_obs(i,j,k-1) > 0.5f) {
			frictional_boundary(sand.v.x(i,  j,k), sand.mu, dvz, k==0 ? 0 : sand.aux.v_obs.x(i,  j,k));
			frictional_boundary(sand.v.x(i+1,j,k), sand.mu, dvz, k==0 ? 0 : sand.aux.v_obs.x(i+1,j,k));
			frictional_boundary(sand.v.y(i,j,  k), sand.mu, dvz, k==0 ? 0 : sand.aux.v_obs.y(i,j,  k));
			frictional_boundary(sand.v.y(i,j+1,k), sand.mu, dvz, k==0 ? 0 : sand.aux.v_obs.y(i,j+1,k));
		}
		if (k > 0) if (k == o || sand.aux.f_obs(i,j,k) > 0.5f) {
			frictional_boundary(sand.v.x(i,  j,k-1), sand.mu, dvz, k==o ? 0 : sand.aux.v_obs.x(i,  j,k-1));
			frictional_boundary(sand.v.x(i+1,j,k-1), sand.mu, dvz, k==o ? 0 : sand.aux.v_obs.x(i+1,j,k-1));
			frictional_boundary(sand.v.y(i,j,  k-1), sand.mu, dvz, k==o ? 0 : sand.aux.v_obs.y(i,j,  k-1));
			frictional_boundary(sand.v.y(i,j+1,k-1), sand.mu, dvz, k==o ? 0 : sand.aux.v_obs.y(i,j+1,k-1));
		}
	}
}

struct Index2 {int i, j; Index2 (int i, int j) : i(i), j(j) {}};
struct Index3 {int i, j, k; Index3 (int i, int j, int k) : i(i), j(j), k(k) {}};

template <> void friction_solve_zb (Sand<2> &sand, float dt) {
	static IntArray<2> rigid;
	int m = sand.m, n = sand.n;
	float dx = sand.dx;
	rigid.initialize(m,n, 0);
	set_all(sand.aux.s, Mat2f(0.f));
	// determine yield condition
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		if (!is_fluid(sand, i,j)) continue;
		Mat2f D(0.f);
		D(0,0) = (sand.v.x(i+1,j) - sand.v.x(i,j))/dx;
		D(1,1) = (sand.v.y(i,j+1) - sand.v.y(i,j))/dx;
		float dvy_dx = (i == 0 || i == m-1) ? 0 : (sand.v.y(i+1,j) + sand.v.y(i+1,j+1) - sand.v.y(i-1,j) - sand.v.y(i-1,j+1))/(4*dx);
		float dvx_dy = (j == 0 || j == n-1) ? 0 : (sand.v.x(i,j+1) + sand.v.x(i+1,j+1) - sand.v.x(i,j-1) - sand.v.x(i+1,j-1))/(4*dx);
		D(0,1) = D(1,0) = (dvy_dx + dvx_dy)*0.5f;
		float d = sqrtf(sqf(D(0,0)) + sqf(D(0,1)) + sqf(D(1,0)) + sqf(D(1,1)));
		float s_rigid = sand.rho_max*d*dx*dx/dt;
		float s_max = sand.mu*sand.aux.p(i+1,j+1) + sand.hacks.cohesion;
		if (sqrtf(3.f)*s_rigid < s_max) {
			rigid(i,j) = 1;
		} else {
			Mat2f s(s_max*D/(sqrtf(1/3.f)*d+epsilon));
			sand.aux.s.xx(i,j) = s(0,0);
			sand.aux.s.yy(i,j) = s(1,1);
			sand.aux.s.xy(i,j)     += s(0,1)/4;
			sand.aux.s.xy(i+1,j)   += s(0,1)/4;
			sand.aux.s.xy(i,j+1)   += s(0,1)/4;
			sand.aux.s.xy(i+1,j+1) += s(0,1)/4;
		}
		//sand.aux.p(i+1,j+1) = (sqrtf(3.f)*s_rigid - sand.mu*sand.aux.p(i+1,j+1))*10;
	}
	// apply friction to nonrigid cells
	for (int j = 0; j < n; j++) for (int i = 1; i < m; i++) {
		if (rigid(i-1,j) || rigid(i,j)) continue;
		sand.v.x(i,j) += dt/sand.rho_max*(sand.aux.s.xx(i,j) - sand.aux.s.xx(i-1,j))/dx;
		sand.v.x(i,j) += dt/sand.rho_max*(sand.aux.s.xy(i,j+1) - sand.aux.s.xy(i,j))/dx;
	}
	for (int j = 1; j < n; j++) for (int i = 0; i < m; i++) {
		if (rigid(i,j-1) || rigid(i,j)) continue;
		sand.v.y(i,j) += dt/sand.rho_max*(sand.aux.s.yy(i,j) - sand.aux.s.yy(i,j-1))/dx;
		sand.v.y(i,j) += dt/sand.rho_max*(sand.aux.s.xy(i+1,j) - sand.aux.s.xy(i,j))/dx;
	}
	// project velocities in rigid cells
	for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		if (!rigid(i,j))
			continue;
		// found new rigid component
		int ncells = 0;
		Vec2f x(0.f);
		Screw<2> t(0.f);
		vector<Index2> group, stack;
		stack.push_back(Index2(i,j));
		while (!stack.empty()) {
			Index2 ix = stack.back();
			stack.pop_back();
			int i = ix.i, j = ix.j; // aliasing! doesn't matter; we won't need the original i, j in here
			group.push_back(ix);
			rigid(i,j) = 0;
			ncells += 1;
			x += Vec2f((i+0.5f)*dx, (j+0.5f)*dx);
			Vec2f v((sand.v.x(i,j) + sand.v.x(i+1,j))*0.5f, (sand.v.y(i,j) + sand.v.y(i,j+1))*0.5f);
			float w = (i+0.5f)*dx*v[1] - (j+0.5f)*dx*v[0];
			t += screw(v,w);
			if (i > 0   && rigid(i-1,j)) stack.push_back(Index2(i-1,j));
			if (i < m-1 && rigid(i+1,j)) stack.push_back(Index2(i+1,j));
			if (j > 0   && rigid(i,j-1)) stack.push_back(Index2(i,j-1));
			if (j < n-1 && rigid(i,j+1)) stack.push_back(Index2(i,j+1));
		}
		x /= ncells;
		t /= ncells;
		t[2] -= x[0]*t[1] - x[1]*t[0];
		for (int c = 0; c < ncells; c++) {
			Index2 ix = group[c];
			int i = ix.i, j = ix.j;
			sand.v.x(i,j) = sand.v.x(i+1,j) = t[0];// + (x[1] - (j+0.5f)*dx)*t[2];
			sand.v.y(i,j) = sand.v.y(i,j+1) = t[1];// + ((i+0.5f)*dx - x[0])*t[2];
		}
	}
}

template <> void friction_solve_zb (Sand<3> &sand, float dt) {
	static IntArray<3> rigid;
	int m = sand.m, n = sand.n, o = sand.o;
	float dx = sand.dx;
	rigid.initialize(m,n,o, 0);
	set_all(sand.aux.s, Mat3f(0.f));
	// determine yield condition
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		if (!is_fluid(sand, i,j,k)) continue;
		Mat3f D(0.f);
		D(0,0) = (sand.v.x(i+1,j,k) - sand.v.x(i,j,k))/dx;
		D(1,1) = (sand.v.y(i,j+1,k) - sand.v.y(i,j,k))/dx;
		D(2,2) = (sand.v.z(i,j,k+1) - sand.v.z(i,j,k))/dx;
		float dvy_dx = (i == 0 || i == m-1) ? 0 : (sand.v.y(i+1,j,k) + sand.v.y(i+1,j+1,k) - sand.v.y(i-1,j,k) - sand.v.y(i-1,j+1,k))/(4*dx);
		float dvx_dy = (j == 0 || j == n-1) ? 0 : (sand.v.x(i,j+1,k) + sand.v.x(i+1,j+1,k) - sand.v.x(i,j-1,k) - sand.v.x(i+1,j-1,k))/(4*dx);
		D(0,1) = D(1,0) = (dvy_dx + dvx_dy)*0.5f;
		float dvz_dx = (i == 0 || i == m-1) ? 0 : (sand.v.z(i+1,j,k) + sand.v.z(i+1,j,k+1) - sand.v.z(i-1,j,k) - sand.v.z(i-1,j,k+1))/(4*dx);
		float dvx_dz = (k == 0 || k == o-1) ? 0 : (sand.v.x(i,j,k+1) + sand.v.x(i+1,j,k+1) - sand.v.x(i,j,k-1) - sand.v.x(i+1,j,k-1))/(4*dx);
		D(0,2) = D(2,0) = (dvz_dx + dvx_dz)*0.5f;
		float dvz_dy = (j == 0 || j == n-1) ? 0 : (sand.v.z(i,j+1,k) + sand.v.z(i,j+1,k+1) - sand.v.z(i,j-1,k) - sand.v.z(i,j-1,k+1))/(4*dx);
		float dvy_dz = (k == 0 || k == o-1) ? 0 : (sand.v.y(i,j,k+1) + sand.v.y(i,j+1,k+1) - sand.v.y(i,j,k-1) - sand.v.y(i,j+1,k-1))/(4*dx);
		D(1,2) = D(2,1) = (dvz_dy + dvy_dz)*0.5f;
		float d = sqrtf(sqf(D(0,0)) + sqf(D(0,1)) + sqf(D(0,2)) + sqf(D(1,0)) + sqf(D(1,1)) + sqf(D(1,2)) + sqf(D(2,0)) + sqf(D(2,1)) + sqf(D(2,2)));
		float s_rigid = sand.rho_max*d*dx*dx/dt;
		float s_max = sand.mu*sand.aux.p(i+1,j+1,k+1) + sand.hacks.cohesion;
		if (sqrtf(3.f)*s_rigid < s_max) {
			rigid(i,j,k) = 1;
		} else {
			Mat3f s(s_max*D/(sqrtf(1/3.f)*d+epsilon));
			sand.aux.s.xx(i,j,k) = s(0,0);
			sand.aux.s.yy(i,j,k) = s(1,1);
			sand.aux.s.zz(i,j,k) = s(2,2);
			sand.aux.s.xy(i,  j,  k) += s(0,1)/4;
			sand.aux.s.xy(i+1,j,  k) += s(0,1)/4;
			sand.aux.s.xy(i,  j+1,k) += s(0,1)/4;
			sand.aux.s.xy(i+1,j+1,k) += s(0,1)/4;
			sand.aux.s.xz(i,  j,k)   += s(0,2)/4;
			sand.aux.s.xz(i+1,j,k)   += s(0,2)/4;
			sand.aux.s.xz(i,  j,k+1) += s(0,2)/4;
			sand.aux.s.xz(i+1,j,k+1) += s(0,2)/4;
			sand.aux.s.yz(i,j,  k)   += s(1,2)/4;
			sand.aux.s.yz(i,j+1,k)   += s(1,2)/4;
			sand.aux.s.yz(i,j,  k+1) += s(1,2)/4;
			sand.aux.s.yz(i,j+1,k+1) += s(1,2)/4;
		}
		sand.aux.p(i+1,j+1,k+1) = D(1,2);//rigid(i,j,k)*10000;
		//sand.aux.p(i+1,j+1) = (sqrtf(3.f)*s_rigid - sand.mu*sand.aux.p(i+1,j+1))*10;
	}
	// apply friction to nonrigid cells
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 1; i < m; i++) {
		if (rigid(i-1,j,k) || rigid(i,j,k)) continue;
		sand.v.x(i,j,k) += dt/sand.rho_max*(sand.aux.s.xx(i,j,k) - sand.aux.s.xx(i-1,j,k))/dx;
		sand.v.x(i,j,k) += dt/sand.rho_max*(sand.aux.s.xy(i,j+1,k) - sand.aux.s.xy(i,j,k))/dx;
		sand.v.x(i,j,k) += dt/sand.rho_max*(sand.aux.s.xz(i,j,k+1) - sand.aux.s.xz(i,j,k))/dx;
	}
	for (int k = 0; k < o; k++) for (int j = 1; j < n; j++) for (int i = 0; i < m; i++) {
		if (rigid(i,j-1,k) || rigid(i,j,k)) continue;
		sand.v.y(i,j,k) += dt/sand.rho_max*(sand.aux.s.yy(i,j,k) - sand.aux.s.yy(i,j-1,k))/dx;
		sand.v.y(i,j,k) += dt/sand.rho_max*(sand.aux.s.xy(i+1,j,k) - sand.aux.s.xy(i,j,k))/dx;
		sand.v.y(i,j,k) += dt/sand.rho_max*(sand.aux.s.yz(i,j,k+1) - sand.aux.s.yz(i,j,k))/dx;
	}
	for (int k = 1; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		if (rigid(i,j,k-1) || rigid(i,j,k)) continue;
		sand.v.z(i,j,k) += dt/sand.rho_max*(sand.aux.s.zz(i,j,k) - sand.aux.s.zz(i,j,k-1))/dx;
		sand.v.z(i,j,k) += dt/sand.rho_max*(sand.aux.s.xz(i+1,j,k) - sand.aux.s.xz(i,j,k))/dx;
		sand.v.z(i,j,k) += dt/sand.rho_max*(sand.aux.s.yz(i,j+1,k) - sand.aux.s.yz(i,j,k))/dx;
	}
	// project velocities in rigid cells
	for (int k = 0; k < o; k++) for (int j = 0; j < n; j++) for (int i = 0; i < m; i++) {
		if (!rigid(i,j,k))
			continue;
		// found new rigid component
		int ncells = 0;
		Vec3f sumx(0.f);
		Screw<3> sumt(0.f);
		vector<Index3> group, stack;
		stack.push_back(Index3(i,j,k));
		while (!stack.empty()) {
			Index3 ix = stack.back();
			stack.pop_back();
			int i = ix.i, j = ix.j, k = ix.k; // aliasing! doesn't matter; we won't need the original i, j in here
			group.push_back(ix);
			rigid(i,j,k) = -1;
			ncells += 1;
			Vec3f x((i+0.5f)*dx, (j+0.5f)*dx, (k+0.5f)*dx);
			sumx += x;
			Vec3f v((sand.v.x(i,j,k) + sand.v.x(i+1,j,k))*0.5f, (sand.v.y(i,j,k) + sand.v.y(i,j+1,k))*0.5f, (sand.v.z(i,j,k) + sand.v.z(i,j,k+1))*0.5f);
			Vec3f w(cross(x, v));
			sumt += screw(v,w);
			if (i > 0   && rigid(i-1,j,k) == 1) stack.push_back(Index3(i-1,j,k));
			if (i < m-1 && rigid(i+1,j,k) == 1) stack.push_back(Index3(i+1,j,k));
			if (j > 0   && rigid(i,j-1,k) == 1) stack.push_back(Index3(i,j-1,k));
			if (j < n-1 && rigid(i,j+1,k) == 1) stack.push_back(Index3(i,j+1,k));
			if (k > 0   && rigid(i,j,k-1) == 1) stack.push_back(Index3(i,j,k-1));
			if (k < o-1 && rigid(i,j,k+1) == 1) stack.push_back(Index3(i,j,k+1));
		}
		Vec3f x(sumx/ncells);
		Screw<3> t (sumt/ncells);
		Vec3f v(lin(t));
		Vec3f w(ang(t) - cross(x,lin(t)));
		for (int c = 0; c < ncells; c++) {
			Index3 ix = group[c];
			int i = ix.i, j = ix.j, k = ix.k;
			Vec3f xcell((i+0.5f)*dx, (j+0.5f)*dx, (k+0.5f)*dx);
			Vec3f vcell(v);// + cross(w, xcell - x));
			sand.v.x(i,j,k) = sand.v.x(i+1,j,k) = vcell[0];
			sand.v.y(i,j,k) = sand.v.y(i,j+1,k) = vcell[1];
			sand.v.z(i,j,k) = sand.v.z(i,j,k+1) = vcell[2];
		}
	}
}
